(window["webpackJsonp"]=window["webpackJsonp"]||[]).push([["chunk-vendors"],{"0048":function(t,e,n){!function(e,n){t.exports=n()}("undefined"!=typeof self&&self,(function(){return function(t){var e={};function n(r){if(e[r])return e[r].exports;var i=e[r]={i:r,l:!1,exports:{}};return t[r].call(i.exports,i,i.exports,n),i.l=!0,i.exports}return n.m=t,n.c=e,n.d=function(t,e,r){n.o(t,e)||Object.defineProperty(t,e,{enumerable:!0,get:r})},n.r=function(t){"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(t,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(t,"__esModule",{value:!0})},n.t=function(t,e){if(1&e&&(t=n(t)),8&e)return t;if(4&e&&"object"==typeof t&&t&&t.__esModule)return t;var r=Object.create(null);if(n.r(r),Object.defineProperty(r,"default",{enumerable:!0,value:t}),2&e&&"string"!=typeof t)for(var i in t)n.d(r,i,function(e){return t[e]}.bind(null,i));return r},n.n=function(t){var e=t&&t.__esModule?function(){return t.default}:function(){return t};return n.d(e,"a",e),e},n.o=function(t,e){return Object.prototype.hasOwnProperty.call(t,e)},n.p="",n(n.s=2)}([function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",(function(){return i})),n.d(e,"__assign",(function(){return o})),n.d(e,"__rest",(function(){return s})),n.d(e,"__decorate",(function(){return a})),n.d(e,"__param",(function(){return c})),n.d(e,"__metadata",(function(){return u})),n.d(e,"__awaiter",(function(){return l})),n.d(e,"__generator",(function(){return f})),n.d(e,"__createBinding",(function(){return d})),n.d(e,"__exportStar",(function(){return h})),n.d(e,"__values",(function(){return p})),n.d(e,"__read",(function(){return y})),n.d(e,"__spread",(function(){return v})),n.d(e,"__spreadArrays",(function(){return b})),n.d(e,"__await",(function(){return m})),n.d(e,"__asyncGenerator",(function(){return g})),n.d(e,"__asyncDelegator",(function(){return _})),n.d(e,"__asyncValues",(function(){return w})),n.d(e,"__makeTemplateObject",(function(){return O})),n.d(e,"__importStar",(function(){return S})),n.d(e,"__importDefault",(function(){return x})),n.d(e,"__classPrivateFieldGet",(function(){return I})),n.d(e,"__classPrivateFieldSet",(function(){return A}));
/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
var r=function(t,e){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return(o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)};function s(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&Object.prototype.propertyIsEnumerable.call(t,r[i])&&(n[r[i]]=t[r[i]])}return n}function a(t,e,n,r){var i,o=arguments.length,s=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)s=Reflect.decorate(t,e,n,r);else for(var a=t.length-1;a>=0;a--)(i=t[a])&&(s=(o<3?i(s):o>3?i(e,n,s):i(e,n))||s);return o>3&&s&&Object.defineProperty(e,n,s),s}function c(t,e){return function(n,r){e(n,r,t)}}function u(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){return new(n||(n=Promise))((function(i,o){function s(t){try{c(r.next(t))}catch(t){o(t)}}function a(t){try{c(r.throw(t))}catch(t){o(t)}}function c(t){var e;t.done?i(t.value):(e=t.value,e instanceof n?e:new n((function(t){t(e)}))).then(s,a)}c((r=r.apply(t,e||[])).next())}))}function f(t,e){var n,r,i,o,s={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:a(0),throw:a(1),return:a(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function a(o){return function(a){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;s;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return s.label++,{value:o[1],done:!1};case 5:s.label++,r=o[1],o=[0];continue;case 7:o=s.ops.pop(),s.trys.pop();continue;default:if(i=s.trys,!((i=i.length>0&&i[i.length-1])||6!==o[0]&&2!==o[0])){s=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){s.label=o[1];break}if(6===o[0]&&s.label<i[1]){s.label=i[1],i=o;break}if(i&&s.label<i[2]){s.label=i[2],s.ops.push(o);break}i[2]&&s.ops.pop(),s.trys.pop();continue}o=e.call(t,s)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,a])}}}function d(t,e,n,r){void 0===r&&(r=n),t[r]=e[n]}function h(t,e){for(var n in t)"default"===n||e.hasOwnProperty(n)||(e[n]=t[n])}function p(t){var e="function"==typeof Symbol&&Symbol.iterator,n=e&&t[e],r=0;if(n)return n.call(t);if(t&&"number"==typeof t.length)return{next:function(){return t&&r>=t.length&&(t=void 0),{value:t&&t[r++],done:!t}}};throw new TypeError(e?"Object is not iterable.":"Symbol.iterator is not defined.")}function y(t,e){var n="function"==typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),s=[];try{for(;(void 0===e||e-- >0)&&!(r=o.next()).done;)s.push(r.value)}catch(t){i={error:t}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return s}function v(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(y(arguments[e]));return t}function b(){for(var t=0,e=0,n=arguments.length;e<n;e++)t+=arguments[e].length;var r=Array(t),i=0;for(e=0;e<n;e++)for(var o=arguments[e],s=0,a=o.length;s<a;s++,i++)r[i]=o[s];return r}function m(t){return this instanceof m?(this.v=t,this):new m(t)}function g(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},s("next"),s("throw"),s("return"),r[Symbol.asyncIterator]=function(){return this},r;function s(t){i[t]&&(r[t]=function(e){return new Promise((function(n,r){o.push([t,e,n,r])>1||a(t,e)}))})}function a(t,e){try{(n=i[t](e)).value instanceof m?Promise.resolve(n.value.v).then(c,u):l(o[0][2],n)}catch(t){l(o[0][3],t)}var n}function c(t){a("next",t)}function u(t){a("throw",t)}function l(t,e){t(e),o.shift(),o.length&&a(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",(function(t){throw t})),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:m(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function w(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t=p(t),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise((function(r,i){(function(t,e,n,r){Promise.resolve(r).then((function(e){t({value:e,done:n})}),e)})(r,i,(e=t[n](e)).done,e.value)}))}}}function O(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function S(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function x(t){return t&&t.__esModule?t:{default:t}}function I(t,e){if(!e.has(t))throw new TypeError("attempted to get private field on non-instance");return e.get(t)}function A(t,e,n){if(!e.has(t))throw new TypeError("attempted to set private field on non-instance");return e.set(t,n),n}},function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0}),e.SequentialHook=e.Hook=e.PreventableBaklavaEvent=e.BaklavaEvent=void 0;var r=n(0),i=function(){function t(){this.listeners=new Map}return t.prototype.addListener=function(t,e){this.listeners.set(t,e)},t.prototype.removeListener=function(t){this.listeners.has(t)&&this.listeners.delete(t)},t.prototype.emit=function(t){this.listeners.forEach((function(e){return e(t)}))},t}();e.BaklavaEvent=i;var o=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return r.__extends(e,t),e.prototype.emit=function(t){var e,n;try{for(var i=r.__values(Array.from(this.listeners.values())),o=i.next();!o.done;o=i.next())if(!1===(0,o.value)(t))return!0}catch(t){e={error:t}}finally{try{o&&!o.done&&(n=i.return)&&n.call(i)}finally{if(e)throw e.error}}return!1},e}(i);e.PreventableBaklavaEvent=o;var s=function(){function t(){this.tapMap=new Map,this.taps=[]}return t.prototype.tap=function(t,e){this.tapMap.has(t)&&this.untap(t),this.tapMap.set(t,e),this.taps.push(e)},t.prototype.untap=function(t){if(this.tapMap.has(t)){var e=this.tapMap.get(t);this.tapMap.delete(t);var n=this.taps.indexOf(e);n>=0&&this.taps.splice(n,1)}},t}();e.Hook=s;var a=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return r.__extends(e,t),e.prototype.execute=function(t){var e,n,i=t;try{for(var o=r.__values(this.taps),s=o.next();!s.done;s=o.next())i=(0,s.value)(i)}catch(t){e={error:t}}finally{try{s&&!s.done&&(n=o.return)&&n.call(o)}finally{if(e)throw e.error}}return i},e}(s);e.SequentialHook=a},function(t,e,n){"use strict";n.r(e),n.d(e,"Engine",(function(){return d})),n.d(e,"calculateOrder",(function(){return l})),n.d(e,"containsCycle",(function(){return f}));var r=n(0),i=n(1);function o(t,e,n,r){return new(n||(n=Promise))((function(i,o){function s(t){try{c(r.next(t))}catch(t){o(t)}}function a(t){try{c(r.throw(t))}catch(t){o(t)}}function c(t){var e;t.done?i(t.value):(e=t.value,e instanceof n?e:new n((function(t){t(e)}))).then(s,a)}c((r=r.apply(t,e||[])).next())}))}function s(t,e){var n,r,i,o,s={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:a(0),throw:a(1),return:a(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function a(o){return function(a){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;s;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return s.label++,{value:o[1],done:!1};case 5:s.label++,r=o[1],o=[0];continue;case 7:o=s.ops.pop(),s.trys.pop();continue;default:if(i=s.trys,!((i=i.length>0&&i[i.length-1])||6!==o[0]&&2!==o[0])){s=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){s.label=o[1];break}if(6===o[0]&&s.label<i[1]){s.label=i[1],i=o;break}if(i&&s.label<i[2]){s.label=i[2],s.ops.push(o);break}i[2]&&s.ops.pop(),s.trys.pop();continue}o=e.call(t,s)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,a])}}}Object.create,Object.create,new Error("timeout while waiting for mutex to become available"),new Error("mutex already locked");var a=new Error("request for lock canceled"),c=function(){function t(t,e){if(void 0===e&&(e=a),this._maxConcurrency=t,this._cancelError=e,this._queue=[],t<=0)throw new Error("semaphore must be initialized to a positive value");this._value=t}return t.prototype.acquire=function(){var t=this,e=this.isLocked(),n=new Promise((function(e,n){return t._queue.push({resolve:e,reject:n})}));return e||this._dispatch(),n},t.prototype.runExclusive=function(t){return o(this,void 0,void 0,(function(){var e,n,r;return s(this,(function(i){switch(i.label){case 0:return[4,this.acquire()];case 1:e=i.sent(),n=e[0],r=e[1],i.label=2;case 2:return i.trys.push([2,,4,5]),[4,t(n)];case 3:return[2,i.sent()];case 4:return r(),[7];case 5:return[2]}}))}))},t.prototype.isLocked=function(){return this._value<=0},t.prototype.release=function(){if(this._maxConcurrency>1)throw new Error("this method is unavailable on semaphores with concurrency > 1; use the scoped release returned by acquire instead");if(this._currentReleaser){var t=this._currentReleaser;this._currentReleaser=void 0,t()}},t.prototype.cancel=function(){var t=this;this._queue.forEach((function(e){return e.reject(t._cancelError)})),this._queue=[]},t.prototype._dispatch=function(){var t=this,e=this._queue.shift();if(e){var n=!1;this._currentReleaser=function(){n||(n=!0,t._value++,t._dispatch())},e.resolve([this._value--,this._currentReleaser])}},t}(),u=function(){function t(t){this._semaphore=new c(1,t)}return t.prototype.acquire=function(){return o(this,void 0,void 0,(function(){var t;return s(this,(function(e){switch(e.label){case 0:return[4,this._semaphore.acquire()];case 1:return t=e.sent(),[2,t[1]]}}))}))},t.prototype.runExclusive=function(t){return this._semaphore.runExclusive((function(){return t()}))},t.prototype.isLocked=function(){return this._semaphore.isLocked()},t.prototype.release=function(){this._semaphore.release()},t.prototype.cancel=function(){return this._semaphore.cancel()},t}();function l(t,e,n){var i=new Map;t.forEach((function(t){i.set(t,e.filter((function(e){return e.to&&e.to.parent===t})).map((function(t){return t.from.parent})))}));var o=n||t.filter((function(t){return e=t.outputInterfaces,[Object,Array].includes((e||{}).constructor)&&!Object.entries(e||{}).length;var e})),s={children:o.map((function(t){return{n:t,children:[]}}))};!function t(e,n,i){var o,s;try{for(var a=Object(r.__values)(e.children),c=a.next();!c.done;c=a.next()){var u=c.value;if(n.includes(u.n))throw new Error("Cycle detected");n.unshift(u.n),u.children=u.children.concat(i.get(u.n).map((function(t){return{n:t,children:new Array}}))),t(u,n,i),n.shift()}}catch(t){o={error:t}}finally{try{c&&!c.done&&(s=a.return)&&s.call(a)}finally{if(o)throw o.error}}}(s,[],i);var a=[],c=[];for(a.push(s);a.length>0;)a.shift().children.forEach((function(t){c.push(t.n),a.push(t)}));for(var u=[];c.length>0;){var l=c.pop();u.includes(l)||u.push(l)}return{calculationOrder:u,rootNodes:o}}function f(t,e){try{return l(t,e),!0}catch(t){return!1}}var d=function(){function t(t){void 0===t&&(t=!1),this.type="EnginePlugin",this.events={beforeCalculate:new i.PreventableBaklavaEvent,calculated:new i.BaklavaEvent},this.hooks={gatherCalculationData:new i.SequentialHook},this.nodeCalculationOrder=[],this.actualRootNodes=[],this.connectionsPerNode=new Map,this.recalculateOrder=!1,this.calculateOnChange=!1,this.calculationInProgress=!1,this.mutex=new u,this._rootNodes=void 0,this.interfaceTypePlugins=[],this.calculateOnChange=t}return Object.defineProperty(t.prototype,"rootNodes",{get:function(){return this._rootNodes},set:function(t){this._rootNodes=t,this.recalculateOrder=!0},enumerable:!1,configurable:!0}),t.prototype.register=function(t){var e=this;this.editor=t,this.editor.plugins.forEach((function(t){"InterfaceTypePlugin"===t.type&&e.interfaceTypePlugins.push(t)})),this.editor.events.usePlugin.addListener(this,(function(t){"InterfaceTypePlugin"===t.type&&e.interfaceTypePlugins.push(t)})),this.editor.events.addNode.addListener(this,(function(t){t.events.update.addListener(e,(function(t){(t.interface&&0===t.interface.connectionCount||t.option)&&e.onChange(!1)})),e.onChange(!0)})),this.editor.events.removeNode.addListener(this,(function(t){t.events.update.removeListener(e)})),this.editor.events.checkConnection.addListener(this,(function(t){if(!e.checkConnection(t.from,t.to))return!1})),this.editor.events.addConnection.addListener(this,(function(t){e.editor.connections.filter((function(e){return e.id!==t.id&&e.to===t.to})).forEach((function(t){return e.editor.removeConnection(t)})),e.onChange(!0)})),this.editor.events.removeConnection.addListener(this,(function(){e.onChange(!0)}))},t.prototype.calculate=function(t){return Object(r.__awaiter)(this,void 0,void 0,(function(){var e=this;return Object(r.__generator)(this,(function(n){switch(n.label){case 0:return[4,this.mutex.runExclusive((function(){return Object(r.__awaiter)(e,void 0,void 0,(function(){return Object(r.__generator)(this,(function(e){switch(e.label){case 0:return[4,this.internalCalculate(t)];case 1:return[2,e.sent()]}}))}))}))];case 1:return[2,n.sent()]}}))}))},t.prototype.calculateOrder=function(){this.calculateNodeTree(),this.recalculateOrder=!1},t.prototype.internalCalculate=function(t){return Object(r.__awaiter)(this,void 0,void 0,(function(){var e,n,i,o,s,a,c,u,l=this;return Object(r.__generator)(this,(function(f){switch(f.label){case 0:if(this.events.beforeCalculate.emit(t))return[2,null];t=this.hooks.gatherCalculationData.execute(t),this.calculationInProgress=!0,this.recalculateOrder&&this.calculateOrder(),e=new Map,f.label=1;case 1:f.trys.push([1,6,7,8]),n=Object(r.__values)(this.nodeCalculationOrder),i=n.next(),f.label=2;case 2:return i.done?[3,5]:[4,(o=i.value).calculate(t)];case 3:s=f.sent(),this.actualRootNodes.includes(o)&&e.set(o,s),this.connectionsPerNode.has(o)&&this.connectionsPerNode.get(o).forEach((function(t){var e=l.interfaceTypePlugins.find((function(e){return e.canConvert(t.from.type,t.to.type)}));t.to.value=e?e.convert(t.from.type,t.to.type,t.from.value):t.from.value})),f.label=4;case 4:return i=n.next(),[3,2];case 5:return[3,8];case 6:return a=f.sent(),c={error:a},[3,8];case 7:try{i&&!i.done&&(u=n.return)&&u.call(n)}finally{if(c)throw c.error}return[7];case 8:return this.calculationInProgress=!1,this.events.calculated.emit(e),[2,e]}}))}))},t.prototype.checkConnection=function(t,e){var n={from:t,to:e,id:"dc",destructed:!1,isInDanger:!1},r=this.editor.connections.concat([n]);return r.filter((function(t){return t.to!==e})),f(this.editor.nodes,r)},t.prototype.onChange=function(t){this.recalculateOrder=this.recalculateOrder||t,this.calculateOnChange&&!this.calculationInProgress&&this.calculate()},t.prototype.calculateNodeTree=function(){var t=this,e=l(this.editor.nodes,this.editor.connections,this.rootNodes),n=e.calculationOrder,r=e.rootNodes;this.nodeCalculationOrder=n,this.actualRootNodes=r,this.connectionsPerNode.clear(),this.editor.nodes.forEach((function(e){t.connectionsPerNode.set(e,t.editor.connections.filter((function(t){return t.from.parent===e})))}))},t}()}])}))},"00ee":function(t,e,n){var r=n("b622"),i=r("toStringTag"),o={};o[i]="z",t.exports="[object z]"===String(o)},"0366":function(t,e,n){var r=n("e330"),i=n("59ed"),o=r(r.bind);t.exports=function(t,e){return i(t),void 0===e?t:o?o(t,e):function(){return t.apply(e,arguments)}}},"06cf":function(t,e,n){var r=n("83ab"),i=n("c65b"),o=n("d1e7"),s=n("5c6c"),a=n("fc6a"),c=n("a04b"),u=n("1a2d"),l=n("0cfb"),f=Object.getOwnPropertyDescriptor;e.f=r?f:function(t,e){if(t=a(t),e=c(e),l)try{return f(t,e)}catch(n){}if(u(t,e))return s(!i(o.f,t,e),t[e])}},"07fa":function(t,e,n){var r=n("50c4");t.exports=function(t){return r(t.length)}},"0b42":function(t,e,n){var r=n("da84"),i=n("e8b5"),o=n("68ee"),s=n("861d"),a=n("b622"),c=a("species"),u=r.Array;t.exports=function(t){var e;return i(t)&&(e=t.constructor,o(e)&&(e===u||i(e.prototype))?e=void 0:s(e)&&(e=e[c],null===e&&(e=void 0))),void 0===e?u:e}},"0c30":function(t,e,n){(function(e,r){t.exports=r(n("2b0e"))})("undefined"!==typeof self&&self,(function(t){return function(t){var e={};function n(r){if(e[r])return e[r].exports;var i=e[r]={i:r,l:!1,exports:{}};return t[r].call(i.exports,i,i.exports,n),i.l=!0,i.exports}return n.m=t,n.c=e,n.d=function(t,e,r){n.o(t,e)||Object.defineProperty(t,e,{enumerable:!0,get:r})},n.r=function(t){"undefined"!==typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(t,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(t,"__esModule",{value:!0})},n.t=function(t,e){if(1&e&&(t=n(t)),8&e)return t;if(4&e&&"object"===typeof t&&t&&t.__esModule)return t;var r=Object.create(null);if(n.r(r),Object.defineProperty(r,"default",{enumerable:!0,value:t}),2&e&&"string"!=typeof t)for(var i in t)n.d(r,i,function(e){return t[e]}.bind(null,i));return r},n.n=function(t){var e=t&&t.__esModule?function(){return t["default"]}:function(){return t};return n.d(e,"a",e),e},n.o=function(t,e){return Object.prototype.hasOwnProperty.call(t,e)},n.p="",n(n.s=112)}([function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",(function(){return i})),n.d(e,"__assign",(function(){return o})),n.d(e,"__rest",(function(){return s})),n.d(e,"__decorate",(function(){return a})),n.d(e,"__param",(function(){return c})),n.d(e,"__metadata",(function(){return u})),n.d(e,"__awaiter",(function(){return l})),n.d(e,"__generator",(function(){return f})),n.d(e,"__createBinding",(function(){return d})),n.d(e,"__exportStar",(function(){return h})),n.d(e,"__values",(function(){return p})),n.d(e,"__read",(function(){return y})),n.d(e,"__spread",(function(){return v})),n.d(e,"__spreadArrays",(function(){return b})),n.d(e,"__await",(function(){return m})),n.d(e,"__asyncGenerator",(function(){return g})),n.d(e,"__asyncDelegator",(function(){return _})),n.d(e,"__asyncValues",(function(){return w})),n.d(e,"__makeTemplateObject",(function(){return O})),n.d(e,"__importStar",(function(){return S})),n.d(e,"__importDefault",(function(){return x})),n.d(e,"__classPrivateFieldGet",(function(){return I})),n.d(e,"__classPrivateFieldSet",(function(){return A}));
/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
var r=function(t,e){return r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])},r(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n],e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t},o.apply(this,arguments)};function s(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"===typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&Object.prototype.propertyIsEnumerable.call(t,r[i])&&(n[r[i]]=t[r[i]])}return n}function a(t,e,n,r){var i,o=arguments.length,s=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"===typeof Reflect&&"function"===typeof Reflect.decorate)s=Reflect.decorate(t,e,n,r);else for(var a=t.length-1;a>=0;a--)(i=t[a])&&(s=(o<3?i(s):o>3?i(e,n,s):i(e,n))||s);return o>3&&s&&Object.defineProperty(e,n,s),s}function c(t,e){return function(n,r){e(n,r,t)}}function u(t,e){if("object"===typeof Reflect&&"function"===typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){function i(t){return t instanceof n?t:new n((function(e){e(t)}))}return new(n||(n=Promise))((function(n,o){function s(t){try{c(r.next(t))}catch(e){o(e)}}function a(t){try{c(r["throw"](t))}catch(e){o(e)}}function c(t){t.done?n(t.value):i(t.value).then(s,a)}c((r=r.apply(t,e||[])).next())}))}function f(t,e){var n,r,i,o,s={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:a(0),throw:a(1),return:a(2)},"function"===typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function a(t){return function(e){return c([t,e])}}function c(o){if(n)throw new TypeError("Generator is already executing.");while(s)try{if(n=1,r&&(i=2&o[0]?r["return"]:o[0]?r["throw"]||((i=r["return"])&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return s.label++,{value:o[1],done:!1};case 5:s.label++,r=o[1],o=[0];continue;case 7:o=s.ops.pop(),s.trys.pop();continue;default:if(i=s.trys,!(i=i.length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){s=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){s.label=o[1];break}if(6===o[0]&&s.label<i[1]){s.label=i[1],i=o;break}if(i&&s.label<i[2]){s.label=i[2],s.ops.push(o);break}i[2]&&s.ops.pop(),s.trys.pop();continue}o=e.call(t,s)}catch(a){o=[6,a],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}}function d(t,e,n,r){void 0===r&&(r=n),t[r]=e[n]}function h(t,e){for(var n in t)"default"===n||e.hasOwnProperty(n)||(e[n]=t[n])}function p(t){var e="function"===typeof Symbol&&Symbol.iterator,n=e&&t[e],r=0;if(n)return n.call(t);if(t&&"number"===typeof t.length)return{next:function(){return t&&r>=t.length&&(t=void 0),{value:t&&t[r++],done:!t}}};throw new TypeError(e?"Object is not iterable.":"Symbol.iterator is not defined.")}function y(t,e){var n="function"===typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),s=[];try{while((void 0===e||e-- >0)&&!(r=o.next()).done)s.push(r.value)}catch(a){i={error:a}}finally{try{r&&!r.done&&(n=o["return"])&&n.call(o)}finally{if(i)throw i.error}}return s}function v(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(y(arguments[e]));return t}function b(){for(var t=0,e=0,n=arguments.length;e<n;e++)t+=arguments[e].length;var r=Array(t),i=0;for(e=0;e<n;e++)for(var o=arguments[e],s=0,a=o.length;s<a;s++,i++)r[i]=o[s];return r}function m(t){return this instanceof m?(this.v=t,this):new m(t)}function g(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},s("next"),s("throw"),s("return"),r[Symbol.asyncIterator]=function(){return this},r;function s(t){i[t]&&(r[t]=function(e){return new Promise((function(n,r){o.push([t,e,n,r])>1||a(t,e)}))})}function a(t,e){try{c(i[t](e))}catch(n){f(o[0][3],n)}}function c(t){t.value instanceof m?Promise.resolve(t.value.v).then(u,l):f(o[0][2],t)}function u(t){a("next",t)}function l(t){a("throw",t)}function f(t,e){t(e),o.shift(),o.length&&a(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",(function(t){throw t})),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:m(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function w(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t="function"===typeof p?p(t):t[Symbol.iterator](),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise((function(r,o){e=t[n](e),i(r,o,e.done,e.value)}))}}function i(t,e,n,r){Promise.resolve(r).then((function(e){t({value:e,done:n})}),e)}}function O(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function S(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function x(t){return t&&t.__esModule?t:{default:t}}function I(t,e){if(!e.has(t))throw new TypeError("attempted to get private field on non-instance");return e.get(t)}function A(t,e,n){if(!e.has(t))throw new TypeError("attempted to set private field on non-instance");return e.set(t,n),n}},function(t,e,n){"use strict";var r=n(5),i=n.n(r);n.d(e,"e",(function(){return i.a}));var o=n(2);n.d(e,"a",(function(){return o["b"]}));n(66);var s=n(67);n.d(e,"b",(function(){return s["a"]}));n(68),n(69),n(70);var a=n(71);n.d(e,"c",(function(){return a["a"]}));n(72);var c=n(73);n.d(e,"d",(function(){return c["a"]}));n(74),n(75),n(76);var u=n(77);n.d(e,"f",(function(){return u["a"]}))},function(t,e,n){"use strict";n.d(e,"a",(function(){return v}));var r=n(5),i=n.n(r);
/**
  * vue-class-component v7.2.3
  * (c) 2015-present Evan You
  * @license MIT
  */
function o(t){return o="function"===typeof Symbol&&"symbol"===typeof Symbol.iterator?function(t){return typeof t}:function(t){return t&&"function"===typeof Symbol&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},o(t)}function s(t,e,n){return e in t?Object.defineProperty(t,e,{value:n,enumerable:!0,configurable:!0,writable:!0}):t[e]=n,t}function a(t){return c(t)||u(t)||l()}function c(t){if(Array.isArray(t)){for(var e=0,n=new Array(t.length);e<t.length;e++)n[e]=t[e];return n}}function u(t){if(Symbol.iterator in Object(t)||"[object Arguments]"===Object.prototype.toString.call(t))return Array.from(t)}function l(){throw new TypeError("Invalid attempt to spread non-iterable instance")}function f(){return"undefined"!==typeof Reflect&&Reflect.defineMetadata&&Reflect.getOwnMetadataKeys}function d(t,e){h(t,e),Object.getOwnPropertyNames(e.prototype).forEach((function(n){h(t.prototype,e.prototype,n)})),Object.getOwnPropertyNames(e).forEach((function(n){h(t,e,n)}))}function h(t,e,n){var r=n?Reflect.getOwnMetadataKeys(e,n):Reflect.getOwnMetadataKeys(e);r.forEach((function(r){var i=n?Reflect.getOwnMetadata(r,e,n):Reflect.getOwnMetadata(r,e);n?Reflect.defineMetadata(r,i,t,n):Reflect.defineMetadata(r,i,t)}))}var p={__proto__:[]},y=p instanceof Array;function v(t){return function(e,n,r){var i="function"===typeof e?e:e.constructor;i.__decorators__||(i.__decorators__=[]),"number"!==typeof r&&(r=void 0),i.__decorators__.push((function(e){return t(e,n,r)}))}}function b(t){var e=o(t);return null==t||"object"!==e&&"function"!==e}function m(t,e){var n=e.prototype._init;e.prototype._init=function(){var e=this,n=Object.getOwnPropertyNames(t);if(t.$options.props)for(var r in t.$options.props)t.hasOwnProperty(r)||n.push(r);n.forEach((function(n){"_"!==n.charAt(0)&&Object.defineProperty(e,n,{get:function(){return t[n]},set:function(e){t[n]=e},configurable:!0})}))};var r=new e;e.prototype._init=n;var i={};return Object.keys(r).forEach((function(t){void 0!==r[t]&&(i[t]=r[t])})),i}var g=["data","beforeCreate","created","beforeMount","mounted","beforeDestroy","destroyed","beforeUpdate","updated","activated","deactivated","render","errorCaptured","serverPrefetch"];function _(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{};e.name=e.name||t._componentTag||t.name;var n=t.prototype;Object.getOwnPropertyNames(n).forEach((function(t){if("constructor"!==t)if(g.indexOf(t)>-1)e[t]=n[t];else{var r=Object.getOwnPropertyDescriptor(n,t);void 0!==r.value?"function"===typeof r.value?(e.methods||(e.methods={}))[t]=r.value:(e.mixins||(e.mixins=[])).push({data:function(){return s({},t,r.value)}}):(r.get||r.set)&&((e.computed||(e.computed={}))[t]={get:r.get,set:r.set})}})),(e.mixins||(e.mixins=[])).push({data:function(){return m(this,t)}});var r=t.__decorators__;r&&(r.forEach((function(t){return t(e)})),delete t.__decorators__);var o=Object.getPrototypeOf(t.prototype),a=o instanceof i.a?o.constructor:i.a,c=a.extend(e);return O(c,t,a),f()&&d(c,t),c}var w={prototype:!0,arguments:!0,callee:!0,caller:!0};function O(t,e,n){Object.getOwnPropertyNames(e).forEach((function(r){if(!w[r]){var i=Object.getOwnPropertyDescriptor(t,r);if(!i||i.configurable){var o=Object.getOwnPropertyDescriptor(e,r);if(!y){if("cid"===r)return;var s=Object.getOwnPropertyDescriptor(n,r);if(!b(o.value)&&s&&s.value===o.value)return}0,Object.defineProperty(t,r,o)}}}))}function S(t){return"function"===typeof t?_(t):function(e){return _(e,t)}}S.registerHooks=function(t){g.push.apply(g,a(t))},e["b"]=S},function(t,e,n){"use strict";function r(t,e,n,r,i,o,s,a){var c,u="function"===typeof t?t.options:t;if(e&&(u.render=e,u.staticRenderFns=n,u._compiled=!0),r&&(u.functional=!0),o&&(u._scopeId="data-v-"+o),s?(c=function(t){t=t||this.$vnode&&this.$vnode.ssrContext||this.parent&&this.parent.$vnode&&this.parent.$vnode.ssrContext,t||"undefined"===typeof __VUE_SSR_CONTEXT__||(t=__VUE_SSR_CONTEXT__),i&&i.call(this,t),t&&t._registeredComponents&&t._registeredComponents.add(s)},u._ssrRegister=c):i&&(c=a?function(){i.call(this,(u.functional?this.parent:this).$root.$options.shadowRoot)}:i),c)if(u.functional){u._injectStyles=c;var l=u.render;u.render=function(t,e){return c.call(e),l(t,e)}}else{var f=u.beforeCreate;u.beforeCreate=f?[].concat(f,c):[c]}return{exports:t,options:u}}n.d(e,"a",(function(){return r}))},function(t,e,n){"use strict";var r=n(78);n.d(e,"a",(function(){return r["a"]}));n(79),n(80),n(81),n(82),n(83),n(84),n(85)},function(e,n){e.exports=t},function(t,e,n){"use strict";n.d(e,"a",(function(){return r}));var r=[]},function(t,e,n){"use strict";function r(t){return"function"!==typeof t||!t.managed&&!t.managedReactive}function i(t){var e=function(){var n=this,r="function"===typeof t?t.call(this):t;for(var i in r=Object.create(r||null),r[o]=Object.create(this[o]||{}),e.managed)r[e.managed[i]]=this[i];var s=function(t){r[e.managedReactive[t]]=a[t],Object.defineProperty(r[o],e.managedReactive[t],{enumerable:!0,configurable:!0,get:function(){return n[t]}})},a=this;for(var i in e.managedReactive)s(i);return r};return e.managed={},e.managedReactive={},e}n.d(e,"b",(function(){return r})),n.d(e,"c",(function(){return i})),n.d(e,"d",(function(){return o})),n.d(e,"a",(function(){return s}));var o="__reactiveInject__";function s(t){Array.isArray(t.inject)||(t.inject=t.inject||{},t.inject[o]={from:o,default:{}})}},function(t,e,n){"use strict";n.d(e,"d",(function(){return r})),n.d(e,"b",(function(){return i})),n.d(e,"a",(function(){return o})),n.d(e,"c",(function(){return s}));var r=function(t){return t instanceof SVGElement&&"getBBox"in t},i=function(t){if(r(t)){var e=t.getBBox(),n=e.width,i=e.height;return!n&&!i}var o=t,s=o.offsetWidth,a=o.offsetHeight;return!(s||a||t.getClientRects().length)},o=function(t){var e,n,r=null===(n=null===(e=t)||void 0===e?void 0:e.ownerDocument)||void 0===n?void 0:n.defaultView;return!!(r&&t instanceof r.Element)},s=function(t){switch(t.tagName){case"INPUT":if("image"!==t.type)break;case"VIDEO":case"AUDIO":case"EMBED":case"OBJECT":case"CANVAS":case"IFRAME":case"IMG":return!0}return!1}},function(t,e,n){"use strict";function r(t){return document.getElementById(t.id)}function i(t){var e=document.getElementById(t.parent.id),n=document.getElementById(t.id),r=null===n||void 0===n?void 0:n.getElementsByClassName("__port");return{node:e,interface:n,port:r&&r.length>0?r[0]:null}}n.d(e,"b",(function(){return r})),n.d(e,"a",(function(){return i}))},function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));var r="undefined"!==typeof Reflect&&"undefined"!==typeof Reflect.getMetadata;function i(t,e,n){if(r&&!Array.isArray(t)&&"function"!==typeof t&&!t.hasOwnProperty("type")&&"undefined"===typeof t.type){var i=Reflect.getMetadata("design:type",e,n);i!==Object&&(t.type=i)}}},function(t,e,n){"use strict";n.d(e,"a",(function(){return r}));var r="undefined"!==typeof window?window:{}},function(t,e,n){"use strict";var r=n(54),i=n(21),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/components/connection/ConnectionView.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return y})),n.d(e,"b",(function(){return p}));var r=n(15),i=n(56),o=n(8),s=n(11),a=new WeakMap,c=/auto|scroll/,u=/^tb|vertical/,l=/msie|trident/i.test(s["a"].navigator&&s["a"].navigator.userAgent),f=function(t){return parseFloat(t||"0")},d=function(t,e,n){return void 0===t&&(t=0),void 0===e&&(e=0),void 0===n&&(n=!1),Object.freeze({inlineSize:(n?e:t)||0,blockSize:(n?t:e)||0})},h=Object.freeze({devicePixelContentBoxSize:d(),borderBoxSize:d(),contentBoxSize:d(),contentRect:new i["a"](0,0,0,0)}),p=function(t,e){if(void 0===e&&(e=!1),a.has(t)&&!e)return a.get(t);if(Object(o["b"])(t))return a.set(t,h),h;var n=getComputedStyle(t),r=Object(o["d"])(t)&&t.ownerSVGElement&&t.getBBox(),s=!l&&"border-box"===n.boxSizing,p=u.test(n.writingMode||""),y=!r&&c.test(n.overflowY||""),v=!r&&c.test(n.overflowX||""),b=r?0:f(n.paddingTop),m=r?0:f(n.paddingRight),g=r?0:f(n.paddingBottom),_=r?0:f(n.paddingLeft),w=r?0:f(n.borderTopWidth),O=r?0:f(n.borderRightWidth),S=r?0:f(n.borderBottomWidth),x=r?0:f(n.borderLeftWidth),I=_+m,A=b+g,T=x+O,j=w+S,E=v?t.offsetHeight-j-t.clientHeight:0,B=y?t.offsetWidth-T-t.clientWidth:0,C=s?I+T:0,M=s?A+j:0,k=r?r.width:f(n.width)-C-B,D=r?r.height:f(n.height)-M-E,N=k+I+B+T,L=D+A+E+j,P=Object.freeze({devicePixelContentBoxSize:d(Math.round(k*devicePixelRatio),Math.round(D*devicePixelRatio),p),borderBoxSize:d(N,L,p),contentBoxSize:d(k,D,p),contentRect:new i["a"](_,b,k,D)});return a.set(t,P),P},y=function(t,e,n){var i=p(t,n),o=i.borderBoxSize,s=i.contentBoxSize,a=i.devicePixelContentBoxSize;switch(e){case r["a"].DEVICE_PIXEL_CONTENT_BOX:return a;case r["a"].BORDER_BOX:return o;default:return s}}},function(t,e,n){"use strict";function r(t){return t.node&&t.interface&&t.port?[t.node.offsetLeft+t.interface.offsetLeft+t.port.offsetLeft+t.port.clientWidth/2,t.node.offsetTop+t.interface.offsetTop+t.port.offsetTop+t.port.clientHeight/2]:[0,0]}n.d(e,"a",(function(){return r}))},function(t,e,n){"use strict";var r;n.d(e,"a",(function(){return r})),function(t){t["BORDER_BOX"]="border-box",t["CONTENT_BOX"]="content-box",t["DEVICE_PIXEL_CONTENT_BOX"]="device-pixel-content-box"}(r||(r={}))},function(t,e,n){"use strict";n.d(e,"a",(function(){return u}));var r=n(50),i=n(108),o=n(109),s=n(6),a=new WeakMap,c=function(t,e){for(var n=0;n<t.length;n+=1)if(t[n].target===e)return n;return-1},u=function(){function t(){}return t.connect=function(t,e){var n=new o["a"](t,e);a.set(t,n)},t.observe=function(t,e,n){var o=a.get(t),u=0===o.observationTargets.length;c(o.observationTargets,e)<0&&(u&&s["a"].push(o),o.observationTargets.push(new i["a"](e,n&&n.box)),Object(r["b"])(1),r["a"].schedule())},t.unobserve=function(t,e){var n=a.get(t),i=c(n.observationTargets,e),o=1===n.observationTargets.length;i>=0&&(o&&s["a"].splice(s["a"].indexOf(n),1),n.observationTargets.splice(i,1),Object(r["b"])(-1))},t.disconnect=function(t){var e=this,n=a.get(t);n.observationTargets.slice().forEach((function(n){return e.unobserve(t,n.target)})),n.activeTargets.splice(0,n.activeTargets.length)},t}()},function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0}),e.SequentialHook=e.Hook=e.PreventableBaklavaEvent=e.BaklavaEvent=void 0;var r=n(0),i=function(){function t(){this.listeners=new Map}return t.prototype.addListener=function(t,e){this.listeners.set(t,e)},t.prototype.removeListener=function(t){this.listeners.has(t)&&this.listeners.delete(t)},t.prototype.emit=function(t){this.listeners.forEach((function(e){return e(t)}))},t}();e.BaklavaEvent=i;var o=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return r.__extends(e,t),e.prototype.emit=function(t){var e,n;try{for(var i=r.__values(Array.from(this.listeners.values())),o=i.next();!o.done;o=i.next()){var s=o.value;if(!1===s(t))return!0}}catch(a){e={error:a}}finally{try{o&&!o.done&&(n=i.return)&&n.call(i)}finally{if(e)throw e.error}}return!1},e}(i);e.PreventableBaklavaEvent=o;var s=function(){function t(){this.tapMap=new Map,this.taps=[]}return t.prototype.tap=function(t,e){this.tapMap.has(t)&&this.untap(t),this.tapMap.set(t,e),this.taps.push(e)},t.prototype.untap=function(t){if(this.tapMap.has(t)){var e=this.tapMap.get(t);this.tapMap.delete(t);var n=this.taps.indexOf(e);n>=0&&this.taps.splice(n,1)}},t}();e.Hook=s;var a=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return r.__extends(e,t),e.prototype.execute=function(t){var e,n,i=t;try{for(var o=r.__values(this.taps),s=o.next();!s.done;s=o.next()){var a=s.value;i=a(i)}}catch(c){e={error:c}}finally{try{s&&!s.done&&(n=o.return)&&n.call(o)}finally{if(e)throw e.error}}return i},e}(s);e.SequentialHook=a},function(t,e,n){"use strict";var r=n(51),i=n(19),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/components/Editor.vue",e["a"]=s.exports},function(t,e,n){"use strict";var r=n(20);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=n(4),s=n(98),a=n(99),c=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.nodeeditor=e,e.selectedNodeViews=[],e.temporaryConnection=null,e.hoveringOver=null,e.selectedNodes=[],e.ctrlPressed=!1,e.draggingStartPoint=null,e.draggingStartPanning=null,e.counter=0,e.contextMenu={items:[],show:!1,x:0,y:0},e}return Object(r["__extends"])(e,t),Object.defineProperty(e.prototype,"styles",{get:function(){return{"transform-origin":"0 0",transform:"scale("+this.plugin.scaling+") translate("+this.plugin.panning.x+"px, "+this.plugin.panning.y+"px)"}},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"backgroundStyle",{get:function(){var t=this.plugin.panning.x*this.plugin.scaling,e=this.plugin.panning.y*this.plugin.scaling,n=this.plugin.scaling*this.plugin.backgroundGrid.gridSize,r=n/this.plugin.backgroundGrid.gridDivision,i=n+"px "+n+"px, "+n+"px "+n+"px",o=this.plugin.scaling>this.plugin.backgroundGrid.subGridVisibleThreshold?", "+r+"px "+r+"px, "+r+"px "+r+"px":"";return{"background-position":"left "+t+"px top "+e+"px","background-size":i+" "+o}},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"nodes",{get:function(){return this.plugin.editor?this.plugin.editor.nodes:[]},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"connections",{get:function(){return this.plugin.editor?this.plugin.editor.connections:[]},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"hasEnginePlugin",{get:function(){var t,e;try{for(var n=Object(r["__values"])(this.plugin.editor.plugins.values()),i=n.next();!i.done;i=n.next()){var o=i.value;if("EnginePlugin"===o.type)return!0}}catch(s){t={error:s}}finally{try{i&&!i.done&&(e=n.return)&&e.call(n)}finally{if(t)throw t.error}}return!1},enumerable:!1,configurable:!0}),e.prototype.mounted=function(){var t=this;this.updateContextMenu(),this.plugin.editor.events.registerNodeType.addListener(this,(function(){return t.updateContextMenu()})),this.plugin.editor.hooks.load.tap(this,(function(e){return t.counter++,e})),this.clipboard=new s["a"](this.plugin.editor),this.history=new a["a"](this.plugin)},e.prototype.updateContextMenu=function(){var t=this,e=Array.from(this.plugin.editor.nodeCategories.keys()).filter((function(t){return"default"!==t})).map((function(e){var n=Array.from(t.plugin.editor.nodeCategories.get(e)).map((function(e){return{value:"addNode:"+e,label:t.plugin.nodeTypeAliases[e]||e}}));return{label:e,submenu:n}})),n=this.plugin.editor.nodeCategories.get("default").map((function(e){return{value:"addNode:"+e,label:t.plugin.nodeTypeAliases[e]||e}})),i=Object(r["__spread"])(e);e.length>0&&n.length>0&&i.push({isDivider:!0}),i.push.apply(i,Object(r["__spread"])(n)),this.contextMenu.items=[{label:"Add Block",submenu:i},{label:"Copy Blocks",value:"copy",disabledFunction:function(){return 0===t.selectedNodes.length}},{label:"Paste Blocks",value:"paste",disabledFunction:function(){return t.clipboard.isEmpty}}]},e.prototype.hoveredOver=function(t){this.hoveringOver=t,t&&this.temporaryConnection?(this.temporaryConnection.to=t,this.temporaryConnection.status=this.plugin.editor.checkConnection(this.temporaryConnection.from,this.temporaryConnection.to)?o["a"].ALLOWED:o["a"].FORBIDDEN,this.hasEnginePlugin&&this.connections.filter((function(e){return e.to===t})).forEach((function(t){t.isInDanger=!0}))):!t&&this.temporaryConnection&&(this.$set(this.temporaryConnection,"to",void 0),this.temporaryConnection.status=o["a"].NONE,this.connections.forEach((function(t){t.isInDanger=!1})))},e.prototype.mouseMoveHandler=function(t){if(this.temporaryConnection)this.temporaryConnection.mx=t.offsetX/this.plugin.scaling-this.plugin.panning.x,this.temporaryConnection.my=t.offsetY/this.plugin.scaling-this.plugin.panning.y;else if(this.draggingStartPoint){var e=t.screenX-this.draggingStartPoint.x,n=t.screenY-this.draggingStartPoint.y;this.plugin.panning.x=this.draggingStartPanning.x+e/this.plugin.scaling,this.plugin.panning.y=this.draggingStartPanning.y+n/this.plugin.scaling}},e.prototype.mouseDown=function(t){var e=this;if(0===t.button)if(this.hoveringOver){var n=this.connections.find((function(t){return t.to===e.hoveringOver}));this.hoveringOver.isInput&&n?(this.temporaryConnection={status:o["a"].NONE,from:n.from},this.plugin.editor.removeConnection(n)):this.temporaryConnection={status:o["a"].NONE,from:this.hoveringOver},this.$set(this.temporaryConnection,"mx",null),this.$set(this.temporaryConnection,"my",null)}else t.target===this.$el&&(this.unselectAllNodes(),this.draggingStartPoint={x:t.screenX,y:t.screenY},this.draggingStartPanning={x:this.plugin.panning.x,y:this.plugin.panning.y})},e.prototype.mouseUp=function(){this.draggingStartPoint=null,this.draggingStartPanning=null;var t=this.temporaryConnection;t&&this.hoveringOver&&this.plugin.editor.addConnection(t.from,t.to),this.temporaryConnection=null},e.prototype.mouseWheel=function(t){t.preventDefault();var e=t.deltaY;1===t.deltaMode&&(e*=32);var n=this.plugin.scaling*(1-e/3e3),r=[t.offsetX/this.plugin.scaling-this.plugin.panning.x,t.offsetY/this.plugin.scaling-this.plugin.panning.y],i=[t.offsetX/n-this.plugin.panning.x,t.offsetY/n-this.plugin.panning.y],o=[i[0]-r[0],i[1]-r[1]];this.plugin.panning.x+=o[0],this.plugin.panning.y+=o[1],this.plugin.scaling=n},e.prototype.keyDown=function(t){var e=this;"Delete"===t.key&&this.selectedNodes.length>0?this.selectedNodes.forEach((function(t){return e.plugin.editor.removeNode(t)})):"Tab"===t.key?t.preventDefault():"Control"===t.key?this.ctrlPressed=!0:"z"===t.key&&t.ctrlKey?this.history.undo():"y"===t.key&&t.ctrlKey&&this.history.redo()},e.prototype.keyUp=function(t){"Control"===t.key&&(this.ctrlPressed=!1)},e.prototype.selectNode=function(t,e){this.ctrlPressed||this.unselectAllNodes(),this.selectedNodes.push(t),this.selectedNodeViews.push(e)},e.prototype.unselectAllNodes=function(){this.selectedNodes.splice(0,this.selectedNodes.length),this.selectedNodeViews.splice(0,this.selectedNodeViews.length)},e.prototype.openContextMenu=function(t){this.contextMenu.show=!0,this.contextMenu.x=t.offsetX,this.contextMenu.y=t.offsetY},e.prototype.onContextMenuClick=function(t){if(t.startsWith("addNode:")){var e=t.substring(t.indexOf(":")+1),n=this.plugin.editor.nodeTypes.get(e);if(n){var r=this.plugin.editor.addNode(new n);r&&(r.position.x=this.contextMenu.x/this.plugin.scaling-this.plugin.panning.x,r.position.y=this.contextMenu.y/this.plugin.scaling-this.plugin.panning.y)}}else"copy"===t&&this.selectedNodes.length>0?this.clipboard.copy(this.selectedNodes):"paste"!==t||this.clipboard.isEmpty||this.clipboard.paste()},Object(r["__decorate"])([Object(i["c"])({type:Object,required:!0}),Object(i["d"])("plugin")],e.prototype,"plugin",void 0),Object(r["__decorate"])([Object(i["d"])("editor")],e.prototype,"nodeeditor",void 0),Object(r["__decorate"])([Object(i["d"])("selectedNodeViews")],e.prototype,"selectedNodeViews",void 0),Object(r["__decorate"])([Object(i["f"])("plugin.nodeTypeAliases")],e.prototype,"updateContextMenu",null),e=Object(r["__decorate"])([i["a"]],e),e}(i["e"]);e["a"]=c},function(t,e,n){"use strict";var r=n(22);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=n(4),s=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r["__extends"])(e,t),e.prototype.mounted=function(){this.plugin.hooks.renderConnection.execute(this)},e.prototype.updated=function(){this.plugin.hooks.renderConnection.execute(this)},Object.defineProperty(e.prototype,"d",{get:function(){var t=Object(r["__read"])(this.transform(this.x1,this.y1),2),e=t[0],n=t[1],i=Object(r["__read"])(this.transform(this.x2,this.y2),2),o=i[0],s=i[1];if(this.plugin.useStraightConnections)return"M "+e+" "+n+" L "+o+" "+s;var a=.3*Math.abs(e-o);return"M "+e+" "+n+" C "+(e+a)+" "+n+", "+(o-a)+" "+s+", "+o+" "+s},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"classes",{get:function(){return{connection:!0,"--temporary":this.isTemporary,"--allowed":this.state===o["a"].ALLOWED,"--forbidden":this.state===o["a"].FORBIDDEN}},enumerable:!1,configurable:!0}),e.prototype.transform=function(t,e){var n=(t+this.plugin.panning.x)*this.plugin.scaling,r=(e+this.plugin.panning.y)*this.plugin.scaling;return[n,r]},Object(r["__decorate"])([Object(i["c"])({type:Number})],e.prototype,"x1",void 0),Object(r["__decorate"])([Object(i["c"])({type:Number})],e.prototype,"y1",void 0),Object(r["__decorate"])([Object(i["c"])({type:Number})],e.prototype,"x2",void 0),Object(r["__decorate"])([Object(i["c"])({type:Number})],e.prototype,"y2",void 0),Object(r["__decorate"])([Object(i["c"])({type:Number,default:o["a"].NONE})],e.prototype,"state",void 0),Object(r["__decorate"])([Object(i["c"])({type:Boolean,default:!1})],e.prototype,"isTemporary",void 0),Object(r["__decorate"])([Object(i["c"])({type:Object})],e.prototype,"connection",void 0),Object(r["__decorate"])([Object(i["b"])("plugin")],e.prototype,"plugin",void 0),e=Object(r["__decorate"])([i["a"]],e),e}(i["e"]);e["a"]=s},function(t,e,n){"use strict";var r=n(24);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=n(100),s=n(12),a=n(9),c=n(4),u=window.ResizeObserver||o["a"],l=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.d={x1:0,y1:0,x2:0,y2:0},e}return Object(r["__extends"])(e,t),Object.defineProperty(e.prototype,"state",{get:function(){return this.connection.isInDanger?c["a"].FORBIDDEN:c["a"].NONE},enumerable:!1,configurable:!0}),e.prototype.mounted=function(){return Object(r["__awaiter"])(this,void 0,void 0,(function(){return Object(r["__generator"])(this,(function(t){switch(t.label){case 0:return[4,this.$nextTick()];case 1:return t.sent(),this.updateCoords(),[2]}}))}))},e.prototype.beforeDestroy=function(){this.resizeObserver.disconnect()},e.prototype.updateCoords=function(){var t=this,e=Object(a["a"])(this.connection.from),n=Object(a["a"])(this.connection.to);e.node&&n.node&&(this.resizeObserver||(this.resizeObserver=new u((function(){t.updateCoords()})),this.resizeObserver.observe(e.node),this.resizeObserver.observe(n.node)));var i=Object(r["__read"])(this.getPortCoordinates(e),2),o=i[0],s=i[1],c=Object(r["__read"])(this.getPortCoordinates(n),2),l=c[0],f=c[1];this.d={x1:o,y1:s,x2:l,y2:f}},e.prototype.getPortCoordinates=function(t){return t.node&&t.interface&&t.port?[t.node.offsetLeft+t.interface.offsetLeft+t.port.offsetLeft+t.port.clientWidth/2,t.node.offsetTop+t.interface.offsetTop+t.port.offsetTop+t.port.clientHeight/2]:[0,0]},Object(r["__decorate"])([Object(i["c"])({type:Object})],e.prototype,"connection",void 0),Object(r["__decorate"])([Object(i["f"])("connection.from.parent.position",{deep:!0}),Object(i["f"])("connection.to.parent.position",{deep:!0})],e.prototype,"updateCoords",null),e=Object(r["__decorate"])([Object(i["a"])({components:{"connection-view":s["a"]}})],e),e}(i["e"]);e["a"]=l},function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));var r=n(13),i=function(){function t(t){var e=Object(r["b"])(t);this.target=t,this.contentRect=e.contentRect,this.borderBoxSize=[e.borderBoxSize],this.contentBoxSize=[e.contentBoxSize],this.devicePixelContentBoxSize=[e.devicePixelContentBoxSize]}return t}()},function(t,e,n){"use strict";var r=n(27);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=n(12),s=n(4),a=n(9),c=n(14),u=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r["__extends"])(e,t),Object.defineProperty(e.prototype,"status",{get:function(){return this.connection?this.connection.status:s["a"].NONE},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"d",{get:function(){if(!this.connection)return{input:[0,0],output:[0,0]};var t=Object(c["a"])(Object(a["a"])(this.connection.from)),e=this.connection.to?Object(c["a"])(Object(a["a"])(this.connection.to)):[this.connection.mx||t[0],this.connection.my||t[1]];return this.connection.from.isInput?{input:e,output:t}:{input:t,output:e}},enumerable:!1,configurable:!0}),Object(r["__decorate"])([Object(i["c"])({type:Object})],e.prototype,"connection",void 0),e=Object(r["__decorate"])([Object(i["a"])({components:{"connection-view":o["a"]}})],e),e}(i["e"]);e["a"]=u},function(t,e,n){"use strict";var r=n(29);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=n(44),s=n.n(o),a=n(110),c=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.draggingStartPosition=null,e.draggingStartPoint=null,e.renaming=!1,e.tempName="",e.contextMenu={show:!1,x:0,y:0,items:[{value:"rename",label:"Rename"},{value:"delete",label:"Delete"}]},e}return Object(r["__extends"])(e,t),Object.defineProperty(e.prototype,"classes",{get:function(){var t;return t={node:!0,"--selected":this.selected,"--dragging":!!this.draggingStartPoint,"--two-column":!!this.data.twoColumn},t["--type-"+Object(a["a"])(this.data.type)]=!0,t[this.data.customClasses]=!0,t},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"styles",{get:function(){return{top:this.data.position.y+"px",left:this.data.position.x+"px",width:this.data.width+"px"}},enumerable:!1,configurable:!0}),e.prototype.mounted=function(){var t=this;this.data.events.addInterface.addListener(this,(function(){return t.update()})),this.data.events.removeInterface.addListener(this,(function(){return t.update()})),this.data.events.addOption.addListener(this,(function(){return t.update()})),this.data.events.removeOption.addListener(this,(function(){return t.update()})),this.plugin.hooks.renderNode.execute(this)},e.prototype.updated=function(){this.plugin.hooks.renderNode.execute(this)},e.prototype.beforeDestroy=function(){this.data.events.addInterface.removeListener(this),this.data.events.removeInterface.removeListener(this),this.data.events.addOption.removeListener(this),this.data.events.removeOption.removeListener(this)},e.prototype.update=function(){this.$forceUpdate()},e.prototype.startDrag=function(t){this.select(),0!==this.selectedNodeViews.length&&void 0!==this.selectedNodeViews[0]||(this.selectedNodeViews.splice(0,this.selectedNodeViews.length),this.selectedNodeViews.push(this)),this.selectedNodeViews.forEach((function(e){e.draggingStartPoint={x:t.screenX,y:t.screenY},e.draggingStartPosition={x:e.data.position.x,y:e.data.position.y},document.addEventListener("mousemove",e.handleMove),document.addEventListener("mouseup",e.stopDrag)}))},e.prototype.select=function(){this.$emit("select",this)},e.prototype.stopDrag=function(){this.selectedNodeViews.forEach((function(t){t.draggingStartPoint=null,t.draggingStartPosition=null,document.removeEventListener("mousemove",t.handleMove),document.removeEventListener("mouseup",t.stopDrag)}))},e.prototype.handleMove=function(t){this.selectedNodeViews.forEach((function(e){if(e.draggingStartPoint){var n=t.screenX-e.draggingStartPoint.x,r=t.screenY-e.draggingStartPoint.y;e.data.position.x=e.draggingStartPosition.x+n/e.plugin.scaling,e.data.position.y=e.draggingStartPosition.y+r/e.plugin.scaling}}))},e.prototype.openContextMenu=function(t){this.contextMenu.show=!0,this.contextMenu.x=t.offsetX,this.contextMenu.y=t.offsetY},e.prototype.onContextMenu=function(t){switch(t){case"delete":this.plugin.editor.removeNode(this.data);break;case"rename":this.tempName=this.data.name,this.renaming=!0}},e.prototype.doneRenaming=function(){this.data.name=this.tempName,this.renaming=!1},e.prototype.openSidebar=function(t){this.plugin.sidebar.nodeId=this.data.id,this.plugin.sidebar.optionName=t,this.plugin.sidebar.visible=!0},Object(r["__decorate"])([Object(i["c"])({type:Object})],e.prototype,"data",void 0),Object(r["__decorate"])([Object(i["c"])({type:Boolean,default:!1})],e.prototype,"selected",void 0),Object(r["__decorate"])([Object(i["b"])("plugin")],e.prototype,"plugin",void 0),Object(r["__decorate"])([Object(i["b"])("selectedNodeViews")],e.prototype,"selectedNodeViews",void 0),e=Object(r["__decorate"])([Object(i["a"])({directives:{ClickOutside:s.a.directive}})],e),e}(i["e"]);e["a"]=c},function(t,e,n){"use strict";var r=n(31);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.value=null,e.isConnected=!1,e}return Object(r["__extends"])(e,t),Object.defineProperty(e.prototype,"classes",{get:function(){return{"node-interface":!0,"--input":this.data.isInput,"--output":!this.data.isInput,"--connected":this.isConnected}},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"displayName",{get:function(){return this.data.displayName||this.name},enumerable:!1,configurable:!0}),e.prototype.beforeMount=function(){var t=this;this.value=this.data.value,this.data.events.setValue.addListener(this,(function(e){t.value=e})),this.data.events.setConnectionCount.addListener(this,(function(e){t.$forceUpdate(),t.isConnected=e>0})),this.data.events.updated.addListener(this,(function(e){t.$forceUpdate()})),this.isConnected=this.data.connectionCount>0},e.prototype.mounted=function(){this.plugin.hooks.renderInterface.execute(this)},e.prototype.updated=function(){this.plugin.hooks.renderInterface.execute(this)},e.prototype.beforeDestroy=function(){this.data.events.setValue.removeListener(this),this.data.events.setConnectionCount.removeListener(this),this.data.events.updated.removeListener(this)},e.prototype.startHover=function(){this.editor.hoveredOver(this.data)},e.prototype.endHover=function(){this.editor.hoveredOver(void 0)},e.prototype.getOptionComponent=function(t){if(t&&this.plugin.options)return this.plugin.options[t]},Object(r["__decorate"])([Object(i["c"])({type:Object,default:function(){return{}}})],e.prototype,"data",void 0),Object(r["__decorate"])([Object(i["c"])({type:String,default:""})],e.prototype,"name",void 0),Object(r["__decorate"])([Object(i["b"])("plugin")],e.prototype,"plugin",void 0),Object(r["__decorate"])([Object(i["b"])("editor")],e.prototype,"editor",void 0),e=Object(r["__decorate"])([i["a"]],e),e}(i["e"]);e["a"]=o},function(t,e,n){"use strict";var r=n(33);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.value=null,e}return Object(r["__extends"])(e,t),Object.defineProperty(e.prototype,"component",{get:function(){if(this.plugin.options&&this.componentName)return this.plugin.options[this.componentName]},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"displayName",{get:function(){return this.option.displayName||this.name},enumerable:!1,configurable:!0}),e.prototype.beforeMount=function(){var t=this;this.value=this.option.value,this.option.events.setValue.addListener(this,(function(e){t.value=e}))},e.prototype.mounted=function(){this.plugin.hooks.renderOption.execute(this)},e.prototype.updated=function(){this.plugin.hooks.renderOption.execute(this)},e.prototype.beforeDestroy=function(){this.option.events.setValue.removeListener(this)},e.prototype.updateValue=function(t){this.option.value=t},Object(r["__decorate"])([Object(i["c"])()],e.prototype,"name",void 0),Object(r["__decorate"])([Object(i["c"])()],e.prototype,"option",void 0),Object(r["__decorate"])([Object(i["c"])()],e.prototype,"componentName",void 0),Object(r["__decorate"])([Object(i["c"])()],e.prototype,"node",void 0),Object(r["__decorate"])([Object(i["b"])("plugin")],e.prototype,"plugin",void 0),e=Object(r["__decorate"])([i["a"]],e),e}(i["e"]);e["a"]=o},function(t,e,n){"use strict";var r=n(35);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=n(44),s=n.n(o),a=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.activeMenu=-1,e.activeMenuResetTimeout=null,e.height=0,e.rootIsFlipped={x:!1,y:!1},e}var n;return Object(r["__extends"])(e,t),n=e,Object.defineProperty(e.prototype,"styles",{get:function(){var t={};return this.isNested||(t.top=(this.flippedY?this.y-this.height:this.y)+"px",t.left=this.x+"px"),t},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"classes",{get:function(){return{"dark-context-menu":!0,"--flipped-x":this.flippedX,"--flipped-y":this.flippedY,"--nested":this.isNested}},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"_items",{get:function(){return this.items.map((function(t){return Object(r["__assign"])(Object(r["__assign"])({},t),{hover:!1})}))},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"flippedX",{get:function(){return this.flippable&&(this.rootIsFlipped.x||this.isFlipped.x)},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"flippedY",{get:function(){return this.flippable&&(this.rootIsFlipped.y||this.isFlipped.y)},enumerable:!1,configurable:!0}),e.prototype.onClick=function(t){!t.submenu&&t.value&&(this.$emit("click",t.value),this.$emit("input",!1))},e.prototype.onChildClick=function(t){this.$emit("click",t),this.activeMenu=-1,this.isNested||this.$emit("input",!1)},e.prototype.onClickOutside=function(t){this.value&&this.$emit("input",!1)},e.prototype.onMouseEnter=function(t,e){this.items[e].submenu&&(this.activeMenu=e,null!==this.activeMenuResetTimeout&&(clearTimeout(this.activeMenuResetTimeout),this.activeMenuResetTimeout=null))},e.prototype.onMouseLeave=function(t,e){var n=this;this.items[e].submenu&&(this.activeMenuResetTimeout=window.setTimeout((function(){n.activeMenu=-1,n.activeMenuResetTimeout=null}),200))},e.prototype.created=function(){this.$options.components?this.$options.components["context-menu"]=i["e"].extend(n):this.$options.components={"context-menu":i["e"].extend(n)}},e.prototype.updateFlipped=function(){this.height=30*this.items.length;var t=this.$parent.$el.offsetWidth,e=this.$parent.$el.offsetHeight;this.rootIsFlipped.x=!this.isNested&&this.x>.75*t,this.rootIsFlipped.y=!this.isNested&&this.y+this.height>e-20},e.prototype.updateDisabledValues=function(){var t=this;this.value&&this.items.forEach((function(e){e.disabledFunction&&t.$set(e,"disabled",e.disabledFunction())}))},Object(r["__decorate"])([Object(i["c"])({type:Boolean,default:!1})],e.prototype,"value",void 0),Object(r["__decorate"])([Object(i["c"])({type:Array,default:function(){return[]}})],e.prototype,"items",void 0),Object(r["__decorate"])([Object(i["c"])({type:Number,default:0})],e.prototype,"x",void 0),Object(r["__decorate"])([Object(i["c"])({type:Number,default:0})],e.prototype,"y",void 0),Object(r["__decorate"])([Object(i["c"])({type:Boolean,default:!1})],e.prototype,"isNested",void 0),Object(r["__decorate"])([Object(i["c"])({type:Object,default:function(){return{x:!1,y:!1}}})],e.prototype,"isFlipped",void 0),Object(r["__decorate"])([Object(i["c"])({type:Boolean,default:!1})],e.prototype,"flippable",void 0),Object(r["__decorate"])([Object(i["f"])("y"),Object(i["f"])("items")],e.prototype,"updateFlipped",null),Object(r["__decorate"])([Object(i["f"])("value",{immediate:!0})],e.prototype,"updateDisabledValues",null),e=n=Object(r["__decorate"])([Object(i["a"])({directives:{ClickOutside:s.a.directive}})],e),e}(i["e"]);e["a"]=a},function(t,e,n){"use strict";var r=n(37);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.width=300,e}return Object(r["__extends"])(e,t),Object.defineProperty(e.prototype,"nodeName",{get:function(){var t=this.plugin.sidebar.nodeId,e=this.plugin.editor.nodes.find((function(e){return e.id===t}));return e?e.name:""},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"styles",{get:function(){return{width:this.width+"px"}},enumerable:!1,configurable:!0}),e.prototype.close=function(){this.plugin.sidebar.visible=!1},e.prototype.startResize=function(){var t=this;window.addEventListener("mousemove",this.onMouseMove),window.addEventListener("mouseup",(function(){window.removeEventListener("mousemove",t.onMouseMove)}),{once:!0})},e.prototype.onMouseMove=function(t){var e=this.$parent.$el.getBoundingClientRect().width;this.width-=t.movementX,this.width<300?this.width=300:this.width>.9*e&&(this.width=.9*e)},Object(r["__decorate"])([Object(i["b"])("plugin")],e.prototype,"plugin",void 0),e=Object(r["__decorate"])([i["a"]],e),e}(i["e"]);e["a"]=o},function(t,e,n){"use strict";var r=n(39);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=n(9),s=n(14),a=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.intervalHandle=0,e.showViewBounds=!1,e.dragging=!1,e.bounds={x1:0,y1:0,x2:0,y2:0},e}return Object(r["__extends"])(e,t),e.prototype.mounted=function(){var t,e=this,n=this.$refs.cv;this.ctx=null!==(t=n.getContext("2d"))&&void 0!==t?t:void 0,this.ctx&&(this.ctx.imageSmoothingQuality="high"),this.intervalHandle=setInterval((function(){return e.updateCanvas()}),250)},e.prototype.beforeDestroy=function(){clearInterval(this.intervalHandle)},e.prototype.updateCanvas=function(){var t,e,n,i,a,c,u,l,f,d;if(this.ctx){var h=new Map,p=new Map;try{for(var y=Object(r["__values"])(this.nodes),v=y.next();!v.done;v=y.next()){var b=v.value,m=Object(o["b"])(b),g=null!==(f=null===m||void 0===m?void 0:m.clientWidth)&&void 0!==f?f:0,_=null!==(d=null===m||void 0===m?void 0:m.clientHeight)&&void 0!==d?d:0;h.set(b,{x1:b.position.x,y1:b.position.y,x2:b.position.x+g,y2:b.position.y+_}),p.set(b,m)}}catch(X){t={error:X}}finally{try{v&&!v.done&&(e=y.return)&&e.call(y)}finally{if(t)throw t.error}}var w={x1:Number.MAX_SAFE_INTEGER,y1:Number.MAX_SAFE_INTEGER,x2:Number.MIN_SAFE_INTEGER,y2:Number.MIN_SAFE_INTEGER};try{for(var O=Object(r["__values"])(h.values()),S=O.next();!S.done;S=O.next()){var x=S.value;x.x1<w.x1&&(w.x1=x.x1),x.y1<w.y1&&(w.y1=x.y1),x.x2>w.x2&&(w.x2=x.x2),x.y2>w.y2&&(w.y2=x.y2)}}catch(J){n={error:J}}finally{try{S&&!S.done&&(i=O.return)&&i.call(O)}finally{if(n)throw n.error}}var I=50;w.x1-=I,w.y1-=I,w.x2+=I,w.y2+=I,this.bounds=w,this.ctx.clearRect(0,0,this.ctx.canvas.width,this.ctx.canvas.height),this.ctx.strokeStyle="white";try{for(var A=Object(r["__values"])(this.connections),T=A.next();!T.done;T=A.next()){var j=T.value,E=(Object(o["a"])(j.to),Object(r["__read"])(Object(s["a"])(Object(o["a"])(j.from)),2)),B=E[0],C=E[1],M=Object(r["__read"])(Object(s["a"])(Object(o["a"])(j.to)),2),k=M[0],D=M[1],N=Object(r["__read"])(this.transformCoordinates(B,C),2),L=N[0],P=N[1],F=Object(r["__read"])(this.transformCoordinates(k,D),2),U=F[0],R=F[1];if(this.ctx.beginPath(),this.ctx.moveTo(L,P),this.plugin.useStraightConnections)this.ctx.lineTo(U,R);else{var V=.3*Math.abs(L-U);this.ctx.bezierCurveTo(L+V,P,U-V,R,U,R)}this.ctx.stroke()}}catch(Z){a={error:Z}}finally{try{T&&!T.done&&(c=A.return)&&c.call(A)}finally{if(a)throw a.error}}this.ctx.strokeStyle="lightgray";try{for(var $=Object(r["__values"])(h.entries()),z=$.next();!z.done;z=$.next()){var Y=Object(r["__read"])(z.value,2),H=(b=Y[0],x=Y[1],Object(r["__read"])(this.transformCoordinates(x.x1,x.y1),2)),W=(L=H[0],P=H[1],Object(r["__read"])(this.transformCoordinates(x.x2,x.y2),2));U=W[0],R=W[1];this.ctx.fillStyle=this.getNodeColor(p.get(b)),this.ctx.beginPath(),this.ctx.rect(L,P,U-L,R-P),this.ctx.fill(),this.ctx.stroke()}}catch(Q){u={error:Q}}finally{try{z&&!z.done&&(l=$.return)&&l.call($)}finally{if(u)throw u.error}}if(this.showViewBounds){var G=this.getViewBounds(),q=Object(r["__read"])(this.transformCoordinates(G.x1,G.y1),2),K=(L=q[0],P=q[1],Object(r["__read"])(this.transformCoordinates(G.x2,G.y2),2));U=K[0],R=K[1];this.ctx.fillStyle="rgba(255, 255, 255, 0.2)",this.ctx.fillRect(L,P,U-L,R-P)}}},e.prototype.transformCoordinates=function(t,e){return[(t-this.bounds.x1)/(this.bounds.x2-this.bounds.x1)*this.ctx.canvas.clientWidth,(e-this.bounds.y1)/(this.bounds.y2-this.bounds.y1)*this.ctx.canvas.clientHeight]},e.prototype.reverseTransform=function(t,e){return[t*(this.bounds.x2-this.bounds.x1)/this.ctx.canvas.clientWidth+this.bounds.x1,e*(this.bounds.y2-this.bounds.y1)/this.ctx.canvas.clientHeight+this.bounds.y1]},e.prototype.getNodeColor=function(t){if(t){var e=t.querySelector(".__content");if(e){var n=this.getComputedColor(e);if(n)return n}var r=this.getComputedColor(t);if(r)return r}return"gray"},e.prototype.getComputedColor=function(t){var e=getComputedStyle(t).backgroundColor;if(e&&"rgba(0, 0, 0, 0)"!==e)return e},e.prototype.getViewBounds=function(){var t=this.$parent.$el.offsetWidth,e=this.$parent.$el.offsetHeight,n=t/this.plugin.scaling-this.plugin.panning.x,r=e/this.plugin.scaling-this.plugin.panning.y;return{x1:-this.plugin.panning.x,y1:-this.plugin.panning.y,x2:n,y2:r}},e.prototype.mousedown=function(t){0===t.button&&(this.dragging=!0,this.mousemove(t))},e.prototype.mousemove=function(t){if(this.dragging){var e=Object(r["__read"])(this.reverseTransform(t.offsetX,t.offsetY),2),n=e[0],i=e[1],o=this.getViewBounds(),s=(o.x1-o.x2)/2,a=(o.y1-o.y2)/2;this.plugin.panning.x=-(n+s),this.plugin.panning.y=-(i+a)}},e.prototype.mouseup=function(t){this.dragging=!1},Object(r["__decorate"])([Object(i["c"])()],e.prototype,"nodes",void 0),Object(r["__decorate"])([Object(i["c"])()],e.prototype,"connections",void 0),Object(r["__decorate"])([Object(i["b"])("plugin")],e.prototype,"plugin",void 0),Object(r["__decorate"])([Object(i["f"])("showViewBounds"),Object(i["f"])("plugin.panning.x"),Object(i["f"])("plugin.panning.y"),Object(i["f"])("plugin.scaling")],e.prototype,"updateCanvas",null),e=Object(r["__decorate"])([i["a"]],e),e}(i["e"]);e["a"]=a},function(t,e,n){"use strict";var r=n(55),i=n(23),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/components/connection/ConnectionWrapper.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));var r=n(8),i=function(t){if(Object(r["b"])(t))return 1/0;var e=0,n=t.parentNode;while(n)e+=1,n=n.parentNode;return e}},function(t,e,n){"use strict";var r=n(58),i=n(26),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/components/connection/TemporaryConnection.vue",e["a"]=s.exports},function(t,e,n){"use strict";var r=n(59),i=n(28),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/components/node/Node.vue",e["a"]=s.exports},function(t,e,n){!function(e,n){t.exports=n()}(0,(function(){var t="undefined"!=typeof window,e="undefined"!=typeof navigator,n=t&&("ontouchstart"in window||e&&navigator.msMaxTouchPoints>0)?["touchstart"]:["click"];function r(t){var e=t.event,n=t.handler;(0,t.middleware)(e)&&n(e)}function i(t,e){var i=function(t){var e="function"==typeof t;if(!e&&"object"!=typeof t)throw new Error("v-click-outside: Binding value must be a function or an object");return{handler:e?t:t.handler,middleware:t.middleware||function(t){return t},events:t.events||n,isActive:!(!1===t.isActive),detectIframe:!(!1===t.detectIframe)}}(e.value),o=i.handler,s=i.middleware,a=i.detectIframe;if(i.isActive){if(t["__v-click-outside"]=i.events.map((function(e){return{event:e,srcTarget:document.documentElement,handler:function(e){return function(t){var e=t.el,n=t.event,i=t.handler,o=t.middleware,s=n.path||n.composedPath&&n.composedPath();(s?s.indexOf(e)<0:!e.contains(n.target))&&r({event:n,handler:i,middleware:o})}({el:t,event:e,handler:o,middleware:s})}}})),a){var c={event:"blur",srcTarget:window,handler:function(e){return function(t){var e=t.el,n=t.event,i=t.handler,o=t.middleware;setTimeout((function(){var t=document.activeElement;t&&"IFRAME"===t.tagName&&!e.contains(t)&&r({event:n,handler:i,middleware:o})}),0)}({el:t,event:e,handler:o,middleware:s})}};t["__v-click-outside"]=[].concat(t["__v-click-outside"],[c])}t["__v-click-outside"].forEach((function(e){var n=e.event,r=e.srcTarget,i=e.handler;return setTimeout((function(){t["__v-click-outside"]&&r.addEventListener(n,i,!1)}),0)}))}}function o(t){(t["__v-click-outside"]||[]).forEach((function(t){return t.srcTarget.removeEventListener(t.event,t.handler,!1)})),delete t["__v-click-outside"]}var s=t?{bind:i,update:function(t,e){var n=e.value,r=e.oldValue;JSON.stringify(n)!==JSON.stringify(r)&&(o(t),i(t,{value:n}))},unbind:o}:{};return{install:function(t){t.directive("click-outside",s)},directive:s}}))},function(t,e,n){"use strict";var r=n(60),i=n(30),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/components/node/NodeInterface.vue",e["a"]=s.exports},function(t,e,n){"use strict";var r=n(61),i=n(32),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/components/node/NodeOption.vue",e["a"]=s.exports},function(t,e,n){"use strict";var r=n(62),i=n(34),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/components/ContextMenu.vue",e["a"]=s.exports},function(t,e,n){"use strict";var r=n(63),i=n(36),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/components/Sidebar.vue",e["a"]=s.exports},function(t,e,n){"use strict";var r=n(64),i=n(38),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/components/Minimap.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return p})),n.d(e,"b",(function(){return y}));var r=n(101),i=n(11),o=n(106),s=0,a=function(){return!!s},c=250,u={attributes:!0,characterData:!0,childList:!0,subtree:!0},l=["resize","load","transitionend","animationend","animationstart","animationiteration","keyup","keydown","mouseup","mousedown","mouseover","mouseout","blur","focus"],f=function(t){return void 0===t&&(t=0),Date.now()+t},d=!1,h=function(){function t(){var t=this;this.stopped=!0,this.listener=function(){return t.schedule()}}return t.prototype.run=function(t){var e=this;if(void 0===t&&(t=c),!d){d=!0;var n=f(t);Object(o["a"])((function(){var i=!1;try{i=Object(r["a"])()}finally{if(d=!1,t=n-f(),!a())return;i?e.run(1e3):t>0?e.run(t):e.start()}}))}},t.prototype.schedule=function(){this.stop(),this.run()},t.prototype.observe=function(){var t=this,e=function(){return t.observer&&t.observer.observe(document.body,u)};document.body?e():i["a"].addEventListener("DOMContentLoaded",e)},t.prototype.start=function(){var t=this;this.stopped&&(this.stopped=!1,this.observer=new MutationObserver(this.listener),this.observe(),l.forEach((function(e){return i["a"].addEventListener(e,t.listener,!0)})))},t.prototype.stop=function(){var t=this;this.stopped||(this.observer&&this.observer.disconnect(),l.forEach((function(e){return i["a"].removeEventListener(e,t.listener,!0)})),this.stopped=!0)},t}(),p=new h,y=function(t){!s&&t>0&&p.start(),s+=t,!s&&p.stop()}},function(t,e,n){"use strict";var r=n(65);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=function(){function t(t,e){this.type=t,"addNode"===t?this.nodeId=e:this.nodeState=e}return t.prototype.undo=function(t){"addNode"===this.type?this.removeNode(t):this.addNode(t)},t.prototype.redo=function(t){"addNode"===this.type&&this.nodeState?this.addNode(t):"removeNode"===this.type&&this.nodeId&&this.removeNode(t)},t.prototype.addNode=function(t){var e=t.nodeTypes.get(this.nodeState.type);if(e){var n=new e;t.addNode(n),n.load(this.nodeState),this.nodeId=n.id}},t.prototype.removeNode=function(t){var e=this,n=t.nodes.find((function(t){return t.id===e.nodeId}));n&&(this.nodeState=n.save(),t.removeNode(n))},t}();e["a"]=r},function(t,e,n){"use strict";var r=function(){function t(t,e){if(this.type=t,"addConnection"===t)this.connectionId=e;else{var n=e;this.connectionState={id:n.id,from:n.from.id,to:n.to.id}}}return t.prototype.undo=function(t){"addConnection"===this.type?this.removeConnection(t):this.addConnection(t)},t.prototype.redo=function(t){"addConnection"===this.type&&this.connectionState?this.addConnection(t):"removeConnection"===this.type&&this.connectionId&&this.removeConnection(t)},t.prototype.addConnection=function(t){var e=t.findNodeInterface(this.connectionState.from),n=t.findNodeInterface(this.connectionState.to);e&&n&&t.addConnection(e,n)},t.prototype.removeConnection=function(t){var e=this,n=t.connections.find((function(t){return t.id===e.connectionId}));n&&(this.connectionState={id:n.id,from:n.from.id,to:n.to.id},t.removeConnection(n))},t}();e["a"]=r},function(t,e,n){"use strict";var r=n(86);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(87);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";n.d(e,"a",(function(){return r}));var r=function(){function t(t,e,n,r){return this.x=t,this.y=e,this.width=n,this.height=r,this.top=this.y,this.left=this.x,this.bottom=this.top+this.height,this.right=this.left+this.width,Object.freeze(this)}return t.prototype.toJSON=function(){var t=this,e=t.x,n=t.y,r=t.top,i=t.right,o=t.bottom,s=t.left,a=t.width,c=t.height;return{x:e,y:n,top:r,right:i,bottom:o,left:s,width:a,height:c}},t.fromRect=function(e){return new t(e.x,e.y,e.width,e.height)},t}()},function(t,e,n){"use strict";n.d(e,"a",(function(){return o}));var r=n(6),i=n(41),o=function(t){r["a"].forEach((function(e){e.activeTargets.splice(0,e.activeTargets.length),e.skippedTargets.splice(0,e.skippedTargets.length),e.observationTargets.forEach((function(n){n.isActive()&&(Object(i["a"])(n.target)>t?e.activeTargets.push(n):e.skippedTargets.push(n))}))}))}},function(t,e,n){"use strict";var r=n(89);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(90);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(91);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(92);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(93);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(94);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(95);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:["node-editor",{"ignore-mouse":!!t.temporaryConnection,"--temporary-connection":!!t.temporaryConnection}],attrs:{tabindex:"-1"},on:{mousemove:function(e){return e.target!==e.currentTarget?null:t.mouseMoveHandler(e)},mousedown:t.mouseDown,mouseup:t.mouseUp,wheel:function(e){return e.target!==e.currentTarget?null:t.mouseWheel(e)},keydown:t.keyDown,keyup:t.keyUp,contextmenu:function(e){return e.target!==e.currentTarget?null:(e.preventDefault(),t.openContextMenu(e))}}},[n("div",{staticClass:"background",style:t.backgroundStyle}),t._v(" "),n("svg",{staticClass:"connections-container"},[t._l(t.connections,(function(e){return n("g",{key:e.id+t.counter.toString()},[t._t("connections",[n(t.plugin.components.connection,{tag:"component",attrs:{connection:e}})],{connection:e})],2)})),t._v(" "),t.temporaryConnection?n(t.plugin.components.tempConnection,{tag:"component",attrs:{connection:t.temporaryConnection}}):t._e()],2),t._v(" "),n("div",{staticClass:"node-container",style:t.styles},t._l(t.nodes,(function(e){return n(t.plugin.components.node,{key:e.id+t.counter.toString(),tag:"component",attrs:{data:e,selected:t.selectedNodes.includes(e)},on:{select:function(n){return t.selectNode(e,n)}}})})),1),t._v(" "),n(t.plugin.components.contextMenu,{tag:"component",attrs:{x:t.contextMenu.x,y:t.contextMenu.y,items:t.contextMenu.items,flippable:""},on:{click:t.onContextMenuClick},model:{value:t.contextMenu.show,callback:function(e){t.$set(t.contextMenu,"show",e)},expression:"contextMenu.show"}}),t._v(" "),n(t.plugin.components.sidebar,{tag:"component"}),t._v(" "),t.plugin.enableMinimap?n(t.plugin.components.minimap,{tag:"component",attrs:{nodes:t.nodes,connections:t.connections}}):t._e()],1)},i=[];r._withStripped=!0},function(t,e,n){"use strict"},function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));var r=n(2);function i(t){return Object(r["a"])((function(e,n){"undefined"===typeof e.inject&&(e.inject={}),Array.isArray(e.inject)||(e.inject[n]=t||n)}))}},function(t,e,n){"use strict";n(2),n(7)},function(t,e,n){"use strict";n(2),n(10)},function(t,e,n){"use strict";n(2),n(10)},function(t,e,n){"use strict";n.d(e,"a",(function(){return o}));var r=n(2),i=n(10);function o(t){return void 0===t&&(t={}),function(e,n){Object(i["a"])(t,e,n),Object(r["a"])((function(e,n){(e.props||(e.props={}))[n]=t}))(e,n)}}},function(t,e,n){"use strict";n(2),n(10)},function(t,e,n){"use strict";n.d(e,"a",(function(){return o}));var r=n(2),i=n(7);function o(t){return Object(r["a"])((function(e,n){var r=e.provide;Object(i["a"])(e),Object(i["b"])(r)&&(r=e.provide=Object(i["c"])(r)),r.managed[n]=t||n}))}},function(t,e,n){"use strict";n(2),n(7)},function(t,e,n){"use strict";n(2)},function(t,e,n){"use strict";n(2)},function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));var r=n(2);function i(t,e){void 0===e&&(e={});var n=e.deep,i=void 0!==n&&n,o=e.immediate,s=void 0!==o&&o;return Object(r["a"])((function(e,n){"object"!==typeof e.watch&&(e.watch=Object.create(null));var r=e.watch;"object"!==typeof r[t]||Array.isArray(r[t])?"undefined"===typeof r[t]&&(r[t]=[]):r[t]=[r[t]],r[t].push({handler:n,deep:i,immediate:s})}))}},function(t,e,n){"use strict";var r;n.d(e,"a",(function(){return r})),function(t){t[t["NONE"]=0]="NONE",t[t["ALLOWED"]=1]="ALLOWED",t[t["FORBIDDEN"]=2]="FORBIDDEN"}(r||(r={}))},function(t,e,n){},function(t,e,n){},function(t,e,n){},function(t,e,n){},function(t,e,n){},function(t,e,n){},function(t,e,n){},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("path",{class:t.classes,attrs:{d:t.d}})},i=[];r._withStripped=!0},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("connection-view",{attrs:{x1:t.d.x1,y1:t.d.y1,x2:t.d.x2,y2:t.d.y2,state:t.state,connection:t.connection}})},i=[];r._withStripped=!0},function(t,e,n){"use strict";n.d(e,"a",(function(){return o}));var r=n(16),i=n(8),o=function(){function t(t){if(0===arguments.length)throw new TypeError("Failed to construct 'ResizeObserver': 1 argument required, but only 0 present.");if("function"!==typeof t)throw new TypeError("Failed to construct 'ResizeObserver': The callback provided as parameter 1 is not a function.");r["a"].connect(this,t)}return t.prototype.observe=function(t,e){if(0===arguments.length)throw new TypeError("Failed to execute 'observe' on 'ResizeObserver': 1 argument required, but only 0 present.");if(!Object(i["a"])(t))throw new TypeError("Failed to execute 'observe' on 'ResizeObserver': parameter 1 is not of type 'Element");r["a"].observe(this,t,e)},t.prototype.unobserve=function(t){if(0===arguments.length)throw new TypeError("Failed to execute 'unobserve' on 'ResizeObserver': 1 argument required, but only 0 present.");if(!Object(i["a"])(t))throw new TypeError("Failed to execute 'unobserve' on 'ResizeObserver': parameter 1 is not of type 'Element");r["a"].unobserve(this,t)},t.prototype.disconnect=function(){r["a"].disconnect(this)},t.toString=function(){return"function ResizeObserver () { [polyfill code] }"},t}()},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("connection-view",{attrs:{x1:t.d.input[0],y1:t.d.input[1],x2:t.d.output[0],y2:t.d.output[1],state:t.status,connection:t.connection,"is-temporary":""}})},i=[];r._withStripped=!0},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.classes,style:t.styles,attrs:{id:t.data.id}},[n("div",{staticClass:"__title",on:{mousedown:function(e){return e.target!==e.currentTarget?null:(e.stopPropagation(),t.startDrag(e))},contextmenu:function(e){return e.target!==e.currentTarget?null:(e.preventDefault(),t.openContextMenu(e))}}},[t.renaming?n("input",{directives:[{name:"model",rawName:"v-model",value:t.tempName,expression:"tempName"},{name:"click-outside",rawName:"v-click-outside",value:t.doneRenaming,expression:"doneRenaming"}],staticClass:"dark-input",attrs:{type:"text",placeholder:"Node Name"},domProps:{value:t.tempName},on:{keydown:function(e){return!e.type.indexOf("key")&&t._k(e.keyCode,"enter",13,e.key,"Enter")?null:t.doneRenaming(e)},input:function(e){e.target.composing||(t.tempName=e.target.value)}}}):n("span",[t._v(t._s(t.data.name))]),t._v(" "),n(t.plugin.components.contextMenu,{tag:"component",attrs:{x:t.contextMenu.x,y:t.contextMenu.y,items:t.contextMenu.items},on:{click:t.onContextMenu},model:{value:t.contextMenu.show,callback:function(e){t.$set(t.contextMenu,"show",e)},expression:"contextMenu.show"}})],1),t._v(" "),n("div",{staticClass:"__content"},[n("div",{staticClass:"__outputs"},t._l(t.data.outputInterfaces,(function(e,r){return n(t.plugin.components.nodeInterface,{key:e.id,tag:"component",attrs:{name:r,data:e}})})),1),t._v(" "),n("div",{staticClass:"__options"},[t._l(t.data.options,(function(e){var r=e[0],i=e[1];return[n(t.plugin.components.nodeOption,{key:r,tag:"component",attrs:{name:r,option:i,componentName:i.optionComponent,node:t.data},on:{openSidebar:function(e){return t.openSidebar(r)}}}),t._v(" "),t.plugin.sidebar.nodeId===t.data.id&&t.plugin.sidebar.optionName===r&&i.sidebarComponent?n("portal",{key:"sb_"+r,attrs:{to:"sidebar"}},[n(t.plugin.components.nodeOption,{key:t.data.id+r,tag:"component",attrs:{name:r,option:i,componentName:i.sidebarComponent,node:t.data}})],1):t._e()]}))],2),t._v(" "),n("div",{staticClass:"__inputs"},t._l(t.data.inputInterfaces,(function(e,r){return n(t.plugin.components.nodeInterface,{key:e.id,tag:"component",attrs:{name:r,data:e}})})),1)])])},i=[];r._withStripped=!0},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:t.classes,attrs:{id:t.data.id}},[n("div",{staticClass:"__port",on:{mouseover:t.startHover,mouseout:t.endHover}}),t._v(" "),t.data.connectionCount>0||!t.data.option||!t.getOptionComponent(t.data.option)?n("span",{staticClass:"align-middle"},[t._v("\n        "+t._s(t.displayName)+"\n    ")]):n(t.getOptionComponent(t.data.option),{tag:"component",attrs:{option:t.data,value:t.value,name:t.displayName},on:{input:function(e){t.data.value=e}}})],1)},i=[];r._withStripped=!0},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n(t.component,{tag:"component",staticClass:"node-option",attrs:{name:t.displayName,node:t.node,value:t.value,option:t.option},on:{input:t.updateValue,openSidebar:function(e){return t.$emit("openSidebar")}}})},i=[];r._withStripped=!0},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{directives:[{name:"show",rawName:"v-show",value:t.value,expression:"value"},{name:"click-outside",rawName:"v-click-outside",value:t.onClickOutside,expression:"onClickOutside"}],class:t.classes,style:t.styles},[t._l(t._items,(function(e,r){return[e.isDivider?n("div",{key:r,staticClass:"divider"}):n("div",{key:r,staticClass:"d-flex align-items-center",class:{item:!0,submenu:!!e.submenu,"--disabled":!!e.disabled},on:{mouseenter:function(e){return t.onMouseEnter(e,r)},mouseleave:function(e){return t.onMouseLeave(e,r)},click:function(n){return n.stopPropagation(),n.preventDefault(),t.onClick(e)}}},[n("div",{staticClass:"flex-fill"},[t._v(t._s(e.label))]),t._v(" "),e.submenu?n("div",{staticClass:"ml-3",staticStyle:{"line-height":"1em"}},[n("svg",{attrs:{width:"13",height:"13",viewBox:"-60 120 250 250"}},[n("path",{attrs:{d:"M160.875 279.5625 L70.875 369.5625 L70.875 189.5625 L160.875 279.5625 Z",stroke:"none",fill:"white"}})])]):t._e(),t._v(" "),e.submenu?n("context-menu",{attrs:{value:t.activeMenu===r,items:e.submenu,"is-nested":!0,"is-flipped":{x:t.flippedX,y:t.flippedY},flippable:t.flippable},on:{click:t.onChildClick}}):t._e()],1)]}))],2)},i=[];r._withStripped=!0},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:["sidebar",{"--open":t.plugin.sidebar.visible}],style:t.styles},[n("div",{staticClass:"__resizer",on:{mousedown:t.startResize}}),t._v(" "),n("div",{staticClass:"d-flex align-items-center"},[n("button",{staticClass:"__close",attrs:{tabindex:"-1"},on:{click:t.close}},[t._v("×")]),t._v(" "),n("div",{staticClass:"ml-2"},[n("b",[t._v(t._s(t.nodeName))])])]),t._v(" "),n("portal-target",{attrs:{name:"sidebar"}})],1)},i=[];r._withStripped=!0},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=this,n=e.$createElement,r=e._self._c||n;return r("canvas",{ref:"cv",staticClass:"minimap",on:{mouseenter:function(t){e.showViewBounds=!0},mouseleave:function(){t.showViewBounds=!1,t.mouseup()},mousedown:function(t){return t.target!==t.currentTarget?null:e.mousedown(t)},mousemove:function(t){return t.target!==t.currentTarget?null:e.mousemove(t)},mouseup:e.mouseup}})},i=[];r._withStripped=!0},function(t,e,n){"use strict";n.d(e,"a",(function(){return u}));var r=n(5),i=n.n(r),o=n(18),s=n(111),a=n.n(s),c={install:function(t,e){i.a!==t&&console.error("Multiple instances of Vue detected\nSee https://github.com/vuetifyjs/vuetify/issues/4068\n\nIf you're seeing \"$attrs is readonly\", it's caused by this"),t.use(a.a),t.component("baklava-editor",o["a"])}},u=c},function(t,e,n){"use strict";n.d(e,"a",(function(){return p}));var r=n(5),i=n.n(r),o=n(17),s=n(43),a=n(46),c=n(45),u=n(40),l=n(42),f=n(47),d=n(48),h=n(49),p=function(){function t(){this.type="ViewPlugin",this.panning={x:0,y:0},this.scaling=1,this.sidebar={visible:!1,nodeId:"",optionName:""},this.useStraightConnections=!1,this.enableMinimap=!1,this.backgroundGrid={gridSize:100,gridDivision:5,subGridVisibleThreshold:.6},this.options={},this.nodeTypeAliases={},this.hooks={renderNode:new o["SequentialHook"],renderOption:new o["SequentialHook"],renderInterface:new o["SequentialHook"],renderConnection:new o["SequentialHook"]},this.components={node:s["a"],nodeOption:a["a"],nodeInterface:c["a"],connection:u["a"],tempConnection:l["a"],contextMenu:f["a"],sidebar:d["a"],minimap:h["a"]}}return t.prototype.register=function(t){var e=this;this.editor=t,this.editor.hooks.load.tap(this,(function(t){return e.panning=t.panning,e.scaling=t.scaling,t})),this.editor.hooks.save.tap(this,(function(t){return t.panning=e.panning,t.scaling=e.scaling,t})),this.editor.events.beforeAddNode.addListener(this,(function(t){var n=t;n.position={x:0,y:0},n.disablePointerEvents=!1,n.twoColumn=n.twoColumn||!1,n.width=n.width||200,n.customClasses=n.customClasses||"",t.hooks.save.tap(e,(function(t){return t.position=n.position,t.width=n.width,t.twoColumn=n.twoColumn,t.customClasses=n.customClasses,t})),t.hooks.load.tap(e,(function(t){return n.position=t.position||{x:0,y:0},n.width=t.width||200,n.twoColumn=t.twoColumn||!1,t}))}))},t.prototype.registerOption=function(t,e){i.a.set(this.options,t,e)},t.prototype.setNodeTypeAlias=function(t,e){e?i.a.set(this.nodeTypeAliases,t,e):this.nodeTypeAliases[t]&&i.a.delete(this.nodeTypeAliases,t)},t}()},function(t,e,n){"use strict";var r=n(0),i=function(){function t(t){this.nodeBuffer="",this.connectionBuffer="",this.editor=t}return Object.defineProperty(t.prototype,"isEmpty",{get:function(){return!this.nodeBuffer},enumerable:!1,configurable:!0}),t.prototype.clear=function(){this.nodeBuffer="",this.connectionBuffer=""},t.prototype.copy=function(t){this.connectionBuffer=JSON.stringify(this.editor.connections.filter((function(e){return t.includes(e.from.parent)&&t.includes(e.to.parent)})).map((function(t){return{from:t.from.id,to:t.to.id}}))),this.nodeBuffer=JSON.stringify(t.map((function(t){return t.save()})))},t.prototype.paste=function(){var t,e,n,i,o=this,s=new Map,a=new Map,c=JSON.parse(this.nodeBuffer),u=JSON.parse(this.connectionBuffer),l=function(t){var e=f.editor.nodeTypes.get(t.type);if(!e)return console.warn("Node type "+t.type+" not registered"),{value:void 0};var n=new e,r=n.id;n.interfaces.forEach((function(t){t.hooks.load.tap(o,(function(e){var n=o.editor.generateId("ni");return s.set(e.id,n),a.set(e.id,r),t.id=n,t.hooks.load.untap(o),e}))})),n.hooks.load.tap(f,(function(t){var e=t;return e.position&&(e.position.x+=10,e.position.y+=10),e})),f.editor.addNode(n),n.load(t),n.id=r,s.set(t.id,r)},f=this;try{for(var d=Object(r["__values"])(c),h=d.next();!h.done;h=d.next()){var p=h.value,y=l(p);if("object"===typeof y)return y.value}}catch(w){t={error:w}}finally{try{h&&!h.done&&(e=d.return)&&e.call(d)}finally{if(t)throw t.error}}var v=function(t){var e=b.editor.nodes.find((function(e){return e.id===a.get(t.from)})),n=b.editor.nodes.find((function(e){return e.id===a.get(t.to)}));if(!e||!n)return"continue";var r=Array.from(e.interfaces.values()).find((function(e){return e.id===s.get(t.from)})),i=Array.from(n.interfaces.values()).find((function(e){return e.id===s.get(t.to)}));if(!r||!i)return"continue";b.editor.addConnection(r,i)},b=this;try{for(var m=Object(r["__values"])(u),g=m.next();!g.done;g=m.next()){var _=g.value;v(_)}}catch(O){n={error:O}}finally{try{g&&!g.done&&(i=m.return)&&i.call(m)}finally{if(n)throw n.error}}},t}();e["a"]=i},function(t,e,n){"use strict";var r=n(52),i=n(53),o=function(){function t(t){var e=this;this.maxSteps=200,this.steps=[],this.changeBySelf=!1,this.currentIndex=-1,this.activeTransaction=!1,this.transactionSteps=[],this.viewPlugin=t,this.viewPlugin.editor.events.addNode.addListener(this,(function(t){e.addStep(new r["a"]("addNode",t.id))})),this.viewPlugin.editor.events.removeNode.addListener(this,(function(t){e.addStep(new r["a"]("removeNode",t.save()))})),this.viewPlugin.editor.events.addConnection.addListener(this,(function(t){e.addStep(new i["a"]("addConnection",t.id))})),this.viewPlugin.editor.events.removeConnection.addListener(this,(function(t){e.addStep(new i["a"]("removeConnection",t))}))}return t.prototype.startTransaction=function(){this.activeTransaction=!0},t.prototype.commitTransaction=function(){this.activeTransaction=!1},t.prototype.undo=function(){0!==this.steps.length&&-1!==this.currentIndex&&(this.changeBySelf=!0,this.steps[this.currentIndex--].undo(this.viewPlugin.editor),this.changeBySelf=!1)},t.prototype.redo=function(){0===this.steps.length||this.currentIndex>=this.steps.length-1||(this.changeBySelf=!0,this.steps[++this.currentIndex].redo(this.viewPlugin.editor),this.changeBySelf=!1)},t.prototype.addStep=function(t){if(!this.changeBySelf)if(this.activeTransaction)this.transactionSteps.push(t);else{this.currentIndex!==this.steps.length-1&&(this.steps=this.steps.slice(0,this.currentIndex+1)),this.steps.push(t),this.currentIndex++;while(this.steps.length>this.maxSteps)this.steps.shift()}},t}();e["a"]=o},function(t,e,n){"use strict";var r=n(88);n.d(e,"a",(function(){return r["a"]}));n(25)},function(t,e,n){"use strict";n.d(e,"a",(function(){return c}));var r=n(102),i=n(103),o=n(104),s=n(105),a=n(57),c=function(){var t=0;Object(a["a"])(t);while(Object(r["a"])())t=Object(s["a"])(),Object(a["a"])(t);return Object(i["a"])()&&Object(o["a"])(),t>0}},function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));var r=n(6),i=function(){return r["a"].some((function(t){return t.activeTargets.length>0}))}},function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));var r=n(6),i=function(){return r["a"].some((function(t){return t.skippedTargets.length>0}))}},function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));var r="ResizeObserver loop completed with undelivered notifications.",i=function(){var t;"function"===typeof ErrorEvent?t=new ErrorEvent("error",{message:r}):(t=document.createEvent("Event"),t.initEvent("error",!1,!1),t.message=r),window.dispatchEvent(t)}},function(t,e,n){"use strict";n.d(e,"a",(function(){return a}));var r=n(6),i=n(25),o=n(41),s=n(13),a=function(){var t=1/0,e=[];r["a"].forEach((function(n){if(0!==n.activeTargets.length){var r=[];n.activeTargets.forEach((function(e){var n=new i["a"](e.target),a=Object(o["a"])(e.target);r.push(n),e.lastReportedSize=Object(s["a"])(e.target,e.observedBox),a<t&&(t=a)})),e.push((function(){n.callback.call(n.observer,r,n.observer)})),n.activeTargets.splice(0,n.activeTargets.length)}}));for(var n=0,a=e;n<a.length;n++){var c=a[n];c()}return t}},function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));var r=n(107),i=function(t){Object(r["a"])((function(){requestAnimationFrame(t)}))}},function(t,e,n){"use strict";var r;n.d(e,"a",(function(){return s}));var i=[],o=function(){return i.splice(0).forEach((function(t){return t()}))},s=function(t){if(!r){var e=0,n=document.createTextNode(""),s={characterData:!0};new MutationObserver((function(){return o()})).observe(n,s),r=function(){n.textContent=""+(e?e--:e++)}}i.push(t),r()}},function(t,e,n){"use strict";n.d(e,"a",(function(){return a}));var r=n(15),i=n(13),o=n(8),s=function(t){return!Object(o["d"])(t)&&!Object(o["c"])(t)&&"inline"===getComputedStyle(t).display},a=function(){function t(t,e){this.target=t,this.observedBox=e||r["a"].CONTENT_BOX,this.lastReportedSize={inlineSize:0,blockSize:0}}return t.prototype.isActive=function(){var t=Object(i["a"])(this.target,this.observedBox,!0);return s(this.target)&&(this.lastReportedSize=t),this.lastReportedSize.inlineSize!==t.inlineSize||this.lastReportedSize.blockSize!==t.blockSize},t}()},function(t,e,n){"use strict";n.d(e,"a",(function(){return r}));var r=function(){function t(t,e){this.activeTargets=[],this.skippedTargets=[],this.observationTargets=[],this.observer=t,this.callback=e}return t}()},function(t,e,n){"use strict";function r(t){return t.replace(" ","-")}n.d(e,"a",(function(){return r}))},function(t,e,n){"use strict";
/*! 
  * portal-vue © Thorsten Lünborg, 2019 
  * 
  * Version: 2.1.7
  * 
  * LICENCE: MIT 
  * 
  * https://github.com/linusborg/portal-vue
  * 
 */function r(t){return t&&"object"===typeof t&&"default"in t?t["default"]:t}Object.defineProperty(e,"__esModule",{value:!0});var i=r(n(5));function o(t){return o="function"===typeof Symbol&&"symbol"===typeof Symbol.iterator?function(t){return typeof t}:function(t){return t&&"function"===typeof Symbol&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},o(t)}function s(t){return a(t)||c(t)||u()}function a(t){if(Array.isArray(t)){for(var e=0,n=new Array(t.length);e<t.length;e++)n[e]=t[e];return n}}function c(t){if(Symbol.iterator in Object(t)||"[object Arguments]"===Object.prototype.toString.call(t))return Array.from(t)}function u(){throw new TypeError("Invalid attempt to spread non-iterable instance")}var l="undefined"!==typeof window;function f(t){return Array.isArray(t)||"object"===o(t)?Object.freeze(t):t}function d(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{};return t.reduce((function(t,n){var r=n.passengers[0],i="function"===typeof r?r(e):n.passengers;return t.concat(i)}),[])}function h(t,e){return t.map((function(t,e){return[e,t]})).sort((function(t,n){return e(t[1],n[1])||t[0]-n[0]})).map((function(t){return t[1]}))}function p(t,e){return e.reduce((function(e,n){return t.hasOwnProperty(n)&&(e[n]=t[n]),e}),{})}var y={},v={},b={},m=i.extend({data:function(){return{transports:y,targets:v,sources:b,trackInstances:l}},methods:{open:function(t){if(l){var e=t.to,n=t.from,r=t.passengers,o=t.order,s=void 0===o?1/0:o;if(e&&n&&r){var a={to:e,from:n,passengers:f(r),order:s},c=Object.keys(this.transports);-1===c.indexOf(e)&&i.set(this.transports,e,[]);var u=this.$_getTransportIndex(a),d=this.transports[e].slice(0);-1===u?d.push(a):d[u]=a,this.transports[e]=h(d,(function(t,e){return t.order-e.order}))}}},close:function(t){var e=arguments.length>1&&void 0!==arguments[1]&&arguments[1],n=t.to,r=t.from;if(n&&(r||!1!==e)&&this.transports[n])if(e)this.transports[n]=[];else{var i=this.$_getTransportIndex(t);if(i>=0){var o=this.transports[n].slice(0);o.splice(i,1),this.transports[n]=o}}},registerTarget:function(t,e,n){l&&(this.trackInstances&&!n&&this.targets[t]&&console.warn("[portal-vue]: Target ".concat(t," already exists")),this.$set(this.targets,t,Object.freeze([e])))},unregisterTarget:function(t){this.$delete(this.targets,t)},registerSource:function(t,e,n){l&&(this.trackInstances&&!n&&this.sources[t]&&console.warn("[portal-vue]: source ".concat(t," already exists")),this.$set(this.sources,t,Object.freeze([e])))},unregisterSource:function(t){this.$delete(this.sources,t)},hasTarget:function(t){return!(!this.targets[t]||!this.targets[t][0])},hasSource:function(t){return!(!this.sources[t]||!this.sources[t][0])},hasContentFor:function(t){return!!this.transports[t]&&!!this.transports[t].length},$_getTransportIndex:function(t){var e=t.to,n=t.from;for(var r in this.transports[e])if(this.transports[e][r].from===n)return+r;return-1}}}),g=new m(y),_=1,w=i.extend({name:"portal",props:{disabled:{type:Boolean},name:{type:String,default:function(){return String(_++)}},order:{type:Number,default:0},slim:{type:Boolean},slotProps:{type:Object,default:function(){return{}}},tag:{type:String,default:"DIV"},to:{type:String,default:function(){return String(Math.round(1e7*Math.random()))}}},created:function(){var t=this;this.$nextTick((function(){g.registerSource(t.name,t)}))},mounted:function(){this.disabled||this.sendUpdate()},updated:function(){this.disabled?this.clear():this.sendUpdate()},beforeDestroy:function(){g.unregisterSource(this.name),this.clear()},watch:{to:function(t,e){e&&e!==t&&this.clear(e),this.sendUpdate()}},methods:{clear:function(t){var e={from:this.name,to:t||this.to};g.close(e)},normalizeSlots:function(){return this.$scopedSlots.default?[this.$scopedSlots.default]:this.$slots.default},normalizeOwnChildren:function(t){return"function"===typeof t?t(this.slotProps):t},sendUpdate:function(){var t=this.normalizeSlots();if(t){var e={from:this.name,to:this.to,passengers:s(t),order:this.order};g.open(e)}else this.clear()}},render:function(t){var e=this.$slots.default||this.$scopedSlots.default||[],n=this.tag;return e&&this.disabled?e.length<=1&&this.slim?this.normalizeOwnChildren(e)[0]:t(n,[this.normalizeOwnChildren(e)]):this.slim?t():t(n,{class:{"v-portal":!0},style:{display:"none"},key:"v-portal-placeholder"})}}),O=i.extend({name:"portalTarget",props:{multiple:{type:Boolean,default:!1},name:{type:String,required:!0},slim:{type:Boolean,default:!1},slotProps:{type:Object,default:function(){return{}}},tag:{type:String,default:"div"},transition:{type:[String,Object,Function]}},data:function(){return{transports:g.transports,firstRender:!0}},created:function(){var t=this;this.$nextTick((function(){g.registerTarget(t.name,t)}))},watch:{ownTransports:function(){this.$emit("change",this.children().length>0)},name:function(t,e){g.unregisterTarget(e),g.registerTarget(t,this)}},mounted:function(){var t=this;this.transition&&this.$nextTick((function(){t.firstRender=!1}))},beforeDestroy:function(){g.unregisterTarget(this.name)},computed:{ownTransports:function(){var t=this.transports[this.name]||[];return this.multiple?t:0===t.length?[]:[t[t.length-1]]},passengers:function(){return d(this.ownTransports,this.slotProps)}},methods:{children:function(){return 0!==this.passengers.length?this.passengers:this.$scopedSlots.default?this.$scopedSlots.default(this.slotProps):this.$slots.default||[]},noWrapper:function(){var t=this.slim&&!this.transition;return t&&this.children().length>1&&console.warn("[portal-vue]: PortalTarget with `slim` option received more than one child element."),t}},render:function(t){var e=this.noWrapper(),n=this.children(),r=this.transition||this.tag;return e?n[0]:this.slim&&!r?t():t(r,{props:{tag:this.transition&&this.tag?this.tag:void 0},class:{"vue-portal-target":!0}},n)}}),S=0,x=["disabled","name","order","slim","slotProps","tag","to"],I=["multiple","transition"],A=i.extend({name:"MountingPortal",inheritAttrs:!1,props:{append:{type:[Boolean,String]},bail:{type:Boolean},mountTo:{type:String,required:!0},disabled:{type:Boolean},name:{type:String,default:function(){return"mounted_"+String(S++)}},order:{type:Number,default:0},slim:{type:Boolean},slotProps:{type:Object,default:function(){return{}}},tag:{type:String,default:"DIV"},to:{type:String,default:function(){return String(Math.round(1e7*Math.random()))}},multiple:{type:Boolean,default:!1},targetSlim:{type:Boolean},targetSlotProps:{type:Object,default:function(){return{}}},targetTag:{type:String,default:"div"},transition:{type:[String,Object,Function]}},created:function(){if("undefined"!==typeof document){var t=document.querySelector(this.mountTo);if(t){var e=this.$props;if(g.targets[e.name])e.bail?console.warn("[portal-vue]: Target ".concat(e.name," is already mounted.\n        Aborting because 'bail: true' is set")):this.portalTarget=g.targets[e.name];else{var n=e.append;if(n){var r="string"===typeof n?n:"DIV",i=document.createElement(r);t.appendChild(i),t=i}var o=p(this.$props,I);o.slim=this.targetSlim,o.tag=this.targetTag,o.slotProps=this.targetSlotProps,o.name=this.to,this.portalTarget=new O({el:t,parent:this.$parent||this,propsData:o})}}else console.error("[portal-vue]: Mount Point '".concat(this.mountTo,"' not found in document"))}},beforeDestroy:function(){var t=this.portalTarget;if(this.append){var e=t.$el;e.parentNode.removeChild(e)}t.$destroy()},render:function(t){if(!this.portalTarget)return console.warn("[portal-vue] Target wasn't mounted"),t();if(!this.$scopedSlots.manual){var e=p(this.$props,x);return t(w,{props:e,attrs:this.$attrs,on:this.$listeners,scopedSlots:this.$scopedSlots},this.$slots.default)}var n=this.$scopedSlots.manual({to:this.to});return Array.isArray(n)&&(n=n[0]),n||t()}});function T(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{};t.component(e.portalName||"Portal",w),t.component(e.portalTargetName||"PortalTarget",O),t.component(e.MountingPortalName||"MountingPortal",A)}var j={install:T};e.default=j,e.Portal=w,e.PortalTarget=O,e.MountingPortal=A,e.Wormhole=g},function(t,e,n){"use strict";n.r(e),n.d(e,"Components",(function(){return y}));var r=n(18);n.d(e,"Editor",(function(){return r["a"]}));var i=n(12),o=n(40),s=n(42),a=n(43),c=n(45),u=n(46),l=n(47),f=n(48),d=n(49),h=n(96);n.d(e,"BaklavaVuePlugin",(function(){return h["a"]}));var p=n(97);n.d(e,"ViewPlugin",(function(){return p["a"]}));var y={Connection:i["a"],ConnectionWrapper:o["a"],TemporaryConnection:s["a"],Node:a["a"],NodeInterface:c["a"],NodeOption:u["a"],ContextMenu:l["a"],Sidebar:f["a"],Minimap:d["a"]}}])}))},"0cfb":function(t,e,n){var r=n("83ab"),i=n("d039"),o=n("cc12");t.exports=!r&&!i((function(){return 7!=Object.defineProperty(o("div"),"a",{get:function(){return 7}}).a}))},"0d51":function(t,e,n){var r=n("da84"),i=r.String;t.exports=function(t){try{return i(t)}catch(e){return"Object"}}},"159b":function(t,e,n){var r=n("da84"),i=n("fdbc"),o=n("785a"),s=n("17c2"),a=n("9112"),c=function(t){if(t&&t.forEach!==s)try{a(t,"forEach",s)}catch(e){t.forEach=s}};for(var u in i)i[u]&&c(r[u]&&r[u].prototype);c(o)},1626:function(t,e){t.exports=function(t){return"function"==typeof t}},"17c2":function(t,e,n){"use strict";var r=n("b727").forEach,i=n("a640"),o=i("forEach");t.exports=o?[].forEach:function(t){return r(this,t,arguments.length>1?arguments[1]:void 0)}},"19aa":function(t,e,n){var r=n("da84"),i=n("3a9b"),o=r.TypeError;t.exports=function(t,e){if(i(e,t))return t;throw o("Incorrect invocation")}},"1a2d":function(t,e,n){var r=n("e330"),i=n("7b0b"),o=r({}.hasOwnProperty);t.exports=Object.hasOwn||function(t,e){return o(i(t),e)}},"1be4":function(t,e,n){var r=n("d066");t.exports=r("document","documentElement")},"1c7e":function(t,e,n){var r=n("b622"),i=r("iterator"),o=!1;try{var s=0,a={next:function(){return{done:!!s++}},return:function(){o=!0}};a[i]=function(){return this},Array.from(a,(function(){throw 2}))}catch(c){}t.exports=function(t,e){if(!e&&!o)return!1;var n=!1;try{var r={};r[i]=function(){return{next:function(){return{done:n=!0}}}},t(r)}catch(c){}return n}},"1cdc":function(t,e,n){var r=n("342f");t.exports=/(?:ipad|iphone|ipod).*applewebkit/i.test(r)},"1d80":function(t,e,n){var r=n("da84"),i=r.TypeError;t.exports=function(t){if(void 0==t)throw i("Can't call method on "+t);return t}},"1fb5":function(t,e,n){"use strict";e.byteLength=l,e.toByteArray=d,e.fromByteArray=y;for(var r=[],i=[],o="undefined"!==typeof Uint8Array?Uint8Array:Array,s="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/",a=0,c=s.length;a<c;++a)r[a]=s[a],i[s.charCodeAt(a)]=a;function u(t){var e=t.length;if(e%4>0)throw new Error("Invalid string. Length must be a multiple of 4");var n=t.indexOf("=");-1===n&&(n=e);var r=n===e?0:4-n%4;return[n,r]}function l(t){var e=u(t),n=e[0],r=e[1];return 3*(n+r)/4-r}function f(t,e,n){return 3*(e+n)/4-n}function d(t){var e,n,r=u(t),s=r[0],a=r[1],c=new o(f(t,s,a)),l=0,d=a>0?s-4:s;for(n=0;n<d;n+=4)e=i[t.charCodeAt(n)]<<18|i[t.charCodeAt(n+1)]<<12|i[t.charCodeAt(n+2)]<<6|i[t.charCodeAt(n+3)],c[l++]=e>>16&255,c[l++]=e>>8&255,c[l++]=255&e;return 2===a&&(e=i[t.charCodeAt(n)]<<2|i[t.charCodeAt(n+1)]>>4,c[l++]=255&e),1===a&&(e=i[t.charCodeAt(n)]<<10|i[t.charCodeAt(n+1)]<<4|i[t.charCodeAt(n+2)]>>2,c[l++]=e>>8&255,c[l++]=255&e),c}function h(t){return r[t>>18&63]+r[t>>12&63]+r[t>>6&63]+r[63&t]}function p(t,e,n){for(var r,i=[],o=e;o<n;o+=3)r=(t[o]<<16&16711680)+(t[o+1]<<8&65280)+(255&t[o+2]),i.push(h(r));return i.join("")}function y(t){for(var e,n=t.length,i=n%3,o=[],s=16383,a=0,c=n-i;a<c;a+=s)o.push(p(t,a,a+s>c?c:a+s));return 1===i?(e=t[n-1],o.push(r[e>>2]+r[e<<4&63]+"==")):2===i&&(e=(t[n-2]<<8)+t[n-1],o.push(r[e>>10]+r[e>>4&63]+r[e<<2&63]+"=")),o.join("")}i["-".charCodeAt(0)]=62,i["_".charCodeAt(0)]=63},2266:function(t,e,n){var r=n("da84"),i=n("0366"),o=n("c65b"),s=n("825a"),a=n("0d51"),c=n("e95a"),u=n("07fa"),l=n("3a9b"),f=n("9a1f"),d=n("35a1"),h=n("2a62"),p=r.TypeError,y=function(t,e){this.stopped=t,this.result=e},v=y.prototype;t.exports=function(t,e,n){var r,b,m,g,_,w,O,S=n&&n.that,x=!(!n||!n.AS_ENTRIES),I=!(!n||!n.IS_ITERATOR),A=!(!n||!n.INTERRUPTED),T=i(e,S),j=function(t){return r&&h(r,"normal",t),new y(!0,t)},E=function(t){return x?(s(t),A?T(t[0],t[1],j):T(t[0],t[1])):A?T(t,j):T(t)};if(I)r=t;else{if(b=d(t),!b)throw p(a(t)+" is not iterable");if(c(b)){for(m=0,g=u(t);g>m;m++)if(_=E(t[m]),_&&l(v,_))return _;return new y(!1)}r=f(t,b)}w=r.next;while(!(O=o(w,r)).done){try{_=E(O.value)}catch(B){h(r,"throw",B)}if("object"==typeof _&&_&&l(v,_))return _}return new y(!1)}},"23cb":function(t,e,n){var r=n("5926"),i=Math.max,o=Math.min;t.exports=function(t,e){var n=r(t);return n<0?i(n+e,0):o(n,e)}},"23e7":function(t,e,n){var r=n("da84"),i=n("06cf").f,o=n("9112"),s=n("6eeb"),a=n("ce4e"),c=n("e893"),u=n("94ca");t.exports=function(t,e){var n,l,f,d,h,p,y=t.target,v=t.global,b=t.stat;if(l=v?r:b?r[y]||a(y,{}):(r[y]||{}).prototype,l)for(f in e){if(h=e[f],t.noTargetGet?(p=i(l,f),d=p&&p.value):d=l[f],n=u(v?f:y+(b?".":"#")+f,t.forced),!n&&void 0!==d){if(typeof h==typeof d)continue;c(h,d)}(t.sham||d&&d.sham)&&o(h,"sham",!0),s(l,f,h,t)}}},"241c":function(t,e,n){var r=n("ca84"),i=n("7839"),o=i.concat("length","prototype");e.f=Object.getOwnPropertyNames||function(t){return r(t,o)}},2626:function(t,e,n){"use strict";var r=n("d066"),i=n("9bf2"),o=n("b622"),s=n("83ab"),a=o("species");t.exports=function(t){var e=r(t),n=i.f;s&&e&&!e[a]&&n(e,a,{configurable:!0,get:function(){return this}})}},2877:function(t,e,n){"use strict";function r(t,e,n,r,i,o,s,a){var c,u="function"===typeof t?t.options:t;if(e&&(u.render=e,u.staticRenderFns=n,u._compiled=!0),r&&(u.functional=!0),o&&(u._scopeId="data-v-"+o),s?(c=function(t){t=t||this.$vnode&&this.$vnode.ssrContext||this.parent&&this.parent.$vnode&&this.parent.$vnode.ssrContext,t||"undefined"===typeof __VUE_SSR_CONTEXT__||(t=__VUE_SSR_CONTEXT__),i&&i.call(this,t),t&&t._registeredComponents&&t._registeredComponents.add(s)},u._ssrRegister=c):i&&(c=a?function(){i.call(this,(u.functional?this.parent:this).$root.$options.shadowRoot)}:i),c)if(u.functional){u._injectStyles=c;var l=u.render;u.render=function(t,e){return c.call(e),l(t,e)}}else{var f=u.beforeCreate;u.beforeCreate=f?[].concat(f,c):[c]}return{exports:t,options:u}}n.d(e,"a",(function(){return r}))},"2a62":function(t,e,n){var r=n("c65b"),i=n("825a"),o=n("dc4a");t.exports=function(t,e,n){var s,a;i(t);try{if(s=o(t,"return"),!s){if("throw"===e)throw n;return n}s=r(s,t)}catch(c){a=!0,s=c}if("throw"===e)throw n;if(a)throw s;return i(s),n}},"2b0e":function(t,e,n){"use strict";n.r(e),function(t){
/*!
 * Vue.js v2.6.14
 * (c) 2014-2021 Evan You
 * Released under the MIT License.
 */
var n=Object.freeze({});function r(t){return void 0===t||null===t}function i(t){return void 0!==t&&null!==t}function o(t){return!0===t}function s(t){return!1===t}function a(t){return"string"===typeof t||"number"===typeof t||"symbol"===typeof t||"boolean"===typeof t}function c(t){return null!==t&&"object"===typeof t}var u=Object.prototype.toString;function l(t){return"[object Object]"===u.call(t)}function f(t){return"[object RegExp]"===u.call(t)}function d(t){var e=parseFloat(String(t));return e>=0&&Math.floor(e)===e&&isFinite(t)}function h(t){return i(t)&&"function"===typeof t.then&&"function"===typeof t.catch}function p(t){return null==t?"":Array.isArray(t)||l(t)&&t.toString===u?JSON.stringify(t,null,2):String(t)}function y(t){var e=parseFloat(t);return isNaN(e)?t:e}function v(t,e){for(var n=Object.create(null),r=t.split(","),i=0;i<r.length;i++)n[r[i]]=!0;return e?function(t){return n[t.toLowerCase()]}:function(t){return n[t]}}v("slot,component",!0);var b=v("key,ref,slot,slot-scope,is");function m(t,e){if(t.length){var n=t.indexOf(e);if(n>-1)return t.splice(n,1)}}var g=Object.prototype.hasOwnProperty;function _(t,e){return g.call(t,e)}function w(t){var e=Object.create(null);return function(n){var r=e[n];return r||(e[n]=t(n))}}var O=/-(\w)/g,S=w((function(t){return t.replace(O,(function(t,e){return e?e.toUpperCase():""}))})),x=w((function(t){return t.charAt(0).toUpperCase()+t.slice(1)})),I=/\B([A-Z])/g,A=w((function(t){return t.replace(I,"-$1").toLowerCase()}));function T(t,e){function n(n){var r=arguments.length;return r?r>1?t.apply(e,arguments):t.call(e,n):t.call(e)}return n._length=t.length,n}function j(t,e){return t.bind(e)}var E=Function.prototype.bind?j:T;function B(t,e){e=e||0;var n=t.length-e,r=new Array(n);while(n--)r[n]=t[n+e];return r}function C(t,e){for(var n in e)t[n]=e[n];return t}function M(t){for(var e={},n=0;n<t.length;n++)t[n]&&C(e,t[n]);return e}function k(t,e,n){}var D=function(t,e,n){return!1},N=function(t){return t};function L(t,e){if(t===e)return!0;var n=c(t),r=c(e);if(!n||!r)return!n&&!r&&String(t)===String(e);try{var i=Array.isArray(t),o=Array.isArray(e);if(i&&o)return t.length===e.length&&t.every((function(t,n){return L(t,e[n])}));if(t instanceof Date&&e instanceof Date)return t.getTime()===e.getTime();if(i||o)return!1;var s=Object.keys(t),a=Object.keys(e);return s.length===a.length&&s.every((function(n){return L(t[n],e[n])}))}catch(u){return!1}}function P(t,e){for(var n=0;n<t.length;n++)if(L(t[n],e))return n;return-1}function F(t){var e=!1;return function(){e||(e=!0,t.apply(this,arguments))}}var U="data-server-rendered",R=["component","directive","filter"],V=["beforeCreate","created","beforeMount","mounted","beforeUpdate","updated","beforeDestroy","destroyed","activated","deactivated","errorCaptured","serverPrefetch"],$={optionMergeStrategies:Object.create(null),silent:!1,productionTip:!1,devtools:!1,performance:!1,errorHandler:null,warnHandler:null,ignoredElements:[],keyCodes:Object.create(null),isReservedTag:D,isReservedAttr:D,isUnknownElement:D,getTagNamespace:k,parsePlatformTagName:N,mustUseProp:D,async:!0,_lifecycleHooks:V},z=/a-zA-Z\u00B7\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u037D\u037F-\u1FFF\u200C-\u200D\u203F-\u2040\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD/;function Y(t){var e=(t+"").charCodeAt(0);return 36===e||95===e}function H(t,e,n,r){Object.defineProperty(t,e,{value:n,enumerable:!!r,writable:!0,configurable:!0})}var W=new RegExp("[^"+z.source+".$_\\d]");function G(t){if(!W.test(t)){var e=t.split(".");return function(t){for(var n=0;n<e.length;n++){if(!t)return;t=t[e[n]]}return t}}}var q,K="__proto__"in{},X="undefined"!==typeof window,J="undefined"!==typeof WXEnvironment&&!!WXEnvironment.platform,Z=J&&WXEnvironment.platform.toLowerCase(),Q=X&&window.navigator.userAgent.toLowerCase(),tt=Q&&/msie|trident/.test(Q),et=Q&&Q.indexOf("msie 9.0")>0,nt=Q&&Q.indexOf("edge/")>0,rt=(Q&&Q.indexOf("android"),Q&&/iphone|ipad|ipod|ios/.test(Q)||"ios"===Z),it=(Q&&/chrome\/\d+/.test(Q),Q&&/phantomjs/.test(Q),Q&&Q.match(/firefox\/(\d+)/)),ot={}.watch,st=!1;if(X)try{var at={};Object.defineProperty(at,"passive",{get:function(){st=!0}}),window.addEventListener("test-passive",null,at)}catch(xs){}var ct=function(){return void 0===q&&(q=!X&&!J&&"undefined"!==typeof t&&(t["process"]&&"server"===t["process"].env.VUE_ENV)),q},ut=X&&window.__VUE_DEVTOOLS_GLOBAL_HOOK__;function lt(t){return"function"===typeof t&&/native code/.test(t.toString())}var ft,dt="undefined"!==typeof Symbol&&lt(Symbol)&&"undefined"!==typeof Reflect&&lt(Reflect.ownKeys);ft="undefined"!==typeof Set&&lt(Set)?Set:function(){function t(){this.set=Object.create(null)}return t.prototype.has=function(t){return!0===this.set[t]},t.prototype.add=function(t){this.set[t]=!0},t.prototype.clear=function(){this.set=Object.create(null)},t}();var ht=k,pt=0,yt=function(){this.id=pt++,this.subs=[]};yt.prototype.addSub=function(t){this.subs.push(t)},yt.prototype.removeSub=function(t){m(this.subs,t)},yt.prototype.depend=function(){yt.target&&yt.target.addDep(this)},yt.prototype.notify=function(){var t=this.subs.slice();for(var e=0,n=t.length;e<n;e++)t[e].update()},yt.target=null;var vt=[];function bt(t){vt.push(t),yt.target=t}function mt(){vt.pop(),yt.target=vt[vt.length-1]}var gt=function(t,e,n,r,i,o,s,a){this.tag=t,this.data=e,this.children=n,this.text=r,this.elm=i,this.ns=void 0,this.context=o,this.fnContext=void 0,this.fnOptions=void 0,this.fnScopeId=void 0,this.key=e&&e.key,this.componentOptions=s,this.componentInstance=void 0,this.parent=void 0,this.raw=!1,this.isStatic=!1,this.isRootInsert=!0,this.isComment=!1,this.isCloned=!1,this.isOnce=!1,this.asyncFactory=a,this.asyncMeta=void 0,this.isAsyncPlaceholder=!1},_t={child:{configurable:!0}};_t.child.get=function(){return this.componentInstance},Object.defineProperties(gt.prototype,_t);var wt=function(t){void 0===t&&(t="");var e=new gt;return e.text=t,e.isComment=!0,e};function Ot(t){return new gt(void 0,void 0,void 0,String(t))}function St(t){var e=new gt(t.tag,t.data,t.children&&t.children.slice(),t.text,t.elm,t.context,t.componentOptions,t.asyncFactory);return e.ns=t.ns,e.isStatic=t.isStatic,e.key=t.key,e.isComment=t.isComment,e.fnContext=t.fnContext,e.fnOptions=t.fnOptions,e.fnScopeId=t.fnScopeId,e.asyncMeta=t.asyncMeta,e.isCloned=!0,e}var xt=Array.prototype,It=Object.create(xt),At=["push","pop","shift","unshift","splice","sort","reverse"];At.forEach((function(t){var e=xt[t];H(It,t,(function(){var n=[],r=arguments.length;while(r--)n[r]=arguments[r];var i,o=e.apply(this,n),s=this.__ob__;switch(t){case"push":case"unshift":i=n;break;case"splice":i=n.slice(2);break}return i&&s.observeArray(i),s.dep.notify(),o}))}));var Tt=Object.getOwnPropertyNames(It),jt=!0;function Et(t){jt=t}var Bt=function(t){this.value=t,this.dep=new yt,this.vmCount=0,H(t,"__ob__",this),Array.isArray(t)?(K?Ct(t,It):Mt(t,It,Tt),this.observeArray(t)):this.walk(t)};function Ct(t,e){t.__proto__=e}function Mt(t,e,n){for(var r=0,i=n.length;r<i;r++){var o=n[r];H(t,o,e[o])}}function kt(t,e){var n;if(c(t)&&!(t instanceof gt))return _(t,"__ob__")&&t.__ob__ instanceof Bt?n=t.__ob__:jt&&!ct()&&(Array.isArray(t)||l(t))&&Object.isExtensible(t)&&!t._isVue&&(n=new Bt(t)),e&&n&&n.vmCount++,n}function Dt(t,e,n,r,i){var o=new yt,s=Object.getOwnPropertyDescriptor(t,e);if(!s||!1!==s.configurable){var a=s&&s.get,c=s&&s.set;a&&!c||2!==arguments.length||(n=t[e]);var u=!i&&kt(n);Object.defineProperty(t,e,{enumerable:!0,configurable:!0,get:function(){var e=a?a.call(t):n;return yt.target&&(o.depend(),u&&(u.dep.depend(),Array.isArray(e)&&Pt(e))),e},set:function(e){var r=a?a.call(t):n;e===r||e!==e&&r!==r||a&&!c||(c?c.call(t,e):n=e,u=!i&&kt(e),o.notify())}})}}function Nt(t,e,n){if(Array.isArray(t)&&d(e))return t.length=Math.max(t.length,e),t.splice(e,1,n),n;if(e in t&&!(e in Object.prototype))return t[e]=n,n;var r=t.__ob__;return t._isVue||r&&r.vmCount?n:r?(Dt(r.value,e,n),r.dep.notify(),n):(t[e]=n,n)}function Lt(t,e){if(Array.isArray(t)&&d(e))t.splice(e,1);else{var n=t.__ob__;t._isVue||n&&n.vmCount||_(t,e)&&(delete t[e],n&&n.dep.notify())}}function Pt(t){for(var e=void 0,n=0,r=t.length;n<r;n++)e=t[n],e&&e.__ob__&&e.__ob__.dep.depend(),Array.isArray(e)&&Pt(e)}Bt.prototype.walk=function(t){for(var e=Object.keys(t),n=0;n<e.length;n++)Dt(t,e[n])},Bt.prototype.observeArray=function(t){for(var e=0,n=t.length;e<n;e++)kt(t[e])};var Ft=$.optionMergeStrategies;function Ut(t,e){if(!e)return t;for(var n,r,i,o=dt?Reflect.ownKeys(e):Object.keys(e),s=0;s<o.length;s++)n=o[s],"__ob__"!==n&&(r=t[n],i=e[n],_(t,n)?r!==i&&l(r)&&l(i)&&Ut(r,i):Nt(t,n,i));return t}function Rt(t,e,n){return n?function(){var r="function"===typeof e?e.call(n,n):e,i="function"===typeof t?t.call(n,n):t;return r?Ut(r,i):i}:e?t?function(){return Ut("function"===typeof e?e.call(this,this):e,"function"===typeof t?t.call(this,this):t)}:e:t}function Vt(t,e){var n=e?t?t.concat(e):Array.isArray(e)?e:[e]:t;return n?$t(n):n}function $t(t){for(var e=[],n=0;n<t.length;n++)-1===e.indexOf(t[n])&&e.push(t[n]);return e}function zt(t,e,n,r){var i=Object.create(t||null);return e?C(i,e):i}Ft.data=function(t,e,n){return n?Rt(t,e,n):e&&"function"!==typeof e?t:Rt(t,e)},V.forEach((function(t){Ft[t]=Vt})),R.forEach((function(t){Ft[t+"s"]=zt})),Ft.watch=function(t,e,n,r){if(t===ot&&(t=void 0),e===ot&&(e=void 0),!e)return Object.create(t||null);if(!t)return e;var i={};for(var o in C(i,t),e){var s=i[o],a=e[o];s&&!Array.isArray(s)&&(s=[s]),i[o]=s?s.concat(a):Array.isArray(a)?a:[a]}return i},Ft.props=Ft.methods=Ft.inject=Ft.computed=function(t,e,n,r){if(!t)return e;var i=Object.create(null);return C(i,t),e&&C(i,e),i},Ft.provide=Rt;var Yt=function(t,e){return void 0===e?t:e};function Ht(t,e){var n=t.props;if(n){var r,i,o,s={};if(Array.isArray(n)){r=n.length;while(r--)i=n[r],"string"===typeof i&&(o=S(i),s[o]={type:null})}else if(l(n))for(var a in n)i=n[a],o=S(a),s[o]=l(i)?i:{type:i};else 0;t.props=s}}function Wt(t,e){var n=t.inject;if(n){var r=t.inject={};if(Array.isArray(n))for(var i=0;i<n.length;i++)r[n[i]]={from:n[i]};else if(l(n))for(var o in n){var s=n[o];r[o]=l(s)?C({from:o},s):{from:s}}else 0}}function Gt(t){var e=t.directives;if(e)for(var n in e){var r=e[n];"function"===typeof r&&(e[n]={bind:r,update:r})}}function qt(t,e,n){if("function"===typeof e&&(e=e.options),Ht(e,n),Wt(e,n),Gt(e),!e._base&&(e.extends&&(t=qt(t,e.extends,n)),e.mixins))for(var r=0,i=e.mixins.length;r<i;r++)t=qt(t,e.mixins[r],n);var o,s={};for(o in t)a(o);for(o in e)_(t,o)||a(o);function a(r){var i=Ft[r]||Yt;s[r]=i(t[r],e[r],n,r)}return s}function Kt(t,e,n,r){if("string"===typeof n){var i=t[e];if(_(i,n))return i[n];var o=S(n);if(_(i,o))return i[o];var s=x(o);if(_(i,s))return i[s];var a=i[n]||i[o]||i[s];return a}}function Xt(t,e,n,r){var i=e[t],o=!_(n,t),s=n[t],a=ee(Boolean,i.type);if(a>-1)if(o&&!_(i,"default"))s=!1;else if(""===s||s===A(t)){var c=ee(String,i.type);(c<0||a<c)&&(s=!0)}if(void 0===s){s=Jt(r,i,t);var u=jt;Et(!0),kt(s),Et(u)}return s}function Jt(t,e,n){if(_(e,"default")){var r=e.default;return t&&t.$options.propsData&&void 0===t.$options.propsData[n]&&void 0!==t._props[n]?t._props[n]:"function"===typeof r&&"Function"!==Qt(e.type)?r.call(t):r}}var Zt=/^\s*function (\w+)/;function Qt(t){var e=t&&t.toString().match(Zt);return e?e[1]:""}function te(t,e){return Qt(t)===Qt(e)}function ee(t,e){if(!Array.isArray(e))return te(e,t)?0:-1;for(var n=0,r=e.length;n<r;n++)if(te(e[n],t))return n;return-1}function ne(t,e,n){bt();try{if(e){var r=e;while(r=r.$parent){var i=r.$options.errorCaptured;if(i)for(var o=0;o<i.length;o++)try{var s=!1===i[o].call(r,t,e,n);if(s)return}catch(xs){ie(xs,r,"errorCaptured hook")}}}ie(t,e,n)}finally{mt()}}function re(t,e,n,r,i){var o;try{o=n?t.apply(e,n):t.call(e),o&&!o._isVue&&h(o)&&!o._handled&&(o.catch((function(t){return ne(t,r,i+" (Promise/async)")})),o._handled=!0)}catch(xs){ne(xs,r,i)}return o}function ie(t,e,n){if($.errorHandler)try{return $.errorHandler.call(null,t,e,n)}catch(xs){xs!==t&&oe(xs,null,"config.errorHandler")}oe(t,e,n)}function oe(t,e,n){if(!X&&!J||"undefined"===typeof console)throw t;console.error(t)}var se,ae=!1,ce=[],ue=!1;function le(){ue=!1;var t=ce.slice(0);ce.length=0;for(var e=0;e<t.length;e++)t[e]()}if("undefined"!==typeof Promise&&lt(Promise)){var fe=Promise.resolve();se=function(){fe.then(le),rt&&setTimeout(k)},ae=!0}else if(tt||"undefined"===typeof MutationObserver||!lt(MutationObserver)&&"[object MutationObserverConstructor]"!==MutationObserver.toString())se="undefined"!==typeof setImmediate&&lt(setImmediate)?function(){setImmediate(le)}:function(){setTimeout(le,0)};else{var de=1,he=new MutationObserver(le),pe=document.createTextNode(String(de));he.observe(pe,{characterData:!0}),se=function(){de=(de+1)%2,pe.data=String(de)},ae=!0}function ye(t,e){var n;if(ce.push((function(){if(t)try{t.call(e)}catch(xs){ne(xs,e,"nextTick")}else n&&n(e)})),ue||(ue=!0,se()),!t&&"undefined"!==typeof Promise)return new Promise((function(t){n=t}))}var ve=new ft;function be(t){me(t,ve),ve.clear()}function me(t,e){var n,r,i=Array.isArray(t);if(!(!i&&!c(t)||Object.isFrozen(t)||t instanceof gt)){if(t.__ob__){var o=t.__ob__.dep.id;if(e.has(o))return;e.add(o)}if(i){n=t.length;while(n--)me(t[n],e)}else{r=Object.keys(t),n=r.length;while(n--)me(t[r[n]],e)}}}var ge=w((function(t){var e="&"===t.charAt(0);t=e?t.slice(1):t;var n="~"===t.charAt(0);t=n?t.slice(1):t;var r="!"===t.charAt(0);return t=r?t.slice(1):t,{name:t,once:n,capture:r,passive:e}}));function _e(t,e){function n(){var t=arguments,r=n.fns;if(!Array.isArray(r))return re(r,null,arguments,e,"v-on handler");for(var i=r.slice(),o=0;o<i.length;o++)re(i[o],null,t,e,"v-on handler")}return n.fns=t,n}function we(t,e,n,i,s,a){var c,u,l,f;for(c in t)u=t[c],l=e[c],f=ge(c),r(u)||(r(l)?(r(u.fns)&&(u=t[c]=_e(u,a)),o(f.once)&&(u=t[c]=s(f.name,u,f.capture)),n(f.name,u,f.capture,f.passive,f.params)):u!==l&&(l.fns=u,t[c]=l));for(c in e)r(t[c])&&(f=ge(c),i(f.name,e[c],f.capture))}function Oe(t,e,n){var s;t instanceof gt&&(t=t.data.hook||(t.data.hook={}));var a=t[e];function c(){n.apply(this,arguments),m(s.fns,c)}r(a)?s=_e([c]):i(a.fns)&&o(a.merged)?(s=a,s.fns.push(c)):s=_e([a,c]),s.merged=!0,t[e]=s}function Se(t,e,n){var o=e.options.props;if(!r(o)){var s={},a=t.attrs,c=t.props;if(i(a)||i(c))for(var u in o){var l=A(u);xe(s,c,u,l,!0)||xe(s,a,u,l,!1)}return s}}function xe(t,e,n,r,o){if(i(e)){if(_(e,n))return t[n]=e[n],o||delete e[n],!0;if(_(e,r))return t[n]=e[r],o||delete e[r],!0}return!1}function Ie(t){for(var e=0;e<t.length;e++)if(Array.isArray(t[e]))return Array.prototype.concat.apply([],t);return t}function Ae(t){return a(t)?[Ot(t)]:Array.isArray(t)?je(t):void 0}function Te(t){return i(t)&&i(t.text)&&s(t.isComment)}function je(t,e){var n,s,c,u,l=[];for(n=0;n<t.length;n++)s=t[n],r(s)||"boolean"===typeof s||(c=l.length-1,u=l[c],Array.isArray(s)?s.length>0&&(s=je(s,(e||"")+"_"+n),Te(s[0])&&Te(u)&&(l[c]=Ot(u.text+s[0].text),s.shift()),l.push.apply(l,s)):a(s)?Te(u)?l[c]=Ot(u.text+s):""!==s&&l.push(Ot(s)):Te(s)&&Te(u)?l[c]=Ot(u.text+s.text):(o(t._isVList)&&i(s.tag)&&r(s.key)&&i(e)&&(s.key="__vlist"+e+"_"+n+"__"),l.push(s)));return l}function Ee(t){var e=t.$options.provide;e&&(t._provided="function"===typeof e?e.call(t):e)}function Be(t){var e=Ce(t.$options.inject,t);e&&(Et(!1),Object.keys(e).forEach((function(n){Dt(t,n,e[n])})),Et(!0))}function Ce(t,e){if(t){for(var n=Object.create(null),r=dt?Reflect.ownKeys(t):Object.keys(t),i=0;i<r.length;i++){var o=r[i];if("__ob__"!==o){var s=t[o].from,a=e;while(a){if(a._provided&&_(a._provided,s)){n[o]=a._provided[s];break}a=a.$parent}if(!a)if("default"in t[o]){var c=t[o].default;n[o]="function"===typeof c?c.call(e):c}else 0}}return n}}function Me(t,e){if(!t||!t.length)return{};for(var n={},r=0,i=t.length;r<i;r++){var o=t[r],s=o.data;if(s&&s.attrs&&s.attrs.slot&&delete s.attrs.slot,o.context!==e&&o.fnContext!==e||!s||null==s.slot)(n.default||(n.default=[])).push(o);else{var a=s.slot,c=n[a]||(n[a]=[]);"template"===o.tag?c.push.apply(c,o.children||[]):c.push(o)}}for(var u in n)n[u].every(ke)&&delete n[u];return n}function ke(t){return t.isComment&&!t.asyncFactory||" "===t.text}function De(t){return t.isComment&&t.asyncFactory}function Ne(t,e,r){var i,o=Object.keys(e).length>0,s=t?!!t.$stable:!o,a=t&&t.$key;if(t){if(t._normalized)return t._normalized;if(s&&r&&r!==n&&a===r.$key&&!o&&!r.$hasNormal)return r;for(var c in i={},t)t[c]&&"$"!==c[0]&&(i[c]=Le(e,c,t[c]))}else i={};for(var u in e)u in i||(i[u]=Pe(e,u));return t&&Object.isExtensible(t)&&(t._normalized=i),H(i,"$stable",s),H(i,"$key",a),H(i,"$hasNormal",o),i}function Le(t,e,n){var r=function(){var t=arguments.length?n.apply(null,arguments):n({});t=t&&"object"===typeof t&&!Array.isArray(t)?[t]:Ae(t);var e=t&&t[0];return t&&(!e||1===t.length&&e.isComment&&!De(e))?void 0:t};return n.proxy&&Object.defineProperty(t,e,{get:r,enumerable:!0,configurable:!0}),r}function Pe(t,e){return function(){return t[e]}}function Fe(t,e){var n,r,o,s,a;if(Array.isArray(t)||"string"===typeof t)for(n=new Array(t.length),r=0,o=t.length;r<o;r++)n[r]=e(t[r],r);else if("number"===typeof t)for(n=new Array(t),r=0;r<t;r++)n[r]=e(r+1,r);else if(c(t))if(dt&&t[Symbol.iterator]){n=[];var u=t[Symbol.iterator](),l=u.next();while(!l.done)n.push(e(l.value,n.length)),l=u.next()}else for(s=Object.keys(t),n=new Array(s.length),r=0,o=s.length;r<o;r++)a=s[r],n[r]=e(t[a],a,r);return i(n)||(n=[]),n._isVList=!0,n}function Ue(t,e,n,r){var i,o=this.$scopedSlots[t];o?(n=n||{},r&&(n=C(C({},r),n)),i=o(n)||("function"===typeof e?e():e)):i=this.$slots[t]||("function"===typeof e?e():e);var s=n&&n.slot;return s?this.$createElement("template",{slot:s},i):i}function Re(t){return Kt(this.$options,"filters",t,!0)||N}function Ve(t,e){return Array.isArray(t)?-1===t.indexOf(e):t!==e}function $e(t,e,n,r,i){var o=$.keyCodes[e]||n;return i&&r&&!$.keyCodes[e]?Ve(i,r):o?Ve(o,t):r?A(r)!==e:void 0===t}function ze(t,e,n,r,i){if(n)if(c(n)){var o;Array.isArray(n)&&(n=M(n));var s=function(s){if("class"===s||"style"===s||b(s))o=t;else{var a=t.attrs&&t.attrs.type;o=r||$.mustUseProp(e,a,s)?t.domProps||(t.domProps={}):t.attrs||(t.attrs={})}var c=S(s),u=A(s);if(!(c in o)&&!(u in o)&&(o[s]=n[s],i)){var l=t.on||(t.on={});l["update:"+s]=function(t){n[s]=t}}};for(var a in n)s(a)}else;return t}function Ye(t,e){var n=this._staticTrees||(this._staticTrees=[]),r=n[t];return r&&!e||(r=n[t]=this.$options.staticRenderFns[t].call(this._renderProxy,null,this),We(r,"__static__"+t,!1)),r}function He(t,e,n){return We(t,"__once__"+e+(n?"_"+n:""),!0),t}function We(t,e,n){if(Array.isArray(t))for(var r=0;r<t.length;r++)t[r]&&"string"!==typeof t[r]&&Ge(t[r],e+"_"+r,n);else Ge(t,e,n)}function Ge(t,e,n){t.isStatic=!0,t.key=e,t.isOnce=n}function qe(t,e){if(e)if(l(e)){var n=t.on=t.on?C({},t.on):{};for(var r in e){var i=n[r],o=e[r];n[r]=i?[].concat(i,o):o}}else;return t}function Ke(t,e,n,r){e=e||{$stable:!n};for(var i=0;i<t.length;i++){var o=t[i];Array.isArray(o)?Ke(o,e,n):o&&(o.proxy&&(o.fn.proxy=!0),e[o.key]=o.fn)}return r&&(e.$key=r),e}function Xe(t,e){for(var n=0;n<e.length;n+=2){var r=e[n];"string"===typeof r&&r&&(t[e[n]]=e[n+1])}return t}function Je(t,e){return"string"===typeof t?e+t:t}function Ze(t){t._o=He,t._n=y,t._s=p,t._l=Fe,t._t=Ue,t._q=L,t._i=P,t._m=Ye,t._f=Re,t._k=$e,t._b=ze,t._v=Ot,t._e=wt,t._u=Ke,t._g=qe,t._d=Xe,t._p=Je}function Qe(t,e,r,i,s){var a,c=this,u=s.options;_(i,"_uid")?(a=Object.create(i),a._original=i):(a=i,i=i._original);var l=o(u._compiled),f=!l;this.data=t,this.props=e,this.children=r,this.parent=i,this.listeners=t.on||n,this.injections=Ce(u.inject,i),this.slots=function(){return c.$slots||Ne(t.scopedSlots,c.$slots=Me(r,i)),c.$slots},Object.defineProperty(this,"scopedSlots",{enumerable:!0,get:function(){return Ne(t.scopedSlots,this.slots())}}),l&&(this.$options=u,this.$slots=this.slots(),this.$scopedSlots=Ne(t.scopedSlots,this.$slots)),u._scopeId?this._c=function(t,e,n,r){var o=hn(a,t,e,n,r,f);return o&&!Array.isArray(o)&&(o.fnScopeId=u._scopeId,o.fnContext=i),o}:this._c=function(t,e,n,r){return hn(a,t,e,n,r,f)}}function tn(t,e,r,o,s){var a=t.options,c={},u=a.props;if(i(u))for(var l in u)c[l]=Xt(l,u,e||n);else i(r.attrs)&&nn(c,r.attrs),i(r.props)&&nn(c,r.props);var f=new Qe(r,c,s,o,t),d=a.render.call(null,f._c,f);if(d instanceof gt)return en(d,r,f.parent,a,f);if(Array.isArray(d)){for(var h=Ae(d)||[],p=new Array(h.length),y=0;y<h.length;y++)p[y]=en(h[y],r,f.parent,a,f);return p}}function en(t,e,n,r,i){var o=St(t);return o.fnContext=n,o.fnOptions=r,e.slot&&((o.data||(o.data={})).slot=e.slot),o}function nn(t,e){for(var n in e)t[S(n)]=e[n]}Ze(Qe.prototype);var rn={init:function(t,e){if(t.componentInstance&&!t.componentInstance._isDestroyed&&t.data.keepAlive){var n=t;rn.prepatch(n,n)}else{var r=t.componentInstance=an(t,Cn);r.$mount(e?t.elm:void 0,e)}},prepatch:function(t,e){var n=e.componentOptions,r=e.componentInstance=t.componentInstance;Ln(r,n.propsData,n.listeners,e,n.children)},insert:function(t){var e=t.context,n=t.componentInstance;n._isMounted||(n._isMounted=!0,Rn(n,"mounted")),t.data.keepAlive&&(e._isMounted?Qn(n):Fn(n,!0))},destroy:function(t){var e=t.componentInstance;e._isDestroyed||(t.data.keepAlive?Un(e,!0):e.$destroy())}},on=Object.keys(rn);function sn(t,e,n,s,a){if(!r(t)){var u=n.$options._base;if(c(t)&&(t=u.extend(t)),"function"===typeof t){var l;if(r(t.cid)&&(l=t,t=Sn(l,u),void 0===t))return On(l,e,n,s,a);e=e||{},Or(t),i(e.model)&&ln(t.options,e);var f=Se(e,t,a);if(o(t.options.functional))return tn(t,f,e,n,s);var d=e.on;if(e.on=e.nativeOn,o(t.options.abstract)){var h=e.slot;e={},h&&(e.slot=h)}cn(e);var p=t.options.name||a,y=new gt("vue-component-"+t.cid+(p?"-"+p:""),e,void 0,void 0,void 0,n,{Ctor:t,propsData:f,listeners:d,tag:a,children:s},l);return y}}}function an(t,e){var n={_isComponent:!0,_parentVnode:t,parent:e},r=t.data.inlineTemplate;return i(r)&&(n.render=r.render,n.staticRenderFns=r.staticRenderFns),new t.componentOptions.Ctor(n)}function cn(t){for(var e=t.hook||(t.hook={}),n=0;n<on.length;n++){var r=on[n],i=e[r],o=rn[r];i===o||i&&i._merged||(e[r]=i?un(o,i):o)}}function un(t,e){var n=function(n,r){t(n,r),e(n,r)};return n._merged=!0,n}function ln(t,e){var n=t.model&&t.model.prop||"value",r=t.model&&t.model.event||"input";(e.attrs||(e.attrs={}))[n]=e.model.value;var o=e.on||(e.on={}),s=o[r],a=e.model.callback;i(s)?(Array.isArray(s)?-1===s.indexOf(a):s!==a)&&(o[r]=[a].concat(s)):o[r]=a}var fn=1,dn=2;function hn(t,e,n,r,i,s){return(Array.isArray(n)||a(n))&&(i=r,r=n,n=void 0),o(s)&&(i=dn),pn(t,e,n,r,i)}function pn(t,e,n,r,o){if(i(n)&&i(n.__ob__))return wt();if(i(n)&&i(n.is)&&(e=n.is),!e)return wt();var s,a,c;(Array.isArray(r)&&"function"===typeof r[0]&&(n=n||{},n.scopedSlots={default:r[0]},r.length=0),o===dn?r=Ae(r):o===fn&&(r=Ie(r)),"string"===typeof e)?(a=t.$vnode&&t.$vnode.ns||$.getTagNamespace(e),s=$.isReservedTag(e)?new gt($.parsePlatformTagName(e),n,r,void 0,void 0,t):n&&n.pre||!i(c=Kt(t.$options,"components",e))?new gt(e,n,r,void 0,void 0,t):sn(c,n,t,r,e)):s=sn(e,n,t,r);return Array.isArray(s)?s:i(s)?(i(a)&&yn(s,a),i(n)&&vn(n),s):wt()}function yn(t,e,n){if(t.ns=e,"foreignObject"===t.tag&&(e=void 0,n=!0),i(t.children))for(var s=0,a=t.children.length;s<a;s++){var c=t.children[s];i(c.tag)&&(r(c.ns)||o(n)&&"svg"!==c.tag)&&yn(c,e,n)}}function vn(t){c(t.style)&&be(t.style),c(t.class)&&be(t.class)}function bn(t){t._vnode=null,t._staticTrees=null;var e=t.$options,r=t.$vnode=e._parentVnode,i=r&&r.context;t.$slots=Me(e._renderChildren,i),t.$scopedSlots=n,t._c=function(e,n,r,i){return hn(t,e,n,r,i,!1)},t.$createElement=function(e,n,r,i){return hn(t,e,n,r,i,!0)};var o=r&&r.data;Dt(t,"$attrs",o&&o.attrs||n,null,!0),Dt(t,"$listeners",e._parentListeners||n,null,!0)}var mn,gn=null;function _n(t){Ze(t.prototype),t.prototype.$nextTick=function(t){return ye(t,this)},t.prototype._render=function(){var t,e=this,n=e.$options,r=n.render,i=n._parentVnode;i&&(e.$scopedSlots=Ne(i.data.scopedSlots,e.$slots,e.$scopedSlots)),e.$vnode=i;try{gn=e,t=r.call(e._renderProxy,e.$createElement)}catch(xs){ne(xs,e,"render"),t=e._vnode}finally{gn=null}return Array.isArray(t)&&1===t.length&&(t=t[0]),t instanceof gt||(t=wt()),t.parent=i,t}}function wn(t,e){return(t.__esModule||dt&&"Module"===t[Symbol.toStringTag])&&(t=t.default),c(t)?e.extend(t):t}function On(t,e,n,r,i){var o=wt();return o.asyncFactory=t,o.asyncMeta={data:e,context:n,children:r,tag:i},o}function Sn(t,e){if(o(t.error)&&i(t.errorComp))return t.errorComp;if(i(t.resolved))return t.resolved;var n=gn;if(n&&i(t.owners)&&-1===t.owners.indexOf(n)&&t.owners.push(n),o(t.loading)&&i(t.loadingComp))return t.loadingComp;if(n&&!i(t.owners)){var s=t.owners=[n],a=!0,u=null,l=null;n.$on("hook:destroyed",(function(){return m(s,n)}));var f=function(t){for(var e=0,n=s.length;e<n;e++)s[e].$forceUpdate();t&&(s.length=0,null!==u&&(clearTimeout(u),u=null),null!==l&&(clearTimeout(l),l=null))},d=F((function(n){t.resolved=wn(n,e),a?s.length=0:f(!0)})),p=F((function(e){i(t.errorComp)&&(t.error=!0,f(!0))})),y=t(d,p);return c(y)&&(h(y)?r(t.resolved)&&y.then(d,p):h(y.component)&&(y.component.then(d,p),i(y.error)&&(t.errorComp=wn(y.error,e)),i(y.loading)&&(t.loadingComp=wn(y.loading,e),0===y.delay?t.loading=!0:u=setTimeout((function(){u=null,r(t.resolved)&&r(t.error)&&(t.loading=!0,f(!1))}),y.delay||200)),i(y.timeout)&&(l=setTimeout((function(){l=null,r(t.resolved)&&p(null)}),y.timeout)))),a=!1,t.loading?t.loadingComp:t.resolved}}function xn(t){if(Array.isArray(t))for(var e=0;e<t.length;e++){var n=t[e];if(i(n)&&(i(n.componentOptions)||De(n)))return n}}function In(t){t._events=Object.create(null),t._hasHookEvent=!1;var e=t.$options._parentListeners;e&&En(t,e)}function An(t,e){mn.$on(t,e)}function Tn(t,e){mn.$off(t,e)}function jn(t,e){var n=mn;return function r(){var i=e.apply(null,arguments);null!==i&&n.$off(t,r)}}function En(t,e,n){mn=t,we(e,n||{},An,Tn,jn,t),mn=void 0}function Bn(t){var e=/^hook:/;t.prototype.$on=function(t,n){var r=this;if(Array.isArray(t))for(var i=0,o=t.length;i<o;i++)r.$on(t[i],n);else(r._events[t]||(r._events[t]=[])).push(n),e.test(t)&&(r._hasHookEvent=!0);return r},t.prototype.$once=function(t,e){var n=this;function r(){n.$off(t,r),e.apply(n,arguments)}return r.fn=e,n.$on(t,r),n},t.prototype.$off=function(t,e){var n=this;if(!arguments.length)return n._events=Object.create(null),n;if(Array.isArray(t)){for(var r=0,i=t.length;r<i;r++)n.$off(t[r],e);return n}var o,s=n._events[t];if(!s)return n;if(!e)return n._events[t]=null,n;var a=s.length;while(a--)if(o=s[a],o===e||o.fn===e){s.splice(a,1);break}return n},t.prototype.$emit=function(t){var e=this,n=e._events[t];if(n){n=n.length>1?B(n):n;for(var r=B(arguments,1),i='event handler for "'+t+'"',o=0,s=n.length;o<s;o++)re(n[o],e,r,e,i)}return e}}var Cn=null;function Mn(t){var e=Cn;return Cn=t,function(){Cn=e}}function kn(t){var e=t.$options,n=e.parent;if(n&&!e.abstract){while(n.$options.abstract&&n.$parent)n=n.$parent;n.$children.push(t)}t.$parent=n,t.$root=n?n.$root:t,t.$children=[],t.$refs={},t._watcher=null,t._inactive=null,t._directInactive=!1,t._isMounted=!1,t._isDestroyed=!1,t._isBeingDestroyed=!1}function Dn(t){t.prototype._update=function(t,e){var n=this,r=n.$el,i=n._vnode,o=Mn(n);n._vnode=t,n.$el=i?n.__patch__(i,t):n.__patch__(n.$el,t,e,!1),o(),r&&(r.__vue__=null),n.$el&&(n.$el.__vue__=n),n.$vnode&&n.$parent&&n.$vnode===n.$parent._vnode&&(n.$parent.$el=n.$el)},t.prototype.$forceUpdate=function(){var t=this;t._watcher&&t._watcher.update()},t.prototype.$destroy=function(){var t=this;if(!t._isBeingDestroyed){Rn(t,"beforeDestroy"),t._isBeingDestroyed=!0;var e=t.$parent;!e||e._isBeingDestroyed||t.$options.abstract||m(e.$children,t),t._watcher&&t._watcher.teardown();var n=t._watchers.length;while(n--)t._watchers[n].teardown();t._data.__ob__&&t._data.__ob__.vmCount--,t._isDestroyed=!0,t.__patch__(t._vnode,null),Rn(t,"destroyed"),t.$off(),t.$el&&(t.$el.__vue__=null),t.$vnode&&(t.$vnode.parent=null)}}}function Nn(t,e,n){var r;return t.$el=e,t.$options.render||(t.$options.render=wt),Rn(t,"beforeMount"),r=function(){t._update(t._render(),n)},new rr(t,r,k,{before:function(){t._isMounted&&!t._isDestroyed&&Rn(t,"beforeUpdate")}},!0),n=!1,null==t.$vnode&&(t._isMounted=!0,Rn(t,"mounted")),t}function Ln(t,e,r,i,o){var s=i.data.scopedSlots,a=t.$scopedSlots,c=!!(s&&!s.$stable||a!==n&&!a.$stable||s&&t.$scopedSlots.$key!==s.$key||!s&&t.$scopedSlots.$key),u=!!(o||t.$options._renderChildren||c);if(t.$options._parentVnode=i,t.$vnode=i,t._vnode&&(t._vnode.parent=i),t.$options._renderChildren=o,t.$attrs=i.data.attrs||n,t.$listeners=r||n,e&&t.$options.props){Et(!1);for(var l=t._props,f=t.$options._propKeys||[],d=0;d<f.length;d++){var h=f[d],p=t.$options.props;l[h]=Xt(h,p,e,t)}Et(!0),t.$options.propsData=e}r=r||n;var y=t.$options._parentListeners;t.$options._parentListeners=r,En(t,r,y),u&&(t.$slots=Me(o,i.context),t.$forceUpdate())}function Pn(t){while(t&&(t=t.$parent))if(t._inactive)return!0;return!1}function Fn(t,e){if(e){if(t._directInactive=!1,Pn(t))return}else if(t._directInactive)return;if(t._inactive||null===t._inactive){t._inactive=!1;for(var n=0;n<t.$children.length;n++)Fn(t.$children[n]);Rn(t,"activated")}}function Un(t,e){if((!e||(t._directInactive=!0,!Pn(t)))&&!t._inactive){t._inactive=!0;for(var n=0;n<t.$children.length;n++)Un(t.$children[n]);Rn(t,"deactivated")}}function Rn(t,e){bt();var n=t.$options[e],r=e+" hook";if(n)for(var i=0,o=n.length;i<o;i++)re(n[i],t,null,t,r);t._hasHookEvent&&t.$emit("hook:"+e),mt()}var Vn=[],$n=[],zn={},Yn=!1,Hn=!1,Wn=0;function Gn(){Wn=Vn.length=$n.length=0,zn={},Yn=Hn=!1}var qn=0,Kn=Date.now;if(X&&!tt){var Xn=window.performance;Xn&&"function"===typeof Xn.now&&Kn()>document.createEvent("Event").timeStamp&&(Kn=function(){return Xn.now()})}function Jn(){var t,e;for(qn=Kn(),Hn=!0,Vn.sort((function(t,e){return t.id-e.id})),Wn=0;Wn<Vn.length;Wn++)t=Vn[Wn],t.before&&t.before(),e=t.id,zn[e]=null,t.run();var n=$n.slice(),r=Vn.slice();Gn(),tr(n),Zn(r),ut&&$.devtools&&ut.emit("flush")}function Zn(t){var e=t.length;while(e--){var n=t[e],r=n.vm;r._watcher===n&&r._isMounted&&!r._isDestroyed&&Rn(r,"updated")}}function Qn(t){t._inactive=!1,$n.push(t)}function tr(t){for(var e=0;e<t.length;e++)t[e]._inactive=!0,Fn(t[e],!0)}function er(t){var e=t.id;if(null==zn[e]){if(zn[e]=!0,Hn){var n=Vn.length-1;while(n>Wn&&Vn[n].id>t.id)n--;Vn.splice(n+1,0,t)}else Vn.push(t);Yn||(Yn=!0,ye(Jn))}}var nr=0,rr=function(t,e,n,r,i){this.vm=t,i&&(t._watcher=this),t._watchers.push(this),r?(this.deep=!!r.deep,this.user=!!r.user,this.lazy=!!r.lazy,this.sync=!!r.sync,this.before=r.before):this.deep=this.user=this.lazy=this.sync=!1,this.cb=n,this.id=++nr,this.active=!0,this.dirty=this.lazy,this.deps=[],this.newDeps=[],this.depIds=new ft,this.newDepIds=new ft,this.expression="","function"===typeof e?this.getter=e:(this.getter=G(e),this.getter||(this.getter=k)),this.value=this.lazy?void 0:this.get()};rr.prototype.get=function(){var t;bt(this);var e=this.vm;try{t=this.getter.call(e,e)}catch(xs){if(!this.user)throw xs;ne(xs,e,'getter for watcher "'+this.expression+'"')}finally{this.deep&&be(t),mt(),this.cleanupDeps()}return t},rr.prototype.addDep=function(t){var e=t.id;this.newDepIds.has(e)||(this.newDepIds.add(e),this.newDeps.push(t),this.depIds.has(e)||t.addSub(this))},rr.prototype.cleanupDeps=function(){var t=this.deps.length;while(t--){var e=this.deps[t];this.newDepIds.has(e.id)||e.removeSub(this)}var n=this.depIds;this.depIds=this.newDepIds,this.newDepIds=n,this.newDepIds.clear(),n=this.deps,this.deps=this.newDeps,this.newDeps=n,this.newDeps.length=0},rr.prototype.update=function(){this.lazy?this.dirty=!0:this.sync?this.run():er(this)},rr.prototype.run=function(){if(this.active){var t=this.get();if(t!==this.value||c(t)||this.deep){var e=this.value;if(this.value=t,this.user){var n='callback for watcher "'+this.expression+'"';re(this.cb,this.vm,[t,e],this.vm,n)}else this.cb.call(this.vm,t,e)}}},rr.prototype.evaluate=function(){this.value=this.get(),this.dirty=!1},rr.prototype.depend=function(){var t=this.deps.length;while(t--)this.deps[t].depend()},rr.prototype.teardown=function(){if(this.active){this.vm._isBeingDestroyed||m(this.vm._watchers,this);var t=this.deps.length;while(t--)this.deps[t].removeSub(this);this.active=!1}};var ir={enumerable:!0,configurable:!0,get:k,set:k};function or(t,e,n){ir.get=function(){return this[e][n]},ir.set=function(t){this[e][n]=t},Object.defineProperty(t,n,ir)}function sr(t){t._watchers=[];var e=t.$options;e.props&&ar(t,e.props),e.methods&&yr(t,e.methods),e.data?cr(t):kt(t._data={},!0),e.computed&&fr(t,e.computed),e.watch&&e.watch!==ot&&vr(t,e.watch)}function ar(t,e){var n=t.$options.propsData||{},r=t._props={},i=t.$options._propKeys=[],o=!t.$parent;o||Et(!1);var s=function(o){i.push(o);var s=Xt(o,e,n,t);Dt(r,o,s),o in t||or(t,"_props",o)};for(var a in e)s(a);Et(!0)}function cr(t){var e=t.$options.data;e=t._data="function"===typeof e?ur(e,t):e||{},l(e)||(e={});var n=Object.keys(e),r=t.$options.props,i=(t.$options.methods,n.length);while(i--){var o=n[i];0,r&&_(r,o)||Y(o)||or(t,"_data",o)}kt(e,!0)}function ur(t,e){bt();try{return t.call(e,e)}catch(xs){return ne(xs,e,"data()"),{}}finally{mt()}}var lr={lazy:!0};function fr(t,e){var n=t._computedWatchers=Object.create(null),r=ct();for(var i in e){var o=e[i],s="function"===typeof o?o:o.get;0,r||(n[i]=new rr(t,s||k,k,lr)),i in t||dr(t,i,o)}}function dr(t,e,n){var r=!ct();"function"===typeof n?(ir.get=r?hr(e):pr(n),ir.set=k):(ir.get=n.get?r&&!1!==n.cache?hr(e):pr(n.get):k,ir.set=n.set||k),Object.defineProperty(t,e,ir)}function hr(t){return function(){var e=this._computedWatchers&&this._computedWatchers[t];if(e)return e.dirty&&e.evaluate(),yt.target&&e.depend(),e.value}}function pr(t){return function(){return t.call(this,this)}}function yr(t,e){t.$options.props;for(var n in e)t[n]="function"!==typeof e[n]?k:E(e[n],t)}function vr(t,e){for(var n in e){var r=e[n];if(Array.isArray(r))for(var i=0;i<r.length;i++)br(t,n,r[i]);else br(t,n,r)}}function br(t,e,n,r){return l(n)&&(r=n,n=n.handler),"string"===typeof n&&(n=t[n]),t.$watch(e,n,r)}function mr(t){var e={get:function(){return this._data}},n={get:function(){return this._props}};Object.defineProperty(t.prototype,"$data",e),Object.defineProperty(t.prototype,"$props",n),t.prototype.$set=Nt,t.prototype.$delete=Lt,t.prototype.$watch=function(t,e,n){var r=this;if(l(e))return br(r,t,e,n);n=n||{},n.user=!0;var i=new rr(r,t,e,n);if(n.immediate){var o='callback for immediate watcher "'+i.expression+'"';bt(),re(e,r,[i.value],r,o),mt()}return function(){i.teardown()}}}var gr=0;function _r(t){t.prototype._init=function(t){var e=this;e._uid=gr++,e._isVue=!0,t&&t._isComponent?wr(e,t):e.$options=qt(Or(e.constructor),t||{},e),e._renderProxy=e,e._self=e,kn(e),In(e),bn(e),Rn(e,"beforeCreate"),Be(e),sr(e),Ee(e),Rn(e,"created"),e.$options.el&&e.$mount(e.$options.el)}}function wr(t,e){var n=t.$options=Object.create(t.constructor.options),r=e._parentVnode;n.parent=e.parent,n._parentVnode=r;var i=r.componentOptions;n.propsData=i.propsData,n._parentListeners=i.listeners,n._renderChildren=i.children,n._componentTag=i.tag,e.render&&(n.render=e.render,n.staticRenderFns=e.staticRenderFns)}function Or(t){var e=t.options;if(t.super){var n=Or(t.super),r=t.superOptions;if(n!==r){t.superOptions=n;var i=Sr(t);i&&C(t.extendOptions,i),e=t.options=qt(n,t.extendOptions),e.name&&(e.components[e.name]=t)}}return e}function Sr(t){var e,n=t.options,r=t.sealedOptions;for(var i in n)n[i]!==r[i]&&(e||(e={}),e[i]=n[i]);return e}function xr(t){this._init(t)}function Ir(t){t.use=function(t){var e=this._installedPlugins||(this._installedPlugins=[]);if(e.indexOf(t)>-1)return this;var n=B(arguments,1);return n.unshift(this),"function"===typeof t.install?t.install.apply(t,n):"function"===typeof t&&t.apply(null,n),e.push(t),this}}function Ar(t){t.mixin=function(t){return this.options=qt(this.options,t),this}}function Tr(t){t.cid=0;var e=1;t.extend=function(t){t=t||{};var n=this,r=n.cid,i=t._Ctor||(t._Ctor={});if(i[r])return i[r];var o=t.name||n.options.name;var s=function(t){this._init(t)};return s.prototype=Object.create(n.prototype),s.prototype.constructor=s,s.cid=e++,s.options=qt(n.options,t),s["super"]=n,s.options.props&&jr(s),s.options.computed&&Er(s),s.extend=n.extend,s.mixin=n.mixin,s.use=n.use,R.forEach((function(t){s[t]=n[t]})),o&&(s.options.components[o]=s),s.superOptions=n.options,s.extendOptions=t,s.sealedOptions=C({},s.options),i[r]=s,s}}function jr(t){var e=t.options.props;for(var n in e)or(t.prototype,"_props",n)}function Er(t){var e=t.options.computed;for(var n in e)dr(t.prototype,n,e[n])}function Br(t){R.forEach((function(e){t[e]=function(t,n){return n?("component"===e&&l(n)&&(n.name=n.name||t,n=this.options._base.extend(n)),"directive"===e&&"function"===typeof n&&(n={bind:n,update:n}),this.options[e+"s"][t]=n,n):this.options[e+"s"][t]}}))}function Cr(t){return t&&(t.Ctor.options.name||t.tag)}function Mr(t,e){return Array.isArray(t)?t.indexOf(e)>-1:"string"===typeof t?t.split(",").indexOf(e)>-1:!!f(t)&&t.test(e)}function kr(t,e){var n=t.cache,r=t.keys,i=t._vnode;for(var o in n){var s=n[o];if(s){var a=s.name;a&&!e(a)&&Dr(n,o,r,i)}}}function Dr(t,e,n,r){var i=t[e];!i||r&&i.tag===r.tag||i.componentInstance.$destroy(),t[e]=null,m(n,e)}_r(xr),mr(xr),Bn(xr),Dn(xr),_n(xr);var Nr=[String,RegExp,Array],Lr={name:"keep-alive",abstract:!0,props:{include:Nr,exclude:Nr,max:[String,Number]},methods:{cacheVNode:function(){var t=this,e=t.cache,n=t.keys,r=t.vnodeToCache,i=t.keyToCache;if(r){var o=r.tag,s=r.componentInstance,a=r.componentOptions;e[i]={name:Cr(a),tag:o,componentInstance:s},n.push(i),this.max&&n.length>parseInt(this.max)&&Dr(e,n[0],n,this._vnode),this.vnodeToCache=null}}},created:function(){this.cache=Object.create(null),this.keys=[]},destroyed:function(){for(var t in this.cache)Dr(this.cache,t,this.keys)},mounted:function(){var t=this;this.cacheVNode(),this.$watch("include",(function(e){kr(t,(function(t){return Mr(e,t)}))})),this.$watch("exclude",(function(e){kr(t,(function(t){return!Mr(e,t)}))}))},updated:function(){this.cacheVNode()},render:function(){var t=this.$slots.default,e=xn(t),n=e&&e.componentOptions;if(n){var r=Cr(n),i=this,o=i.include,s=i.exclude;if(o&&(!r||!Mr(o,r))||s&&r&&Mr(s,r))return e;var a=this,c=a.cache,u=a.keys,l=null==e.key?n.Ctor.cid+(n.tag?"::"+n.tag:""):e.key;c[l]?(e.componentInstance=c[l].componentInstance,m(u,l),u.push(l)):(this.vnodeToCache=e,this.keyToCache=l),e.data.keepAlive=!0}return e||t&&t[0]}},Pr={KeepAlive:Lr};function Fr(t){var e={get:function(){return $}};Object.defineProperty(t,"config",e),t.util={warn:ht,extend:C,mergeOptions:qt,defineReactive:Dt},t.set=Nt,t.delete=Lt,t.nextTick=ye,t.observable=function(t){return kt(t),t},t.options=Object.create(null),R.forEach((function(e){t.options[e+"s"]=Object.create(null)})),t.options._base=t,C(t.options.components,Pr),Ir(t),Ar(t),Tr(t),Br(t)}Fr(xr),Object.defineProperty(xr.prototype,"$isServer",{get:ct}),Object.defineProperty(xr.prototype,"$ssrContext",{get:function(){return this.$vnode&&this.$vnode.ssrContext}}),Object.defineProperty(xr,"FunctionalRenderContext",{value:Qe}),xr.version="2.6.14";var Ur=v("style,class"),Rr=v("input,textarea,option,select,progress"),Vr=function(t,e,n){return"value"===n&&Rr(t)&&"button"!==e||"selected"===n&&"option"===t||"checked"===n&&"input"===t||"muted"===n&&"video"===t},$r=v("contenteditable,draggable,spellcheck"),zr=v("events,caret,typing,plaintext-only"),Yr=function(t,e){return Kr(e)||"false"===e?"false":"contenteditable"===t&&zr(e)?e:"true"},Hr=v("allowfullscreen,async,autofocus,autoplay,checked,compact,controls,declare,default,defaultchecked,defaultmuted,defaultselected,defer,disabled,enabled,formnovalidate,hidden,indeterminate,inert,ismap,itemscope,loop,multiple,muted,nohref,noresize,noshade,novalidate,nowrap,open,pauseonexit,readonly,required,reversed,scoped,seamless,selected,sortable,truespeed,typemustmatch,visible"),Wr="http://www.w3.org/1999/xlink",Gr=function(t){return":"===t.charAt(5)&&"xlink"===t.slice(0,5)},qr=function(t){return Gr(t)?t.slice(6,t.length):""},Kr=function(t){return null==t||!1===t};function Xr(t){var e=t.data,n=t,r=t;while(i(r.componentInstance))r=r.componentInstance._vnode,r&&r.data&&(e=Jr(r.data,e));while(i(n=n.parent))n&&n.data&&(e=Jr(e,n.data));return Zr(e.staticClass,e.class)}function Jr(t,e){return{staticClass:Qr(t.staticClass,e.staticClass),class:i(t.class)?[t.class,e.class]:e.class}}function Zr(t,e){return i(t)||i(e)?Qr(t,ti(e)):""}function Qr(t,e){return t?e?t+" "+e:t:e||""}function ti(t){return Array.isArray(t)?ei(t):c(t)?ni(t):"string"===typeof t?t:""}function ei(t){for(var e,n="",r=0,o=t.length;r<o;r++)i(e=ti(t[r]))&&""!==e&&(n&&(n+=" "),n+=e);return n}function ni(t){var e="";for(var n in t)t[n]&&(e&&(e+=" "),e+=n);return e}var ri={svg:"http://www.w3.org/2000/svg",math:"http://www.w3.org/1998/Math/MathML"},ii=v("html,body,base,head,link,meta,style,title,address,article,aside,footer,header,h1,h2,h3,h4,h5,h6,hgroup,nav,section,div,dd,dl,dt,figcaption,figure,picture,hr,img,li,main,ol,p,pre,ul,a,b,abbr,bdi,bdo,br,cite,code,data,dfn,em,i,kbd,mark,q,rp,rt,rtc,ruby,s,samp,small,span,strong,sub,sup,time,u,var,wbr,area,audio,map,track,video,embed,object,param,source,canvas,script,noscript,del,ins,caption,col,colgroup,table,thead,tbody,td,th,tr,button,datalist,fieldset,form,input,label,legend,meter,optgroup,option,output,progress,select,textarea,details,dialog,menu,menuitem,summary,content,element,shadow,template,blockquote,iframe,tfoot"),oi=v("svg,animate,circle,clippath,cursor,defs,desc,ellipse,filter,font-face,foreignobject,g,glyph,image,line,marker,mask,missing-glyph,path,pattern,polygon,polyline,rect,switch,symbol,text,textpath,tspan,use,view",!0),si=function(t){return ii(t)||oi(t)};function ai(t){return oi(t)?"svg":"math"===t?"math":void 0}var ci=Object.create(null);function ui(t){if(!X)return!0;if(si(t))return!1;if(t=t.toLowerCase(),null!=ci[t])return ci[t];var e=document.createElement(t);return t.indexOf("-")>-1?ci[t]=e.constructor===window.HTMLUnknownElement||e.constructor===window.HTMLElement:ci[t]=/HTMLUnknownElement/.test(e.toString())}var li=v("text,number,password,search,email,tel,url");function fi(t){if("string"===typeof t){var e=document.querySelector(t);return e||document.createElement("div")}return t}function di(t,e){var n=document.createElement(t);return"select"!==t||e.data&&e.data.attrs&&void 0!==e.data.attrs.multiple&&n.setAttribute("multiple","multiple"),n}function hi(t,e){return document.createElementNS(ri[t],e)}function pi(t){return document.createTextNode(t)}function yi(t){return document.createComment(t)}function vi(t,e,n){t.insertBefore(e,n)}function bi(t,e){t.removeChild(e)}function mi(t,e){t.appendChild(e)}function gi(t){return t.parentNode}function _i(t){return t.nextSibling}function wi(t){return t.tagName}function Oi(t,e){t.textContent=e}function Si(t,e){t.setAttribute(e,"")}var xi=Object.freeze({createElement:di,createElementNS:hi,createTextNode:pi,createComment:yi,insertBefore:vi,removeChild:bi,appendChild:mi,parentNode:gi,nextSibling:_i,tagName:wi,setTextContent:Oi,setStyleScope:Si}),Ii={create:function(t,e){Ai(e)},update:function(t,e){t.data.ref!==e.data.ref&&(Ai(t,!0),Ai(e))},destroy:function(t){Ai(t,!0)}};function Ai(t,e){var n=t.data.ref;if(i(n)){var r=t.context,o=t.componentInstance||t.elm,s=r.$refs;e?Array.isArray(s[n])?m(s[n],o):s[n]===o&&(s[n]=void 0):t.data.refInFor?Array.isArray(s[n])?s[n].indexOf(o)<0&&s[n].push(o):s[n]=[o]:s[n]=o}}var Ti=new gt("",{},[]),ji=["create","activate","update","remove","destroy"];function Ei(t,e){return t.key===e.key&&t.asyncFactory===e.asyncFactory&&(t.tag===e.tag&&t.isComment===e.isComment&&i(t.data)===i(e.data)&&Bi(t,e)||o(t.isAsyncPlaceholder)&&r(e.asyncFactory.error))}function Bi(t,e){if("input"!==t.tag)return!0;var n,r=i(n=t.data)&&i(n=n.attrs)&&n.type,o=i(n=e.data)&&i(n=n.attrs)&&n.type;return r===o||li(r)&&li(o)}function Ci(t,e,n){var r,o,s={};for(r=e;r<=n;++r)o=t[r].key,i(o)&&(s[o]=r);return s}function Mi(t){var e,n,s={},c=t.modules,u=t.nodeOps;for(e=0;e<ji.length;++e)for(s[ji[e]]=[],n=0;n<c.length;++n)i(c[n][ji[e]])&&s[ji[e]].push(c[n][ji[e]]);function l(t){return new gt(u.tagName(t).toLowerCase(),{},[],void 0,t)}function f(t,e){function n(){0===--n.listeners&&d(t)}return n.listeners=e,n}function d(t){var e=u.parentNode(t);i(e)&&u.removeChild(e,t)}function h(t,e,n,r,s,a,c){if(i(t.elm)&&i(a)&&(t=a[c]=St(t)),t.isRootInsert=!s,!p(t,e,n,r)){var l=t.data,f=t.children,d=t.tag;i(d)?(t.elm=t.ns?u.createElementNS(t.ns,d):u.createElement(d,t),O(t),g(t,f,e),i(l)&&w(t,e),m(n,t.elm,r)):o(t.isComment)?(t.elm=u.createComment(t.text),m(n,t.elm,r)):(t.elm=u.createTextNode(t.text),m(n,t.elm,r))}}function p(t,e,n,r){var s=t.data;if(i(s)){var a=i(t.componentInstance)&&s.keepAlive;if(i(s=s.hook)&&i(s=s.init)&&s(t,!1),i(t.componentInstance))return y(t,e),m(n,t.elm,r),o(a)&&b(t,e,n,r),!0}}function y(t,e){i(t.data.pendingInsert)&&(e.push.apply(e,t.data.pendingInsert),t.data.pendingInsert=null),t.elm=t.componentInstance.$el,_(t)?(w(t,e),O(t)):(Ai(t),e.push(t))}function b(t,e,n,r){var o,a=t;while(a.componentInstance)if(a=a.componentInstance._vnode,i(o=a.data)&&i(o=o.transition)){for(o=0;o<s.activate.length;++o)s.activate[o](Ti,a);e.push(a);break}m(n,t.elm,r)}function m(t,e,n){i(t)&&(i(n)?u.parentNode(n)===t&&u.insertBefore(t,e,n):u.appendChild(t,e))}function g(t,e,n){if(Array.isArray(e)){0;for(var r=0;r<e.length;++r)h(e[r],n,t.elm,null,!0,e,r)}else a(t.text)&&u.appendChild(t.elm,u.createTextNode(String(t.text)))}function _(t){while(t.componentInstance)t=t.componentInstance._vnode;return i(t.tag)}function w(t,n){for(var r=0;r<s.create.length;++r)s.create[r](Ti,t);e=t.data.hook,i(e)&&(i(e.create)&&e.create(Ti,t),i(e.insert)&&n.push(t))}function O(t){var e;if(i(e=t.fnScopeId))u.setStyleScope(t.elm,e);else{var n=t;while(n)i(e=n.context)&&i(e=e.$options._scopeId)&&u.setStyleScope(t.elm,e),n=n.parent}i(e=Cn)&&e!==t.context&&e!==t.fnContext&&i(e=e.$options._scopeId)&&u.setStyleScope(t.elm,e)}function S(t,e,n,r,i,o){for(;r<=i;++r)h(n[r],o,t,e,!1,n,r)}function x(t){var e,n,r=t.data;if(i(r))for(i(e=r.hook)&&i(e=e.destroy)&&e(t),e=0;e<s.destroy.length;++e)s.destroy[e](t);if(i(e=t.children))for(n=0;n<t.children.length;++n)x(t.children[n])}function I(t,e,n){for(;e<=n;++e){var r=t[e];i(r)&&(i(r.tag)?(A(r),x(r)):d(r.elm))}}function A(t,e){if(i(e)||i(t.data)){var n,r=s.remove.length+1;for(i(e)?e.listeners+=r:e=f(t.elm,r),i(n=t.componentInstance)&&i(n=n._vnode)&&i(n.data)&&A(n,e),n=0;n<s.remove.length;++n)s.remove[n](t,e);i(n=t.data.hook)&&i(n=n.remove)?n(t,e):e()}else d(t.elm)}function T(t,e,n,o,s){var a,c,l,f,d=0,p=0,y=e.length-1,v=e[0],b=e[y],m=n.length-1,g=n[0],_=n[m],w=!s;while(d<=y&&p<=m)r(v)?v=e[++d]:r(b)?b=e[--y]:Ei(v,g)?(E(v,g,o,n,p),v=e[++d],g=n[++p]):Ei(b,_)?(E(b,_,o,n,m),b=e[--y],_=n[--m]):Ei(v,_)?(E(v,_,o,n,m),w&&u.insertBefore(t,v.elm,u.nextSibling(b.elm)),v=e[++d],_=n[--m]):Ei(b,g)?(E(b,g,o,n,p),w&&u.insertBefore(t,b.elm,v.elm),b=e[--y],g=n[++p]):(r(a)&&(a=Ci(e,d,y)),c=i(g.key)?a[g.key]:j(g,e,d,y),r(c)?h(g,o,t,v.elm,!1,n,p):(l=e[c],Ei(l,g)?(E(l,g,o,n,p),e[c]=void 0,w&&u.insertBefore(t,l.elm,v.elm)):h(g,o,t,v.elm,!1,n,p)),g=n[++p]);d>y?(f=r(n[m+1])?null:n[m+1].elm,S(t,f,n,p,m,o)):p>m&&I(e,d,y)}function j(t,e,n,r){for(var o=n;o<r;o++){var s=e[o];if(i(s)&&Ei(t,s))return o}}function E(t,e,n,a,c,l){if(t!==e){i(e.elm)&&i(a)&&(e=a[c]=St(e));var f=e.elm=t.elm;if(o(t.isAsyncPlaceholder))i(e.asyncFactory.resolved)?M(t.elm,e,n):e.isAsyncPlaceholder=!0;else if(o(e.isStatic)&&o(t.isStatic)&&e.key===t.key&&(o(e.isCloned)||o(e.isOnce)))e.componentInstance=t.componentInstance;else{var d,h=e.data;i(h)&&i(d=h.hook)&&i(d=d.prepatch)&&d(t,e);var p=t.children,y=e.children;if(i(h)&&_(e)){for(d=0;d<s.update.length;++d)s.update[d](t,e);i(d=h.hook)&&i(d=d.update)&&d(t,e)}r(e.text)?i(p)&&i(y)?p!==y&&T(f,p,y,n,l):i(y)?(i(t.text)&&u.setTextContent(f,""),S(f,null,y,0,y.length-1,n)):i(p)?I(p,0,p.length-1):i(t.text)&&u.setTextContent(f,""):t.text!==e.text&&u.setTextContent(f,e.text),i(h)&&i(d=h.hook)&&i(d=d.postpatch)&&d(t,e)}}}function B(t,e,n){if(o(n)&&i(t.parent))t.parent.data.pendingInsert=e;else for(var r=0;r<e.length;++r)e[r].data.hook.insert(e[r])}var C=v("attrs,class,staticClass,staticStyle,key");function M(t,e,n,r){var s,a=e.tag,c=e.data,u=e.children;if(r=r||c&&c.pre,e.elm=t,o(e.isComment)&&i(e.asyncFactory))return e.isAsyncPlaceholder=!0,!0;if(i(c)&&(i(s=c.hook)&&i(s=s.init)&&s(e,!0),i(s=e.componentInstance)))return y(e,n),!0;if(i(a)){if(i(u))if(t.hasChildNodes())if(i(s=c)&&i(s=s.domProps)&&i(s=s.innerHTML)){if(s!==t.innerHTML)return!1}else{for(var l=!0,f=t.firstChild,d=0;d<u.length;d++){if(!f||!M(f,u[d],n,r)){l=!1;break}f=f.nextSibling}if(!l||f)return!1}else g(e,u,n);if(i(c)){var h=!1;for(var p in c)if(!C(p)){h=!0,w(e,n);break}!h&&c["class"]&&be(c["class"])}}else t.data!==e.text&&(t.data=e.text);return!0}return function(t,e,n,a){if(!r(e)){var c=!1,f=[];if(r(t))c=!0,h(e,f);else{var d=i(t.nodeType);if(!d&&Ei(t,e))E(t,e,f,null,null,a);else{if(d){if(1===t.nodeType&&t.hasAttribute(U)&&(t.removeAttribute(U),n=!0),o(n)&&M(t,e,f))return B(e,f,!0),t;t=l(t)}var p=t.elm,y=u.parentNode(p);if(h(e,f,p._leaveCb?null:y,u.nextSibling(p)),i(e.parent)){var v=e.parent,b=_(e);while(v){for(var m=0;m<s.destroy.length;++m)s.destroy[m](v);if(v.elm=e.elm,b){for(var g=0;g<s.create.length;++g)s.create[g](Ti,v);var w=v.data.hook.insert;if(w.merged)for(var O=1;O<w.fns.length;O++)w.fns[O]()}else Ai(v);v=v.parent}}i(y)?I([t],0,0):i(t.tag)&&x(t)}}return B(e,f,c),e.elm}i(t)&&x(t)}}var ki={create:Di,update:Di,destroy:function(t){Di(t,Ti)}};function Di(t,e){(t.data.directives||e.data.directives)&&Ni(t,e)}function Ni(t,e){var n,r,i,o=t===Ti,s=e===Ti,a=Pi(t.data.directives,t.context),c=Pi(e.data.directives,e.context),u=[],l=[];for(n in c)r=a[n],i=c[n],r?(i.oldValue=r.value,i.oldArg=r.arg,Ui(i,"update",e,t),i.def&&i.def.componentUpdated&&l.push(i)):(Ui(i,"bind",e,t),i.def&&i.def.inserted&&u.push(i));if(u.length){var f=function(){for(var n=0;n<u.length;n++)Ui(u[n],"inserted",e,t)};o?Oe(e,"insert",f):f()}if(l.length&&Oe(e,"postpatch",(function(){for(var n=0;n<l.length;n++)Ui(l[n],"componentUpdated",e,t)})),!o)for(n in a)c[n]||Ui(a[n],"unbind",t,t,s)}var Li=Object.create(null);function Pi(t,e){var n,r,i=Object.create(null);if(!t)return i;for(n=0;n<t.length;n++)r=t[n],r.modifiers||(r.modifiers=Li),i[Fi(r)]=r,r.def=Kt(e.$options,"directives",r.name,!0);return i}function Fi(t){return t.rawName||t.name+"."+Object.keys(t.modifiers||{}).join(".")}function Ui(t,e,n,r,i){var o=t.def&&t.def[e];if(o)try{o(n.elm,t,n,r,i)}catch(xs){ne(xs,n.context,"directive "+t.name+" "+e+" hook")}}var Ri=[Ii,ki];function Vi(t,e){var n=e.componentOptions;if((!i(n)||!1!==n.Ctor.options.inheritAttrs)&&(!r(t.data.attrs)||!r(e.data.attrs))){var o,s,a,c=e.elm,u=t.data.attrs||{},l=e.data.attrs||{};for(o in i(l.__ob__)&&(l=e.data.attrs=C({},l)),l)s=l[o],a=u[o],a!==s&&$i(c,o,s,e.data.pre);for(o in(tt||nt)&&l.value!==u.value&&$i(c,"value",l.value),u)r(l[o])&&(Gr(o)?c.removeAttributeNS(Wr,qr(o)):$r(o)||c.removeAttribute(o))}}function $i(t,e,n,r){r||t.tagName.indexOf("-")>-1?zi(t,e,n):Hr(e)?Kr(n)?t.removeAttribute(e):(n="allowfullscreen"===e&&"EMBED"===t.tagName?"true":e,t.setAttribute(e,n)):$r(e)?t.setAttribute(e,Yr(e,n)):Gr(e)?Kr(n)?t.removeAttributeNS(Wr,qr(e)):t.setAttributeNS(Wr,e,n):zi(t,e,n)}function zi(t,e,n){if(Kr(n))t.removeAttribute(e);else{if(tt&&!et&&"TEXTAREA"===t.tagName&&"placeholder"===e&&""!==n&&!t.__ieph){var r=function(e){e.stopImmediatePropagation(),t.removeEventListener("input",r)};t.addEventListener("input",r),t.__ieph=!0}t.setAttribute(e,n)}}var Yi={create:Vi,update:Vi};function Hi(t,e){var n=e.elm,o=e.data,s=t.data;if(!(r(o.staticClass)&&r(o.class)&&(r(s)||r(s.staticClass)&&r(s.class)))){var a=Xr(e),c=n._transitionClasses;i(c)&&(a=Qr(a,ti(c))),a!==n._prevClass&&(n.setAttribute("class",a),n._prevClass=a)}}var Wi,Gi={create:Hi,update:Hi},qi="__r",Ki="__c";function Xi(t){if(i(t[qi])){var e=tt?"change":"input";t[e]=[].concat(t[qi],t[e]||[]),delete t[qi]}i(t[Ki])&&(t.change=[].concat(t[Ki],t.change||[]),delete t[Ki])}function Ji(t,e,n){var r=Wi;return function i(){var o=e.apply(null,arguments);null!==o&&to(t,i,n,r)}}var Zi=ae&&!(it&&Number(it[1])<=53);function Qi(t,e,n,r){if(Zi){var i=qn,o=e;e=o._wrapper=function(t){if(t.target===t.currentTarget||t.timeStamp>=i||t.timeStamp<=0||t.target.ownerDocument!==document)return o.apply(this,arguments)}}Wi.addEventListener(t,e,st?{capture:n,passive:r}:n)}function to(t,e,n,r){(r||Wi).removeEventListener(t,e._wrapper||e,n)}function eo(t,e){if(!r(t.data.on)||!r(e.data.on)){var n=e.data.on||{},i=t.data.on||{};Wi=e.elm,Xi(n),we(n,i,Qi,to,Ji,e.context),Wi=void 0}}var no,ro={create:eo,update:eo};function io(t,e){if(!r(t.data.domProps)||!r(e.data.domProps)){var n,o,s=e.elm,a=t.data.domProps||{},c=e.data.domProps||{};for(n in i(c.__ob__)&&(c=e.data.domProps=C({},c)),a)n in c||(s[n]="");for(n in c){if(o=c[n],"textContent"===n||"innerHTML"===n){if(e.children&&(e.children.length=0),o===a[n])continue;1===s.childNodes.length&&s.removeChild(s.childNodes[0])}if("value"===n&&"PROGRESS"!==s.tagName){s._value=o;var u=r(o)?"":String(o);oo(s,u)&&(s.value=u)}else if("innerHTML"===n&&oi(s.tagName)&&r(s.innerHTML)){no=no||document.createElement("div"),no.innerHTML="<svg>"+o+"</svg>";var l=no.firstChild;while(s.firstChild)s.removeChild(s.firstChild);while(l.firstChild)s.appendChild(l.firstChild)}else if(o!==a[n])try{s[n]=o}catch(xs){}}}}function oo(t,e){return!t.composing&&("OPTION"===t.tagName||so(t,e)||ao(t,e))}function so(t,e){var n=!0;try{n=document.activeElement!==t}catch(xs){}return n&&t.value!==e}function ao(t,e){var n=t.value,r=t._vModifiers;if(i(r)){if(r.number)return y(n)!==y(e);if(r.trim)return n.trim()!==e.trim()}return n!==e}var co={create:io,update:io},uo=w((function(t){var e={},n=/;(?![^(]*\))/g,r=/:(.+)/;return t.split(n).forEach((function(t){if(t){var n=t.split(r);n.length>1&&(e[n[0].trim()]=n[1].trim())}})),e}));function lo(t){var e=fo(t.style);return t.staticStyle?C(t.staticStyle,e):e}function fo(t){return Array.isArray(t)?M(t):"string"===typeof t?uo(t):t}function ho(t,e){var n,r={};if(e){var i=t;while(i.componentInstance)i=i.componentInstance._vnode,i&&i.data&&(n=lo(i.data))&&C(r,n)}(n=lo(t.data))&&C(r,n);var o=t;while(o=o.parent)o.data&&(n=lo(o.data))&&C(r,n);return r}var po,yo=/^--/,vo=/\s*!important$/,bo=function(t,e,n){if(yo.test(e))t.style.setProperty(e,n);else if(vo.test(n))t.style.setProperty(A(e),n.replace(vo,""),"important");else{var r=go(e);if(Array.isArray(n))for(var i=0,o=n.length;i<o;i++)t.style[r]=n[i];else t.style[r]=n}},mo=["Webkit","Moz","ms"],go=w((function(t){if(po=po||document.createElement("div").style,t=S(t),"filter"!==t&&t in po)return t;for(var e=t.charAt(0).toUpperCase()+t.slice(1),n=0;n<mo.length;n++){var r=mo[n]+e;if(r in po)return r}}));function _o(t,e){var n=e.data,o=t.data;if(!(r(n.staticStyle)&&r(n.style)&&r(o.staticStyle)&&r(o.style))){var s,a,c=e.elm,u=o.staticStyle,l=o.normalizedStyle||o.style||{},f=u||l,d=fo(e.data.style)||{};e.data.normalizedStyle=i(d.__ob__)?C({},d):d;var h=ho(e,!0);for(a in f)r(h[a])&&bo(c,a,"");for(a in h)s=h[a],s!==f[a]&&bo(c,a,null==s?"":s)}}var wo={create:_o,update:_o},Oo=/\s+/;function So(t,e){if(e&&(e=e.trim()))if(t.classList)e.indexOf(" ")>-1?e.split(Oo).forEach((function(e){return t.classList.add(e)})):t.classList.add(e);else{var n=" "+(t.getAttribute("class")||"")+" ";n.indexOf(" "+e+" ")<0&&t.setAttribute("class",(n+e).trim())}}function xo(t,e){if(e&&(e=e.trim()))if(t.classList)e.indexOf(" ")>-1?e.split(Oo).forEach((function(e){return t.classList.remove(e)})):t.classList.remove(e),t.classList.length||t.removeAttribute("class");else{var n=" "+(t.getAttribute("class")||"")+" ",r=" "+e+" ";while(n.indexOf(r)>=0)n=n.replace(r," ");n=n.trim(),n?t.setAttribute("class",n):t.removeAttribute("class")}}function Io(t){if(t){if("object"===typeof t){var e={};return!1!==t.css&&C(e,Ao(t.name||"v")),C(e,t),e}return"string"===typeof t?Ao(t):void 0}}var Ao=w((function(t){return{enterClass:t+"-enter",enterToClass:t+"-enter-to",enterActiveClass:t+"-enter-active",leaveClass:t+"-leave",leaveToClass:t+"-leave-to",leaveActiveClass:t+"-leave-active"}})),To=X&&!et,jo="transition",Eo="animation",Bo="transition",Co="transitionend",Mo="animation",ko="animationend";To&&(void 0===window.ontransitionend&&void 0!==window.onwebkittransitionend&&(Bo="WebkitTransition",Co="webkitTransitionEnd"),void 0===window.onanimationend&&void 0!==window.onwebkitanimationend&&(Mo="WebkitAnimation",ko="webkitAnimationEnd"));var Do=X?window.requestAnimationFrame?window.requestAnimationFrame.bind(window):setTimeout:function(t){return t()};function No(t){Do((function(){Do(t)}))}function Lo(t,e){var n=t._transitionClasses||(t._transitionClasses=[]);n.indexOf(e)<0&&(n.push(e),So(t,e))}function Po(t,e){t._transitionClasses&&m(t._transitionClasses,e),xo(t,e)}function Fo(t,e,n){var r=Ro(t,e),i=r.type,o=r.timeout,s=r.propCount;if(!i)return n();var a=i===jo?Co:ko,c=0,u=function(){t.removeEventListener(a,l),n()},l=function(e){e.target===t&&++c>=s&&u()};setTimeout((function(){c<s&&u()}),o+1),t.addEventListener(a,l)}var Uo=/\b(transform|all)(,|$)/;function Ro(t,e){var n,r=window.getComputedStyle(t),i=(r[Bo+"Delay"]||"").split(", "),o=(r[Bo+"Duration"]||"").split(", "),s=Vo(i,o),a=(r[Mo+"Delay"]||"").split(", "),c=(r[Mo+"Duration"]||"").split(", "),u=Vo(a,c),l=0,f=0;e===jo?s>0&&(n=jo,l=s,f=o.length):e===Eo?u>0&&(n=Eo,l=u,f=c.length):(l=Math.max(s,u),n=l>0?s>u?jo:Eo:null,f=n?n===jo?o.length:c.length:0);var d=n===jo&&Uo.test(r[Bo+"Property"]);return{type:n,timeout:l,propCount:f,hasTransform:d}}function Vo(t,e){while(t.length<e.length)t=t.concat(t);return Math.max.apply(null,e.map((function(e,n){return $o(e)+$o(t[n])})))}function $o(t){return 1e3*Number(t.slice(0,-1).replace(",","."))}function zo(t,e){var n=t.elm;i(n._leaveCb)&&(n._leaveCb.cancelled=!0,n._leaveCb());var o=Io(t.data.transition);if(!r(o)&&!i(n._enterCb)&&1===n.nodeType){var s=o.css,a=o.type,u=o.enterClass,l=o.enterToClass,f=o.enterActiveClass,d=o.appearClass,h=o.appearToClass,p=o.appearActiveClass,v=o.beforeEnter,b=o.enter,m=o.afterEnter,g=o.enterCancelled,_=o.beforeAppear,w=o.appear,O=o.afterAppear,S=o.appearCancelled,x=o.duration,I=Cn,A=Cn.$vnode;while(A&&A.parent)I=A.context,A=A.parent;var T=!I._isMounted||!t.isRootInsert;if(!T||w||""===w){var j=T&&d?d:u,E=T&&p?p:f,B=T&&h?h:l,C=T&&_||v,M=T&&"function"===typeof w?w:b,k=T&&O||m,D=T&&S||g,N=y(c(x)?x.enter:x);0;var L=!1!==s&&!et,P=Wo(M),U=n._enterCb=F((function(){L&&(Po(n,B),Po(n,E)),U.cancelled?(L&&Po(n,j),D&&D(n)):k&&k(n),n._enterCb=null}));t.data.show||Oe(t,"insert",(function(){var e=n.parentNode,r=e&&e._pending&&e._pending[t.key];r&&r.tag===t.tag&&r.elm._leaveCb&&r.elm._leaveCb(),M&&M(n,U)})),C&&C(n),L&&(Lo(n,j),Lo(n,E),No((function(){Po(n,j),U.cancelled||(Lo(n,B),P||(Ho(N)?setTimeout(U,N):Fo(n,a,U)))}))),t.data.show&&(e&&e(),M&&M(n,U)),L||P||U()}}}function Yo(t,e){var n=t.elm;i(n._enterCb)&&(n._enterCb.cancelled=!0,n._enterCb());var o=Io(t.data.transition);if(r(o)||1!==n.nodeType)return e();if(!i(n._leaveCb)){var s=o.css,a=o.type,u=o.leaveClass,l=o.leaveToClass,f=o.leaveActiveClass,d=o.beforeLeave,h=o.leave,p=o.afterLeave,v=o.leaveCancelled,b=o.delayLeave,m=o.duration,g=!1!==s&&!et,_=Wo(h),w=y(c(m)?m.leave:m);0;var O=n._leaveCb=F((function(){n.parentNode&&n.parentNode._pending&&(n.parentNode._pending[t.key]=null),g&&(Po(n,l),Po(n,f)),O.cancelled?(g&&Po(n,u),v&&v(n)):(e(),p&&p(n)),n._leaveCb=null}));b?b(S):S()}function S(){O.cancelled||(!t.data.show&&n.parentNode&&((n.parentNode._pending||(n.parentNode._pending={}))[t.key]=t),d&&d(n),g&&(Lo(n,u),Lo(n,f),No((function(){Po(n,u),O.cancelled||(Lo(n,l),_||(Ho(w)?setTimeout(O,w):Fo(n,a,O)))}))),h&&h(n,O),g||_||O())}}function Ho(t){return"number"===typeof t&&!isNaN(t)}function Wo(t){if(r(t))return!1;var e=t.fns;return i(e)?Wo(Array.isArray(e)?e[0]:e):(t._length||t.length)>1}function Go(t,e){!0!==e.data.show&&zo(e)}var qo=X?{create:Go,activate:Go,remove:function(t,e){!0!==t.data.show?Yo(t,e):e()}}:{},Ko=[Yi,Gi,ro,co,wo,qo],Xo=Ko.concat(Ri),Jo=Mi({nodeOps:xi,modules:Xo});et&&document.addEventListener("selectionchange",(function(){var t=document.activeElement;t&&t.vmodel&&os(t,"input")}));var Zo={inserted:function(t,e,n,r){"select"===n.tag?(r.elm&&!r.elm._vOptions?Oe(n,"postpatch",(function(){Zo.componentUpdated(t,e,n)})):Qo(t,e,n.context),t._vOptions=[].map.call(t.options,ns)):("textarea"===n.tag||li(t.type))&&(t._vModifiers=e.modifiers,e.modifiers.lazy||(t.addEventListener("compositionstart",rs),t.addEventListener("compositionend",is),t.addEventListener("change",is),et&&(t.vmodel=!0)))},componentUpdated:function(t,e,n){if("select"===n.tag){Qo(t,e,n.context);var r=t._vOptions,i=t._vOptions=[].map.call(t.options,ns);if(i.some((function(t,e){return!L(t,r[e])}))){var o=t.multiple?e.value.some((function(t){return es(t,i)})):e.value!==e.oldValue&&es(e.value,i);o&&os(t,"change")}}}};function Qo(t,e,n){ts(t,e,n),(tt||nt)&&setTimeout((function(){ts(t,e,n)}),0)}function ts(t,e,n){var r=e.value,i=t.multiple;if(!i||Array.isArray(r)){for(var o,s,a=0,c=t.options.length;a<c;a++)if(s=t.options[a],i)o=P(r,ns(s))>-1,s.selected!==o&&(s.selected=o);else if(L(ns(s),r))return void(t.selectedIndex!==a&&(t.selectedIndex=a));i||(t.selectedIndex=-1)}}function es(t,e){return e.every((function(e){return!L(e,t)}))}function ns(t){return"_value"in t?t._value:t.value}function rs(t){t.target.composing=!0}function is(t){t.target.composing&&(t.target.composing=!1,os(t.target,"input"))}function os(t,e){var n=document.createEvent("HTMLEvents");n.initEvent(e,!0,!0),t.dispatchEvent(n)}function ss(t){return!t.componentInstance||t.data&&t.data.transition?t:ss(t.componentInstance._vnode)}var as={bind:function(t,e,n){var r=e.value;n=ss(n);var i=n.data&&n.data.transition,o=t.__vOriginalDisplay="none"===t.style.display?"":t.style.display;r&&i?(n.data.show=!0,zo(n,(function(){t.style.display=o}))):t.style.display=r?o:"none"},update:function(t,e,n){var r=e.value,i=e.oldValue;if(!r!==!i){n=ss(n);var o=n.data&&n.data.transition;o?(n.data.show=!0,r?zo(n,(function(){t.style.display=t.__vOriginalDisplay})):Yo(n,(function(){t.style.display="none"}))):t.style.display=r?t.__vOriginalDisplay:"none"}},unbind:function(t,e,n,r,i){i||(t.style.display=t.__vOriginalDisplay)}},cs={model:Zo,show:as},us={name:String,appear:Boolean,css:Boolean,mode:String,type:String,enterClass:String,leaveClass:String,enterToClass:String,leaveToClass:String,enterActiveClass:String,leaveActiveClass:String,appearClass:String,appearActiveClass:String,appearToClass:String,duration:[Number,String,Object]};function ls(t){var e=t&&t.componentOptions;return e&&e.Ctor.options.abstract?ls(xn(e.children)):t}function fs(t){var e={},n=t.$options;for(var r in n.propsData)e[r]=t[r];var i=n._parentListeners;for(var o in i)e[S(o)]=i[o];return e}function ds(t,e){if(/\d-keep-alive$/.test(e.tag))return t("keep-alive",{props:e.componentOptions.propsData})}function hs(t){while(t=t.parent)if(t.data.transition)return!0}function ps(t,e){return e.key===t.key&&e.tag===t.tag}var ys=function(t){return t.tag||De(t)},vs=function(t){return"show"===t.name},bs={name:"transition",props:us,abstract:!0,render:function(t){var e=this,n=this.$slots.default;if(n&&(n=n.filter(ys),n.length)){0;var r=this.mode;0;var i=n[0];if(hs(this.$vnode))return i;var o=ls(i);if(!o)return i;if(this._leaving)return ds(t,i);var s="__transition-"+this._uid+"-";o.key=null==o.key?o.isComment?s+"comment":s+o.tag:a(o.key)?0===String(o.key).indexOf(s)?o.key:s+o.key:o.key;var c=(o.data||(o.data={})).transition=fs(this),u=this._vnode,l=ls(u);if(o.data.directives&&o.data.directives.some(vs)&&(o.data.show=!0),l&&l.data&&!ps(o,l)&&!De(l)&&(!l.componentInstance||!l.componentInstance._vnode.isComment)){var f=l.data.transition=C({},c);if("out-in"===r)return this._leaving=!0,Oe(f,"afterLeave",(function(){e._leaving=!1,e.$forceUpdate()})),ds(t,i);if("in-out"===r){if(De(o))return u;var d,h=function(){d()};Oe(c,"afterEnter",h),Oe(c,"enterCancelled",h),Oe(f,"delayLeave",(function(t){d=t}))}}return i}}},ms=C({tag:String,moveClass:String},us);delete ms.mode;var gs={props:ms,beforeMount:function(){var t=this,e=this._update;this._update=function(n,r){var i=Mn(t);t.__patch__(t._vnode,t.kept,!1,!0),t._vnode=t.kept,i(),e.call(t,n,r)}},render:function(t){for(var e=this.tag||this.$vnode.data.tag||"span",n=Object.create(null),r=this.prevChildren=this.children,i=this.$slots.default||[],o=this.children=[],s=fs(this),a=0;a<i.length;a++){var c=i[a];if(c.tag)if(null!=c.key&&0!==String(c.key).indexOf("__vlist"))o.push(c),n[c.key]=c,(c.data||(c.data={})).transition=s;else;}if(r){for(var u=[],l=[],f=0;f<r.length;f++){var d=r[f];d.data.transition=s,d.data.pos=d.elm.getBoundingClientRect(),n[d.key]?u.push(d):l.push(d)}this.kept=t(e,null,u),this.removed=l}return t(e,null,o)},updated:function(){var t=this.prevChildren,e=this.moveClass||(this.name||"v")+"-move";t.length&&this.hasMove(t[0].elm,e)&&(t.forEach(_s),t.forEach(ws),t.forEach(Os),this._reflow=document.body.offsetHeight,t.forEach((function(t){if(t.data.moved){var n=t.elm,r=n.style;Lo(n,e),r.transform=r.WebkitTransform=r.transitionDuration="",n.addEventListener(Co,n._moveCb=function t(r){r&&r.target!==n||r&&!/transform$/.test(r.propertyName)||(n.removeEventListener(Co,t),n._moveCb=null,Po(n,e))})}})))},methods:{hasMove:function(t,e){if(!To)return!1;if(this._hasMove)return this._hasMove;var n=t.cloneNode();t._transitionClasses&&t._transitionClasses.forEach((function(t){xo(n,t)})),So(n,e),n.style.display="none",this.$el.appendChild(n);var r=Ro(n);return this.$el.removeChild(n),this._hasMove=r.hasTransform}}};function _s(t){t.elm._moveCb&&t.elm._moveCb(),t.elm._enterCb&&t.elm._enterCb()}function ws(t){t.data.newPos=t.elm.getBoundingClientRect()}function Os(t){var e=t.data.pos,n=t.data.newPos,r=e.left-n.left,i=e.top-n.top;if(r||i){t.data.moved=!0;var o=t.elm.style;o.transform=o.WebkitTransform="translate("+r+"px,"+i+"px)",o.transitionDuration="0s"}}var Ss={Transition:bs,TransitionGroup:gs};xr.config.mustUseProp=Vr,xr.config.isReservedTag=si,xr.config.isReservedAttr=Ur,xr.config.getTagNamespace=ai,xr.config.isUnknownElement=ui,C(xr.options.directives,cs),C(xr.options.components,Ss),xr.prototype.__patch__=X?Jo:k,xr.prototype.$mount=function(t,e){return t=t&&X?fi(t):void 0,Nn(this,t,e)},X&&setTimeout((function(){$.devtools&&ut&&ut.emit("init",xr)}),0),e["default"]=xr}.call(this,n("c8ba"))},"2ba4":function(t,e){var n=Function.prototype,r=n.apply,i=n.bind,o=n.call;t.exports="object"==typeof Reflect&&Reflect.apply||(i?o.bind(r):function(){return o.apply(r,arguments)})},"2cf4":function(t,e,n){var r,i,o,s,a=n("da84"),c=n("2ba4"),u=n("0366"),l=n("1626"),f=n("1a2d"),d=n("d039"),h=n("1be4"),p=n("f36a"),y=n("cc12"),v=n("1cdc"),b=n("605d"),m=a.setImmediate,g=a.clearImmediate,_=a.process,w=a.Dispatch,O=a.Function,S=a.MessageChannel,x=a.String,I=0,A={},T="onreadystatechange";try{r=a.location}catch(M){}var j=function(t){if(f(A,t)){var e=A[t];delete A[t],e()}},E=function(t){return function(){j(t)}},B=function(t){j(t.data)},C=function(t){a.postMessage(x(t),r.protocol+"//"+r.host)};m&&g||(m=function(t){var e=p(arguments,1);return A[++I]=function(){c(l(t)?t:O(t),void 0,e)},i(I),I},g=function(t){delete A[t]},b?i=function(t){_.nextTick(E(t))}:w&&w.now?i=function(t){w.now(E(t))}:S&&!v?(o=new S,s=o.port2,o.port1.onmessage=B,i=u(s.postMessage,s)):a.addEventListener&&l(a.postMessage)&&!a.importScripts&&r&&"file:"!==r.protocol&&!d(C)?(i=C,a.addEventListener("message",B,!1)):i=T in y("script")?function(t){h.appendChild(y("script"))[T]=function(){h.removeChild(this),j(t)}}:function(t){setTimeout(E(t),0)}),t.exports={set:m,clear:g}},"2d00":function(t,e,n){var r,i,o=n("da84"),s=n("342f"),a=o.process,c=o.Deno,u=a&&a.versions||c&&c.version,l=u&&u.v8;l&&(r=l.split("."),i=r[0]>0&&r[0]<4?1:+(r[0]+r[1])),!i&&s&&(r=s.match(/Edge\/(\d+)/),(!r||r[1]>=74)&&(r=s.match(/Chrome\/(\d+)/),r&&(i=+r[1]))),t.exports=i},"320c":function(t,e,n){"use strict";
/*
object-assign
(c) Sindre Sorhus
@license MIT
*/var r=Object.getOwnPropertySymbols,i=Object.prototype.hasOwnProperty,o=Object.prototype.propertyIsEnumerable;function s(t){if(null===t||void 0===t)throw new TypeError("Object.assign cannot be called with null or undefined");return Object(t)}function a(){try{if(!Object.assign)return!1;var t=new String("abc");if(t[5]="de","5"===Object.getOwnPropertyNames(t)[0])return!1;for(var e={},n=0;n<10;n++)e["_"+String.fromCharCode(n)]=n;var r=Object.getOwnPropertyNames(e).map((function(t){return e[t]}));if("0123456789"!==r.join(""))return!1;var i={};return"abcdefghijklmnopqrst".split("").forEach((function(t){i[t]=t})),"abcdefghijklmnopqrst"===Object.keys(Object.assign({},i)).join("")}catch(o){return!1}}t.exports=a()?Object.assign:function(t,e){for(var n,a,c=s(t),u=1;u<arguments.length;u++){for(var l in n=Object(arguments[u]),n)i.call(n,l)&&(c[l]=n[l]);if(r){a=r(n);for(var f=0;f<a.length;f++)o.call(n,a[f])&&(c[a[f]]=n[a[f]])}}return c}},"342f":function(t,e,n){var r=n("d066");t.exports=r("navigator","userAgent")||""},"35a1":function(t,e,n){var r=n("f5df"),i=n("dc4a"),o=n("3f8c"),s=n("b622"),a=s("iterator");t.exports=function(t){if(void 0!=t)return i(t,a)||i(t,"@@iterator")||o[r(t)]}},"37e8":function(t,e,n){var r=n("83ab"),i=n("9bf2"),o=n("825a"),s=n("fc6a"),a=n("df75");t.exports=r?Object.defineProperties:function(t,e){o(t);var n,r=s(e),c=a(e),u=c.length,l=0;while(u>l)i.f(t,n=c[l++],r[n]);return t}},"3a9b":function(t,e,n){var r=n("e330");t.exports=r({}.isPrototypeOf)},"3bbe":function(t,e,n){var r=n("da84"),i=n("1626"),o=r.String,s=r.TypeError;t.exports=function(t){if("object"==typeof t||i(t))return t;throw s("Can't set "+o(t)+" as a prototype")}},"3eea":function(t,e,n){"use strict";n.r(e),n.d(e,"memcpy",(function(){return u})),n.d(e,"joinUint8Arrays",(function(){return l})),n.d(e,"toArrayBufferView",(function(){return f})),n.d(e,"toInt8Array",(function(){return d})),n.d(e,"toInt16Array",(function(){return h})),n.d(e,"toInt32Array",(function(){return p})),n.d(e,"toBigInt64Array",(function(){return y})),n.d(e,"toUint8Array",(function(){return v})),n.d(e,"toUint16Array",(function(){return b})),n.d(e,"toUint32Array",(function(){return m})),n.d(e,"toBigUint64Array",(function(){return g})),n.d(e,"toFloat32Array",(function(){return _})),n.d(e,"toFloat64Array",(function(){return w})),n.d(e,"toUint8ClampedArray",(function(){return O})),n.d(e,"toArrayBufferViewIterator",(function(){return x})),n.d(e,"toInt8ArrayIterator",(function(){return I})),n.d(e,"toInt16ArrayIterator",(function(){return A})),n.d(e,"toInt32ArrayIterator",(function(){return T})),n.d(e,"toUint8ArrayIterator",(function(){return j})),n.d(e,"toUint16ArrayIterator",(function(){return E})),n.d(e,"toUint32ArrayIterator",(function(){return B})),n.d(e,"toFloat32ArrayIterator",(function(){return C})),n.d(e,"toFloat64ArrayIterator",(function(){return M})),n.d(e,"toUint8ClampedArrayIterator",(function(){return k})),n.d(e,"toArrayBufferViewAsyncIterator",(function(){return D})),n.d(e,"toInt8ArrayAsyncIterator",(function(){return N})),n.d(e,"toInt16ArrayAsyncIterator",(function(){return L})),n.d(e,"toInt32ArrayAsyncIterator",(function(){return P})),n.d(e,"toUint8ArrayAsyncIterator",(function(){return F})),n.d(e,"toUint16ArrayAsyncIterator",(function(){return U})),n.d(e,"toUint32ArrayAsyncIterator",(function(){return R})),n.d(e,"toFloat32ArrayAsyncIterator",(function(){return V})),n.d(e,"toFloat64ArrayAsyncIterator",(function(){return $})),n.d(e,"toUint8ClampedArrayAsyncIterator",(function(){return z})),n.d(e,"rebaseValueOffsets",(function(){return Y})),n.d(e,"compareArrayLike",(function(){return H}));var r=n("a6b2"),i=n("5217"),o=n("841f"),s=r["a"].ByteBuffer;const a="undefined"!==typeof SharedArrayBuffer?SharedArrayBuffer:ArrayBuffer;function c(t){let e,n,r,i,o=t[0]?[t[0]]:[];for(let s,a,c=0,u=0,l=t.length;++c<l;)s=o[u],a=t[c],!s||!a||s.buffer!==a.buffer||a.byteOffset<s.byteOffset?a&&(o[++u]=a):(({byteOffset:e,byteLength:r}=s),({byteOffset:n,byteLength:i}=a),e+r<n||n+i<e?a&&(o[++u]=a):o[u]=new Uint8Array(s.buffer,e,n-e+i));return o}function u(t,e,n=0,r=e.byteLength){const i=t.byteLength,o=new Uint8Array(t.buffer,t.byteOffset,i),s=new Uint8Array(e.buffer,e.byteOffset,Math.min(r,i));return o.set(s,n),t}function l(t,e){let n,r,i,o=c(t),s=o.reduce((t,e)=>t+e.byteLength,0),a=0,l=-1,f=Math.min(e||1/0,s);for(let c=o.length;++l<c;){if(n=o[l],r=n.subarray(0,Math.min(n.length,f-a)),f<=a+r.length){r.length<n.length?o[l]=n.subarray(r.length):r.length===n.length&&l++,i?u(i,r,a):i=r;break}u(i||(i=new Uint8Array(f)),r,a),a+=r.length}return[i||new Uint8Array(0),o.slice(l),s-(i?i.byteLength:0)]}function f(t,e){let n=Object(o["j"])(e)?e.value:e;return n instanceof t?t===Uint8Array?new t(n.buffer,n.byteOffset,n.byteLength):n:n?("string"===typeof n&&(n=Object(i["b"])(n)),n instanceof ArrayBuffer||n instanceof a?new t(n):n instanceof s?f(t,n.bytes()):ArrayBuffer.isView(n)?n.byteLength<=0?new t(0):new t(n.buffer,n.byteOffset,n.byteLength/t.BYTES_PER_ELEMENT):t.from(n)):new t(0)}const d=t=>f(Int8Array,t),h=t=>f(Int16Array,t),p=t=>f(Int32Array,t),y=t=>f(o["b"],t),v=t=>f(Uint8Array,t),b=t=>f(Uint16Array,t),m=t=>f(Uint32Array,t),g=t=>f(o["d"],t),_=t=>f(Float32Array,t),w=t=>f(Float64Array,t),O=t=>f(Uint8ClampedArray,t),S=t=>(t.next(),t);function*x(t,e){const n=function*(t){yield t},r="string"===typeof e||ArrayBuffer.isView(e)||e instanceof ArrayBuffer||e instanceof a?n(e):Object(o["i"])(e)?e:n(e);yield*S(function*(e){let n=null;do{n=e.next(yield f(t,n))}while(!n.done)}(r[Symbol.iterator]()))}const I=t=>x(Int8Array,t),A=t=>x(Int16Array,t),T=t=>x(Int32Array,t),j=t=>x(Uint8Array,t),E=t=>x(Uint16Array,t),B=t=>x(Uint32Array,t),C=t=>x(Float32Array,t),M=t=>x(Float64Array,t),k=t=>x(Uint8ClampedArray,t);async function*D(t,e){if(Object(o["l"])(e))return yield*D(t,await e);const n=async function*(t){yield await t},r=async function*(t){yield*S(function*(t){let e=null;do{e=t.next(yield e&&e.value)}while(!e.done)}(t[Symbol.iterator]()))},i="string"===typeof e||ArrayBuffer.isView(e)||e instanceof ArrayBuffer||e instanceof a?n(e):Object(o["i"])(e)?r(e):Object(o["f"])(e)?e:n(e);yield*S(async function*(e){let n=null;do{n=await e.next(yield f(t,n))}while(!n.done)}(i[Symbol.asyncIterator]()))}const N=t=>D(Int8Array,t),L=t=>D(Int16Array,t),P=t=>D(Int32Array,t),F=t=>D(Uint8Array,t),U=t=>D(Uint16Array,t),R=t=>D(Uint32Array,t),V=t=>D(Float32Array,t),$=t=>D(Float64Array,t),z=t=>D(Uint8ClampedArray,t);function Y(t,e,n){if(0!==t){n=n.slice(0,e+1);for(let r=-1;++r<=e;)n[r]+=t}return n}function H(t,e){let n=0,r=t.length;if(r!==e.length)return!1;if(r>0)do{if(t[n]!==e[n])return!1}while(++n<r);return!0}},"3f8c":function(t,e){t.exports={}},"44ad":function(t,e,n){var r=n("da84"),i=n("e330"),o=n("d039"),s=n("c6b6"),a=r.Object,c=i("".split);t.exports=o((function(){return!a("z").propertyIsEnumerable(0)}))?function(t){return"String"==s(t)?c(t,""):a(t)}:a},"44d2":function(t,e,n){var r=n("b622"),i=n("7c73"),o=n("9bf2"),s=r("unscopables"),a=Array.prototype;void 0==a[s]&&o.f(a,s,{configurable:!0,value:i(null)}),t.exports=function(t){a[s][t]=!0}},"44de":function(t,e,n){var r=n("da84");t.exports=function(t,e){var n=r.console;n&&n.error&&(1==arguments.length?n.error(t):n.error(t,e))}},4840:function(t,e,n){var r=n("825a"),i=n("5087"),o=n("b622"),s=o("species");t.exports=function(t,e){var n,o=r(t).constructor;return void 0===o||void 0==(n=r(o)[s])?e:i(n)}},"485a":function(t,e,n){var r=n("da84"),i=n("c65b"),o=n("1626"),s=n("861d"),a=r.TypeError;t.exports=function(t,e){var n,r;if("string"===e&&o(n=t.toString)&&!s(r=i(n,t)))return r;if(o(n=t.valueOf)&&!s(r=i(n,t)))return r;if("string"!==e&&o(n=t.toString)&&!s(r=i(n,t)))return r;throw a("Can't convert object to primitive value")}},4930:function(t,e,n){var r=n("2d00"),i=n("d039");t.exports=!!Object.getOwnPropertySymbols&&!i((function(){var t=Symbol();return!String(t)||!(Object(t)instanceof Symbol)||!Symbol.sham&&r&&r<41}))},"4cec":function(t,e,n){"use strict";t.exports=n("a93d")},"4d64":function(t,e,n){var r=n("fc6a"),i=n("23cb"),o=n("07fa"),s=function(t){return function(e,n,s){var a,c=r(e),u=o(c),l=i(s,u);if(t&&n!=n){while(u>l)if(a=c[l++],a!=a)return!0}else for(;u>l;l++)if((t||l in c)&&c[l]===n)return t||l||0;return!t&&-1}};t.exports={includes:s(!0),indexOf:s(!1)}},5087:function(t,e,n){var r=n("da84"),i=n("68ee"),o=n("0d51"),s=r.TypeError;t.exports=function(t){if(i(t))return t;throw s(o(t)+" is not a constructor")}},"50c4":function(t,e,n){var r=n("5926"),i=Math.min;t.exports=function(t){return t>0?i(r(t),9007199254740991):0}},5217:function(t,e,n){"use strict";(function(t){n.d(e,"a",(function(){return a})),n.d(e,"b",(function(){return c}));var r=n("3eea"),i=n("b4b6");const o="function"===typeof t?t:null,s="function"===typeof TextDecoder&&"function"===typeof TextEncoder,a=(t=>{if(s||!o){const e=new t("utf-8");return t=>e.decode(t)}return t=>{const{buffer:e,byteOffset:n,length:i}=Object(r["toUint8Array"])(t);return o.from(e,n,i).toString()}})("undefined"!==typeof TextDecoder?TextDecoder:i["a"]),c=(t=>{if(s||!o){const e=new t;return t=>e.encode(t)}return(t="")=>Object(r["toUint8Array"])(o.from(t,"utf8"))})("undefined"!==typeof TextEncoder?TextEncoder:i["b"])}).call(this,n("b639").Buffer)},5692:function(t,e,n){var r=n("c430"),i=n("c6cd");(t.exports=function(t,e){return i[t]||(i[t]=void 0!==e?e:{})})("versions",[]).push({version:"3.19.3",mode:r?"pure":"global",copyright:"© 2021 Denis Pushkarev (zloirock.ru)"})},"56ef":function(t,e,n){var r=n("d066"),i=n("e330"),o=n("241c"),s=n("7418"),a=n("825a"),c=i([].concat);t.exports=r("Reflect","ownKeys")||function(t){var e=o.f(a(t)),n=s.f;return n?c(e,n(t)):e}},5926:function(t,e){var n=Math.ceil,r=Math.floor;t.exports=function(t){var e=+t;return e!==e||0===e?0:(e>0?r:n)(e)}},"59ed":function(t,e,n){var r=n("da84"),i=n("1626"),o=n("0d51"),s=r.TypeError;t.exports=function(t){if(i(t))return t;throw s(o(t)+" is not a function")}},"5c6c":function(t,e){t.exports=function(t,e){return{enumerable:!(1&t),configurable:!(2&t),writable:!(4&t),value:e}}},"5e77":function(t,e,n){var r=n("83ab"),i=n("1a2d"),o=Function.prototype,s=r&&Object.getOwnPropertyDescriptor,a=i(o,"name"),c=a&&"something"===function(){}.name,u=a&&(!r||r&&s(o,"name").configurable);t.exports={EXISTS:a,PROPER:c,CONFIGURABLE:u}},"605d":function(t,e,n){var r=n("c6b6"),i=n("da84");t.exports="process"==r(i.process)},6069:function(t,e){t.exports="object"==typeof window},"60da":function(t,e,n){"use strict";var r=n("83ab"),i=n("e330"),o=n("c65b"),s=n("d039"),a=n("df75"),c=n("7418"),u=n("d1e7"),l=n("7b0b"),f=n("44ad"),d=Object.assign,h=Object.defineProperty,p=i([].concat);t.exports=!d||s((function(){if(r&&1!==d({b:1},d(h({},"a",{enumerable:!0,get:function(){h(this,"b",{value:3,enumerable:!1})}}),{b:2})).b)return!0;var t={},e={},n=Symbol(),i="abcdefghijklmnopqrst";return t[n]=7,i.split("").forEach((function(t){e[t]=t})),7!=d({},t)[n]||a(d({},e)).join("")!=i}))?function(t,e){var n=l(t),i=arguments.length,s=1,d=c.f,h=u.f;while(i>s){var y,v=f(arguments[s++]),b=d?p(a(v),d(v)):a(v),m=b.length,g=0;while(m>g)y=b[g++],r&&!o(h,v,y)||(n[y]=v[y])}return n}:d},"65f0":function(t,e,n){var r=n("0b42");t.exports=function(t,e){return new(r(t))(0===e?0:e)}},"68ee":function(t,e,n){var r=n("e330"),i=n("d039"),o=n("1626"),s=n("f5df"),a=n("d066"),c=n("8925"),u=function(){},l=[],f=a("Reflect","construct"),d=/^\s*(?:class|function)\b/,h=r(d.exec),p=!d.exec(u),y=function(t){if(!o(t))return!1;try{return f(u,l,t),!0}catch(e){return!1}},v=function(t){if(!o(t))return!1;switch(s(t)){case"AsyncFunction":case"GeneratorFunction":case"AsyncGeneratorFunction":return!1}return p||!!h(d,c(t))};t.exports=!f||i((function(){var t;return y(y.call)||!y(Object)||!y((function(){t=!0}))||t}))?v:y},"69f3":function(t,e,n){var r,i,o,s=n("7f9a"),a=n("da84"),c=n("e330"),u=n("861d"),l=n("9112"),f=n("1a2d"),d=n("c6cd"),h=n("f772"),p=n("d012"),y="Object already initialized",v=a.TypeError,b=a.WeakMap,m=function(t){return o(t)?i(t):r(t,{})},g=function(t){return function(e){var n;if(!u(e)||(n=i(e)).type!==t)throw v("Incompatible receiver, "+t+" required");return n}};if(s||d.state){var _=d.state||(d.state=new b),w=c(_.get),O=c(_.has),S=c(_.set);r=function(t,e){if(O(_,t))throw new v(y);return e.facade=t,S(_,t,e),e},i=function(t){return w(_,t)||{}},o=function(t){return O(_,t)}}else{var x=h("state");p[x]=!0,r=function(t,e){if(f(t,x))throw new v(y);return e.facade=t,l(t,x,e),e},i=function(t){return f(t,x)?t[x]:{}},o=function(t){return f(t,x)}}t.exports={set:r,get:i,has:o,enforce:m,getterFor:g}},"6bfb":function(t,e,n){"use strict";var r=n("3eea");e["a"]={fromIterable(t){return i(o(t))},fromAsyncIterable(t){return i(s(t))},fromDOMStream(t){return i(a(t))},fromNodeStream(t){return i(f(t))},toDOMStream(t,e){throw new Error('"toDOMStream" not available in this environment')},toNodeStream(t,e){throw new Error('"toNodeStream" not available in this environment')}};const i=t=>(t.next(),t);function*o(t){let e,n,i,o,s=!1,a=[],c=0;function u(){return"peek"===i?Object(r["joinUint8Arrays"])(a,o)[0]:([n,a,c]=Object(r["joinUint8Arrays"])(a,o),n)}({cmd:i,size:o}=yield null);let l=Object(r["toUint8ArrayIterator"])(t)[Symbol.iterator]();try{do{if(({done:e,value:n}=isNaN(o-c)?l.next(void 0):l.next(o-c)),!e&&n.byteLength>0&&(a.push(n),c+=n.byteLength),e||o<=c)do{({cmd:i,size:o}=yield u())}while(o<c)}while(!e)}catch(f){(s=!0)&&"function"===typeof l.throw&&l.throw(f)}finally{!1===s&&"function"===typeof l.return&&l.return()}}async function*s(t){let e,n,i,o,s=!1,a=[],c=0;function u(){return"peek"===i?Object(r["joinUint8Arrays"])(a,o)[0]:([n,a,c]=Object(r["joinUint8Arrays"])(a,o),n)}({cmd:i,size:o}=yield null);let l=Object(r["toUint8ArrayAsyncIterator"])(t)[Symbol.asyncIterator]();try{do{if(({done:e,value:n}=isNaN(o-c)?await l.next(void 0):await l.next(o-c)),!e&&n.byteLength>0&&(a.push(n),c+=n.byteLength),e||o<=c)do{({cmd:i,size:o}=yield u())}while(o<c)}while(!e)}catch(f){(s=!0)&&"function"===typeof l.throw&&await l.throw(f)}finally{!1===s&&"function"===typeof l.return&&await l.return()}}async function*a(t){let e,n,i,o=!1,s=!1,a=[],u=0;function l(){return"peek"===n?Object(r["joinUint8Arrays"])(a,i)[0]:([e,a,u]=Object(r["joinUint8Arrays"])(a,i),e)}({cmd:n,size:i}=yield null);let f=new c(t);try{do{if(({done:o,value:e}=isNaN(i-u)?await f["read"](void 0):await f["read"](i-u)),!o&&e.byteLength>0&&(a.push(Object(r["toUint8Array"])(e)),u+=e.byteLength),o||i<=u)do{({cmd:n,size:i}=yield l())}while(i<u)}while(!o)}catch(d){(s=!0)&&await f["cancel"](d)}finally{!1===s?await f["cancel"]():t["locked"]&&f.releaseLock()}}class c{constructor(t){this.source=t,this.byobReader=null,this.defaultReader=null;try{this.supportsBYOB=!!(this.reader=this.getBYOBReader())}catch(e){this.supportsBYOB=!(this.reader=this.getDefaultReader())}}get closed(){return this.reader?this.reader["closed"].catch(()=>{}):Promise.resolve()}releaseLock(){this.reader&&this.reader.releaseLock(),this.reader=this.byobReader=this.defaultReader=null}async cancel(t){const{reader:e,source:n}=this;e&&await e["cancel"](t).catch(()=>{}),n&&n["locked"]&&this.releaseLock()}async read(t){if(0===t)return{done:null==this.reader,value:new Uint8Array(0)};const e=this.supportsBYOB&&"number"===typeof t?await this.readFromBYOBReader(t):await this.getDefaultReader().read();return!e.done&&(e.value=Object(r["toUint8Array"])(e)),e}getDefaultReader(){return this.byobReader&&this.releaseLock(),this.defaultReader||(this.defaultReader=this.source["getReader"](),this.defaultReader["closed"].catch(()=>{})),this.reader=this.defaultReader}getBYOBReader(){return this.defaultReader&&this.releaseLock(),this.byobReader||(this.byobReader=this.source["getReader"]({mode:"byob"}),this.byobReader["closed"].catch(()=>{})),this.reader=this.byobReader}async readFromBYOBReader(t){return await u(this.getBYOBReader(),new ArrayBuffer(t),0,t)}}async function u(t,e,n,r){if(n>=r)return{done:!1,value:new Uint8Array(e,0,r)};const{done:i,value:o}=await t.read(new Uint8Array(e,n,r-n));return(n+=o.byteLength)<r&&!i?await u(t,o.buffer,n,r):{done:i,value:new Uint8Array(o.buffer,0,n)}}const l=(t,e)=>{let n,r=t=>n([e,t]);return[e,r,new Promise(i=>(n=i)&&t["once"](e,r))]};async function*f(t){let e,n,i,o=[],s="error",a=!1,c=null,u=0,f=[];function d(){return"peek"===e?Object(r["joinUint8Arrays"])(f,n)[0]:([i,f,u]=Object(r["joinUint8Arrays"])(f,n),i)}if(({cmd:e,size:n}=yield null),t["isTTY"])return yield new Uint8Array(0);try{o[0]=l(t,"end"),o[1]=l(t,"error");do{if(o[2]=l(t,"readable"),[s,c]=await Promise.race(o.map(t=>t[2])),"error"===s)break;if((a="end"===s)||(isFinite(n-u)?(i=Object(r["toUint8Array"])(t["read"](n-u)),i.byteLength<n-u&&(i=Object(r["toUint8Array"])(t["read"](void 0)))):i=Object(r["toUint8Array"])(t["read"](void 0)),i.byteLength>0&&(f.push(i),u+=i.byteLength)),a||n<=u)do{({cmd:e,size:n}=yield d())}while(n<u)}while(!a)}finally{await h(o,"error"===s?c:null)}function h(e,n){return i=f=null,new Promise(async(r,i)=>{for(const[n,s]of e)t["off"](n,s);try{const e=t["destroy"];e&&e.call(t,n),n=void 0}catch(o){n=o||n}finally{null!=n?i(n):r()}})}}},"6eeb":function(t,e,n){var r=n("da84"),i=n("1626"),o=n("1a2d"),s=n("9112"),a=n("ce4e"),c=n("8925"),u=n("69f3"),l=n("5e77").CONFIGURABLE,f=u.get,d=u.enforce,h=String(String).split("String");(t.exports=function(t,e,n,c){var u,f=!!c&&!!c.unsafe,p=!!c&&!!c.enumerable,y=!!c&&!!c.noTargetGet,v=c&&void 0!==c.name?c.name:e;i(n)&&("Symbol("===String(v).slice(0,7)&&(v="["+String(v).replace(/^Symbol\(([^)]*)\)/,"$1")+"]"),(!o(n,"name")||l&&n.name!==v)&&s(n,"name",v),u=d(n),u.source||(u.source=h.join("string"==typeof v?v:""))),t!==r?(f?!y&&t[e]&&(p=!0):delete t[e],p?t[e]=n:s(t,e,n)):p?t[e]=n:a(e,n)})(Function.prototype,"toString",(function(){return i(this)&&f(this).source||c(this)}))},7418:function(t,e){e.f=Object.getOwnPropertySymbols},7839:function(t,e){t.exports=["constructor","hasOwnProperty","isPrototypeOf","propertyIsEnumerable","toLocaleString","toString","valueOf"]},"785a":function(t,e,n){var r=n("cc12"),i=r("span").classList,o=i&&i.constructor&&i.constructor.prototype;t.exports=o===Object.prototype?void 0:o},"7b0b":function(t,e,n){var r=n("da84"),i=n("1d80"),o=r.Object;t.exports=function(t){return o(i(t))}},"7c73":function(t,e,n){var r,i=n("825a"),o=n("37e8"),s=n("7839"),a=n("d012"),c=n("1be4"),u=n("cc12"),l=n("f772"),f=">",d="<",h="prototype",p="script",y=l("IE_PROTO"),v=function(){},b=function(t){return d+p+f+t+d+"/"+p+f},m=function(t){t.write(b("")),t.close();var e=t.parentWindow.Object;return t=null,e},g=function(){var t,e=u("iframe"),n="java"+p+":";return e.style.display="none",c.appendChild(e),e.src=String(n),t=e.contentWindow.document,t.open(),t.write(b("document.F=Object")),t.close(),t.F},_=function(){try{r=new ActiveXObject("htmlfile")}catch(e){}_="undefined"!=typeof document?document.domain&&r?m(r):g():m(r);var t=s.length;while(t--)delete _[h][s[t]];return _()};a[y]=!0,t.exports=Object.create||function(t,e){var n;return null!==t?(v[h]=i(t),n=new v,v[h]=null,n[y]=t):n=_(),void 0===e?n:o(n,e)}},"7dd0":function(t,e,n){"use strict";var r=n("23e7"),i=n("c65b"),o=n("c430"),s=n("5e77"),a=n("1626"),c=n("9ed3"),u=n("e163"),l=n("d2bb"),f=n("d44e"),d=n("9112"),h=n("6eeb"),p=n("b622"),y=n("3f8c"),v=n("ae93"),b=s.PROPER,m=s.CONFIGURABLE,g=v.IteratorPrototype,_=v.BUGGY_SAFARI_ITERATORS,w=p("iterator"),O="keys",S="values",x="entries",I=function(){return this};t.exports=function(t,e,n,s,p,v,A){c(n,e,s);var T,j,E,B=function(t){if(t===p&&N)return N;if(!_&&t in k)return k[t];switch(t){case O:return function(){return new n(this,t)};case S:return function(){return new n(this,t)};case x:return function(){return new n(this,t)}}return function(){return new n(this)}},C=e+" Iterator",M=!1,k=t.prototype,D=k[w]||k["@@iterator"]||p&&k[p],N=!_&&D||B(p),L="Array"==e&&k.entries||D;if(L&&(T=u(L.call(new t)),T!==Object.prototype&&T.next&&(o||u(T)===g||(l?l(T,g):a(T[w])||h(T,w,I)),f(T,C,!0,!0),o&&(y[C]=I))),b&&p==S&&D&&D.name!==S&&(!o&&m?d(k,"name",S):(M=!0,N=function(){return i(D,this)})),p)if(j={values:B(S),keys:v?N:B(O),entries:B(x)},A)for(E in j)(_||M||!(E in k))&&h(k,E,j[E]);else r({target:e,proto:!0,forced:_||M},j);return o&&!A||k[w]===N||h(k,w,N,{name:p}),y[e]=N,j}},"7f9a":function(t,e,n){var r=n("da84"),i=n("1626"),o=n("8925"),s=r.WeakMap;t.exports=i(s)&&/native code/.test(o(s))},"825a":function(t,e,n){var r=n("da84"),i=n("861d"),o=r.String,s=r.TypeError;t.exports=function(t){if(i(t))return t;throw s(o(t)+" is not an object")}},"83ab":function(t,e,n){var r=n("d039");t.exports=!r((function(){return 7!=Object.defineProperty({},1,{get:function(){return 7}})[1]}))},"841f":function(t,e,n){"use strict";n.d(e,"a",(function(){return i})),n.d(e,"c",(function(){return o})),n.d(e,"b",(function(){return s})),n.d(e,"d",(function(){return c})),n.d(e,"k",(function(){return h})),n.d(e,"l",(function(){return p})),n.d(e,"i",(function(){return y})),n.d(e,"f",(function(){return v})),n.d(e,"e",(function(){return b})),n.d(e,"j",(function(){return m})),n.d(e,"h",(function(){return g})),n.d(e,"g",(function(){return _})),n.d(e,"o",(function(){return w})),n.d(e,"m",(function(){return O})),n.d(e,"p",(function(){return S})),n.d(e,"n",(function(){return x}));var r=n("f673");const[i,o]=(()=>{const t=()=>{throw new Error("BigInt is not available in this environment")};function e(){throw t()}return e.asIntN=()=>{throw t()},e.asUintN=()=>{throw t()},"undefined"!==typeof BigInt?[BigInt,!0]:[e,!1]})(),[s,a]=(()=>{const t=()=>{throw new Error("BigInt64Array is not available in this environment")};class e{static get BYTES_PER_ELEMENT(){return 8}static of(){throw t()}static from(){throw t()}constructor(){throw t()}}return"undefined"!==typeof BigInt64Array?[BigInt64Array,!0]:[e,!1]})(),[c,u]=(()=>{const t=()=>{throw new Error("BigUint64Array is not available in this environment")};class e{static get BYTES_PER_ELEMENT(){return 8}static of(){throw t()}static from(){throw t()}constructor(){throw t()}}return"undefined"!==typeof BigUint64Array?[BigUint64Array,!0]:[e,!1]})(),l=t=>"number"===typeof t,f=t=>"boolean"===typeof t,d=t=>"function"===typeof t,h=t=>null!=t&&Object(t)===t,p=t=>h(t)&&d(t.then),y=t=>h(t)&&d(t[Symbol.iterator]),v=t=>h(t)&&d(t[Symbol.asyncIterator]),b=t=>h(t)&&h(t["schema"]),m=t=>h(t)&&"done"in t&&"value"in t,g=t=>h(t)&&d(t["stat"])&&l(t["fd"]),_=t=>h(t)&&O(t["body"]),w=t=>h(t)&&d(t["abort"])&&d(t["getWriter"])&&!(t instanceof r["d"]),O=t=>h(t)&&d(t["cancel"])&&d(t["getReader"])&&!(t instanceof r["d"]),S=t=>h(t)&&d(t["end"])&&d(t["write"])&&f(t["writable"])&&!(t instanceof r["d"]),x=t=>h(t)&&d(t["read"])&&d(t["pipe"])&&f(t["readable"])&&!(t instanceof r["d"])},"861d":function(t,e,n){var r=n("1626");t.exports=function(t){return"object"==typeof t?null!==t:r(t)}},8925:function(t,e,n){var r=n("e330"),i=n("1626"),o=n("c6cd"),s=r(Function.toString);i(o.inspectSource)||(o.inspectSource=function(t){return s(t)}),t.exports=o.inspectSource},"90e3":function(t,e,n){var r=n("e330"),i=0,o=Math.random(),s=r(1..toString);t.exports=function(t){return"Symbol("+(void 0===t?"":t)+")_"+s(++i+o,36)}},9112:function(t,e,n){var r=n("83ab"),i=n("9bf2"),o=n("5c6c");t.exports=r?function(t,e,n){return i.f(t,e,o(1,n))}:function(t,e,n){return t[e]=n,t}},9152:function(t,e){
/*! ieee754. BSD-3-Clause License. Feross Aboukhadijeh <https://feross.org/opensource> */
e.read=function(t,e,n,r,i){var o,s,a=8*i-r-1,c=(1<<a)-1,u=c>>1,l=-7,f=n?i-1:0,d=n?-1:1,h=t[e+f];for(f+=d,o=h&(1<<-l)-1,h>>=-l,l+=a;l>0;o=256*o+t[e+f],f+=d,l-=8);for(s=o&(1<<-l)-1,o>>=-l,l+=r;l>0;s=256*s+t[e+f],f+=d,l-=8);if(0===o)o=1-u;else{if(o===c)return s?NaN:1/0*(h?-1:1);s+=Math.pow(2,r),o-=u}return(h?-1:1)*s*Math.pow(2,o-r)},e.write=function(t,e,n,r,i,o){var s,a,c,u=8*o-i-1,l=(1<<u)-1,f=l>>1,d=23===i?Math.pow(2,-24)-Math.pow(2,-77):0,h=r?0:o-1,p=r?1:-1,y=e<0||0===e&&1/e<0?1:0;for(e=Math.abs(e),isNaN(e)||e===1/0?(a=isNaN(e)?1:0,s=l):(s=Math.floor(Math.log(e)/Math.LN2),e*(c=Math.pow(2,-s))<1&&(s--,c*=2),e+=s+f>=1?d/c:d*Math.pow(2,1-f),e*c>=2&&(s++,c/=2),s+f>=l?(a=0,s=l):s+f>=1?(a=(e*c-1)*Math.pow(2,i),s+=f):(a=e*Math.pow(2,f-1)*Math.pow(2,i),s=0));i>=8;t[n+h]=255&a,h+=p,a/=256,i-=8);for(s=s<<i|a,u+=i;u>0;t[n+h]=255&s,h+=p,s/=256,u-=8);t[n+h-p]|=128*y}},"94ca":function(t,e,n){var r=n("d039"),i=n("1626"),o=/#|\.prototype\./,s=function(t,e){var n=c[a(t)];return n==l||n!=u&&(i(e)?r(e):!!e)},a=s.normalize=function(t){return String(t).replace(o,".").toLowerCase()},c=s.data={},u=s.NATIVE="N",l=s.POLYFILL="P";t.exports=s},"9a1f":function(t,e,n){var r=n("da84"),i=n("c65b"),o=n("59ed"),s=n("825a"),a=n("0d51"),c=n("35a1"),u=r.TypeError;t.exports=function(t,e){var n=arguments.length<2?c(t):e;if(o(n))return s(i(n,t));throw u(a(t)+" is not iterable")}},"9bf2":function(t,e,n){var r=n("da84"),i=n("83ab"),o=n("0cfb"),s=n("825a"),a=n("a04b"),c=r.TypeError,u=Object.defineProperty;e.f=i?u:function(t,e,n){if(s(t),e=a(e),s(n),o)try{return u(t,e,n)}catch(r){}if("get"in n||"set"in n)throw c("Accessors not supported");return"value"in n&&(t[e]=n.value),t}},"9ed3":function(t,e,n){"use strict";var r=n("ae93").IteratorPrototype,i=n("7c73"),o=n("5c6c"),s=n("d44e"),a=n("3f8c"),c=function(){return this};t.exports=function(t,e,n,u){var l=e+" Iterator";return t.prototype=i(r,{next:o(+!u,n)}),s(t,l,!1,!0),a[l]=c,t}},"9f16":function(t,e,n){(function(e,r){t.exports=r(n("2b0e"))})("undefined"!==typeof self&&self,(function(t){return function(t){var e={};function n(r){if(e[r])return e[r].exports;var i=e[r]={i:r,l:!1,exports:{}};return t[r].call(i.exports,i,i.exports,n),i.l=!0,i.exports}return n.m=t,n.c=e,n.d=function(t,e,r){n.o(t,e)||Object.defineProperty(t,e,{enumerable:!0,get:r})},n.r=function(t){"undefined"!==typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(t,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(t,"__esModule",{value:!0})},n.t=function(t,e){if(1&e&&(t=n(t)),8&e)return t;if(4&e&&"object"===typeof t&&t&&t.__esModule)return t;var r=Object.create(null);if(n.r(r),Object.defineProperty(r,"default",{enumerable:!0,value:t}),2&e&&"string"!=typeof t)for(var i in t)n.d(r,i,function(e){return t[e]}.bind(null,i));return r},n.n=function(t){var e=t&&t.__esModule?function(){return t["default"]}:function(){return t};return n.d(e,"a",e),e},n.o=function(t,e){return Object.prototype.hasOwnProperty.call(t,e)},n.p="",n(n.s=68)}([function(t,e,n){"use strict";n.d(e,"d",(function(){return i})),n.d(e,"a",(function(){return o})),n.d(e,"c",(function(){return s})),n.d(e,"b",(function(){return a})),n.d(e,"e",(function(){return c})),n.d(e,"f",(function(){return u}));
/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
var r=function(t,e){return r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])},r(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n],e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t},o.apply(this,arguments)};function s(t,e,n,r){var i,o=arguments.length,s=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"===typeof Reflect&&"function"===typeof Reflect.decorate)s=Reflect.decorate(t,e,n,r);else for(var a=t.length-1;a>=0;a--)(i=t[a])&&(s=(o<3?i(s):o>3?i(e,n,s):i(e,n))||s);return o>3&&s&&Object.defineProperty(e,n,s),s}function a(t,e,n,r){function i(t){return t instanceof n?t:new n((function(e){e(t)}))}return new(n||(n=Promise))((function(n,o){function s(t){try{c(r.next(t))}catch(e){o(e)}}function a(t){try{c(r["throw"](t))}catch(e){o(e)}}function c(t){t.done?n(t.value):i(t.value).then(s,a)}c((r=r.apply(t,e||[])).next())}))}function c(t,e){var n,r,i,o,s={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:a(0),throw:a(1),return:a(2)},"function"===typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function a(t){return function(e){return c([t,e])}}function c(o){if(n)throw new TypeError("Generator is already executing.");while(s)try{if(n=1,r&&(i=2&o[0]?r["return"]:o[0]?r["throw"]||((i=r["return"])&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return s.label++,{value:o[1],done:!1};case 5:s.label++,r=o[1],o=[0];continue;case 7:o=s.ops.pop(),s.trys.pop();continue;default:if(i=s.trys,!(i=i.length>0&&i[i.length-1])&&(6===o[0]||2===o[0])){s=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){s.label=o[1];break}if(6===o[0]&&s.label<i[1]){s.label=i[1],i=o;break}if(i&&s.label<i[2]){s.label=i[2],s.ops.push(o);break}i[2]&&s.ops.pop(),s.trys.pop();continue}o=e.call(t,s)}catch(a){o=[6,a],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}}function u(t,e){var n="function"===typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),s=[];try{while((void 0===e||e-- >0)&&!(r=o.next()).done)s.push(r.value)}catch(a){i={error:a}}finally{try{r&&!r.done&&(n=o["return"])&&n.call(o)}finally{if(i)throw i.error}}return s}},function(t,e,n){"use strict";var r=n(6),i=n.n(r);n.d(e,"c",(function(){return i.a}));var o=n(2);n.d(e,"a",(function(){return o["b"]}));n(39),n(40),n(41),n(42),n(43);var s=n(44);n.d(e,"b",(function(){return s["a"]}));n(45),n(46),n(47),n(48),n(49);var a=n(50);n.d(e,"d",(function(){return a["a"]}))},function(t,e,n){"use strict";n.d(e,"a",(function(){return v}));var r=n(6),i=n.n(r);
/**
  * vue-class-component v7.2.3
  * (c) 2015-present Evan You
  * @license MIT
  */
function o(t){return o="function"===typeof Symbol&&"symbol"===typeof Symbol.iterator?function(t){return typeof t}:function(t){return t&&"function"===typeof Symbol&&t.constructor===Symbol&&t!==Symbol.prototype?"symbol":typeof t},o(t)}function s(t,e,n){return e in t?Object.defineProperty(t,e,{value:n,enumerable:!0,configurable:!0,writable:!0}):t[e]=n,t}function a(t){return c(t)||u(t)||l()}function c(t){if(Array.isArray(t)){for(var e=0,n=new Array(t.length);e<t.length;e++)n[e]=t[e];return n}}function u(t){if(Symbol.iterator in Object(t)||"[object Arguments]"===Object.prototype.toString.call(t))return Array.from(t)}function l(){throw new TypeError("Invalid attempt to spread non-iterable instance")}function f(){return"undefined"!==typeof Reflect&&Reflect.defineMetadata&&Reflect.getOwnMetadataKeys}function d(t,e){h(t,e),Object.getOwnPropertyNames(e.prototype).forEach((function(n){h(t.prototype,e.prototype,n)})),Object.getOwnPropertyNames(e).forEach((function(n){h(t,e,n)}))}function h(t,e,n){var r=n?Reflect.getOwnMetadataKeys(e,n):Reflect.getOwnMetadataKeys(e);r.forEach((function(r){var i=n?Reflect.getOwnMetadata(r,e,n):Reflect.getOwnMetadata(r,e);n?Reflect.defineMetadata(r,i,t,n):Reflect.defineMetadata(r,i,t)}))}var p={__proto__:[]},y=p instanceof Array;function v(t){return function(e,n,r){var i="function"===typeof e?e:e.constructor;i.__decorators__||(i.__decorators__=[]),"number"!==typeof r&&(r=void 0),i.__decorators__.push((function(e){return t(e,n,r)}))}}function b(t){var e=o(t);return null==t||"object"!==e&&"function"!==e}function m(t,e){var n=e.prototype._init;e.prototype._init=function(){var e=this,n=Object.getOwnPropertyNames(t);if(t.$options.props)for(var r in t.$options.props)t.hasOwnProperty(r)||n.push(r);n.forEach((function(n){"_"!==n.charAt(0)&&Object.defineProperty(e,n,{get:function(){return t[n]},set:function(e){t[n]=e},configurable:!0})}))};var r=new e;e.prototype._init=n;var i={};return Object.keys(r).forEach((function(t){void 0!==r[t]&&(i[t]=r[t])})),i}var g=["data","beforeCreate","created","beforeMount","mounted","beforeDestroy","destroyed","beforeUpdate","updated","activated","deactivated","render","errorCaptured","serverPrefetch"];function _(t){var e=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{};e.name=e.name||t._componentTag||t.name;var n=t.prototype;Object.getOwnPropertyNames(n).forEach((function(t){if("constructor"!==t)if(g.indexOf(t)>-1)e[t]=n[t];else{var r=Object.getOwnPropertyDescriptor(n,t);void 0!==r.value?"function"===typeof r.value?(e.methods||(e.methods={}))[t]=r.value:(e.mixins||(e.mixins=[])).push({data:function(){return s({},t,r.value)}}):(r.get||r.set)&&((e.computed||(e.computed={}))[t]={get:r.get,set:r.set})}})),(e.mixins||(e.mixins=[])).push({data:function(){return m(this,t)}});var r=t.__decorators__;r&&(r.forEach((function(t){return t(e)})),delete t.__decorators__);var o=Object.getPrototypeOf(t.prototype),a=o instanceof i.a?o.constructor:i.a,c=a.extend(e);return O(c,t,a),f()&&d(c,t),c}var w={prototype:!0,arguments:!0,callee:!0,caller:!0};function O(t,e,n){Object.getOwnPropertyNames(e).forEach((function(r){if(!w[r]){var i=Object.getOwnPropertyDescriptor(t,r);if(!i||i.configurable){var o=Object.getOwnPropertyDescriptor(e,r);if(!y){if("cid"===r)return;var s=Object.getOwnPropertyDescriptor(n,r);if(!b(o.value)&&s&&s.value===o.value)return}0,Object.defineProperty(t,r,o)}}}))}function S(t){return"function"===typeof t?_(t):function(e){return _(e,t)}}S.registerHooks=function(t){g.push.apply(g,a(t))},e["b"]=S},function(t,e,n){"use strict";function r(t,e,n,r,i,o,s,a){var c,u="function"===typeof t?t.options:t;if(e&&(u.render=e,u.staticRenderFns=n,u._compiled=!0),r&&(u.functional=!0),o&&(u._scopeId="data-v-"+o),s?(c=function(t){t=t||this.$vnode&&this.$vnode.ssrContext||this.parent&&this.parent.$vnode&&this.parent.$vnode.ssrContext,t||"undefined"===typeof __VUE_SSR_CONTEXT__||(t=__VUE_SSR_CONTEXT__),i&&i.call(this,t),t&&t._registeredComponents&&t._registeredComponents.add(s)},u._ssrRegister=c):i&&(c=a?function(){i.call(this,(u.functional?this.parent:this).$root.$options.shadowRoot)}:i),c)if(u.functional){u._injectStyles=c;var l=u.render;u.render=function(t,e){return c.call(e),l(t,e)}}else{var f=u.beforeCreate;u.beforeCreate=f?[].concat(f,c):[c]}return{exports:t,options:u}}n.d(e,"a",(function(){return r}))},function(t,e,n){"use strict";function r(t){return"function"!==typeof t||!t.managed&&!t.managedReactive}function i(t){var e=function(){var n=this,r="function"===typeof t?t.call(this):t;for(var i in r=Object.create(r||null),r[o]=Object.create(this[o]||{}),e.managed)r[e.managed[i]]=this[i];var s=function(t){r[e.managedReactive[t]]=a[t],Object.defineProperty(r[o],e.managedReactive[t],{enumerable:!0,configurable:!0,get:function(){return n[t]}})},a=this;for(var i in e.managedReactive)s(i);return r};return e.managed={},e.managedReactive={},e}n.d(e,"b",(function(){return r})),n.d(e,"c",(function(){return i})),n.d(e,"d",(function(){return o})),n.d(e,"a",(function(){return s}));var o="__reactiveInject__";function s(t){Array.isArray(t.inject)||(t.inject=t.inject||{},t.inject[o]={from:o,default:{}})}},function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));var r="undefined"!==typeof Reflect&&"undefined"!==typeof Reflect.getMetadata;function i(t,e,n){if(r&&!Array.isArray(t)&&"function"!==typeof t&&!t.hasOwnProperty("type")&&"undefined"===typeof t.type){var i=Reflect.getMetadata("design:type",e,n);i!==Object&&(t.type=i)}}},function(e,n){e.exports=t},function(t,e,n){"use strict";var r=n(32),i=n(18),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/Arrow.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return o}));var r=n(0),i=n(1),o=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.MAX_STRING_LENGTH=9,e.editMode=!1,e.invalid=!1,e.tempValue="0",e}return Object(r["d"])(e,t),Object.defineProperty(e.prototype,"v",{get:function(){return"string"===typeof this.value?parseFloat(this.value):"number"===typeof this.value?this.value:0},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"stringRepresentation",{get:function(){var t=this.v.toFixed(3);return t.length>this.MAX_STRING_LENGTH?this.v.toExponential(this.MAX_STRING_LENGTH-5):t},enumerable:!1,configurable:!0}),e.prototype.setValue=function(t){this.validate(t)&&this.$emit("input",t)},e.prototype.resetInvalid=function(){this.invalid=!1},e.prototype.enterEditMode=function(){return Object(r["b"])(this,void 0,void 0,(function(){return Object(r["e"])(this,(function(t){switch(t.label){case 0:return this.tempValue=this.v.toFixed(3),this.editMode=!0,[4,this.$nextTick()];case 1:return t.sent(),this.$refs.input.focus(),[2]}}))}))},e.prototype.leaveEditMode=function(){var t=parseFloat(this.tempValue);this.validate(t)?(this.$emit("input",t),this.editMode=!1):this.invalid=!0},e.prototype.validate=function(t){return!Number.isNaN(t)&&(!("number"===typeof this.option.min&&t<this.option.min)&&!("number"===typeof this.option.max&&t>this.option.max))},Object(r["c"])([Object(i["b"])()],e.prototype,"value",void 0),Object(r["c"])([Object(i["b"])({type:String})],e.prototype,"name",void 0),Object(r["c"])([Object(i["b"])({type:Object})],e.prototype,"option",void 0),Object(r["c"])([Object(i["d"])("tempValue")],e.prototype,"resetInvalid",null),e=Object(r["c"])([i["a"]],e),e}(i["c"])},function(t,e,n){"use strict";n.r(e);var r=n(37);n.d(e,"ButtonOption",(function(){return r["a"]}));var i=n(51);n.d(e,"CheckboxOption",(function(){return i["a"]}));var o=n(53);n.d(e,"InputOption",(function(){return o["a"]}));var s=n(55);n.d(e,"IntegerOption",(function(){return s["a"]}));var a=n(58);n.d(e,"NumberOption",(function(){return a["a"]}));var c=n(60);n.d(e,"SelectOption",(function(){return c["a"]}));var u=n(62);n.d(e,"SliderOption",(function(){return u["a"]}));var l=n(64);n.d(e,"TextOption",(function(){return l["a"]}))},function(t,e,n){"use strict";var r=n(11);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r["d"])(e,t),Object(r["c"])([Object(i["b"])({type:String})],e.prototype,"name",void 0),e=Object(r["c"])([i["a"]],e),e}(i["c"]);e["a"]=o},function(t,e,n){"use strict";var r=n(13);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r["d"])(e,t),Object(r["c"])([Object(i["b"])({type:Boolean})],e.prototype,"value",void 0),Object(r["c"])([Object(i["b"])({type:String})],e.prototype,"name",void 0),e=Object(r["c"])([i["a"]],e),e}(i["c"]);e["a"]=o},function(t,e,n){"use strict";var r=n(15);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r["d"])(e,t),Object.defineProperty(e.prototype,"listeners",{get:function(){var t=this;return Object(r["a"])(Object(r["a"])({},this.$listeners),{input:function(e){return t.$emit("input",e.target.value)}})},enumerable:!1,configurable:!0}),Object(r["c"])([Object(i["b"])({type:String,default:""})],e.prototype,"value",void 0),Object(r["c"])([Object(i["b"])({type:String})],e.prototype,"name",void 0),e=Object(r["c"])([i["a"]],e),e}(i["c"]);e["a"]=o},function(t,e,n){"use strict";var r=n(17);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=n(7),s=n(8),a=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r["d"])(e,t),Object.defineProperty(e.prototype,"v",{get:function(){return"string"===typeof this.value?parseInt(this.value,10):"number"===typeof this.value?Math.floor(this.value):0},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"stringRepresentation",{get:function(){var t=this.v.toString();return t.length>this.MAX_STRING_LENGTH?this.v.toExponential(this.MAX_STRING_LENGTH-5):t},enumerable:!1,configurable:!0}),e.prototype.increment=function(){this.setValue(this.v+1)},e.prototype.decrement=function(){this.setValue(this.v-1)},e.prototype.leaveEditMode=function(){var t=parseInt(this.tempValue,10);this.validate(t)?(this.$emit("input",t),this.editMode=!1):this.invalid=!0},e=Object(r["c"])([Object(i["a"])({components:{"i-arrow":o["a"]}})],e),e}(s["a"]);e["a"]=a},function(t,e,n){"use strict";var r=n(19);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r["d"])(e,t),e=Object(r["c"])([i["a"]],e),e}(i["c"]);e["a"]=o},function(t,e,n){"use strict";var r=n(21);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=n(7),s=n(8),a=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r["d"])(e,t),e.prototype.increment=function(){this.setValue(this.v+.1)},e.prototype.decrement=function(){this.setValue(this.v-.1)},e=Object(r["c"])([Object(i["a"])({components:{"i-arrow":o["a"]}})],e),e}(s["a"]);e["a"]=a},function(t,e,n){"use strict";var r=n(23);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=n(7),s=n(67),a=n.n(s),c=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.open=!1,e.items=[],e}return Object(r["d"])(e,t),Object.defineProperty(e.prototype,"isAdvancedMode",{get:function(){return!this.items.every((function(t){return"string"===typeof t}))},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"selectedText",{get:function(){var t,e;return this.value?this.isAdvancedMode?null!==(e=null===(t=this.getItemByValue(this.value))||void 0===t?void 0:t.text)&&void 0!==e?e:"":this.value:""},enumerable:!1,configurable:!0}),e.prototype.mounted=function(){var t=this;this.items=this.option.items||[],this.option.events.updated.addListener(this,(function(){t.items=t.option.items||[]}))},e.prototype.beforeDestroy=function(){this.option.events.updated.removeListener(this)},e.prototype.isSelected=function(t){return this.isAdvancedMode?t.value===this.value:t===this.value},e.prototype.setSelected=function(t){this.$emit("input",this.isAdvancedMode?t.value:t)},e.prototype.getItemByValue=function(t){return this.items.find((function(e){return e.value===t}))},Object(r["c"])([Object(i["b"])({type:String})],e.prototype,"name",void 0),Object(r["c"])([Object(i["b"])()],e.prototype,"value",void 0),Object(r["c"])([Object(i["b"])({type:Object})],e.prototype,"option",void 0),e=Object(r["c"])([Object(i["a"])({components:{"i-arrow":o["a"]},directives:{ClickOutside:a.a.directive}})],e),e}(i["c"]);e["a"]=c},function(t,e,n){"use strict";var r=n(25);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=n(8),s=function(t){function e(){var e=null!==t&&t.apply(this,arguments)||this;return e.didSlide=!1,e.isMouseDown=!1,e}return Object(r["d"])(e,t),Object.defineProperty(e.prototype,"min",{get:function(){return this.option.min||0},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"max",{get:function(){return this.option.max||1},enumerable:!1,configurable:!0}),Object.defineProperty(e.prototype,"percentage",{get:function(){return Math.min(100,Math.max(0,100*this.v/(this.max-this.min)))},enumerable:!1,configurable:!0}),e.prototype.mousedown=function(){this.editMode||(this.isMouseDown=!0)},e.prototype.mouseup=function(){this.editMode||(this.didSlide||this.enterEditMode(),this.isMouseDown=!1,this.didSlide=!1)},e.prototype.mouseleave=function(t){this.editMode||(this.isMouseDown&&(t.offsetX>=this.$el.clientWidth?this.$emit("input",this.max):t.offsetX<=0&&this.$emit("input",this.min)),this.isMouseDown=!1,this.didSlide=!1)},e.prototype.mousemove=function(t){if(!this.editMode){var e=Math.max(this.min,Math.min(this.max,(this.max-this.min)*(t.offsetX/this.$el.clientWidth)+this.min));this.isMouseDown&&(this.$emit("input",e),this.didSlide=!0)}},e=Object(r["c"])([i["a"]],e),e}(o["a"]);e["a"]=s},function(t,e,n){"use strict";var r=n(27);e["a"]=r["a"]},function(t,e,n){"use strict";var r=n(0),i=n(1),o=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return Object(r["d"])(e,t),Object.defineProperty(e.prototype,"sanitized",{get:function(){return String(this.value)},enumerable:!1,configurable:!0}),Object(r["c"])([Object(i["b"])({default:""})],e.prototype,"value",void 0),e=Object(r["c"])([i["a"]],e),e}(i["c"]);e["a"]=o},function(t,e,n){"use strict";var r=n(38);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(52);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(54);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(56);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(57);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(59);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(61);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(63);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(65);n.d(e,"a",(function(){return r["a"]})),n.d(e,"b",(function(){return r["b"]}))},function(t,e,n){"use strict";var r=n(28),i=n(10),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/ButtonOption.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("button",{staticClass:"dark-button --block",on:{click:function(e){return t.$emit("openSidebar")}}},[t._v(t._s(t.name))])},i=[];r._withStripped=!0},function(t,e,n){"use strict"},function(t,e,n){"use strict";n(2)},function(t,e,n){"use strict";n(2),n(4)},function(t,e,n){"use strict";n(2),n(5)},function(t,e,n){"use strict";n(2),n(5)},function(t,e,n){"use strict";n.d(e,"a",(function(){return o}));var r=n(2),i=n(5);function o(t){return void 0===t&&(t={}),function(e,n){Object(i["a"])(t,e,n),Object(r["a"])((function(e,n){(e.props||(e.props={}))[n]=t}))(e,n)}}},function(t,e,n){"use strict";n(2),n(5)},function(t,e,n){"use strict";n(2),n(4)},function(t,e,n){"use strict";n(2),n(4)},function(t,e,n){"use strict";n(2)},function(t,e,n){"use strict";n(2)},function(t,e,n){"use strict";n.d(e,"a",(function(){return i}));var r=n(2);function i(t,e){void 0===e&&(e={});var n=e.deep,i=void 0!==n&&n,o=e.immediate,s=void 0!==o&&o;return Object(r["a"])((function(e,n){"object"!==typeof e.watch&&(e.watch=Object.create(null));var r=e.watch;"object"!==typeof r[t]||Array.isArray(r[t])?"undefined"===typeof r[t]&&(r[t]=[]):r[t]=[r[t]],r[t].push({handler:n,deep:i,immediate:s})}))}},function(t,e,n){"use strict";var r=n(29),i=n(12),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/CheckboxOption.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{class:{"dark-checkbox":!0,"--checked":t.value},on:{click:function(e){return t.$emit("input",!t.value)}}},[n("div",{staticClass:"__checkmark-container"},[n("svg",{attrs:{xmlns:"http://www.w3.org/2000/svg",width:"18",height:"18",viewBox:"0 0 18 18"}},[n("path",{staticClass:"__checkmark",attrs:{d:"M 6 5 L 6 10 L 16 10",transform:"rotate(-45 10 10)"}})])]),t._v(" "),n("div",{staticClass:"ml-2"},[t._v(t._s(t.name))])])},i=[];r._withStripped=!0},function(t,e,n){"use strict";var r=n(30),i=n(14),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/InputOption.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",[n("input",t._g({staticClass:"dark-input",attrs:{type:"text",placeholder:t.name},domProps:{value:t.value}},t.listeners))])},i=[];r._withStripped=!0},function(t,e,n){"use strict";var r=n(31),i=n(16),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/IntegerOption.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{staticClass:"dark-num-input"},[n("div",{staticClass:"__button --dec",on:{click:t.decrement}},[n("i-arrow")],1),t._v(" "),t.editMode?n("div",{staticClass:"__content"},[n("input",{directives:[{name:"model",rawName:"v-model",value:t.tempValue,expression:"tempValue"}],ref:"input",attrs:{type:"number"},domProps:{value:t.tempValue},on:{blur:t.leaveEditMode,input:function(e){e.target.composing||(t.tempValue=e.target.value)}}})]):n("div",{staticClass:"__content",on:{click:t.enterEditMode}},[n("div",{staticClass:"__label .text-truncate"},[t._v(t._s(t.name))]),t._v(" "),n("div",{staticClass:"__value"},[t._v(t._s(t.stringRepresentation))])]),t._v(" "),n("div",{staticClass:"__button --inc",on:{click:t.increment}},[n("i-arrow")],1)])},i=[];r._withStripped=!0},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("svg",{attrs:{xmlns:"http://www.w3.org/2000/svg",width:"24",height:"24",viewBox:"0 0 24 24"}},[n("path",{attrs:{d:"M7.41 8.59L12 13.17l4.59-4.58L18 10l-6 6-6-6 1.41-1.41z"}})])},i=[];r._withStripped=!0},function(t,e,n){"use strict";var r=n(33),i=n(20),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/NumberOption.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{staticClass:"dark-num-input"},[n("div",{staticClass:"__button --dec",on:{click:t.decrement}},[n("i-arrow")],1),t._v(" "),t.editMode?n("div",{staticClass:"__content"},[n("input",{directives:[{name:"model",rawName:"v-model",value:t.tempValue,expression:"tempValue"}],ref:"input",staticClass:"dark-input",class:{"--invalid":t.invalid},staticStyle:{"text-align":"right"},attrs:{type:"number"},domProps:{value:t.tempValue},on:{blur:t.leaveEditMode,keydown:function(e){return!e.type.indexOf("key")&&t._k(e.keyCode,"enter",13,e.key,"Enter")?null:t.leaveEditMode(e)},input:function(e){e.target.composing||(t.tempValue=e.target.value)}}})]):n("div",{staticClass:"__content",on:{click:t.enterEditMode}},[n("div",{staticClass:"__label .text-truncate"},[t._v(t._s(t.name))]),t._v(" "),n("div",{staticClass:"__value"},[t._v(t._s(t.stringRepresentation))])]),t._v(" "),n("div",{staticClass:"__button --inc",on:{click:t.increment}},[n("i-arrow")],1)])},i=[];r._withStripped=!0},function(t,e,n){"use strict";var r=n(34),i=n(22),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/SelectOption.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{directives:[{name:"click-outside",rawName:"v-click-outside",value:function(){t.open=!1},expression:"() => { open = false; }"}],class:["dark-select",{"--open":t.open}],on:{click:function(e){t.open=!t.open}}},[n("div",{staticClass:"__selected"},[n("div",{staticClass:"__text"},[t._v(t._s(t.selectedText))]),t._v(" "),n("div",{staticClass:"__icon"},[n("i-arrow")],1)]),t._v(" "),n("transition",{attrs:{name:"slide-fade"}},[n("div",{directives:[{name:"show",rawName:"v-show",value:t.open,expression:"open"}],staticClass:"__dropdown"},[n("div",{staticClass:"item --header"},[t._v(t._s(t.name))]),t._v(" "),t._l(t.items,(function(e,r){return n("div",{key:r,class:["item",{"--active":t.isSelected(e)}],on:{click:function(n){return t.setSelected(e)}}},[t._v(t._s(t.isAdvancedMode?e.text:e))])}))],2)])],1)},i=[];r._withStripped=!0},function(t,e,n){"use strict";var r=n(35),i=n(24),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/SliderOption.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",{staticClass:"dark-slider",class:{"ignore-mouse":!t.editMode},on:{mousedown:t.mousedown,mouseup:t.mouseup,mousemove:t.mousemove,mouseleave:t.mouseleave}},[n("div",{staticClass:"__slider",style:{width:t.percentage+"%"}}),t._v(" "),t.editMode?n("div",{staticClass:"__content"},[n("input",{directives:[{name:"model",rawName:"v-model",value:t.tempValue,expression:"tempValue"}],ref:"input",staticClass:"dark-input",class:{"--invalid":t.invalid},staticStyle:{"text-align":"right"},attrs:{type:"number"},domProps:{value:t.tempValue},on:{blur:t.leaveEditMode,keydown:function(e){return!e.type.indexOf("key")&&t._k(e.keyCode,"enter",13,e.key,"Enter")?null:t.leaveEditMode(e)},input:function(e){e.target.composing||(t.tempValue=e.target.value)}}})]):n("div",{staticClass:"__content"},[n("div",{staticClass:"__label"},[t._v(t._s(t.name))]),t._v(" "),n("div",{staticClass:"__value"},[t._v(t._s(t.stringRepresentation))])])])},i=[];r._withStripped=!0},function(t,e,n){"use strict";var r=n(36),i=n(26),o=n(3),s=Object(o["a"])(i["a"],r["a"],r["b"],!1,null,null,null);s.options.__file="src/TextOption.vue",e["a"]=s.exports},function(t,e,n){"use strict";n.d(e,"a",(function(){return r})),n.d(e,"b",(function(){return i}));var r=function(){var t=this,e=t.$createElement,n=t._self._c||e;return n("div",[t._v(t._s(t.sanitized))])},i=[];r._withStripped=!0},function(t,e,n){"use strict";n.d(e,"a",(function(){return o}));var r=n(0),i=n(9),o=function(){function t(){this.type="OptionPlugin"}return t.prototype.register=function(t){var e=this;t.events.usePlugin.addListener(this,(function(t){"ViewPlugin"===t.type&&e.registerOptions(t)})),t.plugins.forEach((function(t){"ViewPlugin"===t.type&&e.registerOptions(t)}))},t.prototype.registerOptions=function(t){Object.entries(i).forEach((function(e){var n=Object(r["f"])(e,2),i=n[0],o=n[1];t.registerOption(i,o)}))},t}()},function(t,e,n){!function(e,n){t.exports=n()}(0,(function(){var t="undefined"!=typeof window,e="undefined"!=typeof navigator,n=t&&("ontouchstart"in window||e&&navigator.msMaxTouchPoints>0)?["touchstart"]:["click"];function r(t){var e=t.event,n=t.handler;(0,t.middleware)(e)&&n(e)}function i(t,e){var i=function(t){var e="function"==typeof t;if(!e&&"object"!=typeof t)throw new Error("v-click-outside: Binding value must be a function or an object");return{handler:e?t:t.handler,middleware:t.middleware||function(t){return t},events:t.events||n,isActive:!(!1===t.isActive),detectIframe:!(!1===t.detectIframe)}}(e.value),o=i.handler,s=i.middleware,a=i.detectIframe;if(i.isActive){if(t["__v-click-outside"]=i.events.map((function(e){return{event:e,srcTarget:document.documentElement,handler:function(e){return function(t){var e=t.el,n=t.event,i=t.handler,o=t.middleware,s=n.path||n.composedPath&&n.composedPath();(s?s.indexOf(e)<0:!e.contains(n.target))&&r({event:n,handler:i,middleware:o})}({el:t,event:e,handler:o,middleware:s})}}})),a){var c={event:"blur",srcTarget:window,handler:function(e){return function(t){var e=t.el,n=t.event,i=t.handler,o=t.middleware;setTimeout((function(){var t=document.activeElement;t&&"IFRAME"===t.tagName&&!e.contains(t)&&r({event:n,handler:i,middleware:o})}),0)}({el:t,event:e,handler:o,middleware:s})}};t["__v-click-outside"]=[].concat(t["__v-click-outside"],[c])}t["__v-click-outside"].forEach((function(e){var n=e.event,r=e.srcTarget,i=e.handler;return setTimeout((function(){t["__v-click-outside"]&&r.addEventListener(n,i,!1)}),0)}))}}function o(t){(t["__v-click-outside"]||[]).forEach((function(t){return t.srcTarget.removeEventListener(t.event,t.handler,!1)})),delete t["__v-click-outside"]}var s=t?{bind:i,update:function(t,e){var n=e.value,r=e.oldValue;JSON.stringify(n)!==JSON.stringify(r)&&(o(t),i(t,{value:n}))},unbind:o}:{};return{install:function(t){t.directive("click-outside",s)},directive:s}}))},function(t,e,n){"use strict";n.r(e);var r=n(9);n.d(e,"ButtonOption",(function(){return r["ButtonOption"]})),n.d(e,"CheckboxOption",(function(){return r["CheckboxOption"]})),n.d(e,"InputOption",(function(){return r["InputOption"]})),n.d(e,"IntegerOption",(function(){return r["IntegerOption"]})),n.d(e,"NumberOption",(function(){return r["NumberOption"]})),n.d(e,"SelectOption",(function(){return r["SelectOption"]})),n.d(e,"SliderOption",(function(){return r["SliderOption"]})),n.d(e,"TextOption",(function(){return r["TextOption"]}));var i=n(66);n.d(e,"OptionPlugin",(function(){return i["a"]}))}])}))},a04b:function(t,e,n){var r=n("c04e"),i=n("d9b5");t.exports=function(t){var e=r(t,"string");return i(e)?e:e+""}},a4b4:function(t,e,n){var r=n("342f");t.exports=/web0s(?!.*chrome)/i.test(r)},a640:function(t,e,n){"use strict";var r=n("d039");t.exports=function(t,e){var n=[][t];return!!n&&r((function(){n.call(null,e||function(){throw 1},1)}))}},a6b2:function(t,e,n){"use strict";n.d(e,"a",(function(){return r}));var r={};r.Offset,r.Table,r.SIZEOF_SHORT=2,r.SIZEOF_INT=4,r.FILE_IDENTIFIER_LENGTH=4,r.Encoding={UTF8_BYTES:1,UTF16_STRING:2},r.int32=new Int32Array(2),r.float32=new Float32Array(r.int32.buffer),r.float64=new Float64Array(r.int32.buffer),r.isLittleEndian=1===new Uint16Array(new Uint8Array([1,0]).buffer)[0],r.Long=function(t,e){this.low=0|t,this.high=0|e},r.Long.create=function(t,e){return 0==t&&0==e?r.Long.ZERO:new r.Long(t,e)},r.Long.prototype.toFloat64=function(){return(this.low>>>0)+4294967296*this.high},r.Long.prototype.equals=function(t){return this.low==t.low&&this.high==t.high},r.Long.ZERO=new r.Long(0,0),r.Builder=function(t){if(t)e=t;else var e=1024;this.bb=r.ByteBuffer.allocate(e),this.space=e,this.minalign=1,this.vtable=null,this.vtable_in_use=0,this.isNested=!1,this.object_start=0,this.vtables=[],this.vector_num_elems=0,this.force_defaults=!1},r.Builder.prototype.clear=function(){this.bb.clear(),this.space=this.bb.capacity(),this.minalign=1,this.vtable=null,this.vtable_in_use=0,this.isNested=!1,this.object_start=0,this.vtables=[],this.vector_num_elems=0,this.force_defaults=!1},r.Builder.prototype.forceDefaults=function(t){this.force_defaults=t},r.Builder.prototype.dataBuffer=function(){return this.bb},r.Builder.prototype.asUint8Array=function(){return this.bb.bytes().subarray(this.bb.position(),this.bb.position()+this.offset())},r.Builder.prototype.prep=function(t,e){t>this.minalign&&(this.minalign=t);var n=1+~(this.bb.capacity()-this.space+e)&t-1;while(this.space<n+t+e){var i=this.bb.capacity();this.bb=r.Builder.growByteBuffer(this.bb),this.space+=this.bb.capacity()-i}this.pad(n)},r.Builder.prototype.pad=function(t){for(var e=0;e<t;e++)this.bb.writeInt8(--this.space,0)},r.Builder.prototype.writeInt8=function(t){this.bb.writeInt8(this.space-=1,t)},r.Builder.prototype.writeInt16=function(t){this.bb.writeInt16(this.space-=2,t)},r.Builder.prototype.writeInt32=function(t){this.bb.writeInt32(this.space-=4,t)},r.Builder.prototype.writeInt64=function(t){this.bb.writeInt64(this.space-=8,t)},r.Builder.prototype.writeFloat32=function(t){this.bb.writeFloat32(this.space-=4,t)},r.Builder.prototype.writeFloat64=function(t){this.bb.writeFloat64(this.space-=8,t)},r.Builder.prototype.addInt8=function(t){this.prep(1,0),this.writeInt8(t)},r.Builder.prototype.addInt16=function(t){this.prep(2,0),this.writeInt16(t)},r.Builder.prototype.addInt32=function(t){this.prep(4,0),this.writeInt32(t)},r.Builder.prototype.addInt64=function(t){this.prep(8,0),this.writeInt64(t)},r.Builder.prototype.addFloat32=function(t){this.prep(4,0),this.writeFloat32(t)},r.Builder.prototype.addFloat64=function(t){this.prep(8,0),this.writeFloat64(t)},r.Builder.prototype.addFieldInt8=function(t,e,n){(this.force_defaults||e!=n)&&(this.addInt8(e),this.slot(t))},r.Builder.prototype.addFieldInt16=function(t,e,n){(this.force_defaults||e!=n)&&(this.addInt16(e),this.slot(t))},r.Builder.prototype.addFieldInt32=function(t,e,n){(this.force_defaults||e!=n)&&(this.addInt32(e),this.slot(t))},r.Builder.prototype.addFieldInt64=function(t,e,n){!this.force_defaults&&e.equals(n)||(this.addInt64(e),this.slot(t))},r.Builder.prototype.addFieldFloat32=function(t,e,n){(this.force_defaults||e!=n)&&(this.addFloat32(e),this.slot(t))},r.Builder.prototype.addFieldFloat64=function(t,e,n){(this.force_defaults||e!=n)&&(this.addFloat64(e),this.slot(t))},r.Builder.prototype.addFieldOffset=function(t,e,n){(this.force_defaults||e!=n)&&(this.addOffset(e),this.slot(t))},r.Builder.prototype.addFieldStruct=function(t,e,n){e!=n&&(this.nested(e),this.slot(t))},r.Builder.prototype.nested=function(t){if(t!=this.offset())throw new Error("FlatBuffers: struct must be serialized inline.")},r.Builder.prototype.notNested=function(){if(this.isNested)throw new Error("FlatBuffers: object serialization must not be nested.")},r.Builder.prototype.slot=function(t){this.vtable[t]=this.offset()},r.Builder.prototype.offset=function(){return this.bb.capacity()-this.space},r.Builder.growByteBuffer=function(t){var e=t.capacity();if(3221225472&e)throw new Error("FlatBuffers: cannot grow buffer beyond 2 gigabytes.");var n=e<<1,i=r.ByteBuffer.allocate(n);return i.setPosition(n-e),i.bytes().set(t.bytes(),n-e),i},r.Builder.prototype.addOffset=function(t){this.prep(r.SIZEOF_INT,0),this.writeInt32(this.offset()-t+r.SIZEOF_INT)},r.Builder.prototype.startObject=function(t){this.notNested(),null==this.vtable&&(this.vtable=[]),this.vtable_in_use=t;for(var e=0;e<t;e++)this.vtable[e]=0;this.isNested=!0,this.object_start=this.offset()},r.Builder.prototype.endObject=function(){if(null==this.vtable||!this.isNested)throw new Error("FlatBuffers: endObject called without startObject");this.addInt32(0);for(var t=this.offset(),e=this.vtable_in_use-1;e>=0&&0==this.vtable[e];e--);for(var n=e+1;e>=0;e--)this.addInt16(0!=this.vtable[e]?t-this.vtable[e]:0);var i=2;this.addInt16(t-this.object_start);var o=(n+i)*r.SIZEOF_SHORT;this.addInt16(o);var s=0,a=this.space;t:for(e=0;e<this.vtables.length;e++){var c=this.bb.capacity()-this.vtables[e];if(o==this.bb.readInt16(c)){for(var u=r.SIZEOF_SHORT;u<o;u+=r.SIZEOF_SHORT)if(this.bb.readInt16(a+u)!=this.bb.readInt16(c+u))continue t;s=this.vtables[e];break}}return s?(this.space=this.bb.capacity()-t,this.bb.writeInt32(this.space,s-t)):(this.vtables.push(this.offset()),this.bb.writeInt32(this.bb.capacity()-t,this.offset()-t)),this.isNested=!1,t},r.Builder.prototype.finish=function(t,e){if(e){var n=e;if(this.prep(this.minalign,r.SIZEOF_INT+r.FILE_IDENTIFIER_LENGTH),n.length!=r.FILE_IDENTIFIER_LENGTH)throw new Error("FlatBuffers: file identifier must be length "+r.FILE_IDENTIFIER_LENGTH);for(var i=r.FILE_IDENTIFIER_LENGTH-1;i>=0;i--)this.writeInt8(n.charCodeAt(i))}this.prep(this.minalign,r.SIZEOF_INT),this.addOffset(t),this.bb.setPosition(this.space)},r.Builder.prototype.requiredField=function(t,e){var n=this.bb.capacity()-t,r=n-this.bb.readInt32(n),i=0!=this.bb.readInt16(r+e);if(!i)throw new Error("FlatBuffers: field "+e+" must be set")},r.Builder.prototype.startVector=function(t,e,n){this.notNested(),this.vector_num_elems=e,this.prep(r.SIZEOF_INT,t*e),this.prep(n,t*e)},r.Builder.prototype.endVector=function(){return this.writeInt32(this.vector_num_elems),this.offset()},r.Builder.prototype.createString=function(t){if(t instanceof Uint8Array)var e=t;else{e=[];var n=0;while(n<t.length){var r,i=t.charCodeAt(n++);if(i<55296||i>=56320)r=i;else{var o=t.charCodeAt(n++);r=(i<<10)+o+-56613888}r<128?e.push(r):(r<2048?e.push(r>>6&31|192):(r<65536?e.push(r>>12&15|224):e.push(r>>18&7|240,r>>12&63|128),e.push(r>>6&63|128)),e.push(63&r|128))}}this.addInt8(0),this.startVector(1,e.length,1),this.bb.setPosition(this.space-=e.length);n=0;for(var s=this.space,a=this.bb.bytes();n<e.length;n++)a[s++]=e[n];return this.endVector()},r.Builder.prototype.createLong=function(t,e){return r.Long.create(t,e)},r.ByteBuffer=function(t){this.bytes_=t,this.position_=0},r.ByteBuffer.allocate=function(t){return new r.ByteBuffer(new Uint8Array(t))},r.ByteBuffer.prototype.clear=function(){this.position_=0},r.ByteBuffer.prototype.bytes=function(){return this.bytes_},r.ByteBuffer.prototype.position=function(){return this.position_},r.ByteBuffer.prototype.setPosition=function(t){this.position_=t},r.ByteBuffer.prototype.capacity=function(){return this.bytes_.length},r.ByteBuffer.prototype.readInt8=function(t){return this.readUint8(t)<<24>>24},r.ByteBuffer.prototype.readUint8=function(t){return this.bytes_[t]},r.ByteBuffer.prototype.readInt16=function(t){return this.readUint16(t)<<16>>16},r.ByteBuffer.prototype.readUint16=function(t){return this.bytes_[t]|this.bytes_[t+1]<<8},r.ByteBuffer.prototype.readInt32=function(t){return this.bytes_[t]|this.bytes_[t+1]<<8|this.bytes_[t+2]<<16|this.bytes_[t+3]<<24},r.ByteBuffer.prototype.readUint32=function(t){return this.readInt32(t)>>>0},r.ByteBuffer.prototype.readInt64=function(t){return new r.Long(this.readInt32(t),this.readInt32(t+4))},r.ByteBuffer.prototype.readUint64=function(t){return new r.Long(this.readUint32(t),this.readUint32(t+4))},r.ByteBuffer.prototype.readFloat32=function(t){return r.int32[0]=this.readInt32(t),r.float32[0]},r.ByteBuffer.prototype.readFloat64=function(t){return r.int32[r.isLittleEndian?0:1]=this.readInt32(t),r.int32[r.isLittleEndian?1:0]=this.readInt32(t+4),r.float64[0]},r.ByteBuffer.prototype.writeInt8=function(t,e){this.bytes_[t]=e},r.ByteBuffer.prototype.writeUint8=function(t,e){this.bytes_[t]=e},r.ByteBuffer.prototype.writeInt16=function(t,e){this.bytes_[t]=e,this.bytes_[t+1]=e>>8},r.ByteBuffer.prototype.writeUint16=function(t,e){this.bytes_[t]=e,this.bytes_[t+1]=e>>8},r.ByteBuffer.prototype.writeInt32=function(t,e){this.bytes_[t]=e,this.bytes_[t+1]=e>>8,this.bytes_[t+2]=e>>16,this.bytes_[t+3]=e>>24},r.ByteBuffer.prototype.writeUint32=function(t,e){this.bytes_[t]=e,this.bytes_[t+1]=e>>8,this.bytes_[t+2]=e>>16,this.bytes_[t+3]=e>>24},r.ByteBuffer.prototype.writeInt64=function(t,e){this.writeInt32(t,e.low),this.writeInt32(t+4,e.high)},r.ByteBuffer.prototype.writeUint64=function(t,e){this.writeUint32(t,e.low),this.writeUint32(t+4,e.high)},r.ByteBuffer.prototype.writeFloat32=function(t,e){r.float32[0]=e,this.writeInt32(t,r.int32[0])},r.ByteBuffer.prototype.writeFloat64=function(t,e){r.float64[0]=e,this.writeInt32(t,r.int32[r.isLittleEndian?0:1]),this.writeInt32(t+4,r.int32[r.isLittleEndian?1:0])},r.ByteBuffer.prototype.getBufferIdentifier=function(){if(this.bytes_.length<this.position_+r.SIZEOF_INT+r.FILE_IDENTIFIER_LENGTH)throw new Error("FlatBuffers: ByteBuffer is too short to contain an identifier.");for(var t="",e=0;e<r.FILE_IDENTIFIER_LENGTH;e++)t+=String.fromCharCode(this.readInt8(this.position_+r.SIZEOF_INT+e));return t},r.ByteBuffer.prototype.__offset=function(t,e){var n=t-this.readInt32(t);return e<this.readInt16(n)?this.readInt16(n+e):0},r.ByteBuffer.prototype.__union=function(t,e){return t.bb_pos=e+this.readInt32(e),t.bb=this,t},r.ByteBuffer.prototype.__string=function(t,e){t+=this.readInt32(t);var n=this.readInt32(t),i="",o=0;if(t+=r.SIZEOF_INT,e===r.Encoding.UTF8_BYTES)return this.bytes_.subarray(t,t+n);while(o<n){var s,a=this.readUint8(t+o++);if(a<192)s=a;else{var c=this.readUint8(t+o++);if(a<224)s=(31&a)<<6|63&c;else{var u=this.readUint8(t+o++);if(a<240)s=(15&a)<<12|(63&c)<<6|63&u;else{var l=this.readUint8(t+o++);s=(7&a)<<18|(63&c)<<12|(63&u)<<6|63&l}}}s<65536?i+=String.fromCharCode(s):(s-=65536,i+=String.fromCharCode(55296+(s>>10),56320+(1023&s)))}return i},r.ByteBuffer.prototype.__indirect=function(t){return t+this.readInt32(t)},r.ByteBuffer.prototype.__vector=function(t){return t+this.readInt32(t)+r.SIZEOF_INT},r.ByteBuffer.prototype.__vector_len=function(t){return this.readInt32(t+this.readInt32(t))},r.ByteBuffer.prototype.__has_identifier=function(t){if(t.length!=r.FILE_IDENTIFIER_LENGTH)throw new Error("FlatBuffers: file identifier must be length "+r.FILE_IDENTIFIER_LENGTH);for(var e=0;e<r.FILE_IDENTIFIER_LENGTH;e++)if(t.charCodeAt(e)!=this.readInt8(this.position_+r.SIZEOF_INT+e))return!1;return!0},r.ByteBuffer.prototype.createLong=function(t,e){return r.Long.create(t,e)}},a79d:function(t,e,n){"use strict";var r=n("23e7"),i=n("c430"),o=n("fea9"),s=n("d039"),a=n("d066"),c=n("1626"),u=n("4840"),l=n("cdf9"),f=n("6eeb"),d=!!o&&s((function(){o.prototype["finally"].call({then:function(){}},(function(){}))}));if(r({target:"Promise",proto:!0,real:!0,forced:d},{finally:function(t){var e=u(this,a("Promise")),n=c(t);return this.then(n?function(n){return l(e,t()).then((function(){return n}))}:t,n?function(n){return l(e,t()).then((function(){throw n}))}:t)}}),!i&&c(o)){var h=a("Promise").prototype["finally"];o.prototype["finally"]!==h&&f(o.prototype,"finally",h,{unsafe:!0})}},a93d:function(t,e,n){"use strict";
/** @license React v16.13.1
 * react-is.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var r="function"===typeof Symbol&&Symbol.for,i=r?Symbol.for("react.element"):60103,o=r?Symbol.for("react.portal"):60106,s=r?Symbol.for("react.fragment"):60107,a=r?Symbol.for("react.strict_mode"):60108,c=r?Symbol.for("react.profiler"):60114,u=r?Symbol.for("react.provider"):60109,l=r?Symbol.for("react.context"):60110,f=r?Symbol.for("react.async_mode"):60111,d=r?Symbol.for("react.concurrent_mode"):60111,h=r?Symbol.for("react.forward_ref"):60112,p=r?Symbol.for("react.suspense"):60113,y=r?Symbol.for("react.suspense_list"):60120,v=r?Symbol.for("react.memo"):60115,b=r?Symbol.for("react.lazy"):60116,m=r?Symbol.for("react.block"):60121,g=r?Symbol.for("react.fundamental"):60117,_=r?Symbol.for("react.responder"):60118,w=r?Symbol.for("react.scope"):60119;function O(t){if("object"===typeof t&&null!==t){var e=t.$$typeof;switch(e){case i:switch(t=t.type,t){case f:case d:case s:case c:case a:case p:return t;default:switch(t=t&&t.$$typeof,t){case l:case h:case b:case v:case u:return t;default:return e}}case o:return e}}}function S(t){return O(t)===d}e.AsyncMode=f,e.ConcurrentMode=d,e.ContextConsumer=l,e.ContextProvider=u,e.Element=i,e.ForwardRef=h,e.Fragment=s,e.Lazy=b,e.Memo=v,e.Portal=o,e.Profiler=c,e.StrictMode=a,e.Suspense=p,e.isAsyncMode=function(t){return S(t)||O(t)===f},e.isConcurrentMode=S,e.isContextConsumer=function(t){return O(t)===l},e.isContextProvider=function(t){return O(t)===u},e.isElement=function(t){return"object"===typeof t&&null!==t&&t.$$typeof===i},e.isForwardRef=function(t){return O(t)===h},e.isFragment=function(t){return O(t)===s},e.isLazy=function(t){return O(t)===b},e.isMemo=function(t){return O(t)===v},e.isPortal=function(t){return O(t)===o},e.isProfiler=function(t){return O(t)===c},e.isStrictMode=function(t){return O(t)===a},e.isSuspense=function(t){return O(t)===p},e.isValidElementType=function(t){return"string"===typeof t||"function"===typeof t||t===s||t===d||t===c||t===a||t===p||t===y||"object"===typeof t&&null!==t&&(t.$$typeof===b||t.$$typeof===v||t.$$typeof===u||t.$$typeof===l||t.$$typeof===h||t.$$typeof===g||t.$$typeof===_||t.$$typeof===w||t.$$typeof===m)},e.typeOf=O},ab5b:function(t,e,n){"use strict";t.exports=n("be24")},ae93:function(t,e,n){"use strict";var r,i,o,s=n("d039"),a=n("1626"),c=n("7c73"),u=n("e163"),l=n("6eeb"),f=n("b622"),d=n("c430"),h=f("iterator"),p=!1;[].keys&&(o=[].keys(),"next"in o?(i=u(u(o)),i!==Object.prototype&&(r=i)):p=!0);var y=void 0==r||s((function(){var t={};return r[h].call(t)!==t}));y?r={}:d&&(r=c(r)),a(r[h])||l(r,h,(function(){return this})),t.exports={IteratorPrototype:r,BUGGY_SAFARI_ITERATORS:p}},b041:function(t,e,n){"use strict";var r=n("00ee"),i=n("f5df");t.exports=r?{}.toString:function(){return"[object "+i(this)+"]"}},b0c0:function(t,e,n){var r=n("83ab"),i=n("5e77").EXISTS,o=n("e330"),s=n("9bf2").f,a=Function.prototype,c=o(a.toString),u=/function\b(?:\s|\/\*[\S\s]*?\*\/|\/\/[^\n\r]*[\n\r]+)*([^\s(/]*)/,l=o(u.exec),f="name";r&&!i&&s(a,f,{configurable:!0,get:function(){try{return l(u,c(this))[1]}catch(t){return""}}})},b4b6:function(t,e,n){"use strict";function r(t,e,n){return e<=t&&t<=n}function i(t){if(void 0===t)return{};if(t===Object(t))return t;throw TypeError("Could not convert argument to dictionary")}function o(t){var e=String(t),n=e.length,r=0,i=[];while(r<n){var o=e.charCodeAt(r);if(o<55296||o>57343)i.push(o);else if(56320<=o&&o<=57343)i.push(65533);else if(55296<=o&&o<=56319)if(r===n-1)i.push(65533);else{var s=t.charCodeAt(r+1);if(56320<=s&&s<=57343){var a=1023&o,c=1023&s;i.push(65536+(a<<10)+c),r+=1}else i.push(65533)}r+=1}return i}function s(t){for(var e="",n=0;n<t.length;++n){var r=t[n];r<=65535?e+=String.fromCharCode(r):(r-=65536,e+=String.fromCharCode(55296+(r>>10),56320+(1023&r)))}return e}n.d(e,"b",(function(){return y})),n.d(e,"a",(function(){return p}));var a=-1;function c(t){this.tokens=[].slice.call(t)}c.prototype={endOfStream:function(){return!this.tokens.length},read:function(){return this.tokens.length?this.tokens.shift():a},prepend:function(t){if(Array.isArray(t)){var e=t;while(e.length)this.tokens.unshift(e.pop())}else this.tokens.unshift(t)},push:function(t){if(Array.isArray(t)){var e=t;while(e.length)this.tokens.push(e.shift())}else this.tokens.push(t)}};var u=-1;function l(t,e){if(t)throw TypeError("Decoder error");return e||65533}function f(){}function d(){}f.prototype={handler:function(t,e){}},d.prototype={handler:function(t,e){}};var h="utf-8";function p(t,e){if(!(this instanceof p))return new p(t,e);if(t=void 0!==t?String(t).toLowerCase():h,t!==h)throw new Error("Encoding not supported. Only utf-8 is supported");e=i(e),this._streaming=!1,this._BOMseen=!1,this._decoder=null,this._fatal=Boolean(e["fatal"]),this._ignoreBOM=Boolean(e["ignoreBOM"]),Object.defineProperty(this,"encoding",{value:"utf-8"}),Object.defineProperty(this,"fatal",{value:this._fatal}),Object.defineProperty(this,"ignoreBOM",{value:this._ignoreBOM})}function y(t,e){if(!(this instanceof y))return new y(t,e);if(t=void 0!==t?String(t).toLowerCase():h,t!==h)throw new Error("Encoding not supported. Only utf-8 is supported");e=i(e),this._streaming=!1,this._encoder=null,this._options={fatal:Boolean(e["fatal"])},Object.defineProperty(this,"encoding",{value:"utf-8"})}function v(t){var e=t.fatal,n=0,i=0,o=0,s=128,c=191;this.handler=function(t,f){if(f===a&&0!==o)return o=0,l(e);if(f===a)return u;if(0===o){if(r(f,0,127))return f;if(r(f,194,223))o=1,n=f-192;else if(r(f,224,239))224===f&&(s=160),237===f&&(c=159),o=2,n=f-224;else{if(!r(f,240,244))return l(e);240===f&&(s=144),244===f&&(c=143),o=3,n=f-240}return n<<=6*o,null}if(!r(f,s,c))return n=o=i=0,s=128,c=191,t.prepend(f),l(e);if(s=128,c=191,i+=1,n+=f-128<<6*(o-i),i!==o)return null;var d=n;return n=o=i=0,d}}function b(t){t.fatal;this.handler=function(t,e){if(e===a)return u;if(r(e,0,127))return e;var n,i;r(e,128,2047)?(n=1,i=192):r(e,2048,65535)?(n=2,i=224):r(e,65536,1114111)&&(n=3,i=240);var o=[(e>>6*n)+i];while(n>0){var s=e>>6*(n-1);o.push(128|63&s),n-=1}return o}}p.prototype={decode:function(t,e){var n;n="object"===typeof t&&t instanceof ArrayBuffer?new Uint8Array(t):"object"===typeof t&&"buffer"in t&&t.buffer instanceof ArrayBuffer?new Uint8Array(t.buffer,t.byteOffset,t.byteLength):new Uint8Array(0),e=i(e),this._streaming||(this._decoder=new v({fatal:this._fatal}),this._BOMseen=!1),this._streaming=Boolean(e["stream"]);var r,o=new c(n),a=[];while(!o.endOfStream()){if(r=this._decoder.handler(o,o.read()),r===u)break;null!==r&&(Array.isArray(r)?a.push.apply(a,r):a.push(r))}if(!this._streaming){do{if(r=this._decoder.handler(o,o.read()),r===u)break;null!==r&&(Array.isArray(r)?a.push.apply(a,r):a.push(r))}while(!o.endOfStream());this._decoder=null}return a.length&&(-1===["utf-8"].indexOf(this.encoding)||this._ignoreBOM||this._BOMseen||(65279===a[0]?(this._BOMseen=!0,a.shift()):this._BOMseen=!0)),s(a)}},y.prototype={encode:function(t,e){t=t?String(t):"",e=i(e),this._streaming||(this._encoder=new b(this._options)),this._streaming=Boolean(e["stream"]);var n,r=[],s=new c(o(t));while(!s.endOfStream()){if(n=this._encoder.handler(s,s.read()),n===u)break;Array.isArray(n)?r.push.apply(r,n):r.push(n)}if(!this._streaming){while(1){if(n=this._encoder.handler(s,s.read()),n===u)break;Array.isArray(n)?r.push.apply(r,n):r.push(n)}this._encoder=null}return new Uint8Array(r)}}},b575:function(t,e,n){var r,i,o,s,a,c,u,l,f=n("da84"),d=n("0366"),h=n("06cf").f,p=n("2cf4").set,y=n("1cdc"),v=n("d4c3"),b=n("a4b4"),m=n("605d"),g=f.MutationObserver||f.WebKitMutationObserver,_=f.document,w=f.process,O=f.Promise,S=h(f,"queueMicrotask"),x=S&&S.value;x||(r=function(){var t,e;m&&(t=w.domain)&&t.exit();while(i){e=i.fn,i=i.next;try{e()}catch(n){throw i?s():o=void 0,n}}o=void 0,t&&t.enter()},y||m||b||!g||!_?!v&&O&&O.resolve?(u=O.resolve(void 0),u.constructor=O,l=d(u.then,u),s=function(){l(r)}):m?s=function(){w.nextTick(r)}:(p=d(p,f),s=function(){p(r)}):(a=!0,c=_.createTextNode(""),new g(r).observe(c,{characterData:!0}),s=function(){c.data=a=!a})),t.exports=x||function(t){var e={fn:t,next:void 0};o&&(o.next=e),i||(i=e,s()),o=e}},b622:function(t,e,n){var r=n("da84"),i=n("5692"),o=n("1a2d"),s=n("90e3"),a=n("4930"),c=n("fdbf"),u=i("wks"),l=r.Symbol,f=l&&l["for"],d=c?l:l&&l.withoutSetter||s;t.exports=function(t){if(!o(u,t)||!a&&"string"!=typeof u[t]){var e="Symbol."+t;a&&o(l,t)?u[t]=l[t]:u[t]=c&&f?f(e):d(e)}return u[t]}},b639:function(t,e,n){"use strict";(function(t){
/*!
 * The buffer module from node.js, for the browser.
 *
 * @author   Feross Aboukhadijeh <http://feross.org>
 * @license  MIT
 */
var r=n("1fb5"),i=n("9152"),o=n("e3db");function s(){try{var t=new Uint8Array(1);return t.__proto__={__proto__:Uint8Array.prototype,foo:function(){return 42}},42===t.foo()&&"function"===typeof t.subarray&&0===t.subarray(1,1).byteLength}catch(e){return!1}}function a(){return u.TYPED_ARRAY_SUPPORT?2147483647:1073741823}function c(t,e){if(a()<e)throw new RangeError("Invalid typed array length");return u.TYPED_ARRAY_SUPPORT?(t=new Uint8Array(e),t.__proto__=u.prototype):(null===t&&(t=new u(e)),t.length=e),t}function u(t,e,n){if(!u.TYPED_ARRAY_SUPPORT&&!(this instanceof u))return new u(t,e,n);if("number"===typeof t){if("string"===typeof e)throw new Error("If encoding is specified then the first argument must be a string");return h(this,t)}return l(this,t,e,n)}function l(t,e,n,r){if("number"===typeof e)throw new TypeError('"value" argument must not be a number');return"undefined"!==typeof ArrayBuffer&&e instanceof ArrayBuffer?v(t,e,n,r):"string"===typeof e?p(t,e,n):b(t,e)}function f(t){if("number"!==typeof t)throw new TypeError('"size" argument must be a number');if(t<0)throw new RangeError('"size" argument must not be negative')}function d(t,e,n,r){return f(e),e<=0?c(t,e):void 0!==n?"string"===typeof r?c(t,e).fill(n,r):c(t,e).fill(n):c(t,e)}function h(t,e){if(f(e),t=c(t,e<0?0:0|m(e)),!u.TYPED_ARRAY_SUPPORT)for(var n=0;n<e;++n)t[n]=0;return t}function p(t,e,n){if("string"===typeof n&&""!==n||(n="utf8"),!u.isEncoding(n))throw new TypeError('"encoding" must be a valid string encoding');var r=0|_(e,n);t=c(t,r);var i=t.write(e,n);return i!==r&&(t=t.slice(0,i)),t}function y(t,e){var n=e.length<0?0:0|m(e.length);t=c(t,n);for(var r=0;r<n;r+=1)t[r]=255&e[r];return t}function v(t,e,n,r){if(e.byteLength,n<0||e.byteLength<n)throw new RangeError("'offset' is out of bounds");if(e.byteLength<n+(r||0))throw new RangeError("'length' is out of bounds");return e=void 0===n&&void 0===r?new Uint8Array(e):void 0===r?new Uint8Array(e,n):new Uint8Array(e,n,r),u.TYPED_ARRAY_SUPPORT?(t=e,t.__proto__=u.prototype):t=y(t,e),t}function b(t,e){if(u.isBuffer(e)){var n=0|m(e.length);return t=c(t,n),0===t.length?t:(e.copy(t,0,0,n),t)}if(e){if("undefined"!==typeof ArrayBuffer&&e.buffer instanceof ArrayBuffer||"length"in e)return"number"!==typeof e.length||et(e.length)?c(t,0):y(t,e);if("Buffer"===e.type&&o(e.data))return y(t,e.data)}throw new TypeError("First argument must be a string, Buffer, ArrayBuffer, Array, or array-like object.")}function m(t){if(t>=a())throw new RangeError("Attempt to allocate Buffer larger than maximum size: 0x"+a().toString(16)+" bytes");return 0|t}function g(t){return+t!=t&&(t=0),u.alloc(+t)}function _(t,e){if(u.isBuffer(t))return t.length;if("undefined"!==typeof ArrayBuffer&&"function"===typeof ArrayBuffer.isView&&(ArrayBuffer.isView(t)||t instanceof ArrayBuffer))return t.byteLength;"string"!==typeof t&&(t=""+t);var n=t.length;if(0===n)return 0;for(var r=!1;;)switch(e){case"ascii":case"latin1":case"binary":return n;case"utf8":case"utf-8":case void 0:return X(t).length;case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return 2*n;case"hex":return n>>>1;case"base64":return Q(t).length;default:if(r)return X(t).length;e=(""+e).toLowerCase(),r=!0}}function w(t,e,n){var r=!1;if((void 0===e||e<0)&&(e=0),e>this.length)return"";if((void 0===n||n>this.length)&&(n=this.length),n<=0)return"";if(n>>>=0,e>>>=0,n<=e)return"";t||(t="utf8");while(1)switch(t){case"hex":return P(this,e,n);case"utf8":case"utf-8":return M(this,e,n);case"ascii":return N(this,e,n);case"latin1":case"binary":return L(this,e,n);case"base64":return C(this,e,n);case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return F(this,e,n);default:if(r)throw new TypeError("Unknown encoding: "+t);t=(t+"").toLowerCase(),r=!0}}function O(t,e,n){var r=t[e];t[e]=t[n],t[n]=r}function S(t,e,n,r,i){if(0===t.length)return-1;if("string"===typeof n?(r=n,n=0):n>2147483647?n=2147483647:n<-2147483648&&(n=-2147483648),n=+n,isNaN(n)&&(n=i?0:t.length-1),n<0&&(n=t.length+n),n>=t.length){if(i)return-1;n=t.length-1}else if(n<0){if(!i)return-1;n=0}if("string"===typeof e&&(e=u.from(e,r)),u.isBuffer(e))return 0===e.length?-1:x(t,e,n,r,i);if("number"===typeof e)return e&=255,u.TYPED_ARRAY_SUPPORT&&"function"===typeof Uint8Array.prototype.indexOf?i?Uint8Array.prototype.indexOf.call(t,e,n):Uint8Array.prototype.lastIndexOf.call(t,e,n):x(t,[e],n,r,i);throw new TypeError("val must be string, number or Buffer")}function x(t,e,n,r,i){var o,s=1,a=t.length,c=e.length;if(void 0!==r&&(r=String(r).toLowerCase(),"ucs2"===r||"ucs-2"===r||"utf16le"===r||"utf-16le"===r)){if(t.length<2||e.length<2)return-1;s=2,a/=2,c/=2,n/=2}function u(t,e){return 1===s?t[e]:t.readUInt16BE(e*s)}if(i){var l=-1;for(o=n;o<a;o++)if(u(t,o)===u(e,-1===l?0:o-l)){if(-1===l&&(l=o),o-l+1===c)return l*s}else-1!==l&&(o-=o-l),l=-1}else for(n+c>a&&(n=a-c),o=n;o>=0;o--){for(var f=!0,d=0;d<c;d++)if(u(t,o+d)!==u(e,d)){f=!1;break}if(f)return o}return-1}function I(t,e,n,r){n=Number(n)||0;var i=t.length-n;r?(r=Number(r),r>i&&(r=i)):r=i;var o=e.length;if(o%2!==0)throw new TypeError("Invalid hex string");r>o/2&&(r=o/2);for(var s=0;s<r;++s){var a=parseInt(e.substr(2*s,2),16);if(isNaN(a))return s;t[n+s]=a}return s}function A(t,e,n,r){return tt(X(e,t.length-n),t,n,r)}function T(t,e,n,r){return tt(J(e),t,n,r)}function j(t,e,n,r){return T(t,e,n,r)}function E(t,e,n,r){return tt(Q(e),t,n,r)}function B(t,e,n,r){return tt(Z(e,t.length-n),t,n,r)}function C(t,e,n){return 0===e&&n===t.length?r.fromByteArray(t):r.fromByteArray(t.slice(e,n))}function M(t,e,n){n=Math.min(t.length,n);var r=[],i=e;while(i<n){var o,s,a,c,u=t[i],l=null,f=u>239?4:u>223?3:u>191?2:1;if(i+f<=n)switch(f){case 1:u<128&&(l=u);break;case 2:o=t[i+1],128===(192&o)&&(c=(31&u)<<6|63&o,c>127&&(l=c));break;case 3:o=t[i+1],s=t[i+2],128===(192&o)&&128===(192&s)&&(c=(15&u)<<12|(63&o)<<6|63&s,c>2047&&(c<55296||c>57343)&&(l=c));break;case 4:o=t[i+1],s=t[i+2],a=t[i+3],128===(192&o)&&128===(192&s)&&128===(192&a)&&(c=(15&u)<<18|(63&o)<<12|(63&s)<<6|63&a,c>65535&&c<1114112&&(l=c))}null===l?(l=65533,f=1):l>65535&&(l-=65536,r.push(l>>>10&1023|55296),l=56320|1023&l),r.push(l),i+=f}return D(r)}e.Buffer=u,e.SlowBuffer=g,e.INSPECT_MAX_BYTES=50,u.TYPED_ARRAY_SUPPORT=void 0!==t.TYPED_ARRAY_SUPPORT?t.TYPED_ARRAY_SUPPORT:s(),e.kMaxLength=a(),u.poolSize=8192,u._augment=function(t){return t.__proto__=u.prototype,t},u.from=function(t,e,n){return l(null,t,e,n)},u.TYPED_ARRAY_SUPPORT&&(u.prototype.__proto__=Uint8Array.prototype,u.__proto__=Uint8Array,"undefined"!==typeof Symbol&&Symbol.species&&u[Symbol.species]===u&&Object.defineProperty(u,Symbol.species,{value:null,configurable:!0})),u.alloc=function(t,e,n){return d(null,t,e,n)},u.allocUnsafe=function(t){return h(null,t)},u.allocUnsafeSlow=function(t){return h(null,t)},u.isBuffer=function(t){return!(null==t||!t._isBuffer)},u.compare=function(t,e){if(!u.isBuffer(t)||!u.isBuffer(e))throw new TypeError("Arguments must be Buffers");if(t===e)return 0;for(var n=t.length,r=e.length,i=0,o=Math.min(n,r);i<o;++i)if(t[i]!==e[i]){n=t[i],r=e[i];break}return n<r?-1:r<n?1:0},u.isEncoding=function(t){switch(String(t).toLowerCase()){case"hex":case"utf8":case"utf-8":case"ascii":case"latin1":case"binary":case"base64":case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return!0;default:return!1}},u.concat=function(t,e){if(!o(t))throw new TypeError('"list" argument must be an Array of Buffers');if(0===t.length)return u.alloc(0);var n;if(void 0===e)for(e=0,n=0;n<t.length;++n)e+=t[n].length;var r=u.allocUnsafe(e),i=0;for(n=0;n<t.length;++n){var s=t[n];if(!u.isBuffer(s))throw new TypeError('"list" argument must be an Array of Buffers');s.copy(r,i),i+=s.length}return r},u.byteLength=_,u.prototype._isBuffer=!0,u.prototype.swap16=function(){var t=this.length;if(t%2!==0)throw new RangeError("Buffer size must be a multiple of 16-bits");for(var e=0;e<t;e+=2)O(this,e,e+1);return this},u.prototype.swap32=function(){var t=this.length;if(t%4!==0)throw new RangeError("Buffer size must be a multiple of 32-bits");for(var e=0;e<t;e+=4)O(this,e,e+3),O(this,e+1,e+2);return this},u.prototype.swap64=function(){var t=this.length;if(t%8!==0)throw new RangeError("Buffer size must be a multiple of 64-bits");for(var e=0;e<t;e+=8)O(this,e,e+7),O(this,e+1,e+6),O(this,e+2,e+5),O(this,e+3,e+4);return this},u.prototype.toString=function(){var t=0|this.length;return 0===t?"":0===arguments.length?M(this,0,t):w.apply(this,arguments)},u.prototype.equals=function(t){if(!u.isBuffer(t))throw new TypeError("Argument must be a Buffer");return this===t||0===u.compare(this,t)},u.prototype.inspect=function(){var t="",n=e.INSPECT_MAX_BYTES;return this.length>0&&(t=this.toString("hex",0,n).match(/.{2}/g).join(" "),this.length>n&&(t+=" ... ")),"<Buffer "+t+">"},u.prototype.compare=function(t,e,n,r,i){if(!u.isBuffer(t))throw new TypeError("Argument must be a Buffer");if(void 0===e&&(e=0),void 0===n&&(n=t?t.length:0),void 0===r&&(r=0),void 0===i&&(i=this.length),e<0||n>t.length||r<0||i>this.length)throw new RangeError("out of range index");if(r>=i&&e>=n)return 0;if(r>=i)return-1;if(e>=n)return 1;if(e>>>=0,n>>>=0,r>>>=0,i>>>=0,this===t)return 0;for(var o=i-r,s=n-e,a=Math.min(o,s),c=this.slice(r,i),l=t.slice(e,n),f=0;f<a;++f)if(c[f]!==l[f]){o=c[f],s=l[f];break}return o<s?-1:s<o?1:0},u.prototype.includes=function(t,e,n){return-1!==this.indexOf(t,e,n)},u.prototype.indexOf=function(t,e,n){return S(this,t,e,n,!0)},u.prototype.lastIndexOf=function(t,e,n){return S(this,t,e,n,!1)},u.prototype.write=function(t,e,n,r){if(void 0===e)r="utf8",n=this.length,e=0;else if(void 0===n&&"string"===typeof e)r=e,n=this.length,e=0;else{if(!isFinite(e))throw new Error("Buffer.write(string, encoding, offset[, length]) is no longer supported");e|=0,isFinite(n)?(n|=0,void 0===r&&(r="utf8")):(r=n,n=void 0)}var i=this.length-e;if((void 0===n||n>i)&&(n=i),t.length>0&&(n<0||e<0)||e>this.length)throw new RangeError("Attempt to write outside buffer bounds");r||(r="utf8");for(var o=!1;;)switch(r){case"hex":return I(this,t,e,n);case"utf8":case"utf-8":return A(this,t,e,n);case"ascii":return T(this,t,e,n);case"latin1":case"binary":return j(this,t,e,n);case"base64":return E(this,t,e,n);case"ucs2":case"ucs-2":case"utf16le":case"utf-16le":return B(this,t,e,n);default:if(o)throw new TypeError("Unknown encoding: "+r);r=(""+r).toLowerCase(),o=!0}},u.prototype.toJSON=function(){return{type:"Buffer",data:Array.prototype.slice.call(this._arr||this,0)}};var k=4096;function D(t){var e=t.length;if(e<=k)return String.fromCharCode.apply(String,t);var n="",r=0;while(r<e)n+=String.fromCharCode.apply(String,t.slice(r,r+=k));return n}function N(t,e,n){var r="";n=Math.min(t.length,n);for(var i=e;i<n;++i)r+=String.fromCharCode(127&t[i]);return r}function L(t,e,n){var r="";n=Math.min(t.length,n);for(var i=e;i<n;++i)r+=String.fromCharCode(t[i]);return r}function P(t,e,n){var r=t.length;(!e||e<0)&&(e=0),(!n||n<0||n>r)&&(n=r);for(var i="",o=e;o<n;++o)i+=K(t[o]);return i}function F(t,e,n){for(var r=t.slice(e,n),i="",o=0;o<r.length;o+=2)i+=String.fromCharCode(r[o]+256*r[o+1]);return i}function U(t,e,n){if(t%1!==0||t<0)throw new RangeError("offset is not uint");if(t+e>n)throw new RangeError("Trying to access beyond buffer length")}function R(t,e,n,r,i,o){if(!u.isBuffer(t))throw new TypeError('"buffer" argument must be a Buffer instance');if(e>i||e<o)throw new RangeError('"value" argument is out of bounds');if(n+r>t.length)throw new RangeError("Index out of range")}function V(t,e,n,r){e<0&&(e=65535+e+1);for(var i=0,o=Math.min(t.length-n,2);i<o;++i)t[n+i]=(e&255<<8*(r?i:1-i))>>>8*(r?i:1-i)}function $(t,e,n,r){e<0&&(e=4294967295+e+1);for(var i=0,o=Math.min(t.length-n,4);i<o;++i)t[n+i]=e>>>8*(r?i:3-i)&255}function z(t,e,n,r,i,o){if(n+r>t.length)throw new RangeError("Index out of range");if(n<0)throw new RangeError("Index out of range")}function Y(t,e,n,r,o){return o||z(t,e,n,4,34028234663852886e22,-34028234663852886e22),i.write(t,e,n,r,23,4),n+4}function H(t,e,n,r,o){return o||z(t,e,n,8,17976931348623157e292,-17976931348623157e292),i.write(t,e,n,r,52,8),n+8}u.prototype.slice=function(t,e){var n,r=this.length;if(t=~~t,e=void 0===e?r:~~e,t<0?(t+=r,t<0&&(t=0)):t>r&&(t=r),e<0?(e+=r,e<0&&(e=0)):e>r&&(e=r),e<t&&(e=t),u.TYPED_ARRAY_SUPPORT)n=this.subarray(t,e),n.__proto__=u.prototype;else{var i=e-t;n=new u(i,void 0);for(var o=0;o<i;++o)n[o]=this[o+t]}return n},u.prototype.readUIntLE=function(t,e,n){t|=0,e|=0,n||U(t,e,this.length);var r=this[t],i=1,o=0;while(++o<e&&(i*=256))r+=this[t+o]*i;return r},u.prototype.readUIntBE=function(t,e,n){t|=0,e|=0,n||U(t,e,this.length);var r=this[t+--e],i=1;while(e>0&&(i*=256))r+=this[t+--e]*i;return r},u.prototype.readUInt8=function(t,e){return e||U(t,1,this.length),this[t]},u.prototype.readUInt16LE=function(t,e){return e||U(t,2,this.length),this[t]|this[t+1]<<8},u.prototype.readUInt16BE=function(t,e){return e||U(t,2,this.length),this[t]<<8|this[t+1]},u.prototype.readUInt32LE=function(t,e){return e||U(t,4,this.length),(this[t]|this[t+1]<<8|this[t+2]<<16)+16777216*this[t+3]},u.prototype.readUInt32BE=function(t,e){return e||U(t,4,this.length),16777216*this[t]+(this[t+1]<<16|this[t+2]<<8|this[t+3])},u.prototype.readIntLE=function(t,e,n){t|=0,e|=0,n||U(t,e,this.length);var r=this[t],i=1,o=0;while(++o<e&&(i*=256))r+=this[t+o]*i;return i*=128,r>=i&&(r-=Math.pow(2,8*e)),r},u.prototype.readIntBE=function(t,e,n){t|=0,e|=0,n||U(t,e,this.length);var r=e,i=1,o=this[t+--r];while(r>0&&(i*=256))o+=this[t+--r]*i;return i*=128,o>=i&&(o-=Math.pow(2,8*e)),o},u.prototype.readInt8=function(t,e){return e||U(t,1,this.length),128&this[t]?-1*(255-this[t]+1):this[t]},u.prototype.readInt16LE=function(t,e){e||U(t,2,this.length);var n=this[t]|this[t+1]<<8;return 32768&n?4294901760|n:n},u.prototype.readInt16BE=function(t,e){e||U(t,2,this.length);var n=this[t+1]|this[t]<<8;return 32768&n?4294901760|n:n},u.prototype.readInt32LE=function(t,e){return e||U(t,4,this.length),this[t]|this[t+1]<<8|this[t+2]<<16|this[t+3]<<24},u.prototype.readInt32BE=function(t,e){return e||U(t,4,this.length),this[t]<<24|this[t+1]<<16|this[t+2]<<8|this[t+3]},u.prototype.readFloatLE=function(t,e){return e||U(t,4,this.length),i.read(this,t,!0,23,4)},u.prototype.readFloatBE=function(t,e){return e||U(t,4,this.length),i.read(this,t,!1,23,4)},u.prototype.readDoubleLE=function(t,e){return e||U(t,8,this.length),i.read(this,t,!0,52,8)},u.prototype.readDoubleBE=function(t,e){return e||U(t,8,this.length),i.read(this,t,!1,52,8)},u.prototype.writeUIntLE=function(t,e,n,r){if(t=+t,e|=0,n|=0,!r){var i=Math.pow(2,8*n)-1;R(this,t,e,n,i,0)}var o=1,s=0;this[e]=255&t;while(++s<n&&(o*=256))this[e+s]=t/o&255;return e+n},u.prototype.writeUIntBE=function(t,e,n,r){if(t=+t,e|=0,n|=0,!r){var i=Math.pow(2,8*n)-1;R(this,t,e,n,i,0)}var o=n-1,s=1;this[e+o]=255&t;while(--o>=0&&(s*=256))this[e+o]=t/s&255;return e+n},u.prototype.writeUInt8=function(t,e,n){return t=+t,e|=0,n||R(this,t,e,1,255,0),u.TYPED_ARRAY_SUPPORT||(t=Math.floor(t)),this[e]=255&t,e+1},u.prototype.writeUInt16LE=function(t,e,n){return t=+t,e|=0,n||R(this,t,e,2,65535,0),u.TYPED_ARRAY_SUPPORT?(this[e]=255&t,this[e+1]=t>>>8):V(this,t,e,!0),e+2},u.prototype.writeUInt16BE=function(t,e,n){return t=+t,e|=0,n||R(this,t,e,2,65535,0),u.TYPED_ARRAY_SUPPORT?(this[e]=t>>>8,this[e+1]=255&t):V(this,t,e,!1),e+2},u.prototype.writeUInt32LE=function(t,e,n){return t=+t,e|=0,n||R(this,t,e,4,4294967295,0),u.TYPED_ARRAY_SUPPORT?(this[e+3]=t>>>24,this[e+2]=t>>>16,this[e+1]=t>>>8,this[e]=255&t):$(this,t,e,!0),e+4},u.prototype.writeUInt32BE=function(t,e,n){return t=+t,e|=0,n||R(this,t,e,4,4294967295,0),u.TYPED_ARRAY_SUPPORT?(this[e]=t>>>24,this[e+1]=t>>>16,this[e+2]=t>>>8,this[e+3]=255&t):$(this,t,e,!1),e+4},u.prototype.writeIntLE=function(t,e,n,r){if(t=+t,e|=0,!r){var i=Math.pow(2,8*n-1);R(this,t,e,n,i-1,-i)}var o=0,s=1,a=0;this[e]=255&t;while(++o<n&&(s*=256))t<0&&0===a&&0!==this[e+o-1]&&(a=1),this[e+o]=(t/s>>0)-a&255;return e+n},u.prototype.writeIntBE=function(t,e,n,r){if(t=+t,e|=0,!r){var i=Math.pow(2,8*n-1);R(this,t,e,n,i-1,-i)}var o=n-1,s=1,a=0;this[e+o]=255&t;while(--o>=0&&(s*=256))t<0&&0===a&&0!==this[e+o+1]&&(a=1),this[e+o]=(t/s>>0)-a&255;return e+n},u.prototype.writeInt8=function(t,e,n){return t=+t,e|=0,n||R(this,t,e,1,127,-128),u.TYPED_ARRAY_SUPPORT||(t=Math.floor(t)),t<0&&(t=255+t+1),this[e]=255&t,e+1},u.prototype.writeInt16LE=function(t,e,n){return t=+t,e|=0,n||R(this,t,e,2,32767,-32768),u.TYPED_ARRAY_SUPPORT?(this[e]=255&t,this[e+1]=t>>>8):V(this,t,e,!0),e+2},u.prototype.writeInt16BE=function(t,e,n){return t=+t,e|=0,n||R(this,t,e,2,32767,-32768),u.TYPED_ARRAY_SUPPORT?(this[e]=t>>>8,this[e+1]=255&t):V(this,t,e,!1),e+2},u.prototype.writeInt32LE=function(t,e,n){return t=+t,e|=0,n||R(this,t,e,4,2147483647,-2147483648),u.TYPED_ARRAY_SUPPORT?(this[e]=255&t,this[e+1]=t>>>8,this[e+2]=t>>>16,this[e+3]=t>>>24):$(this,t,e,!0),e+4},u.prototype.writeInt32BE=function(t,e,n){return t=+t,e|=0,n||R(this,t,e,4,2147483647,-2147483648),t<0&&(t=4294967295+t+1),u.TYPED_ARRAY_SUPPORT?(this[e]=t>>>24,this[e+1]=t>>>16,this[e+2]=t>>>8,this[e+3]=255&t):$(this,t,e,!1),e+4},u.prototype.writeFloatLE=function(t,e,n){return Y(this,t,e,!0,n)},u.prototype.writeFloatBE=function(t,e,n){return Y(this,t,e,!1,n)},u.prototype.writeDoubleLE=function(t,e,n){return H(this,t,e,!0,n)},u.prototype.writeDoubleBE=function(t,e,n){return H(this,t,e,!1,n)},u.prototype.copy=function(t,e,n,r){if(n||(n=0),r||0===r||(r=this.length),e>=t.length&&(e=t.length),e||(e=0),r>0&&r<n&&(r=n),r===n)return 0;if(0===t.length||0===this.length)return 0;if(e<0)throw new RangeError("targetStart out of bounds");if(n<0||n>=this.length)throw new RangeError("sourceStart out of bounds");if(r<0)throw new RangeError("sourceEnd out of bounds");r>this.length&&(r=this.length),t.length-e<r-n&&(r=t.length-e+n);var i,o=r-n;if(this===t&&n<e&&e<r)for(i=o-1;i>=0;--i)t[i+e]=this[i+n];else if(o<1e3||!u.TYPED_ARRAY_SUPPORT)for(i=0;i<o;++i)t[i+e]=this[i+n];else Uint8Array.prototype.set.call(t,this.subarray(n,n+o),e);return o},u.prototype.fill=function(t,e,n,r){if("string"===typeof t){if("string"===typeof e?(r=e,e=0,n=this.length):"string"===typeof n&&(r=n,n=this.length),1===t.length){var i=t.charCodeAt(0);i<256&&(t=i)}if(void 0!==r&&"string"!==typeof r)throw new TypeError("encoding must be a string");if("string"===typeof r&&!u.isEncoding(r))throw new TypeError("Unknown encoding: "+r)}else"number"===typeof t&&(t&=255);if(e<0||this.length<e||this.length<n)throw new RangeError("Out of range index");if(n<=e)return this;var o;if(e>>>=0,n=void 0===n?this.length:n>>>0,t||(t=0),"number"===typeof t)for(o=e;o<n;++o)this[o]=t;else{var s=u.isBuffer(t)?t:X(new u(t,r).toString()),a=s.length;for(o=0;o<n-e;++o)this[o+e]=s[o%a]}return this};var W=/[^+\/0-9A-Za-z-_]/g;function G(t){if(t=q(t).replace(W,""),t.length<2)return"";while(t.length%4!==0)t+="=";return t}function q(t){return t.trim?t.trim():t.replace(/^\s+|\s+$/g,"")}function K(t){return t<16?"0"+t.toString(16):t.toString(16)}function X(t,e){var n;e=e||1/0;for(var r=t.length,i=null,o=[],s=0;s<r;++s){if(n=t.charCodeAt(s),n>55295&&n<57344){if(!i){if(n>56319){(e-=3)>-1&&o.push(239,191,189);continue}if(s+1===r){(e-=3)>-1&&o.push(239,191,189);continue}i=n;continue}if(n<56320){(e-=3)>-1&&o.push(239,191,189),i=n;continue}n=65536+(i-55296<<10|n-56320)}else i&&(e-=3)>-1&&o.push(239,191,189);if(i=null,n<128){if((e-=1)<0)break;o.push(n)}else if(n<2048){if((e-=2)<0)break;o.push(n>>6|192,63&n|128)}else if(n<65536){if((e-=3)<0)break;o.push(n>>12|224,n>>6&63|128,63&n|128)}else{if(!(n<1114112))throw new Error("Invalid code point");if((e-=4)<0)break;o.push(n>>18|240,n>>12&63|128,n>>6&63|128,63&n|128)}}return o}function J(t){for(var e=[],n=0;n<t.length;++n)e.push(255&t.charCodeAt(n));return e}function Z(t,e){for(var n,r,i,o=[],s=0;s<t.length;++s){if((e-=2)<0)break;n=t.charCodeAt(s),r=n>>8,i=n%256,o.push(i),o.push(r)}return o}function Q(t){return r.toByteArray(G(t))}function tt(t,e,n,r){for(var i=0;i<r;++i){if(i+n>=e.length||i>=t.length)break;e[i+n]=t[i]}return i}function et(t){return t!==t}}).call(this,n("c8ba"))},b727:function(t,e,n){var r=n("0366"),i=n("e330"),o=n("44ad"),s=n("7b0b"),a=n("07fa"),c=n("65f0"),u=i([].push),l=function(t){var e=1==t,n=2==t,i=3==t,l=4==t,f=6==t,d=7==t,h=5==t||f;return function(p,y,v,b){for(var m,g,_=s(p),w=o(_),O=r(y,v),S=a(w),x=0,I=b||c,A=e?I(p,S):n||d?I(p,0):void 0;S>x;x++)if((h||x in w)&&(m=w[x],g=O(m,x,_),t))if(e)A[x]=g;else if(g)switch(t){case 3:return!0;case 5:return m;case 6:return x;case 2:u(A,m)}else switch(t){case 4:return!1;case 7:u(A,m)}return f?-1:i||l?l:A}};t.exports={forEach:l(0),map:l(1),filter:l(2),some:l(3),every:l(4),find:l(5),findIndex:l(6),filterReject:l(7)}},be24:function(t,e,n){"use strict";
/** @license React v16.14.0
 * react.production.min.js
 *
 * Copyright (c) Facebook, Inc. and its affiliates.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */var r=n("320c"),i="function"===typeof Symbol&&Symbol.for,o=i?Symbol.for("react.element"):60103,s=i?Symbol.for("react.portal"):60106,a=i?Symbol.for("react.fragment"):60107,c=i?Symbol.for("react.strict_mode"):60108,u=i?Symbol.for("react.profiler"):60114,l=i?Symbol.for("react.provider"):60109,f=i?Symbol.for("react.context"):60110,d=i?Symbol.for("react.forward_ref"):60112,h=i?Symbol.for("react.suspense"):60113,p=i?Symbol.for("react.memo"):60115,y=i?Symbol.for("react.lazy"):60116,v="function"===typeof Symbol&&Symbol.iterator;function b(t){for(var e="https://reactjs.org/docs/error-decoder.html?invariant="+t,n=1;n<arguments.length;n++)e+="&args[]="+encodeURIComponent(arguments[n]);return"Minified React error #"+t+"; visit "+e+" for the full message or use the non-minified dev environment for full errors and additional helpful warnings."}var m={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},g={};function _(t,e,n){this.props=t,this.context=e,this.refs=g,this.updater=n||m}function w(){}function O(t,e,n){this.props=t,this.context=e,this.refs=g,this.updater=n||m}_.prototype.isReactComponent={},_.prototype.setState=function(t,e){if("object"!==typeof t&&"function"!==typeof t&&null!=t)throw Error(b(85));this.updater.enqueueSetState(this,t,e,"setState")},_.prototype.forceUpdate=function(t){this.updater.enqueueForceUpdate(this,t,"forceUpdate")},w.prototype=_.prototype;var S=O.prototype=new w;S.constructor=O,r(S,_.prototype),S.isPureReactComponent=!0;var x={current:null},I=Object.prototype.hasOwnProperty,A={key:!0,ref:!0,__self:!0,__source:!0};function T(t,e,n){var r,i={},s=null,a=null;if(null!=e)for(r in void 0!==e.ref&&(a=e.ref),void 0!==e.key&&(s=""+e.key),e)I.call(e,r)&&!A.hasOwnProperty(r)&&(i[r]=e[r]);var c=arguments.length-2;if(1===c)i.children=n;else if(1<c){for(var u=Array(c),l=0;l<c;l++)u[l]=arguments[l+2];i.children=u}if(t&&t.defaultProps)for(r in c=t.defaultProps,c)void 0===i[r]&&(i[r]=c[r]);return{$$typeof:o,type:t,key:s,ref:a,props:i,_owner:x.current}}function j(t,e){return{$$typeof:o,type:t.type,key:e,ref:t.ref,props:t.props,_owner:t._owner}}function E(t){return"object"===typeof t&&null!==t&&t.$$typeof===o}function B(t){var e={"=":"=0",":":"=2"};return"$"+(""+t).replace(/[=:]/g,(function(t){return e[t]}))}var C=/\/+/g,M=[];function k(t,e,n,r){if(M.length){var i=M.pop();return i.result=t,i.keyPrefix=e,i.func=n,i.context=r,i.count=0,i}return{result:t,keyPrefix:e,func:n,context:r,count:0}}function D(t){t.result=null,t.keyPrefix=null,t.func=null,t.context=null,t.count=0,10>M.length&&M.push(t)}function N(t,e,n,r){var i=typeof t;"undefined"!==i&&"boolean"!==i||(t=null);var a=!1;if(null===t)a=!0;else switch(i){case"string":case"number":a=!0;break;case"object":switch(t.$$typeof){case o:case s:a=!0}}if(a)return n(r,t,""===e?"."+P(t,0):e),1;if(a=0,e=""===e?".":e+":",Array.isArray(t))for(var c=0;c<t.length;c++){i=t[c];var u=e+P(i,c);a+=N(i,u,n,r)}else if(null===t||"object"!==typeof t?u=null:(u=v&&t[v]||t["@@iterator"],u="function"===typeof u?u:null),"function"===typeof u)for(t=u.call(t),c=0;!(i=t.next()).done;)i=i.value,u=e+P(i,c++),a+=N(i,u,n,r);else if("object"===i)throw n=""+t,Error(b(31,"[object Object]"===n?"object with keys {"+Object.keys(t).join(", ")+"}":n,""));return a}function L(t,e,n){return null==t?0:N(t,"",e,n)}function P(t,e){return"object"===typeof t&&null!==t&&null!=t.key?B(t.key):e.toString(36)}function F(t,e){t.func.call(t.context,e,t.count++)}function U(t,e,n){var r=t.result,i=t.keyPrefix;t=t.func.call(t.context,e,t.count++),Array.isArray(t)?R(t,r,n,(function(t){return t})):null!=t&&(E(t)&&(t=j(t,i+(!t.key||e&&e.key===t.key?"":(""+t.key).replace(C,"$&/")+"/")+n)),r.push(t))}function R(t,e,n,r,i){var o="";null!=n&&(o=(""+n).replace(C,"$&/")+"/"),e=k(e,o,r,i),L(t,U,e),D(e)}var V={current:null};function $(){var t=V.current;if(null===t)throw Error(b(321));return t}var z={ReactCurrentDispatcher:V,ReactCurrentBatchConfig:{suspense:null},ReactCurrentOwner:x,IsSomeRendererActing:{current:!1},assign:r};e.Children={map:function(t,e,n){if(null==t)return t;var r=[];return R(t,r,null,e,n),r},forEach:function(t,e,n){if(null==t)return t;e=k(null,null,e,n),L(t,F,e),D(e)},count:function(t){return L(t,(function(){return null}),null)},toArray:function(t){var e=[];return R(t,e,null,(function(t){return t})),e},only:function(t){if(!E(t))throw Error(b(143));return t}},e.Component=_,e.Fragment=a,e.Profiler=u,e.PureComponent=O,e.StrictMode=c,e.Suspense=h,e.__SECRET_INTERNALS_DO_NOT_USE_OR_YOU_WILL_BE_FIRED=z,e.cloneElement=function(t,e,n){if(null===t||void 0===t)throw Error(b(267,t));var i=r({},t.props),s=t.key,a=t.ref,c=t._owner;if(null!=e){if(void 0!==e.ref&&(a=e.ref,c=x.current),void 0!==e.key&&(s=""+e.key),t.type&&t.type.defaultProps)var u=t.type.defaultProps;for(l in e)I.call(e,l)&&!A.hasOwnProperty(l)&&(i[l]=void 0===e[l]&&void 0!==u?u[l]:e[l])}var l=arguments.length-2;if(1===l)i.children=n;else if(1<l){u=Array(l);for(var f=0;f<l;f++)u[f]=arguments[f+2];i.children=u}return{$$typeof:o,type:t.type,key:s,ref:a,props:i,_owner:c}},e.createContext=function(t,e){return void 0===e&&(e=null),t={$$typeof:f,_calculateChangedBits:e,_currentValue:t,_currentValue2:t,_threadCount:0,Provider:null,Consumer:null},t.Provider={$$typeof:l,_context:t},t.Consumer=t},e.createElement=T,e.createFactory=function(t){var e=T.bind(null,t);return e.type=t,e},e.createRef=function(){return{current:null}},e.forwardRef=function(t){return{$$typeof:d,render:t}},e.isValidElement=E,e.lazy=function(t){return{$$typeof:y,_ctor:t,_status:-1,_result:null}},e.memo=function(t,e){return{$$typeof:p,type:t,compare:void 0===e?null:e}},e.useCallback=function(t,e){return $().useCallback(t,e)},e.useContext=function(t,e){return $().useContext(t,e)},e.useDebugValue=function(){},e.useEffect=function(t,e){return $().useEffect(t,e)},e.useImperativeHandle=function(t,e,n){return $().useImperativeHandle(t,e,n)},e.useLayoutEffect=function(t,e){return $().useLayoutEffect(t,e)},e.useMemo=function(t,e){return $().useMemo(t,e)},e.useReducer=function(t,e,n){return $().useReducer(t,e,n)},e.useRef=function(t){return $().useRef(t)},e.useState=function(t){return $().useState(t)},e.version="16.14.0"},c04e:function(t,e,n){var r=n("da84"),i=n("c65b"),o=n("861d"),s=n("d9b5"),a=n("dc4a"),c=n("485a"),u=n("b622"),l=r.TypeError,f=u("toPrimitive");t.exports=function(t,e){if(!o(t)||s(t))return t;var n,r=a(t,f);if(r){if(void 0===e&&(e="default"),n=i(r,t,e),!o(n)||s(n))return n;throw l("Can't convert object to primitive value")}return void 0===e&&(e="number"),c(t,e)}},c430:function(t,e){t.exports=!1},c65b:function(t,e){var n=Function.prototype.call;t.exports=n.bind?n.bind(n):function(){return n.apply(n,arguments)}},c6b6:function(t,e,n){var r=n("e330"),i=r({}.toString),o=r("".slice);t.exports=function(t){return o(i(t),8,-1)}},c6cd:function(t,e,n){var r=n("da84"),i=n("ce4e"),o="__core-js_shared__",s=r[o]||i(o,{});t.exports=s},c8ba:function(t,e){var n;n=function(){return this}();try{n=n||new Function("return this")()}catch(r){"object"===typeof window&&(n=window)}t.exports=n},ca84:function(t,e,n){var r=n("e330"),i=n("1a2d"),o=n("fc6a"),s=n("4d64").indexOf,a=n("d012"),c=r([].push);t.exports=function(t,e){var n,r=o(t),u=0,l=[];for(n in r)!i(a,n)&&i(r,n)&&c(l,n);while(e.length>u)i(r,n=e[u++])&&(~s(l,n)||c(l,n));return l}},cc12:function(t,e,n){var r=n("da84"),i=n("861d"),o=r.document,s=i(o)&&i(o.createElement);t.exports=function(t){return s?o.createElement(t):{}}},cca6:function(t,e,n){var r=n("23e7"),i=n("60da");r({target:"Object",stat:!0,forced:Object.assign!==i},{assign:i})},cdf9:function(t,e,n){var r=n("825a"),i=n("861d"),o=n("f069");t.exports=function(t,e){if(r(t),i(e)&&e.constructor===t)return e;var n=o.f(t),s=n.resolve;return s(e),n.promise}},ce4e:function(t,e,n){var r=n("da84"),i=Object.defineProperty;t.exports=function(t,e){try{i(r,t,{value:e,configurable:!0,writable:!0})}catch(n){r[t]=e}return e}},d012:function(t,e){t.exports={}},d039:function(t,e){t.exports=function(t){try{return!!t()}catch(e){return!0}}},d066:function(t,e,n){var r=n("da84"),i=n("1626"),o=function(t){return i(t)?t:void 0};t.exports=function(t,e){return arguments.length<2?o(r[t]):r[t]&&r[t][e]}},d092:function(t,e,n){"use strict";n.d(e,"a",(function(){return il}));var r={};n.r(r),n.d(r,"getBool",(function(){return G})),n.d(r,"getBit",(function(){return q})),n.d(r,"setBool",(function(){return K})),n.d(r,"truncateBitmap",(function(){return X})),n.d(r,"packBools",(function(){return J})),n.d(r,"iterateBits",(function(){return Z})),n.d(r,"popcnt_bit_range",(function(){return Q})),n.d(r,"popcnt_array",(function(){return tt})),n.d(r,"popcnt_uint32",(function(){return et}));var i={};n.r(i),n.d(i,"uint16ToFloat64",(function(){return Me})),n.d(i,"float64ToUint16",(function(){return ke}));var o={};n.r(o),n.d(o,"isArrowBigNumSymbol",(function(){return Fe})),n.d(o,"bignumToString",(function(){return Ye})),n.d(o,"bignumToBigInt",(function(){return He})),n.d(o,"BN",(function(){return Ge}));var s={};n.r(s),n.d(s,"clampIndex",(function(){return Mn})),n.d(s,"clampRange",(function(){return Dn})),n.d(s,"createElementComparator",(function(){return Pn}));var a={};n.r(a),n.d(a,"BaseInt64",(function(){return Li})),n.d(a,"Uint64",(function(){return Pi})),n.d(a,"Int64",(function(){return Fi})),n.d(a,"Int128",(function(){return Ui}));n("da6a");var c=n("ab5b"),u=n.n(c);const l=new WeakMap,f=new WeakMap;function d(t){const e=l.get(t);return console.assert(null!=e,"'this' is expected an Event object, but got",t),e}function h(t){null==t.passiveListener?t.event.cancelable&&(t.canceled=!0,"function"===typeof t.event.preventDefault&&t.event.preventDefault()):"undefined"!==typeof console&&"function"===typeof console.error&&console.error("Unable to preventDefault inside passive event listener invocation.",t.passiveListener)}function p(t,e){l.set(this,{eventTarget:t,event:e,eventPhase:2,currentTarget:t,canceled:!1,stopped:!1,immediateStopped:!1,passiveListener:null,timeStamp:e.timeStamp||Date.now()}),Object.defineProperty(this,"isTrusted",{value:!1,enumerable:!0});const n=Object.keys(e);for(let r=0;r<n.length;++r){const t=n[r];t in this||Object.defineProperty(this,t,y(t))}}function y(t){return{get(){return d(this).event[t]},set(e){d(this).event[t]=e},configurable:!0,enumerable:!0}}function v(t){return{value(){const e=d(this).event;return e[t].apply(e,arguments)},configurable:!0,enumerable:!0}}function b(t,e){const n=Object.keys(e);if(0===n.length)return t;function r(e,n){t.call(this,e,n)}r.prototype=Object.create(t.prototype,{constructor:{value:r,configurable:!0,writable:!0}});for(let i=0;i<n.length;++i){const o=n[i];if(!(o in t.prototype)){const t=Object.getOwnPropertyDescriptor(e,o),n="function"===typeof t.value;Object.defineProperty(r.prototype,o,n?v(o):y(o))}}return r}function m(t){if(null==t||t===Object.prototype)return p;let e=f.get(t);return null==e&&(e=b(m(Object.getPrototypeOf(t)),t),f.set(t,e)),e}function g(t,e){const n=m(Object.getPrototypeOf(e));return new n(t,e)}function _(t){return d(t).immediateStopped}function w(t,e){d(t).eventPhase=e}function O(t,e){d(t).currentTarget=e}function S(t,e){d(t).passiveListener=e}p.prototype={get type(){return d(this).event.type},get target(){return d(this).eventTarget},get currentTarget(){return d(this).currentTarget},composedPath(){const t=d(this).currentTarget;return null==t?[]:[t]},get NONE(){return 0},get CAPTURING_PHASE(){return 1},get AT_TARGET(){return 2},get BUBBLING_PHASE(){return 3},get eventPhase(){return d(this).eventPhase},stopPropagation(){const t=d(this);t.stopped=!0,"function"===typeof t.event.stopPropagation&&t.event.stopPropagation()},stopImmediatePropagation(){const t=d(this);t.stopped=!0,t.immediateStopped=!0,"function"===typeof t.event.stopImmediatePropagation&&t.event.stopImmediatePropagation()},get bubbles(){return Boolean(d(this).event.bubbles)},get cancelable(){return Boolean(d(this).event.cancelable)},preventDefault(){h(d(this))},get defaultPrevented(){return d(this).canceled},get composed(){return Boolean(d(this).event.composed)},get timeStamp(){return d(this).timeStamp},get srcElement(){return d(this).eventTarget},get cancelBubble(){return d(this).stopped},set cancelBubble(t){if(!t)return;const e=d(this);e.stopped=!0,"boolean"===typeof e.event.cancelBubble&&(e.event.cancelBubble=!0)},get returnValue(){return!d(this).canceled},set returnValue(t){t||h(d(this))},initEvent(){}},Object.defineProperty(p.prototype,"constructor",{value:p,configurable:!0,writable:!0}),"undefined"!==typeof window&&"undefined"!==typeof window.Event&&(Object.setPrototypeOf(p.prototype,window.Event.prototype),f.set(window.Event.prototype,p));const x=new WeakMap,I=1,A=2,T=3;function j(t){return null!==t&&"object"===typeof t}function E(t){const e=x.get(t);if(null==e)throw new TypeError("'this' is expected an EventTarget object, but got another value.");return e}function B(t){return{get(){const e=E(this);let n=e.get(t);while(null!=n){if(n.listenerType===T)return n.listener;n=n.next}return null},set(e){"function"===typeof e||j(e)||(e=null);const n=E(this);let r=null,i=n.get(t);while(null!=i)i.listenerType===T?null!==r?r.next=i.next:null!==i.next?n.set(t,i.next):n.delete(t):r=i,i=i.next;if(null!==e){const i={listener:e,listenerType:T,passive:!1,once:!1,next:null};null===r?n.set(t,i):r.next=i}},configurable:!0,enumerable:!0}}function C(t,e){Object.defineProperty(t,"on"+e,B(e))}function M(t){function e(){k.call(this)}e.prototype=Object.create(k.prototype,{constructor:{value:e,configurable:!0,writable:!0}});for(let n=0;n<t.length;++n)C(e.prototype,t[n]);return e}function k(){if(!(this instanceof k)){if(1===arguments.length&&Array.isArray(arguments[0]))return M(arguments[0]);if(arguments.length>0){const t=new Array(arguments.length);for(let e=0;e<arguments.length;++e)t[e]=arguments[e];return M(t)}throw new TypeError("Cannot call a class as a function")}x.set(this,new Map)}k.prototype={addEventListener(t,e,n){if(null==e)return;if("function"!==typeof e&&!j(e))throw new TypeError("'listener' should be a function or an object.");const r=E(this),i=j(n),o=i?Boolean(n.capture):Boolean(n),s=o?I:A,a={listener:e,listenerType:s,passive:i&&Boolean(n.passive),once:i&&Boolean(n.once),next:null};let c=r.get(t);if(void 0===c)return void r.set(t,a);let u=null;while(null!=c){if(c.listener===e&&c.listenerType===s)return;u=c,c=c.next}u.next=a},removeEventListener(t,e,n){if(null==e)return;const r=E(this),i=j(n)?Boolean(n.capture):Boolean(n),o=i?I:A;let s=null,a=r.get(t);while(null!=a){if(a.listener===e&&a.listenerType===o)return void(null!==s?s.next=a.next:null!==a.next?r.set(t,a.next):r.delete(t));s=a,a=a.next}},dispatchEvent(t){if(null==t||"string"!==typeof t.type)throw new TypeError('"event.type" should be a string.');const e=E(this),n=t.type;let r=e.get(n);if(null==r)return!0;const i=g(this,t);let o=null;while(null!=r){if(r.once?null!==o?o.next=r.next:null!==r.next?e.set(n,r.next):e.delete(n):o=r,S(i,r.passive?r.listener:null),"function"===typeof r.listener)try{r.listener.call(this,i)}catch(s){"undefined"!==typeof console&&"function"===typeof console.error&&console.error(s)}else r.listenerType!==T&&"function"===typeof r.listener.handleEvent&&r.listener.handleEvent(i);if(_(i))break;r=r.next}return S(i,null),w(i,0),O(i,null),!i.defaultPrevented}},Object.defineProperty(k.prototype,"constructor",{value:k,configurable:!0,writable:!0}),"undefined"!==typeof window&&"undefined"!==typeof window.EventTarget&&Object.setPrototypeOf(k.prototype,window.EventTarget.prototype);var D,N,L=n("6bfb");class P{}(function(t){(function(t){(function(t){(function(t){let e;(function(t){t[t["V1"]=0]="V1",t[t["V2"]=1]="V2",t[t["V3"]=2]="V3",t[t["V4"]=3]="V4"})(e=t.MetadataVersion||(t.MetadataVersion={}))})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))})(D||(D={})),function(t){(function(t){(function(t){(function(t){let e;(function(t){t[t["Sparse"]=0]="Sparse",t[t["Dense"]=1]="Dense"})(e=t.UnionMode||(t.UnionMode={}))})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){let e;(function(t){t[t["HALF"]=0]="HALF",t[t["SINGLE"]=1]="SINGLE",t[t["DOUBLE"]=2]="DOUBLE"})(e=t.Precision||(t.Precision={}))})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){let e;(function(t){t[t["DAY"]=0]="DAY",t[t["MILLISECOND"]=1]="MILLISECOND"})(e=t.DateUnit||(t.DateUnit={}))})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){let e;(function(t){t[t["SECOND"]=0]="SECOND",t[t["MILLISECOND"]=1]="MILLISECOND",t[t["MICROSECOND"]=2]="MICROSECOND",t[t["NANOSECOND"]=3]="NANOSECOND"})(e=t.TimeUnit||(t.TimeUnit={}))})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){let e;(function(t){t[t["YEAR_MONTH"]=0]="YEAR_MONTH",t[t["DAY_TIME"]=1]="DAY_TIME"})(e=t.IntervalUnit||(t.IntervalUnit={}))})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){let e;(function(t){t[t["NONE"]=0]="NONE",t[t["Null"]=1]="Null",t[t["Int"]=2]="Int",t[t["FloatingPoint"]=3]="FloatingPoint",t[t["Binary"]=4]="Binary",t[t["Utf8"]=5]="Utf8",t[t["Bool"]=6]="Bool",t[t["Decimal"]=7]="Decimal",t[t["Date"]=8]="Date",t[t["Time"]=9]="Time",t[t["Timestamp"]=10]="Timestamp",t[t["Interval"]=11]="Interval",t[t["List"]=12]="List",t[t["Struct_"]=13]="Struct_",t[t["Union"]=14]="Union",t[t["FixedSizeBinary"]=15]="FixedSizeBinary",t[t["FixedSizeList"]=16]="FixedSizeList",t[t["Map"]=17]="Map",t[t["Duration"]=18]="Duration",t[t["LargeBinary"]=19]="LargeBinary",t[t["LargeUtf8"]=20]="LargeUtf8",t[t["LargeList"]=21]="LargeList"})(e=t.Type||(t.Type={}))})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){let e;(function(t){t[t["Little"]=0]="Little",t[t["Big"]=1]="Big"})(e=t.Endianness||(t.Endianness={}))})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsNull(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}static startNull(t){t.startObject(0)}static endNull(t){let e=t.endObject();return e}static createNull(t){return e.startNull(t),e.endNull(t)}}t.Null=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsStruct_(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}static startStruct_(t){t.startObject(0)}static endStruct_(t){let e=t.endObject();return e}static createStruct_(t){return e.startStruct_(t),e.endStruct_(t)}}t.Struct_=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsList(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}static startList(t){t.startObject(0)}static endList(t){let e=t.endObject();return e}static createList(t){return e.startList(t),e.endList(t)}}t.List=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsLargeList(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}static startLargeList(t){t.startObject(0)}static endLargeList(t){let e=t.endObject();return e}static createLargeList(t){return e.startLargeList(t),e.endLargeList(t)}}t.LargeList=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsFixedSizeList(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}listSize(){let t=this.bb.__offset(this.bb_pos,4);return t?this.bb.readInt32(this.bb_pos+t):0}static startFixedSizeList(t){t.startObject(1)}static addListSize(t,e){t.addFieldInt32(0,e,0)}static endFixedSizeList(t){let e=t.endObject();return e}static createFixedSizeList(t,n){return e.startFixedSizeList(t),e.addListSize(t,n),e.endFixedSizeList(t)}}t.FixedSizeList=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsMap(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}keysSorted(){let t=this.bb.__offset(this.bb_pos,4);return!!t&&!!this.bb.readInt8(this.bb_pos+t)}static startMap(t){t.startObject(1)}static addKeysSorted(t,e){t.addFieldInt8(0,+e,0)}static endMap(t){let e=t.endObject();return e}static createMap(t,n){return e.startMap(t),e.addKeysSorted(t,n),e.endMap(t)}}t.Map=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsUnion(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}mode(){let e=this.bb.__offset(this.bb_pos,4);return e?this.bb.readInt16(this.bb_pos+e):t.apache.arrow.flatbuf.UnionMode.Sparse}typeIds(t){let e=this.bb.__offset(this.bb_pos,6);return e?this.bb.readInt32(this.bb.__vector(this.bb_pos+e)+4*t):0}typeIdsLength(){let t=this.bb.__offset(this.bb_pos,6);return t?this.bb.__vector_len(this.bb_pos+t):0}typeIdsArray(){let t=this.bb.__offset(this.bb_pos,6);return t?new Int32Array(this.bb.bytes().buffer,this.bb.bytes().byteOffset+this.bb.__vector(this.bb_pos+t),this.bb.__vector_len(this.bb_pos+t)):null}static startUnion(t){t.startObject(2)}static addMode(e,n){e.addFieldInt16(0,n,t.apache.arrow.flatbuf.UnionMode.Sparse)}static addTypeIds(t,e){t.addFieldOffset(1,e,0)}static createTypeIdsVector(t,e){t.startVector(4,e.length,4);for(let n=e.length-1;n>=0;n--)t.addInt32(e[n]);return t.endVector()}static startTypeIdsVector(t,e){t.startVector(4,e,4)}static endUnion(t){let e=t.endObject();return e}static createUnion(t,e,r){return n.startUnion(t),n.addMode(t,e),n.addTypeIds(t,r),n.endUnion(t)}}e.Union=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsInt(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}bitWidth(){let t=this.bb.__offset(this.bb_pos,4);return t?this.bb.readInt32(this.bb_pos+t):0}isSigned(){let t=this.bb.__offset(this.bb_pos,6);return!!t&&!!this.bb.readInt8(this.bb_pos+t)}static startInt(t){t.startObject(2)}static addBitWidth(t,e){t.addFieldInt32(0,e,0)}static addIsSigned(t,e){t.addFieldInt8(1,+e,0)}static endInt(t){let e=t.endObject();return e}static createInt(t,n,r){return e.startInt(t),e.addBitWidth(t,n),e.addIsSigned(t,r),e.endInt(t)}}t.Int=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsFloatingPoint(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}precision(){let e=this.bb.__offset(this.bb_pos,4);return e?this.bb.readInt16(this.bb_pos+e):t.apache.arrow.flatbuf.Precision.HALF}static startFloatingPoint(t){t.startObject(1)}static addPrecision(e,n){e.addFieldInt16(0,n,t.apache.arrow.flatbuf.Precision.HALF)}static endFloatingPoint(t){let e=t.endObject();return e}static createFloatingPoint(t,e){return n.startFloatingPoint(t),n.addPrecision(t,e),n.endFloatingPoint(t)}}e.FloatingPoint=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsUtf8(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}static startUtf8(t){t.startObject(0)}static endUtf8(t){let e=t.endObject();return e}static createUtf8(t){return e.startUtf8(t),e.endUtf8(t)}}t.Utf8=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsBinary(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}static startBinary(t){t.startObject(0)}static endBinary(t){let e=t.endObject();return e}static createBinary(t){return e.startBinary(t),e.endBinary(t)}}t.Binary=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsLargeUtf8(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}static startLargeUtf8(t){t.startObject(0)}static endLargeUtf8(t){let e=t.endObject();return e}static createLargeUtf8(t){return e.startLargeUtf8(t),e.endLargeUtf8(t)}}t.LargeUtf8=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsLargeBinary(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}static startLargeBinary(t){t.startObject(0)}static endLargeBinary(t){let e=t.endObject();return e}static createLargeBinary(t){return e.startLargeBinary(t),e.endLargeBinary(t)}}t.LargeBinary=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsFixedSizeBinary(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}byteWidth(){let t=this.bb.__offset(this.bb_pos,4);return t?this.bb.readInt32(this.bb_pos+t):0}static startFixedSizeBinary(t){t.startObject(1)}static addByteWidth(t,e){t.addFieldInt32(0,e,0)}static endFixedSizeBinary(t){let e=t.endObject();return e}static createFixedSizeBinary(t,n){return e.startFixedSizeBinary(t),e.addByteWidth(t,n),e.endFixedSizeBinary(t)}}t.FixedSizeBinary=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsBool(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}static startBool(t){t.startObject(0)}static endBool(t){let e=t.endObject();return e}static createBool(t){return e.startBool(t),e.endBool(t)}}t.Bool=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsDecimal(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}precision(){let t=this.bb.__offset(this.bb_pos,4);return t?this.bb.readInt32(this.bb_pos+t):0}scale(){let t=this.bb.__offset(this.bb_pos,6);return t?this.bb.readInt32(this.bb_pos+t):0}static startDecimal(t){t.startObject(2)}static addPrecision(t,e){t.addFieldInt32(0,e,0)}static addScale(t,e){t.addFieldInt32(1,e,0)}static endDecimal(t){let e=t.endObject();return e}static createDecimal(t,n,r){return e.startDecimal(t),e.addPrecision(t,n),e.addScale(t,r),e.endDecimal(t)}}t.Decimal=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsDate(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}unit(){let e=this.bb.__offset(this.bb_pos,4);return e?this.bb.readInt16(this.bb_pos+e):t.apache.arrow.flatbuf.DateUnit.MILLISECOND}static startDate(t){t.startObject(1)}static addUnit(e,n){e.addFieldInt16(0,n,t.apache.arrow.flatbuf.DateUnit.MILLISECOND)}static endDate(t){let e=t.endObject();return e}static createDate(t,e){return n.startDate(t),n.addUnit(t,e),n.endDate(t)}}e.Date=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsTime(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}unit(){let e=this.bb.__offset(this.bb_pos,4);return e?this.bb.readInt16(this.bb_pos+e):t.apache.arrow.flatbuf.TimeUnit.MILLISECOND}bitWidth(){let t=this.bb.__offset(this.bb_pos,6);return t?this.bb.readInt32(this.bb_pos+t):32}static startTime(t){t.startObject(2)}static addUnit(e,n){e.addFieldInt16(0,n,t.apache.arrow.flatbuf.TimeUnit.MILLISECOND)}static addBitWidth(t,e){t.addFieldInt32(1,e,32)}static endTime(t){let e=t.endObject();return e}static createTime(t,e,r){return n.startTime(t),n.addUnit(t,e),n.addBitWidth(t,r),n.endTime(t)}}e.Time=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsTimestamp(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}unit(){let e=this.bb.__offset(this.bb_pos,4);return e?this.bb.readInt16(this.bb_pos+e):t.apache.arrow.flatbuf.TimeUnit.SECOND}timezone(t){let e=this.bb.__offset(this.bb_pos,6);return e?this.bb.__string(this.bb_pos+e,t):null}static startTimestamp(t){t.startObject(2)}static addUnit(e,n){e.addFieldInt16(0,n,t.apache.arrow.flatbuf.TimeUnit.SECOND)}static addTimezone(t,e){t.addFieldOffset(1,e,0)}static endTimestamp(t){let e=t.endObject();return e}static createTimestamp(t,e,r){return n.startTimestamp(t),n.addUnit(t,e),n.addTimezone(t,r),n.endTimestamp(t)}}e.Timestamp=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsInterval(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}unit(){let e=this.bb.__offset(this.bb_pos,4);return e?this.bb.readInt16(this.bb_pos+e):t.apache.arrow.flatbuf.IntervalUnit.YEAR_MONTH}static startInterval(t){t.startObject(1)}static addUnit(e,n){e.addFieldInt16(0,n,t.apache.arrow.flatbuf.IntervalUnit.YEAR_MONTH)}static endInterval(t){let e=t.endObject();return e}static createInterval(t,e){return n.startInterval(t),n.addUnit(t,e),n.endInterval(t)}}e.Interval=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsDuration(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}unit(){let e=this.bb.__offset(this.bb_pos,4);return e?this.bb.readInt16(this.bb_pos+e):t.apache.arrow.flatbuf.TimeUnit.MILLISECOND}static startDuration(t){t.startObject(1)}static addUnit(e,n){e.addFieldInt16(0,n,t.apache.arrow.flatbuf.TimeUnit.MILLISECOND)}static endDuration(t){let e=t.endObject();return e}static createDuration(t,e){return n.startDuration(t),n.addUnit(t,e),n.endDuration(t)}}e.Duration=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsKeyValue(t,n){return(n||new e).__init(t.readInt32(t.position())+t.position(),t)}key(t){let e=this.bb.__offset(this.bb_pos,4);return e?this.bb.__string(this.bb_pos+e,t):null}value(t){let e=this.bb.__offset(this.bb_pos,6);return e?this.bb.__string(this.bb_pos+e,t):null}static startKeyValue(t){t.startObject(2)}static addKey(t,e){t.addFieldOffset(0,e,0)}static addValue(t,e){t.addFieldOffset(1,e,0)}static endKeyValue(t){let e=t.endObject();return e}static createKeyValue(t,n,r){return e.startKeyValue(t),e.addKey(t,n),e.addValue(t,r),e.endKeyValue(t)}}t.KeyValue=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsDictionaryEncoding(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}id(){let t=this.bb.__offset(this.bb_pos,4);return t?this.bb.readInt64(this.bb_pos+t):this.bb.createLong(0,0)}indexType(e){let n=this.bb.__offset(this.bb_pos,6);return n?(e||new t.apache.arrow.flatbuf.Int).__init(this.bb.__indirect(this.bb_pos+n),this.bb):null}isOrdered(){let t=this.bb.__offset(this.bb_pos,8);return!!t&&!!this.bb.readInt8(this.bb_pos+t)}static startDictionaryEncoding(t){t.startObject(3)}static addId(t,e){t.addFieldInt64(0,e,t.createLong(0,0))}static addIndexType(t,e){t.addFieldOffset(1,e,0)}static addIsOrdered(t,e){t.addFieldInt8(2,+e,0)}static endDictionaryEncoding(t){let e=t.endObject();return e}static createDictionaryEncoding(t,e,r,i){return n.startDictionaryEncoding(t),n.addId(t,e),n.addIndexType(t,r),n.addIsOrdered(t,i),n.endDictionaryEncoding(t)}}e.DictionaryEncoding=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsField(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}name(t){let e=this.bb.__offset(this.bb_pos,4);return e?this.bb.__string(this.bb_pos+e,t):null}nullable(){let t=this.bb.__offset(this.bb_pos,6);return!!t&&!!this.bb.readInt8(this.bb_pos+t)}typeType(){let e=this.bb.__offset(this.bb_pos,8);return e?this.bb.readUint8(this.bb_pos+e):t.apache.arrow.flatbuf.Type.NONE}type(t){let e=this.bb.__offset(this.bb_pos,10);return e?this.bb.__union(t,this.bb_pos+e):null}dictionary(e){let n=this.bb.__offset(this.bb_pos,12);return n?(e||new t.apache.arrow.flatbuf.DictionaryEncoding).__init(this.bb.__indirect(this.bb_pos+n),this.bb):null}children(e,n){let r=this.bb.__offset(this.bb_pos,14);return r?(n||new t.apache.arrow.flatbuf.Field).__init(this.bb.__indirect(this.bb.__vector(this.bb_pos+r)+4*e),this.bb):null}childrenLength(){let t=this.bb.__offset(this.bb_pos,14);return t?this.bb.__vector_len(this.bb_pos+t):0}customMetadata(e,n){let r=this.bb.__offset(this.bb_pos,16);return r?(n||new t.apache.arrow.flatbuf.KeyValue).__init(this.bb.__indirect(this.bb.__vector(this.bb_pos+r)+4*e),this.bb):null}customMetadataLength(){let t=this.bb.__offset(this.bb_pos,16);return t?this.bb.__vector_len(this.bb_pos+t):0}static startField(t){t.startObject(7)}static addName(t,e){t.addFieldOffset(0,e,0)}static addNullable(t,e){t.addFieldInt8(1,+e,0)}static addTypeType(e,n){e.addFieldInt8(2,n,t.apache.arrow.flatbuf.Type.NONE)}static addType(t,e){t.addFieldOffset(3,e,0)}static addDictionary(t,e){t.addFieldOffset(4,e,0)}static addChildren(t,e){t.addFieldOffset(5,e,0)}static createChildrenVector(t,e){t.startVector(4,e.length,4);for(let n=e.length-1;n>=0;n--)t.addOffset(e[n]);return t.endVector()}static startChildrenVector(t,e){t.startVector(4,e,4)}static addCustomMetadata(t,e){t.addFieldOffset(6,e,0)}static createCustomMetadataVector(t,e){t.startVector(4,e.length,4);for(let n=e.length-1;n>=0;n--)t.addOffset(e[n]);return t.endVector()}static startCustomMetadataVector(t,e){t.startVector(4,e,4)}static endField(t){let e=t.endObject();return e}static createField(t,e,r,i,o,s,a,c){return n.startField(t),n.addName(t,e),n.addNullable(t,r),n.addTypeType(t,i),n.addType(t,o),n.addDictionary(t,s),n.addChildren(t,a),n.addCustomMetadata(t,c),n.endField(t)}}e.Field=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}offset(){return this.bb.readInt64(this.bb_pos)}length(){return this.bb.readInt64(this.bb_pos+8)}static createBuffer(t,e,n){return t.prep(8,16),t.writeInt64(n),t.writeInt64(e),t.offset()}}t.Buffer=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsSchema(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}endianness(){let e=this.bb.__offset(this.bb_pos,4);return e?this.bb.readInt16(this.bb_pos+e):t.apache.arrow.flatbuf.Endianness.Little}fields(e,n){let r=this.bb.__offset(this.bb_pos,6);return r?(n||new t.apache.arrow.flatbuf.Field).__init(this.bb.__indirect(this.bb.__vector(this.bb_pos+r)+4*e),this.bb):null}fieldsLength(){let t=this.bb.__offset(this.bb_pos,6);return t?this.bb.__vector_len(this.bb_pos+t):0}customMetadata(e,n){let r=this.bb.__offset(this.bb_pos,8);return r?(n||new t.apache.arrow.flatbuf.KeyValue).__init(this.bb.__indirect(this.bb.__vector(this.bb_pos+r)+4*e),this.bb):null}customMetadataLength(){let t=this.bb.__offset(this.bb_pos,8);return t?this.bb.__vector_len(this.bb_pos+t):0}static startSchema(t){t.startObject(3)}static addEndianness(e,n){e.addFieldInt16(0,n,t.apache.arrow.flatbuf.Endianness.Little)}static addFields(t,e){t.addFieldOffset(1,e,0)}static createFieldsVector(t,e){t.startVector(4,e.length,4);for(let n=e.length-1;n>=0;n--)t.addOffset(e[n]);return t.endVector()}static startFieldsVector(t,e){t.startVector(4,e,4)}static addCustomMetadata(t,e){t.addFieldOffset(2,e,0)}static createCustomMetadataVector(t,e){t.startVector(4,e.length,4);for(let n=e.length-1;n>=0;n--)t.addOffset(e[n]);return t.endVector()}static startCustomMetadataVector(t,e){t.startVector(4,e,4)}static endSchema(t){let e=t.endObject();return e}static finishSchemaBuffer(t,e){t.finish(e)}static createSchema(t,e,r,i){return n.startSchema(t),n.addEndianness(t,e),n.addFields(t,r),n.addCustomMetadata(t,i),n.endSchema(t)}}e.Schema=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(D||(D={})),function(t){(function(t){(function(t){(function(t){t.Schema=D.apache.arrow.flatbuf.Schema})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(N||(N={})),function(t){(function(t){(function(t){(function(t){let e;(function(t){t[t["NONE"]=0]="NONE",t[t["Schema"]=1]="Schema",t[t["DictionaryBatch"]=2]="DictionaryBatch",t[t["RecordBatch"]=3]="RecordBatch",t[t["Tensor"]=4]="Tensor",t[t["SparseTensor"]=5]="SparseTensor"})(e=t.MessageHeader||(t.MessageHeader={}))})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(N||(N={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}length(){return this.bb.readInt64(this.bb_pos)}nullCount(){return this.bb.readInt64(this.bb_pos+8)}static createFieldNode(t,e,n){return t.prep(8,16),t.writeInt64(n),t.writeInt64(e),t.offset()}}t.FieldNode=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(N||(N={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsRecordBatch(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}length(){let t=this.bb.__offset(this.bb_pos,4);return t?this.bb.readInt64(this.bb_pos+t):this.bb.createLong(0,0)}nodes(e,n){let r=this.bb.__offset(this.bb_pos,6);return r?(n||new t.apache.arrow.flatbuf.FieldNode).__init(this.bb.__vector(this.bb_pos+r)+16*e,this.bb):null}nodesLength(){let t=this.bb.__offset(this.bb_pos,6);return t?this.bb.__vector_len(this.bb_pos+t):0}buffers(t,e){let n=this.bb.__offset(this.bb_pos,8);return n?(e||new D.apache.arrow.flatbuf.Buffer).__init(this.bb.__vector(this.bb_pos+n)+16*t,this.bb):null}buffersLength(){let t=this.bb.__offset(this.bb_pos,8);return t?this.bb.__vector_len(this.bb_pos+t):0}static startRecordBatch(t){t.startObject(3)}static addLength(t,e){t.addFieldInt64(0,e,t.createLong(0,0))}static addNodes(t,e){t.addFieldOffset(1,e,0)}static startNodesVector(t,e){t.startVector(16,e,8)}static addBuffers(t,e){t.addFieldOffset(2,e,0)}static startBuffersVector(t,e){t.startVector(16,e,8)}static endRecordBatch(t){let e=t.endObject();return e}static createRecordBatch(t,e,r,i){return n.startRecordBatch(t),n.addLength(t,e),n.addNodes(t,r),n.addBuffers(t,i),n.endRecordBatch(t)}}e.RecordBatch=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(N||(N={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsDictionaryBatch(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}id(){let t=this.bb.__offset(this.bb_pos,4);return t?this.bb.readInt64(this.bb_pos+t):this.bb.createLong(0,0)}data(e){let n=this.bb.__offset(this.bb_pos,6);return n?(e||new t.apache.arrow.flatbuf.RecordBatch).__init(this.bb.__indirect(this.bb_pos+n),this.bb):null}isDelta(){let t=this.bb.__offset(this.bb_pos,8);return!!t&&!!this.bb.readInt8(this.bb_pos+t)}static startDictionaryBatch(t){t.startObject(3)}static addId(t,e){t.addFieldInt64(0,e,t.createLong(0,0))}static addData(t,e){t.addFieldOffset(1,e,0)}static addIsDelta(t,e){t.addFieldInt8(2,+e,0)}static endDictionaryBatch(t){let e=t.endObject();return e}static createDictionaryBatch(t,e,r,i){return n.startDictionaryBatch(t),n.addId(t,e),n.addData(t,r),n.addIsDelta(t,i),n.endDictionaryBatch(t)}}e.DictionaryBatch=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(N||(N={})),function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsMessage(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}version(){let t=this.bb.__offset(this.bb_pos,4);return t?this.bb.readInt16(this.bb_pos+t):D.apache.arrow.flatbuf.MetadataVersion.V1}headerType(){let e=this.bb.__offset(this.bb_pos,6);return e?this.bb.readUint8(this.bb_pos+e):t.apache.arrow.flatbuf.MessageHeader.NONE}header(t){let e=this.bb.__offset(this.bb_pos,8);return e?this.bb.__union(t,this.bb_pos+e):null}bodyLength(){let t=this.bb.__offset(this.bb_pos,10);return t?this.bb.readInt64(this.bb_pos+t):this.bb.createLong(0,0)}customMetadata(t,e){let n=this.bb.__offset(this.bb_pos,12);return n?(e||new D.apache.arrow.flatbuf.KeyValue).__init(this.bb.__indirect(this.bb.__vector(this.bb_pos+n)+4*t),this.bb):null}customMetadataLength(){let t=this.bb.__offset(this.bb_pos,12);return t?this.bb.__vector_len(this.bb_pos+t):0}static startMessage(t){t.startObject(5)}static addVersion(t,e){t.addFieldInt16(0,e,D.apache.arrow.flatbuf.MetadataVersion.V1)}static addHeaderType(e,n){e.addFieldInt8(1,n,t.apache.arrow.flatbuf.MessageHeader.NONE)}static addHeader(t,e){t.addFieldOffset(2,e,0)}static addBodyLength(t,e){t.addFieldInt64(3,e,t.createLong(0,0))}static addCustomMetadata(t,e){t.addFieldOffset(4,e,0)}static createCustomMetadataVector(t,e){t.startVector(4,e.length,4);for(let n=e.length-1;n>=0;n--)t.addOffset(e[n]);return t.endVector()}static startCustomMetadataVector(t,e){t.startVector(4,e,4)}static endMessage(t){let e=t.endObject();return e}static finishMessageBuffer(t,e){t.finish(e)}static createMessage(t,e,r,i,o,s){return n.startMessage(t),n.addVersion(t,e),n.addHeaderType(t,r),n.addHeader(t,i),n.addBodyLength(t,o),n.addCustomMetadata(t,s),n.endMessage(t)}}e.Message=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(N||(N={}));D.apache.arrow.flatbuf.Type;var F,U,R=D.apache.arrow.flatbuf.DateUnit,V=D.apache.arrow.flatbuf.TimeUnit,$=D.apache.arrow.flatbuf.Precision,z=D.apache.arrow.flatbuf.UnionMode,Y=D.apache.arrow.flatbuf.IntervalUnit,H=N.apache.arrow.flatbuf.MessageHeader,W=D.apache.arrow.flatbuf.MetadataVersion;function G(t,e,n,r){return 0!==(n&1<<r)}function q(t,e,n,r){return(n&1<<r)>>r}function K(t,e,n){return n?!!(t[e>>3]|=1<<e%8)||!0:!(t[e>>3]&=~(1<<e%8))&&!1}function X(t,e,n){const r=n.byteLength+7&-8;if(t>0||n.byteLength<r){const i=new Uint8Array(r);return i.set(t%8===0?n.subarray(t>>3):J(Z(n,t,e,null,G)).subarray(0,r)),i}return n}function J(t){let e=[],n=0,r=0,i=0;for(const s of t)s&&(i|=1<<r),8===++r&&(e[n++]=i,i=r=0);(0===n||r>0)&&(e[n++]=i);let o=new Uint8Array(e.length+7&-8);return o.set(e),o}function*Z(t,e,n,r,i){let o=e%8,s=e>>3,a=0,c=n;for(;c>0;o=0){let e=t[s++];do{yield i(r,a++,e,o)}while(--c>0&&++o<8)}}function Q(t,e,n){if(n-e<=0)return 0;if(n-e<8){let r=0;for(const i of Z(t,e,n-e,t,q))r+=i;return r}const r=n>>3<<3,i=e+(e%8===0?0:8-e%8);return Q(t,e,i)+Q(t,r,n)+tt(t,i>>3,r-i>>3)}function tt(t,e,n){let r=0,i=0|e;const o=new DataView(t.buffer,t.byteOffset,t.byteLength),s=void 0===n?t.byteLength:i+n;while(s-i>=4)r+=et(o.getUint32(i)),i+=4;while(s-i>=2)r+=et(o.getUint16(i)),i+=2;while(s-i>=1)r+=et(o.getUint8(i)),i+=1;return r}function et(t){let e=0|t;return e-=e>>>1&1431655765,e=(858993459&e)+(e>>>2&858993459),16843009*(e+(e>>>4)&252645135)>>>24}(function(t){t[t["NONE"]=0]="NONE",t[t["Null"]=1]="Null",t[t["Int"]=2]="Int",t[t["Float"]=3]="Float",t[t["Binary"]=4]="Binary",t[t["Utf8"]=5]="Utf8",t[t["Bool"]=6]="Bool",t[t["Decimal"]=7]="Decimal",t[t["Date"]=8]="Date",t[t["Time"]=9]="Time",t[t["Timestamp"]=10]="Timestamp",t[t["Interval"]=11]="Interval",t[t["List"]=12]="List",t[t["Struct"]=13]="Struct",t[t["Union"]=14]="Union",t[t["FixedSizeBinary"]=15]="FixedSizeBinary",t[t["FixedSizeList"]=16]="FixedSizeList",t[t["Map"]=17]="Map",t[t["Dictionary"]=-1]="Dictionary",t[t["Int8"]=-2]="Int8",t[t["Int16"]=-3]="Int16",t[t["Int32"]=-4]="Int32",t[t["Int64"]=-5]="Int64",t[t["Uint8"]=-6]="Uint8",t[t["Uint16"]=-7]="Uint16",t[t["Uint32"]=-8]="Uint32",t[t["Uint64"]=-9]="Uint64",t[t["Float16"]=-10]="Float16",t[t["Float32"]=-11]="Float32",t[t["Float64"]=-12]="Float64",t[t["DateDay"]=-13]="DateDay",t[t["DateMillisecond"]=-14]="DateMillisecond",t[t["TimestampSecond"]=-15]="TimestampSecond",t[t["TimestampMillisecond"]=-16]="TimestampMillisecond",t[t["TimestampMicrosecond"]=-17]="TimestampMicrosecond",t[t["TimestampNanosecond"]=-18]="TimestampNanosecond",t[t["TimeSecond"]=-19]="TimeSecond",t[t["TimeMillisecond"]=-20]="TimeMillisecond",t[t["TimeMicrosecond"]=-21]="TimeMicrosecond",t[t["TimeNanosecond"]=-22]="TimeNanosecond",t[t["DenseUnion"]=-23]="DenseUnion",t[t["SparseUnion"]=-24]="SparseUnion",t[t["IntervalDayTime"]=-25]="IntervalDayTime",t[t["IntervalYearMonth"]=-26]="IntervalYearMonth"})(F||(F={})),function(t){t[t["OFFSET"]=0]="OFFSET",t[t["DATA"]=1]="DATA",t[t["VALIDITY"]=2]="VALIDITY",t[t["TYPE"]=3]="TYPE"}(U||(U={}));class nt{visitMany(t,...e){return t.map((t,n)=>this.visit(t,...e.map(t=>t[n])))}visit(...t){return this.getVisitFn(t[0],!1).apply(this,t)}getVisitFn(t,e=!0){return rt(this,t,e)}visitNull(t,...e){return null}visitBool(t,...e){return null}visitInt(t,...e){return null}visitFloat(t,...e){return null}visitUtf8(t,...e){return null}visitBinary(t,...e){return null}visitFixedSizeBinary(t,...e){return null}visitDate(t,...e){return null}visitTimestamp(t,...e){return null}visitTime(t,...e){return null}visitDecimal(t,...e){return null}visitList(t,...e){return null}visitStruct(t,...e){return null}visitUnion(t,...e){return null}visitDictionary(t,...e){return null}visitInterval(t,...e){return null}visitFixedSizeList(t,...e){return null}visitMap(t,...e){return null}}function rt(t,e,n=!0){let r=null,i=F.NONE;switch(e instanceof ie||e instanceof P?i=it(e.type):e instanceof Ot?i=it(e):"number"!==typeof(i=e)&&(i=F[e]),i){case F.Null:r=t.visitNull;break;case F.Bool:r=t.visitBool;break;case F.Int:r=t.visitInt;break;case F.Int8:r=t.visitInt8||t.visitInt;break;case F.Int16:r=t.visitInt16||t.visitInt;break;case F.Int32:r=t.visitInt32||t.visitInt;break;case F.Int64:r=t.visitInt64||t.visitInt;break;case F.Uint8:r=t.visitUint8||t.visitInt;break;case F.Uint16:r=t.visitUint16||t.visitInt;break;case F.Uint32:r=t.visitUint32||t.visitInt;break;case F.Uint64:r=t.visitUint64||t.visitInt;break;case F.Float:r=t.visitFloat;break;case F.Float16:r=t.visitFloat16||t.visitFloat;break;case F.Float32:r=t.visitFloat32||t.visitFloat;break;case F.Float64:r=t.visitFloat64||t.visitFloat;break;case F.Utf8:r=t.visitUtf8;break;case F.Binary:r=t.visitBinary;break;case F.FixedSizeBinary:r=t.visitFixedSizeBinary;break;case F.Date:r=t.visitDate;break;case F.DateDay:r=t.visitDateDay||t.visitDate;break;case F.DateMillisecond:r=t.visitDateMillisecond||t.visitDate;break;case F.Timestamp:r=t.visitTimestamp;break;case F.TimestampSecond:r=t.visitTimestampSecond||t.visitTimestamp;break;case F.TimestampMillisecond:r=t.visitTimestampMillisecond||t.visitTimestamp;break;case F.TimestampMicrosecond:r=t.visitTimestampMicrosecond||t.visitTimestamp;break;case F.TimestampNanosecond:r=t.visitTimestampNanosecond||t.visitTimestamp;break;case F.Time:r=t.visitTime;break;case F.TimeSecond:r=t.visitTimeSecond||t.visitTime;break;case F.TimeMillisecond:r=t.visitTimeMillisecond||t.visitTime;break;case F.TimeMicrosecond:r=t.visitTimeMicrosecond||t.visitTime;break;case F.TimeNanosecond:r=t.visitTimeNanosecond||t.visitTime;break;case F.Decimal:r=t.visitDecimal;break;case F.List:r=t.visitList;break;case F.Struct:r=t.visitStruct;break;case F.Union:r=t.visitUnion;break;case F.DenseUnion:r=t.visitDenseUnion||t.visitUnion;break;case F.SparseUnion:r=t.visitSparseUnion||t.visitUnion;break;case F.Dictionary:r=t.visitDictionary;break;case F.Interval:r=t.visitInterval;break;case F.IntervalDayTime:r=t.visitIntervalDayTime||t.visitInterval;break;case F.IntervalYearMonth:r=t.visitIntervalYearMonth||t.visitInterval;break;case F.FixedSizeList:r=t.visitFixedSizeList;break;case F.Map:r=t.visitMap;break}if("function"===typeof r)return r;if(!n)return()=>null;throw new Error(`Unrecognized type '${F[i]}'`)}function it(t){switch(t.typeId){case F.Null:return F.Null;case F.Int:const{bitWidth:e,isSigned:n}=t;switch(e){case 8:return n?F.Int8:F.Uint8;case 16:return n?F.Int16:F.Uint16;case 32:return n?F.Int32:F.Uint32;case 64:return n?F.Int64:F.Uint64}return F.Int;case F.Float:switch(t.precision){case $.HALF:return F.Float16;case $.SINGLE:return F.Float32;case $.DOUBLE:return F.Float64}return F.Float;case F.Binary:return F.Binary;case F.Utf8:return F.Utf8;case F.Bool:return F.Bool;case F.Decimal:return F.Decimal;case F.Time:switch(t.unit){case V.SECOND:return F.TimeSecond;case V.MILLISECOND:return F.TimeMillisecond;case V.MICROSECOND:return F.TimeMicrosecond;case V.NANOSECOND:return F.TimeNanosecond}return F.Time;case F.Timestamp:switch(t.unit){case V.SECOND:return F.TimestampSecond;case V.MILLISECOND:return F.TimestampMillisecond;case V.MICROSECOND:return F.TimestampMicrosecond;case V.NANOSECOND:return F.TimestampNanosecond}return F.Timestamp;case F.Date:switch(t.unit){case R.DAY:return F.DateDay;case R.MILLISECOND:return F.DateMillisecond}return F.Date;case F.Interval:switch(t.unit){case Y.DAY_TIME:return F.IntervalDayTime;case Y.YEAR_MONTH:return F.IntervalYearMonth}return F.Interval;case F.Map:return F.Map;case F.List:return F.List;case F.Struct:return F.Struct;case F.Union:switch(t.mode){case z.Dense:return F.DenseUnion;case z.Sparse:return F.SparseUnion}return F.Union;case F.FixedSizeBinary:return F.FixedSizeBinary;case F.FixedSizeList:return F.FixedSizeList;case F.Dictionary:return F.Dictionary}throw new Error(`Unrecognized type '${F[t.typeId]}'`)}nt.prototype.visitInt8=null,nt.prototype.visitInt16=null,nt.prototype.visitInt32=null,nt.prototype.visitInt64=null,nt.prototype.visitUint8=null,nt.prototype.visitUint16=null,nt.prototype.visitUint32=null,nt.prototype.visitUint64=null,nt.prototype.visitFloat16=null,nt.prototype.visitFloat32=null,nt.prototype.visitFloat64=null,nt.prototype.visitDateDay=null,nt.prototype.visitDateMillisecond=null,nt.prototype.visitTimestampSecond=null,nt.prototype.visitTimestampMillisecond=null,nt.prototype.visitTimestampMicrosecond=null,nt.prototype.visitTimestampNanosecond=null,nt.prototype.visitTimeSecond=null,nt.prototype.visitTimeMillisecond=null,nt.prototype.visitTimeMicrosecond=null,nt.prototype.visitTimeNanosecond=null,nt.prototype.visitDenseUnion=null,nt.prototype.visitSparseUnion=null,nt.prototype.visitIntervalDayTime=null,nt.prototype.visitIntervalYearMonth=null;class ot extends nt{compareSchemas(t,e){return t===e||e instanceof t.constructor&&wt.compareFields(t.fields,e.fields)}compareFields(t,e){return t===e||Array.isArray(t)&&Array.isArray(e)&&t.length===e.length&&t.every((t,n)=>wt.compareField(t,e[n]))}compareField(t,e){return t===e||e instanceof t.constructor&&t.name===e.name&&t.nullable===e.nullable&&wt.visit(t.type,e.type)}}function st(t,e){return e instanceof t.constructor}function at(t,e){return t===e||st(t,e)}function ct(t,e){return t===e||st(t,e)&&t.bitWidth===e.bitWidth&&t.isSigned===e.isSigned}function ut(t,e){return t===e||st(t,e)&&t.precision===e.precision}function lt(t,e){return t===e||st(t,e)&&t.byteWidth===e.byteWidth}function ft(t,e){return t===e||st(t,e)&&t.unit===e.unit}function dt(t,e){return t===e||st(t,e)&&t.unit===e.unit&&t.timezone===e.timezone}function ht(t,e){return t===e||st(t,e)&&t.unit===e.unit&&t.bitWidth===e.bitWidth}function pt(t,e){return t===e||st(t,e)&&t.children.length===e.children.length&&wt.compareFields(t.children,e.children)}function yt(t,e){return t===e||st(t,e)&&t.children.length===e.children.length&&wt.compareFields(t.children,e.children)}function vt(t,e){return t===e||st(t,e)&&t.mode===e.mode&&t.typeIds.every((t,n)=>t===e.typeIds[n])&&wt.compareFields(t.children,e.children)}function bt(t,e){return t===e||st(t,e)&&t.id===e.id&&t.isOrdered===e.isOrdered&&wt.visit(t.indices,e.indices)&&wt.visit(t.dictionary,e.dictionary)}function mt(t,e){return t===e||st(t,e)&&t.unit===e.unit}function gt(t,e){return t===e||st(t,e)&&t.listSize===e.listSize&&t.children.length===e.children.length&&wt.compareFields(t.children,e.children)}function _t(t,e){return t===e||st(t,e)&&t.keysSorted===e.keysSorted&&t.children.length===e.children.length&&wt.compareFields(t.children,e.children)}ot.prototype.visitNull=at,ot.prototype.visitBool=at,ot.prototype.visitInt=ct,ot.prototype.visitInt8=ct,ot.prototype.visitInt16=ct,ot.prototype.visitInt32=ct,ot.prototype.visitInt64=ct,ot.prototype.visitUint8=ct,ot.prototype.visitUint16=ct,ot.prototype.visitUint32=ct,ot.prototype.visitUint64=ct,ot.prototype.visitFloat=ut,ot.prototype.visitFloat16=ut,ot.prototype.visitFloat32=ut,ot.prototype.visitFloat64=ut,ot.prototype.visitUtf8=at,ot.prototype.visitBinary=at,ot.prototype.visitFixedSizeBinary=lt,ot.prototype.visitDate=ft,ot.prototype.visitDateDay=ft,ot.prototype.visitDateMillisecond=ft,ot.prototype.visitTimestamp=dt,ot.prototype.visitTimestampSecond=dt,ot.prototype.visitTimestampMillisecond=dt,ot.prototype.visitTimestampMicrosecond=dt,ot.prototype.visitTimestampNanosecond=dt,ot.prototype.visitTime=ht,ot.prototype.visitTimeSecond=ht,ot.prototype.visitTimeMillisecond=ht,ot.prototype.visitTimeMicrosecond=ht,ot.prototype.visitTimeNanosecond=ht,ot.prototype.visitDecimal=at,ot.prototype.visitList=pt,ot.prototype.visitStruct=yt,ot.prototype.visitUnion=vt,ot.prototype.visitDenseUnion=vt,ot.prototype.visitSparseUnion=vt,ot.prototype.visitDictionary=bt,ot.prototype.visitInterval=mt,ot.prototype.visitIntervalDayTime=mt,ot.prototype.visitIntervalYearMonth=mt,ot.prototype.visitFixedSizeList=gt,ot.prototype.visitMap=_t;const wt=new ot;class Ot{static isNull(t){return t&&t.typeId===F.Null}static isInt(t){return t&&t.typeId===F.Int}static isFloat(t){return t&&t.typeId===F.Float}static isBinary(t){return t&&t.typeId===F.Binary}static isUtf8(t){return t&&t.typeId===F.Utf8}static isBool(t){return t&&t.typeId===F.Bool}static isDecimal(t){return t&&t.typeId===F.Decimal}static isDate(t){return t&&t.typeId===F.Date}static isTime(t){return t&&t.typeId===F.Time}static isTimestamp(t){return t&&t.typeId===F.Timestamp}static isInterval(t){return t&&t.typeId===F.Interval}static isList(t){return t&&t.typeId===F.List}static isStruct(t){return t&&t.typeId===F.Struct}static isUnion(t){return t&&t.typeId===F.Union}static isFixedSizeBinary(t){return t&&t.typeId===F.FixedSizeBinary}static isFixedSizeList(t){return t&&t.typeId===F.FixedSizeList}static isMap(t){return t&&t.typeId===F.Map}static isDictionary(t){return t&&t.typeId===F.Dictionary}get typeId(){return F.NONE}compareTo(t){return wt.visit(this,t)}}Ot[Symbol.toStringTag]=(t=>(t.children=null,t.ArrayType=Array,t[Symbol.toStringTag]="DataType"))(Ot.prototype);class St extends Ot{toString(){return"Null"}get typeId(){return F.Null}}St[Symbol.toStringTag]=(t=>t[Symbol.toStringTag]="Null")(St.prototype);class xt extends Ot{constructor(t,e){super(),this.isSigned=t,this.bitWidth=e}get typeId(){return F.Int}get ArrayType(){switch(this.bitWidth){case 8:return this.isSigned?Int8Array:Uint8Array;case 16:return this.isSigned?Int16Array:Uint16Array;case 32:return this.isSigned?Int32Array:Uint32Array;case 64:return this.isSigned?Int32Array:Uint32Array}throw new Error(`Unrecognized ${this[Symbol.toStringTag]} type`)}toString(){return`${this.isSigned?"I":"Ui"}nt${this.bitWidth}`}}xt[Symbol.toStringTag]=(t=>(t.isSigned=null,t.bitWidth=null,t[Symbol.toStringTag]="Int"))(xt.prototype);class It extends xt{constructor(){super(!0,8)}}class At extends xt{constructor(){super(!0,16)}}class Tt extends xt{constructor(){super(!0,32)}}class jt extends xt{constructor(){super(!0,64)}}class Et extends xt{constructor(){super(!1,8)}}class Bt extends xt{constructor(){super(!1,16)}}class Ct extends xt{constructor(){super(!1,32)}}class Mt extends xt{constructor(){super(!1,64)}}Object.defineProperty(It.prototype,"ArrayType",{value:Int8Array}),Object.defineProperty(At.prototype,"ArrayType",{value:Int16Array}),Object.defineProperty(Tt.prototype,"ArrayType",{value:Int32Array}),Object.defineProperty(jt.prototype,"ArrayType",{value:Int32Array}),Object.defineProperty(Et.prototype,"ArrayType",{value:Uint8Array}),Object.defineProperty(Bt.prototype,"ArrayType",{value:Uint16Array}),Object.defineProperty(Ct.prototype,"ArrayType",{value:Uint32Array}),Object.defineProperty(Mt.prototype,"ArrayType",{value:Uint32Array});class kt extends Ot{constructor(t){super(),this.precision=t}get typeId(){return F.Float}get ArrayType(){switch(this.precision){case $.HALF:return Uint16Array;case $.SINGLE:return Float32Array;case $.DOUBLE:return Float64Array}throw new Error(`Unrecognized ${this[Symbol.toStringTag]} type`)}toString(){return"Float"+(this.precision<<5||16)}}kt[Symbol.toStringTag]=(t=>(t.precision=null,t[Symbol.toStringTag]="Float"))(kt.prototype);class Dt extends kt{constructor(){super($.HALF)}}class Nt extends kt{constructor(){super($.SINGLE)}}class Lt extends kt{constructor(){super($.DOUBLE)}}Object.defineProperty(Dt.prototype,"ArrayType",{value:Uint16Array}),Object.defineProperty(Nt.prototype,"ArrayType",{value:Float32Array}),Object.defineProperty(Lt.prototype,"ArrayType",{value:Float64Array});class Pt extends Ot{constructor(){super()}get typeId(){return F.Binary}toString(){return"Binary"}}Pt[Symbol.toStringTag]=(t=>(t.ArrayType=Uint8Array,t[Symbol.toStringTag]="Binary"))(Pt.prototype);class Ft extends Ot{constructor(){super()}get typeId(){return F.Utf8}toString(){return"Utf8"}}Ft[Symbol.toStringTag]=(t=>(t.ArrayType=Uint8Array,t[Symbol.toStringTag]="Utf8"))(Ft.prototype);class Ut extends Ot{constructor(){super()}get typeId(){return F.Bool}toString(){return"Bool"}}Ut[Symbol.toStringTag]=(t=>(t.ArrayType=Uint8Array,t[Symbol.toStringTag]="Bool"))(Ut.prototype);class Rt extends Ot{constructor(t,e){super(),this.scale=t,this.precision=e}get typeId(){return F.Decimal}toString(){return`Decimal[${this.precision}e${this.scale>0?"+":""}${this.scale}]`}}Rt[Symbol.toStringTag]=(t=>(t.scale=null,t.precision=null,t.ArrayType=Uint32Array,t[Symbol.toStringTag]="Decimal"))(Rt.prototype);class Vt extends Ot{constructor(t){super(),this.unit=t}get typeId(){return F.Date}toString(){return`Date${32*(this.unit+1)}<${R[this.unit]}>`}}Vt[Symbol.toStringTag]=(t=>(t.unit=null,t.ArrayType=Int32Array,t[Symbol.toStringTag]="Date"))(Vt.prototype);class $t extends Vt{constructor(){super(R.DAY)}}class zt extends Vt{constructor(){super(R.MILLISECOND)}}class Yt extends Ot{constructor(t,e){super(),this.unit=t,this.bitWidth=e}get typeId(){return F.Time}toString(){return`Time${this.bitWidth}<${V[this.unit]}>`}}Yt[Symbol.toStringTag]=(t=>(t.unit=null,t.bitWidth=null,t.ArrayType=Int32Array,t[Symbol.toStringTag]="Time"))(Yt.prototype);class Ht extends Ot{constructor(t,e){super(),this.unit=t,this.timezone=e}get typeId(){return F.Timestamp}toString(){return`Timestamp<${V[this.unit]}${this.timezone?", "+this.timezone:""}>`}}Ht[Symbol.toStringTag]=(t=>(t.unit=null,t.timezone=null,t.ArrayType=Int32Array,t[Symbol.toStringTag]="Timestamp"))(Ht.prototype);class Wt extends Ot{constructor(t){super(),this.unit=t}get typeId(){return F.Interval}toString(){return`Interval<${Y[this.unit]}>`}}Wt[Symbol.toStringTag]=(t=>(t.unit=null,t.ArrayType=Int32Array,t[Symbol.toStringTag]="Interval"))(Wt.prototype);class Gt extends Ot{constructor(t){super(),this.children=[t]}get typeId(){return F.List}toString(){return`List<${this.valueType}>`}get valueType(){return this.children[0].type}get valueField(){return this.children[0]}get ArrayType(){return this.valueType.ArrayType}}Gt[Symbol.toStringTag]=(t=>(t.children=null,t[Symbol.toStringTag]="List"))(Gt.prototype);class qt extends Ot{constructor(t){super(),this.children=t}get typeId(){return F.Struct}toString(){return`Struct<{${this.children.map(t=>`${t.name}:${t.type}`).join(", ")}}>`}}qt[Symbol.toStringTag]=(t=>(t.children=null,t[Symbol.toStringTag]="Struct"))(qt.prototype);class Kt extends Ot{constructor(t,e,n){super(),this.mode=t,this.children=n,this.typeIds=e=Int32Array.from(e),this.typeIdToChildIndex=e.reduce((t,e,n)=>(t[e]=n)&&t||t,Object.create(null))}get typeId(){return F.Union}toString(){return`${this[Symbol.toStringTag]}<${this.children.map(t=>""+t.type).join(" | ")}>`}}Kt[Symbol.toStringTag]=(t=>(t.mode=null,t.typeIds=null,t.children=null,t.typeIdToChildIndex=null,t.ArrayType=Int8Array,t[Symbol.toStringTag]="Union"))(Kt.prototype);class Xt extends Ot{constructor(t){super(),this.byteWidth=t}get typeId(){return F.FixedSizeBinary}toString(){return`FixedSizeBinary[${this.byteWidth}]`}}Xt[Symbol.toStringTag]=(t=>(t.byteWidth=null,t.ArrayType=Uint8Array,t[Symbol.toStringTag]="FixedSizeBinary"))(Xt.prototype);class Jt extends Ot{constructor(t,e){super(),this.listSize=t,this.children=[e]}get typeId(){return F.FixedSizeList}get valueType(){return this.children[0].type}get valueField(){return this.children[0]}get ArrayType(){return this.valueType.ArrayType}toString(){return`FixedSizeList[${this.listSize}]<${this.valueType}>`}}Jt[Symbol.toStringTag]=(t=>(t.children=null,t.listSize=null,t[Symbol.toStringTag]="FixedSizeList"))(Jt.prototype);class Zt extends Ot{constructor(t,e=!1){super(),this.children=[t],this.keysSorted=e}get typeId(){return F.Map}get keyType(){return this.children[0].type.children[0].type}get valueType(){return this.children[0].type.children[1].type}toString(){return`Map<{${this.children[0].type.children.map(t=>`${t.name}:${t.type}`).join(", ")}}>`}}Zt[Symbol.toStringTag]=(t=>(t.children=null,t.keysSorted=null,t[Symbol.toStringTag]="Map_"))(Zt.prototype);const Qt=(t=>()=>++t)(-1);class te extends Ot{constructor(t,e,n,r){super(),this.indices=e,this.dictionary=t,this.isOrdered=r||!1,this.id=null==n?Qt():"number"===typeof n?n:n.low}get typeId(){return F.Dictionary}get children(){return this.dictionary.children}get valueType(){return this.dictionary}get ArrayType(){return this.dictionary.ArrayType}toString(){return`Dictionary<${this.indices}, ${this.dictionary}>`}}function ee(t){let e=t;switch(t.typeId){case F.Decimal:return 4;case F.Timestamp:return 2;case F.Date:return 1+e.unit;case F.Interval:return 1+e.unit;case F.Int:return+(e.bitWidth>32)+1;case F.Time:return+(e.bitWidth>32)+1;case F.FixedSizeList:return e.listSize;case F.FixedSizeBinary:return e.byteWidth;default:return 1}}te[Symbol.toStringTag]=(t=>(t.id=null,t.indices=null,t.isOrdered=null,t.dictionary=null,t[Symbol.toStringTag]="Dictionary"))(te.prototype);var ne=n("3eea");const re=-1;class ie{constructor(t,e,n,r,i,o,s){let a;this.type=t,this.dictionary=s,this.offset=Math.floor(Math.max(e||0,0)),this.length=Math.floor(Math.max(n||0,0)),this._nullCount=Math.floor(Math.max(r||0,-1)),this.childData=(o||[]).map(t=>t instanceof ie?t:t.data),i instanceof ie?(this.stride=i.stride,this.values=i.values,this.typeIds=i.typeIds,this.nullBitmap=i.nullBitmap,this.valueOffsets=i.valueOffsets):(this.stride=ee(t),i&&((a=i[0])&&(this.valueOffsets=a),(a=i[1])&&(this.values=a),(a=i[2])&&(this.nullBitmap=a),(a=i[3])&&(this.typeIds=a)))}get typeId(){return this.type.typeId}get ArrayType(){return this.type.ArrayType}get buffers(){return[this.valueOffsets,this.values,this.nullBitmap,this.typeIds]}get byteLength(){let t=0,{valueOffsets:e,values:n,nullBitmap:r,typeIds:i}=this;return e&&(t+=e.byteLength),n&&(t+=n.byteLength),r&&(t+=r.byteLength),i&&(t+=i.byteLength),this.childData.reduce((t,e)=>t+e.byteLength,t)}get nullCount(){let t,e=this._nullCount;return e<=re&&(t=this.nullBitmap)&&(this._nullCount=e=this.length-Q(t,this.offset,this.offset+this.length)),e}clone(t,e=this.offset,n=this.length,r=this._nullCount,i=this,o=this.childData){return new ie(t,e,n,r,i,o,this.dictionary)}slice(t,e){const{stride:n,typeId:r,childData:i}=this,o=+(0===this._nullCount)-1,s=16===r?n:1,a=this._sliceBuffers(t,e,n,r);return this.clone(this.type,this.offset+t,e,o,a,!i.length||this.valueOffsets?i:this._sliceChildren(i,s*t,s*e))}_changeLengthAndBackfillNullBitmap(t){if(this.typeId===F.Null)return this.clone(this.type,0,t,0);const{length:e,nullCount:n}=this,r=new Uint8Array((t+63&-64)>>3).fill(255,0,e>>3);r[e>>3]=(1<<e-(-8&e))-1,n>0&&r.set(X(this.offset,e,this.nullBitmap),0);const i=this.buffers;return i[U.VALIDITY]=r,this.clone(this.type,0,t,n+(t-e),i)}_sliceBuffers(t,e,n,r){let i,{buffers:o}=this;return(i=o[U.TYPE])&&(o[U.TYPE]=i.subarray(t,t+e)),(i=o[U.OFFSET])&&(o[U.OFFSET]=i.subarray(t,t+e+1))||(i=o[U.DATA])&&(o[U.DATA]=6===r?i:i.subarray(n*t,n*(t+e))),o}_sliceChildren(t,e,n){return t.map(t=>t.slice(e,n))}static new(t,e,n,r,i,o,s){switch(i instanceof ie?i=i.buffers:i||(i=[]),t.typeId){case F.Null:return ie.Null(t,e,n);case F.Int:return ie.Int(t,e,n,r||0,i[U.VALIDITY],i[U.DATA]||[]);case F.Dictionary:return ie.Dictionary(t,e,n,r||0,i[U.VALIDITY],i[U.DATA]||[],s);case F.Float:return ie.Float(t,e,n,r||0,i[U.VALIDITY],i[U.DATA]||[]);case F.Bool:return ie.Bool(t,e,n,r||0,i[U.VALIDITY],i[U.DATA]||[]);case F.Decimal:return ie.Decimal(t,e,n,r||0,i[U.VALIDITY],i[U.DATA]||[]);case F.Date:return ie.Date(t,e,n,r||0,i[U.VALIDITY],i[U.DATA]||[]);case F.Time:return ie.Time(t,e,n,r||0,i[U.VALIDITY],i[U.DATA]||[]);case F.Timestamp:return ie.Timestamp(t,e,n,r||0,i[U.VALIDITY],i[U.DATA]||[]);case F.Interval:return ie.Interval(t,e,n,r||0,i[U.VALIDITY],i[U.DATA]||[]);case F.FixedSizeBinary:return ie.FixedSizeBinary(t,e,n,r||0,i[U.VALIDITY],i[U.DATA]||[]);case F.Binary:return ie.Binary(t,e,n,r||0,i[U.VALIDITY],i[U.OFFSET]||[],i[U.DATA]||[]);case F.Utf8:return ie.Utf8(t,e,n,r||0,i[U.VALIDITY],i[U.OFFSET]||[],i[U.DATA]||[]);case F.List:return ie.List(t,e,n,r||0,i[U.VALIDITY],i[U.OFFSET]||[],(o||[])[0]);case F.FixedSizeList:return ie.FixedSizeList(t,e,n,r||0,i[U.VALIDITY],(o||[])[0]);case F.Struct:return ie.Struct(t,e,n,r||0,i[U.VALIDITY],o||[]);case F.Map:return ie.Map(t,e,n,r||0,i[U.VALIDITY],i[U.OFFSET]||[],(o||[])[0]);case F.Union:return ie.Union(t,e,n,r||0,i[U.VALIDITY],i[U.TYPE]||[],i[U.OFFSET]||o,o)}throw new Error("Unrecognized typeId "+t.typeId)}static Null(t,e,n){return new ie(t,e,n,0)}static Int(t,e,n,r,i,o){return new ie(t,e,n,r,[void 0,Object(ne["toArrayBufferView"])(t.ArrayType,o),Object(ne["toUint8Array"])(i)])}static Dictionary(t,e,n,r,i,o,s){return new ie(t,e,n,r,[void 0,Object(ne["toArrayBufferView"])(t.indices.ArrayType,o),Object(ne["toUint8Array"])(i)],[],s)}static Float(t,e,n,r,i,o){return new ie(t,e,n,r,[void 0,Object(ne["toArrayBufferView"])(t.ArrayType,o),Object(ne["toUint8Array"])(i)])}static Bool(t,e,n,r,i,o){return new ie(t,e,n,r,[void 0,Object(ne["toArrayBufferView"])(t.ArrayType,o),Object(ne["toUint8Array"])(i)])}static Decimal(t,e,n,r,i,o){return new ie(t,e,n,r,[void 0,Object(ne["toArrayBufferView"])(t.ArrayType,o),Object(ne["toUint8Array"])(i)])}static Date(t,e,n,r,i,o){return new ie(t,e,n,r,[void 0,Object(ne["toArrayBufferView"])(t.ArrayType,o),Object(ne["toUint8Array"])(i)])}static Time(t,e,n,r,i,o){return new ie(t,e,n,r,[void 0,Object(ne["toArrayBufferView"])(t.ArrayType,o),Object(ne["toUint8Array"])(i)])}static Timestamp(t,e,n,r,i,o){return new ie(t,e,n,r,[void 0,Object(ne["toArrayBufferView"])(t.ArrayType,o),Object(ne["toUint8Array"])(i)])}static Interval(t,e,n,r,i,o){return new ie(t,e,n,r,[void 0,Object(ne["toArrayBufferView"])(t.ArrayType,o),Object(ne["toUint8Array"])(i)])}static FixedSizeBinary(t,e,n,r,i,o){return new ie(t,e,n,r,[void 0,Object(ne["toArrayBufferView"])(t.ArrayType,o),Object(ne["toUint8Array"])(i)])}static Binary(t,e,n,r,i,o,s){return new ie(t,e,n,r,[Object(ne["toInt32Array"])(o),Object(ne["toUint8Array"])(s),Object(ne["toUint8Array"])(i)])}static Utf8(t,e,n,r,i,o,s){return new ie(t,e,n,r,[Object(ne["toInt32Array"])(o),Object(ne["toUint8Array"])(s),Object(ne["toUint8Array"])(i)])}static List(t,e,n,r,i,o,s){return new ie(t,e,n,r,[Object(ne["toInt32Array"])(o),void 0,Object(ne["toUint8Array"])(i)],[s])}static FixedSizeList(t,e,n,r,i,o){return new ie(t,e,n,r,[void 0,void 0,Object(ne["toUint8Array"])(i)],[o])}static Struct(t,e,n,r,i,o){return new ie(t,e,n,r,[void 0,void 0,Object(ne["toUint8Array"])(i)],o)}static Map(t,e,n,r,i,o,s){return new ie(t,e,n,r,[Object(ne["toInt32Array"])(o),void 0,Object(ne["toUint8Array"])(i)],[s])}static Union(t,e,n,r,i,o,s,a){const c=[void 0,void 0,Object(ne["toUint8Array"])(i),Object(ne["toArrayBufferView"])(t.ArrayType,o)];return t.mode===z.Sparse?new ie(t,e,n,r,c,s):(c[U.OFFSET]=Object(ne["toInt32Array"])(s),new ie(t,e,n,r,c,a))}}ie.prototype.childData=Object.freeze([]);const oe=void 0;function se(t){if(null===t)return"null";if(t===oe)return"undefined";switch(typeof t){case"number":return""+t;case"bigint":return""+t;case"string":return`"${t}"`}return"function"===typeof t[Symbol.toPrimitive]?t[Symbol.toPrimitive]("string"):ArrayBuffer.isView(t)?`[${t}]`:JSON.stringify(t)}var ae=n("841f");function ce(t){if(!t||t.length<=0)return function(t){return!0};let e="",n=t.filter(t=>t===t);return n.length>0&&(e=`\n    switch (x) {${n.map(t=>`\n        case ${ue(t)}:`).join("")}\n            return false;\n    }`),t.length!==n.length&&(e="if (x !== x) return false;\n"+e),new Function("x",e+"\nreturn true;")}function ue(t){return"bigint"!==typeof t?se(t):ae["c"]?se(t)+"n":`"${se(t)}"`}const le=(t,e)=>(t*e+63&-64||64)/e,fe=(t,e=0)=>t.length>=e?t.subarray(0,e):Object(ne["memcpy"])(new t.constructor(e),t,0);class de{constructor(t,e=1){this.buffer=t,this.stride=e,this.BYTES_PER_ELEMENT=t.BYTES_PER_ELEMENT,this.ArrayType=t.constructor,this._resize(this.length=t.length/e|0)}get byteLength(){return this.length*this.stride*this.BYTES_PER_ELEMENT|0}get reservedLength(){return this.buffer.length/this.stride}get reservedByteLength(){return this.buffer.byteLength}set(t,e){return this}append(t){return this.set(this.length,t)}reserve(t){if(t>0){this.length+=t;const e=this.stride,n=this.length*e,r=this.buffer.length;n>=r&&this._resize(le(0===r?1*n:2*n,this.BYTES_PER_ELEMENT))}return this}flush(t=this.length){t=le(t*this.stride,this.BYTES_PER_ELEMENT);const e=fe(this.buffer,t);return this.clear(),e}clear(){return this.length=0,this._resize(0),this}_resize(t){return this.buffer=Object(ne["memcpy"])(new this.ArrayType(t),this.buffer)}}de.prototype.offset=0;class he extends de{last(){return this.get(this.length-1)}get(t){return this.buffer[t]}set(t,e){return this.reserve(t-this.length+1),this.buffer[t*this.stride]=e,this}}class pe extends he{constructor(t=new Uint8Array(0)){super(t,1/8),this.numValid=0}get numInvalid(){return this.length-this.numValid}get(t){return this.buffer[t>>3]>>t%8&1}set(t,e){const{buffer:n}=this.reserve(t-this.length+1),r=t>>3,i=t%8,o=n[r]>>i&1;return e?0===o&&(n[r]|=1<<i,++this.numValid):1===o&&(n[r]&=~(1<<i),--this.numValid),this}clear(){return this.numValid=0,super.clear()}}class ye extends he{constructor(t=new Int32Array(1)){super(t,1)}append(t){return this.set(this.length-1,t)}set(t,e){const n=this.length-1,r=this.reserve(t-n+1).buffer;return n<t++&&r.fill(r[n],n,t),r[t]=r[t-1]+e,this}flush(t=this.length-1){return t>this.length&&this.set(t-1,0),super.flush(t+1)}}class ve extends de{get ArrayType64(){return this._ArrayType64||(this._ArrayType64=this.buffer instanceof Int32Array?ae["b"]:ae["d"])}set(t,e){switch(this.reserve(t-this.length+1),typeof e){case"bigint":this.buffer64[t]=e;break;case"number":this.buffer[t*this.stride]=e;break;default:this.buffer.set(e,t*this.stride)}return this}_resize(t){const e=super._resize(t),n=e.byteLength/(this.BYTES_PER_ELEMENT*this.stride);return ae["c"]&&(this.buffer64=new this.ArrayType64(e.buffer,e.byteOffset,n)),e}}class be{constructor({type:t,nullValues:e}){this.length=0,this.finished=!1,this.type=t,this.children=[],this.nullValues=e,this.stride=ee(t),this._nulls=new pe,e&&e.length>0&&(this._isValid=ce(e))}static new(t){}static throughNode(t){throw new Error('"throughNode" not available in this environment')}static throughDOM(t){throw new Error('"throughDOM" not available in this environment')}static throughIterable(t){return _e(t)}static throughAsyncIterable(t){return we(t)}toVector(){return P.new(this.flush())}get ArrayType(){return this.type.ArrayType}get nullCount(){return this._nulls.numInvalid}get numChildren(){return this.children.length}get byteLength(){let t=0;return this._offsets&&(t+=this._offsets.byteLength),this._values&&(t+=this._values.byteLength),this._nulls&&(t+=this._nulls.byteLength),this._typeIds&&(t+=this._typeIds.byteLength),this.children.reduce((t,e)=>t+e.byteLength,t)}get reservedLength(){return this._nulls.reservedLength}get reservedByteLength(){let t=0;return this._offsets&&(t+=this._offsets.reservedByteLength),this._values&&(t+=this._values.reservedByteLength),this._nulls&&(t+=this._nulls.reservedByteLength),this._typeIds&&(t+=this._typeIds.reservedByteLength),this.children.reduce((t,e)=>t+e.reservedByteLength,t)}get valueOffsets(){return this._offsets?this._offsets.buffer:null}get values(){return this._values?this._values.buffer:null}get nullBitmap(){return this._nulls?this._nulls.buffer:null}get typeIds(){return this._typeIds?this._typeIds.buffer:null}append(t){return this.set(this.length,t)}isValid(t){return this._isValid(t)}set(t,e){return this.setValid(t,this.isValid(e))&&this.setValue(t,e),this}setValue(t,e){this._setValue(this,t,e)}setValid(t,e){return this.length=this._nulls.set(t,+e).length,e}addChild(t,e=""+this.numChildren){throw new Error(`Cannot append children to non-nested type "${this.type}"`)}getChildAt(t){return this.children[t]||null}flush(){const t=[],e=this._values,n=this._offsets,r=this._typeIds,{length:i,nullCount:o}=this;r?(t[U.TYPE]=r.flush(i),n&&(t[U.OFFSET]=n.flush(i))):n?(e&&(t[U.DATA]=e.flush(n.last())),t[U.OFFSET]=n.flush(i)):e&&(t[U.DATA]=e.flush(i)),o>0&&(t[U.VALIDITY]=this._nulls.flush(i));const s=ie.new(this.type,0,i,o,t,this.children.map(t=>t.flush()));return this.clear(),s}finish(){return this.finished=!0,this.children.forEach(t=>t.finish()),this}clear(){return this.length=0,this._offsets&&this._offsets.clear(),this._values&&this._values.clear(),this._nulls&&this._nulls.clear(),this._typeIds&&this._typeIds.clear(),this.children.forEach(t=>t.clear()),this}}be.prototype.length=1,be.prototype.stride=1,be.prototype.children=null,be.prototype.finished=!1,be.prototype.nullValues=null,be.prototype._isValid=()=>!0;class me extends be{constructor(t){super(t),this._values=new he(new this.ArrayType(0),this.stride)}setValue(t,e){const n=this._values;return n.reserve(t-n.length+1),super.setValue(t,e)}}class ge extends be{constructor(t){super(t),this._pendingLength=0,this._offsets=new ye}setValue(t,e){const n=this._pending||(this._pending=new Map),r=n.get(t);r&&(this._pendingLength-=r.length),this._pendingLength+=e.length,n.set(t,e)}setValid(t,e){return!!super.setValid(t,e)||((this._pending||(this._pending=new Map)).set(t,void 0),!1)}clear(){return this._pendingLength=0,this._pending=void 0,super.clear()}flush(){return this._flush(),super.flush()}finish(){return this._flush(),super.finish()}_flush(){const t=this._pending,e=this._pendingLength;return this._pendingLength=0,this._pending=void 0,t&&t.size>0&&this._flushPending(t,e),this}}function _e(t){const{["queueingStrategy"]:e="count"}=t,{["highWaterMark"]:n=("bytes"!==e?1e3:16384)}=t,r="bytes"!==e?"length":"byteLength";return function*(e){let i=0,o=be.new(t);for(const t of e)o.append(t)[r]>=n&&++i&&(yield o.toVector());(o.finish().length>0||0===i)&&(yield o.toVector())}}function we(t){const{["queueingStrategy"]:e="count"}=t,{["highWaterMark"]:n=("bytes"!==e?1e3:16384)}=t,r="bytes"!==e?"length":"byteLength";return async function*(e){let i=0,o=be.new(t);for await(const t of e)o.append(t)[r]>=n&&++i&&(yield o.toVector());(o.finish().length>0||0===i)&&(yield o.toVector())}}class Oe extends be{constructor(t){super(t),this._values=new pe}setValue(t,e){this._values.set(t,+e)}}class Se extends be{setValue(t,e){}setValid(t,e){return this.length=Math.max(t+1,this.length),e}}class xe extends me{}class Ie extends xe{}class Ae extends xe{}class Te extends me{}class je extends be{constructor({type:t,nullValues:e,dictionaryHashFunction:n}){super({type:new te(t.dictionary,t.indices,t.id,t.isOrdered)}),this._nulls=null,this._dictionaryOffset=0,this._keysToIndices=Object.create(null),this.indices=be.new({type:this.type.indices,nullValues:e}),this.dictionary=be.new({type:this.type.dictionary,nullValues:null}),"function"===typeof n&&(this.valueToKey=n)}get values(){return this.indices.values}get nullCount(){return this.indices.nullCount}get nullBitmap(){return this.indices.nullBitmap}get byteLength(){return this.indices.byteLength+this.dictionary.byteLength}get reservedLength(){return this.indices.reservedLength+this.dictionary.reservedLength}get reservedByteLength(){return this.indices.reservedByteLength+this.dictionary.reservedByteLength}isValid(t){return this.indices.isValid(t)}setValid(t,e){const n=this.indices;return e=n.setValid(t,e),this.length=n.length,e}setValue(t,e){let n=this._keysToIndices,r=this.valueToKey(e),i=n[r];return void 0===i&&(n[r]=i=this._dictionaryOffset+this.dictionary.append(e).length-1),this.indices.setValue(t,i)}flush(){const t=this.type,e=this._dictionary,n=this.dictionary.toVector(),r=this.indices.flush().clone(t);return r.dictionary=e?e.concat(n):n,this.finished||(this._dictionaryOffset+=n.length),this._dictionary=r.dictionary,this.clear(),r}finish(){return this.indices.finish(),this.dictionary.finish(),this._dictionaryOffset=0,this._keysToIndices=Object.create(null),super.finish()}clear(){return this.indices.clear(),this.dictionary.clear(),super.clear()}valueToKey(t){return"string"===typeof t?t:""+t}}class Ee extends me{}const Be=new Float64Array(1),Ce=new Uint32Array(Be.buffer);function Me(t){let e=(31744&t)>>10,n=(1023&t)/1024,r=(-1)**((32768&t)>>15);switch(e){case 31:return r*(n?NaN:1/0);case 0:return r*(n?6103515625e-14*n:0)}return r*2**(e-15)*(1+n)}function ke(t){if(t!==t)return 32256;Be[0]=t;let e=(2147483648&Ce[1])>>16&65535,n=2146435072&Ce[1],r=0;return n>=1089470464?Ce[0]>0?n=31744:(n=(2080374784&n)>>16,r=(1048575&Ce[1])>>10):n<=1056964608?(r=1048576+(1048575&Ce[1]),r=1048576+(r<<(n>>20)-998)>>21,n=0):(n=n-1056964608>>10,r=512+(1048575&Ce[1])>>10),e|n|65535&r}class De extends me{}class Ne extends De{setValue(t,e){this._values.set(t,ke(e))}}class Le extends De{setValue(t,e){this._values.set(t,e)}}class Pe extends De{setValue(t,e){this._values.set(t,e)}}const Fe=Symbol.for("isArrowBigNum");function Ue(t,...e){return 0===e.length?Object.setPrototypeOf(Object(ne["toArrayBufferView"])(this["TypedArray"],t),this.constructor.prototype):Object.setPrototypeOf(new this["TypedArray"](t,...e),this.constructor.prototype)}function Re(...t){return Ue.apply(this,t)}function Ve(...t){return Ue.apply(this,t)}function $e(...t){return Ue.apply(this,t)}function ze(t){let e,n,{buffer:r,byteOffset:i,length:o,signed:s}=t,a=new Int32Array(r,i,o),c=0,u=0,l=a.length;while(u<l)n=a[u++],e=a[u++],s||(e>>>=0),c+=(n>>>0)+e*u**32;return c}let Ye,He;function We(t){let e="",n=new Uint32Array(2),r=new Uint16Array(t.buffer,t.byteOffset,t.byteLength/2),i=new Uint32Array((r=new Uint16Array(r).reverse()).buffer),o=-1,s=r.length-1;do{for(n[0]=r[o=0];o<s;)r[o++]=n[1]=n[0]/10,n[0]=(n[0]-10*n[1]<<16)+r[o];r[o]=n[1]=n[0]/10,n[0]=n[0]-10*n[1],e=`${n[0]}${e}`}while(i[0]||i[1]||i[2]||i[3]);return e||"0"}Ue.prototype[Fe]=!0,Ue.prototype.toJSON=function(){return`"${Ye(this)}"`},Ue.prototype.valueOf=function(){return ze(this)},Ue.prototype.toString=function(){return Ye(this)},Ue.prototype[Symbol.toPrimitive]=function(t="default"){switch(t){case"number":return ze(this);case"string":return Ye(this);case"default":return He(this)}return Ye(this)},Object.setPrototypeOf(Re.prototype,Object.create(Int32Array.prototype)),Object.setPrototypeOf(Ve.prototype,Object.create(Uint32Array.prototype)),Object.setPrototypeOf($e.prototype,Object.create(Uint32Array.prototype)),Object.assign(Re.prototype,Ue.prototype,{constructor:Re,signed:!0,TypedArray:Int32Array,BigIntArray:ae["b"]}),Object.assign(Ve.prototype,Ue.prototype,{constructor:Ve,signed:!1,TypedArray:Uint32Array,BigIntArray:ae["d"]}),Object.assign($e.prototype,Ue.prototype,{constructor:$e,signed:!0,TypedArray:Uint32Array,BigIntArray:ae["d"]}),ae["c"]?(He=t=>8===t.byteLength?new t["BigIntArray"](t.buffer,t.byteOffset,1)[0]:We(t),Ye=t=>8===t.byteLength?""+new t["BigIntArray"](t.buffer,t.byteOffset,1)[0]:We(t)):(Ye=We,He=Ye);class Ge{constructor(t,e){return Ge.new(t,e)}static new(t,e){switch(e){case!0:return new Re(t);case!1:return new Ve(t)}switch(t.constructor){case Int8Array:case Int16Array:case Int32Array:case ae["b"]:return new Re(t)}return 16===t.byteLength?new $e(t):new Ve(t)}static signed(t){return new Re(t)}static unsigned(t){return new Ve(t)}static decimal(t){return new $e(t)}}class qe extends me{setValue(t,e){this._values.set(t,e)}}class Ke extends qe{}class Xe extends qe{}class Je extends qe{}class Ze extends qe{constructor(t){t["nullValues"]&&(t["nullValues"]=t["nullValues"].map(rn)),super(t),this._values=new ve(new Int32Array(0),2)}get values64(){return this._values.buffer64}isValid(t){return super.isValid(rn(t))}}class Qe extends qe{}class tn extends qe{}class en extends qe{}class nn extends qe{constructor(t){t["nullValues"]&&(t["nullValues"]=t["nullValues"].map(rn)),super(t),this._values=new ve(new Uint32Array(0),2)}get values64(){return this._values.buffer64}isValid(t){return super.isValid(rn(t))}}const rn=(t=>e=>(ArrayBuffer.isView(e)&&(t.buffer=e.buffer,t.byteOffset=e.byteOffset,t.byteLength=e.byteLength,e=He(t),t.buffer=null),e))({BigIntArray:ae["b"]});class on extends me{}class sn extends on{}class an extends on{}class cn extends on{}class un extends on{}class ln extends me{}class fn extends ln{}class dn extends ln{}class hn extends ln{}class pn extends ln{}class yn extends me{}class vn extends yn{}class bn extends yn{}var mn=n("5217");class gn extends ge{constructor(t){super(t),this._values=new de(new Uint8Array(0))}get byteLength(){let t=this._pendingLength+4*this.length;return this._offsets&&(t+=this._offsets.byteLength),this._values&&(t+=this._values.byteLength),this._nulls&&(t+=this._nulls.byteLength),t}setValue(t,e){return super.setValue(t,Object(ne["toUint8Array"])(e))}_flushPending(t,e){const n=this._offsets,r=this._values.reserve(e).buffer;let i,o=0,s=0,a=0;for([o,i]of t)void 0===i?n.set(o,0):(s=i.length,r.set(i,a),n.set(o,s),a+=s)}}class _n extends ge{constructor(t){super(t),this._values=new de(new Uint8Array(0))}get byteLength(){let t=this._pendingLength+4*this.length;return this._offsets&&(t+=this._offsets.byteLength),this._values&&(t+=this._values.byteLength),this._nulls&&(t+=this._nulls.byteLength),t}setValue(t,e){return super.setValue(t,Object(mn["b"])(e))}_flushPending(t,e){}}_n.prototype._flushPending=gn.prototype._flushPending;class wn{get length(){return this._values.length}get(t){return this._values[t]}clear(){return this._values=null,this}bind(t){return t instanceof P?t:(this._values=t,this)}}const On=Symbol.for("parent"),Sn=Symbol.for("rowIndex"),xn=Symbol.for("keyToIdx"),In=Symbol.for("idxToVal"),An=Symbol.for("nodejs.util.inspect.custom");class Tn{constructor(t,e){this[On]=t,this.size=e}entries(){return this[Symbol.iterator]()}has(t){return void 0!==this.get(t)}get(t){let e=void 0;if(null!==t&&void 0!==t){const n=this[xn]||(this[xn]=new Map);let r=n.get(t);if(void 0!==r){const t=this[In]||(this[In]=new Array(this.size));void 0!==(e=t[r])||(t[r]=e=this.getValue(r))}else if((r=this.getIndex(t))>-1){n.set(t,r);const i=this[In]||(this[In]=new Array(this.size));void 0!==(e=i[r])||(i[r]=e=this.getValue(r))}}return e}set(t,e){if(null!==t&&void 0!==t){const n=this[xn]||(this[xn]=new Map);let r=n.get(t);if(void 0===r&&n.set(t,r=this.getIndex(t)),r>-1){const t=this[In]||(this[In]=new Array(this.size));t[r]=this.setValue(r,e)}}return this}clear(){throw new Error(`Clearing ${this[Symbol.toStringTag]} not supported.`)}delete(t){throw new Error(`Deleting ${this[Symbol.toStringTag]} values not supported.`)}*[Symbol.iterator](){const t=this.keys(),e=this.values(),n=this[xn]||(this[xn]=new Map),r=this[In]||(this[In]=new Array(this.size));for(let i,o,s,a,c=0;!(s=t.next()).done&&!(a=e.next()).done;++c)i=s.value,o=a.value,r[c]=o,n.has(i)||n.set(i,c),yield[i,o]}forEach(t,e){const n=this.keys(),r=this.values(),i=void 0===e?t:(n,r,i)=>t.call(e,n,r,i),o=this[xn]||(this[xn]=new Map),s=this[In]||(this[In]=new Array(this.size));for(let a,c,u,l,f=0;!(u=n.next()).done&&!(l=r.next()).done;++f)a=u.value,c=l.value,s[f]=c,o.has(a)||o.set(a,f),i(c,a,this)}toArray(){return[...this.values()]}toJSON(){const t={};return this.forEach((e,n)=>t[n]=e),t}inspect(){return this.toString()}[An](){return this.toString()}toString(){const t=[];return this.forEach((e,n)=>{n=se(n),e=se(e),t.push(`${n}: ${e}`)}),`{ ${t.join(", ")} }`}}Tn[Symbol.toStringTag]=(t=>(Object.defineProperties(t,{size:{writable:!0,enumerable:!1,configurable:!1,value:0},[On]:{writable:!0,enumerable:!1,configurable:!1,value:null},[Sn]:{writable:!0,enumerable:!1,configurable:!1,value:-1}}),t[Symbol.toStringTag]="Row"))(Tn.prototype);class jn extends Tn{constructor(t){return super(t,t.length),Cn(this)}keys(){return this[On].getChildAt(0)[Symbol.iterator]()}values(){return this[On].getChildAt(1)[Symbol.iterator]()}getKey(t){return this[On].getChildAt(0).get(t)}getIndex(t){return this[On].getChildAt(0).indexOf(t)}getValue(t){return this[On].getChildAt(1).get(t)}setValue(t,e){this[On].getChildAt(1).set(t,e)}}class En extends Tn{constructor(t){return super(t,t.type.children.length),Bn(this)}*keys(){for(const t of this[On].type.children)yield t.name}*values(){for(const t of this[On].type.children)yield this[t.name]}getKey(t){return this[On].type.children[t].name}getIndex(t){return this[On].type.children.findIndex(e=>e.name===t)}getValue(t){return this[On].getChildAt(t).get(this[Sn])}setValue(t,e){return this[On].getChildAt(t).set(this[Sn],e)}}Object.setPrototypeOf(Tn.prototype,Map.prototype);const Bn=(()=>{const t={enumerable:!0,configurable:!1,get:null,set:null};return e=>{let n=-1,r=e[xn]||(e[xn]=new Map);const i=t=>function(){return this.get(t)},o=t=>function(e){return this.set(t,e)};for(const s of e.keys())r.set(s,++n),t.get=i(s),t.set=o(s),e.hasOwnProperty(s)||(t.enumerable=!0,Object.defineProperty(e,s,t)),e.hasOwnProperty(n)||(t.enumerable=!1,Object.defineProperty(e,n,t));return t.get=t.set=null,e}})(),Cn=(()=>{if("undefined"===typeof Proxy)return Bn;const t=Tn.prototype.has,e=Tn.prototype.get,n=Tn.prototype.set,r=Tn.prototype.getKey,i={isExtensible(){return!1},deleteProperty(){return!1},preventExtensions(){return!0},ownKeys(t){return[...t.keys()].map(t=>""+t)},has(t,e){switch(e){case"getKey":case"getIndex":case"getValue":case"setValue":case"toArray":case"toJSON":case"inspect":case"constructor":case"isPrototypeOf":case"propertyIsEnumerable":case"toString":case"toLocaleString":case"valueOf":case"size":case"has":case"get":case"set":case"clear":case"delete":case"keys":case"values":case"entries":case"forEach":case"__proto__":case"__defineGetter__":case"__defineSetter__":case"hasOwnProperty":case"__lookupGetter__":case"__lookupSetter__":case Symbol.iterator:case Symbol.toStringTag:case On:case Sn:case In:case xn:case An:return!0}return"number"!==typeof e||t.has(e)||(e=t.getKey(e)),t.has(e)},get(n,i,o){switch(i){case"getKey":case"getIndex":case"getValue":case"setValue":case"toArray":case"toJSON":case"inspect":case"constructor":case"isPrototypeOf":case"propertyIsEnumerable":case"toString":case"toLocaleString":case"valueOf":case"size":case"has":case"get":case"set":case"clear":case"delete":case"keys":case"values":case"entries":case"forEach":case"__proto__":case"__defineGetter__":case"__defineSetter__":case"hasOwnProperty":case"__lookupGetter__":case"__lookupSetter__":case Symbol.iterator:case Symbol.toStringTag:case On:case Sn:case In:case xn:case An:return Reflect.get(n,i,o)}return"number"!==typeof i||t.call(o,i)||(i=r.call(o,i)),e.call(o,i)},set(e,i,o,s){switch(i){case On:case Sn:case In:case xn:return Reflect.set(e,i,o,s);case"getKey":case"getIndex":case"getValue":case"setValue":case"toArray":case"toJSON":case"inspect":case"constructor":case"isPrototypeOf":case"propertyIsEnumerable":case"toString":case"toLocaleString":case"valueOf":case"size":case"has":case"get":case"set":case"clear":case"delete":case"keys":case"values":case"entries":case"forEach":case"__proto__":case"__defineGetter__":case"__defineSetter__":case"hasOwnProperty":case"__lookupGetter__":case"__lookupSetter__":case Symbol.iterator:case Symbol.toStringTag:return!1}return"number"!==typeof i||t.call(s,i)||(i=r.call(s,i)),!!t.call(s,i)&&!!n.call(s,i,o)}};return t=>new Proxy(t,i)})();function Mn(t,e,n){const r=t.length,i=e>-1?e:r+e%r;return n?n(t,i):i}let kn;function Dn(t,e,n,r){let{length:i=0}=t,o="number"!==typeof e?0:e,s="number"!==typeof n?i:n;return o<0&&(o=(o%i+i)%i),s<0&&(s=(s%i+i)%i),s<o&&(kn=o,o=s,s=kn),s>i&&(s=i),r?r(t,o,s):[o,s]}const Nn=ae["c"]?Object(ae["a"])(0):0,Ln=t=>t!==t;function Pn(t){let e=typeof t;if("object"!==e||null===t)return Ln(t)?Ln:"bigint"!==e?e=>e===t:e=>Nn+e===t;if(t instanceof Date){const e=t.valueOf();return t=>t instanceof Date&&t.valueOf()===e}return ArrayBuffer.isView(t)?e=>!!e&&Object(ne["compareArrayLike"])(t,e):t instanceof Map?Un(t):Array.isArray(t)?Fn(t):t instanceof P?Rn(t):Vn(t)}function Fn(t){const e=[];for(let n=-1,r=t.length;++n<r;)e[n]=Pn(t[n]);return $n(e)}function Un(t){let e=-1;const n=[];return t.forEach(t=>n[++e]=Pn(t)),$n(n)}function Rn(t){const e=[];for(let n=-1,r=t.length;++n<r;)e[n]=Pn(t.get(n));return $n(e)}function Vn(t){const e=Object.keys(t);if(0===e.length)return()=>!1;const n=[];for(let r=-1,i=e.length;++r<i;)n[r]=Pn(t[e[r]]);return $n(n,e)}function $n(t,e){return n=>{if(!n||"object"!==typeof n)return!1;switch(n.constructor){case Array:return zn(t,n);case Map:case jn:case En:return Hn(t,n,n.keys());case Object:case void 0:return Hn(t,n,e||Object.keys(n))}return n instanceof P&&Yn(t,n)}}function zn(t,e){const n=t.length;if(e.length!==n)return!1;for(let r=-1;++r<n;)if(!t[r](e[r]))return!1;return!0}function Yn(t,e){const n=t.length;if(e.length!==n)return!1;for(let r=-1;++r<n;)if(!t[r](e.get(r)))return!1;return!0}function Hn(t,e,n){const r=n[Symbol.iterator](),i=e instanceof Map?e.keys():Object.keys(e)[Symbol.iterator](),o=e instanceof Map?e.values():Object.values(e)[Symbol.iterator]();let s=0,a=t.length,c=o.next(),u=r.next(),l=i.next();for(;s<a&&!u.done&&!l.done&&!c.done;++s,u=r.next(),l=i.next(),c=o.next())if(u.value!==l.value||!t[s](c.value))break;return!!(s===a&&u.done&&l.done&&c.done)||(r.return&&r.return(),i.return&&i.return(),o.return&&o.return(),!1)}class Wn extends P{constructor(t,e=[],n=Gn(e)){super(),this._nullCount=-1,this._type=t,this._chunks=e,this._chunkOffsets=n,this._length=n[n.length-1],this._numChildren=(this._type.children||[]).length}static flatten(...t){return nr(P,t)}static concat(...t){const e=Wn.flatten(...t);return new Wn(e[0].type,e)}get type(){return this._type}get length(){return this._length}get chunks(){return this._chunks}get typeId(){return this._type.typeId}get VectorName(){return`Chunked<${this._type}>`}get data(){return this._chunks[0]?this._chunks[0].data:null}get ArrayType(){return this._type.ArrayType}get numChildren(){return this._numChildren}get stride(){return this._chunks[0]?this._chunks[0].stride:1}get byteLength(){return this._chunks.reduce((t,e)=>t+e.byteLength,0)}get nullCount(){let t=this._nullCount;return t<0&&(this._nullCount=t=this._chunks.reduce((t,{nullCount:e})=>t+e,0)),t}get indices(){if(Ot.isDictionary(this._type)){if(!this._indices){const t=this._chunks;this._indices=1===t.length?t[0].indices:Wn.concat(...t.map(t=>t.indices))}return this._indices}return null}get dictionary(){return Ot.isDictionary(this._type)?this._chunks[this._chunks.length-1].data.dictionary:null}*[Symbol.iterator](){for(const t of this._chunks)yield*t}clone(t=this._chunks){return new Wn(this._type,t)}concat(...t){return this.clone(Wn.flatten(this,...t))}slice(t,e){return Dn(this,t,e,this._sliceInternal)}getChildAt(t){if(t<0||t>=this._numChildren)return null;let e,n,r,i=this._children||(this._children=[]);return(e=i[t])?e:(n=(this._type.children||[])[t])&&(r=this._chunks.map(e=>e.getChildAt(t)).filter(t=>null!=t),r.length>0)?i[t]=new Wn(n.type,r):null}search(t,e){let n=t,r=this._chunkOffsets,i=r.length-1;if(n<0)return null;if(n>=r[i])return null;if(i<=1)return e?e(this,0,n):[0,n];let o=0,s=0,a=0;do{if(o+1===i)return e?e(this,o,n-s):[o,n-s];a=o+(i-o)/2|0,n>=r[a]?o=a:i=a}while(n<r[i]&&n>=(s=r[o]));return null}isValid(t){return!!this.search(t,this.isValidInternal)}get(t){return this.search(t,this.getInternal)}set(t,e){this.search(t,({chunks:t},n,r)=>t[n].set(r,e))}indexOf(t,e){return e&&"number"===typeof e?this.search(e,(e,n,r)=>this.indexOfInternal(e,n,r,t)):this.indexOfInternal(this,0,Math.max(0,e||0),t)}toArray(){const{chunks:t}=this,e=t.length;let n=this._type.ArrayType;if(e<=0)return new n(0);if(e<=1)return t[0].toArray();let r=0,i=new Array(e);for(let a=-1;++a<e;)r+=(i[a]=t[a].toArray()).length;n!==i[0].constructor&&(n=i[0].constructor);let o=new n(r),s=n===Array?Kn:qn;for(let a=-1,c=0;++a<e;)c=s(i[a],o,c);return o}getInternal({_chunks:t},e,n){return t[e].get(n)}isValidInternal({_chunks:t},e,n){return t[e].isValid(n)}indexOfInternal({_chunks:t},e,n,r){let i=e-1,o=t.length,s=n,a=0,c=-1;while(++i<o){if(~(c=t[i].indexOf(r,s)))return a+c;s=0,a+=t[i].length}return-1}_sliceInternal(t,e,n){const r=[],{chunks:i,_chunkOffsets:o}=t;for(let s=-1,a=i.length;++s<a;){const t=i[s],a=t.length,c=o[s];if(c>=n)break;if(e>=c+a)continue;if(c>=e&&c+a<=n){r.push(t);continue}const u=Math.max(0,e-c),l=Math.min(n-c,a);r.push(t.slice(u,l))}return t.clone(r)}}function Gn(t){let e=new Uint32Array((t||[]).length+1),n=e[0]=0,r=e.length;for(let i=0;++i<r;)e[i]=n+=t[i-1].length;return e}const qn=(t,e,n)=>(e.set(t,n),n+t.length),Kn=(t,e,n)=>{let r=n;for(let i=-1,o=t.length;++i<o;)e[r++]=t[i];return r};class Xn extends Wn{constructor(t,e=[],n){if(e=Wn.flatten(...e),super(t.type,e,n),this._field=t,1===e.length&&!(this instanceof Jn))return new Jn(t,e[0],this._chunkOffsets)}static new(t,e,...n){const r=Wn.flatten(Array.isArray(e)?[...e,...n]:e instanceof P?[e,...n]:[P.new(e,...n)]);if("string"===typeof t){const e=r[0].data.type;t=new lr(t,e,!0)}else!t.nullable&&r.some(({nullCount:t})=>t>0)&&(t=t.clone({nullable:!0}));return new Xn(t,r)}get field(){return this._field}get name(){return this._field.name}get nullable(){return this._field.nullable}get metadata(){return this._field.metadata}clone(t=this._chunks){return new Xn(this._field,t)}getChildAt(t){if(t<0||t>=this.numChildren)return null;let e,n,r,i=this._children||(this._children=[]);return(e=i[t])?e:(n=(this.type.children||[])[t])&&(r=this._chunks.map(e=>e.getChildAt(t)).filter(t=>null!=t),r.length>0)?i[t]=new Xn(n,r):null}}class Jn extends Xn{constructor(t,e,n){super(t,[e],n),this._chunk=e}search(t,e){return e?e(this,0,t):[0,t]}isValid(t){return this._chunk.isValid(t)}get(t){return this._chunk.get(t)}set(t,e){this._chunk.set(t,e)}indexOf(t,e){return this._chunk.indexOf(t,e)}}const Zn=Array.isArray,Qn=(t,e)=>ir(t,e,[],0),tr=t=>{const[e,n]=cr(t,[[],[]]);return n.map((t,n)=>t instanceof Xn?Xn.new(t.field.clone(e[n]),t):t instanceof P?Xn.new(e[n],t):Xn.new(e[n],[]))},er=t=>cr(t,[[],[]]),nr=(t,e)=>or(t,e,[],0),rr=(t,e)=>sr(t,e,[],0);function ir(t,e,n,r){let i,o=r,s=-1,a=e.length;while(++s<a)Zn(i=e[s])?o=ir(t,i,n,o).length:i instanceof t&&(n[o++]=i);return n}function or(t,e,n,r){let i,o=r,s=-1,a=e.length;while(++s<a)Zn(i=e[s])?o=or(t,i,n,o).length:i instanceof Wn?o=or(t,i.chunks,n,o).length:i instanceof t&&(n[o++]=i);return n}function sr(t,e,n,r){let i,o=r,s=-1,a=e.length;while(++s<a)Zn(i=e[s])?o=sr(t,i,n,o).length:i instanceof t?o=ir(P,i.schema.fields.map((t,e)=>i.getChildAt(e)),n,o).length:i instanceof P&&(n[o++]=i);return n}const ar=(t,[e,n],r)=>(t[0][r]=e,t[1][r]=n,t);function cr(t,e){let n,r;switch(r=t.length){case 0:return e;case 1:if(n=e[0],!t[0])return e;if(Zn(t[0]))return cr(t[0],e);t[0]instanceof ie||t[0]instanceof P||t[0]instanceof Ot||([n,t]=Object.entries(t[0]).reduce(ar,e));break;default:Zn(n=t[r-1])?t=Zn(t[0])?t[0]:t.slice(0,r-1):(t=Zn(t[0])?t[0]:t,n=[])}let i,o,s=-1,a=-1,c=-1,u=t.length,[l,f]=e;while(++c<u)o=t[c],o instanceof Xn&&(f[++a]=o)?l[++s]=o.field.clone(n[c],o.type,!0):(({[c]:i=c}=n),o instanceof Ot&&(f[++a]=o)?l[++s]=lr.new(i,o,!0):o&&o.type&&(f[++a]=o)&&(o instanceof ie&&(f[a]=o=P.new(o)),l[++s]=lr.new(i,o.type,!0)));return e}class ur{constructor(t=[],e,n){this.fields=t||[],this.metadata=e||new Map,n||(n=dr(t)),this.dictionaries=n}static from(...t){return ur.new(t[0],t[1])}static new(...t){return new ur(er(t)[0])}get[Symbol.toStringTag](){return"Schema"}toString(){return`Schema<{ ${this.fields.map((t,e)=>`${e}: ${t}`).join(", ")} }>`}compareTo(t){return wt.compareSchemas(this,t)}select(...t){const e=t.reduce((t,e)=>(t[e]=!0)&&t,Object.create(null));return new ur(this.fields.filter(t=>e[t.name]),this.metadata)}selectAt(...t){return new ur(t.map(t=>this.fields[t]).filter(Boolean),this.metadata)}assign(...t){const e=t[0]instanceof ur?t[0]:new ur(Qn(lr,t)),n=[...this.fields],r=fr(fr(new Map,this.metadata),e.metadata),i=e.fields.filter(t=>{const e=n.findIndex(e=>e.name===t.name);return!~e||(n[e]=t.clone({metadata:fr(fr(new Map,n[e].metadata),t.metadata)}))&&!1}),o=dr(i,new Map);return new ur([...n,...i],r,new Map([...this.dictionaries,...o]))}}class lr{constructor(t,e,n=!1,r){this.name=t,this.type=e,this.nullable=n,this.metadata=r||new Map}static new(...t){let[e,n,r,i]=t;return t[0]&&"object"===typeof t[0]&&(({name:e}=t[0]),void 0===n&&(n=t[0].type),void 0===r&&(r=t[0].nullable),void 0===i&&(i=t[0].metadata)),new lr(""+e,n,r,i)}get typeId(){return this.type.typeId}get[Symbol.toStringTag](){return"Field"}toString(){return`${this.name}: ${this.type}`}compareTo(t){return wt.compareField(this,t)}clone(...t){let[e,n,r,i]=t;return t[0]&&"object"===typeof t[0]?({name:e=this.name,type:n=this.type,nullable:r=this.nullable,metadata:i=this.metadata}=t[0]):[e=this.name,n=this.type,r=this.nullable,i=this.metadata]=t,lr.new(e,n,r,i)}}function fr(t,e){return new Map([...t||new Map,...e||new Map])}function dr(t,e=new Map){for(let n=-1,r=t.length;++n<r;){const r=t[n],i=r.type;if(Ot.isDictionary(i))if(e.has(i.id)){if(e.get(i.id)!==i.dictionary)throw new Error("Cannot create Schema containing two different dictionaries with the same Id")}else e.set(i.id,i.dictionary);i.children&&i.children.length>0&&dr(i.children,e)}return e}ur.prototype.fields=null,ur.prototype.metadata=null,ur.prototype.dictionaries=null,lr.prototype.type=null,lr.prototype.name=null,lr.prototype.nullable=null,lr.prototype.metadata=null;class hr extends ge{constructor(t){super(t),this._run=new wn,this._offsets=new ye}addChild(t,e="0"){if(this.numChildren>0)throw new Error("ListBuilder can only have one child.");return this.children[this.numChildren]=t,this.type=new Gt(new lr(e,t.type,!0)),this.numChildren-1}clear(){return this._run.clear(),super.clear()}_flushPending(t){const e=this._run,n=this._offsets,r=this._setValue;let i,o=0;for([o,i]of t)void 0===i?n.set(o,0):(n.set(o,i.length),r(this,o,e.bind(i)))}}class pr extends be{constructor(){super(...arguments),this._run=new wn}setValue(t,e){super.setValue(t,this._run.bind(e))}addChild(t,e="0"){if(this.numChildren>0)throw new Error("FixedSizeListBuilder can only have one child.");const n=this.children.push(t);return this.type=new Jt(this.type.listSize,new lr(e,t.type,!0)),n}clear(){return this._run.clear(),super.clear()}}class yr extends ge{set(t,e){return super.set(t,e)}setValue(t,e){e=e instanceof Map?e:new Map(Object.entries(e));const n=this._pending||(this._pending=new Map),r=n.get(t);r&&(this._pendingLength-=r.size),this._pendingLength+=e.size,n.set(t,e)}addChild(t,e=""+this.numChildren){if(this.numChildren>0)throw new Error("ListBuilder can only have one child.");return this.children[this.numChildren]=t,this.type=new Zt(new lr(e,t.type,!0),this.type.keysSorted),this.numChildren-1}_flushPending(t){const e=this._offsets,n=this._setValue;t.forEach((t,r)=>{void 0===t?e.set(r,0):(e.set(r,t.size),n(this,r,t))})}}class vr extends be{addChild(t,e=""+this.numChildren){const n=this.children.push(t);return this.type=new qt([...this.type.children,new lr(e,t.type,!0)]),n}}class br extends be{constructor(t){super(t),this._typeIds=new he(new Int8Array(0),1),"function"===typeof t["valueToChildTypeId"]&&(this._valueToChildTypeId=t["valueToChildTypeId"])}get typeIdToChildIndex(){return this.type.typeIdToChildIndex}append(t,e){return this.set(this.length,t,e)}set(t,e,n){return void 0===n&&(n=this._valueToChildTypeId(this,e,t)),this.setValid(t,this.isValid(e))&&this.setValue(t,e,n),this}setValue(t,e,n){this._typeIds.set(t,n),super.setValue(t,e)}addChild(t,e=""+this.children.length){const n=this.children.push(t),{type:{children:r,mode:i,typeIds:o}}=this,s=[...r,new lr(e,t.type)];return this.type=new Kt(i,[...o,n],s),n}_valueToChildTypeId(t,e,n){throw new Error("Cannot map UnionBuilder value to child typeId. Pass the `childTypeId` as the second argument to unionBuilder.append(), or supply a `valueToChildTypeId` function as part of the UnionBuilder constructor options.")}}class mr extends br{}class gr extends br{constructor(t){super(t),this._offsets=new he(new Int32Array(0))}setValue(t,e,n){const r=this.type.typeIdToChildIndex[n];return this._offsets.set(t,this.getChildAt(r).length),super.setValue(t,e,n)}}class _r extends nt{}const wr=(t,e,n)=>{t[e]=n/864e5|0},Or=(t,e,n)=>{t[e]=n%4294967296|0,t[e+1]=n/4294967296|0},Sr=(t,e,n)=>{t[e]=1e3*n%4294967296|0,t[e+1]=1e3*n/4294967296|0},xr=(t,e,n)=>{t[e]=1e6*n%4294967296|0,t[e+1]=1e6*n/4294967296|0},Ir=(t,e,n,r)=>{const{[n]:i,[n+1]:o}=e;null!=i&&null!=o&&t.set(r.subarray(0,o-i),i)},Ar=({offset:t,values:e},n,r)=>{const i=t+n;r?e[i>>3]|=1<<i%8:e[i>>3]&=~(1<<i%8)},Tr=({values:t},e,n)=>{wr(t,e,n.valueOf())},jr=({values:t},e,n)=>{Or(t,2*e,n.valueOf())},Er=({stride:t,values:e},n,r)=>{e[t*n]=r},Br=({stride:t,values:e},n,r)=>{e[t*n]=ke(r)},Cr=(t,e,n)=>{switch(typeof n){case"bigint":t.values64[e]=n;break;case"number":t.values[e*t.stride]=n;break;default:const r=n,{stride:i,ArrayType:o}=t,s=Object(ne["toArrayBufferView"])(o,r);t.values.set(s.subarray(0,i),i*e)}},Mr=({stride:t,values:e},n,r)=>{e.set(r.subarray(0,t),t*n)},kr=({values:t,valueOffsets:e},n,r)=>Ir(t,e,n,r),Dr=({values:t,valueOffsets:e},n,r)=>{Ir(t,e,n,Object(mn["b"])(r))},Nr=(t,e,n)=>{t.type.bitWidth<64?Er(t,e,n):Cr(t,e,n)},Lr=(t,e,n)=>{t.type.precision!==$.HALF?Er(t,e,n):Br(t,e,n)},Pr=(t,e,n)=>{t.type.unit===R.DAY?Tr(t,e,n):jr(t,e,n)},Fr=({values:t},e,n)=>Or(t,2*e,n/1e3),Ur=({values:t},e,n)=>Or(t,2*e,n),Rr=({values:t},e,n)=>Sr(t,2*e,n),Vr=({values:t},e,n)=>xr(t,2*e,n),$r=(t,e,n)=>{switch(t.type.unit){case V.SECOND:return Fr(t,e,n);case V.MILLISECOND:return Ur(t,e,n);case V.MICROSECOND:return Rr(t,e,n);case V.NANOSECOND:return Vr(t,e,n)}},zr=({values:t,stride:e},n,r)=>{t[e*n]=r},Yr=({values:t,stride:e},n,r)=>{t[e*n]=r},Hr=({values:t},e,n)=>{t.set(n.subarray(0,2),2*e)},Wr=({values:t},e,n)=>{t.set(n.subarray(0,2),2*e)},Gr=(t,e,n)=>{switch(t.type.unit){case V.SECOND:return zr(t,e,n);case V.MILLISECOND:return Yr(t,e,n);case V.MICROSECOND:return Hr(t,e,n);case V.NANOSECOND:return Wr(t,e,n)}},qr=({values:t},e,n)=>{t.set(n.subarray(0,4),4*e)},Kr=(t,e,n)=>{const r=t.getChildAt(0),i=t.valueOffsets;for(let o=-1,s=i[e],a=i[e+1];s<a;)r.set(s++,n.get(++o))},Xr=(t,e,n)=>{const r=t.getChildAt(0),i=t.valueOffsets,o=n instanceof Map?[...n]:Object.entries(n);for(let s=-1,a=i[e],c=i[e+1];a<c;)r.set(a++,o[++s])},Jr=(t,e)=>(n,r,i)=>n&&n.set(t,e[i]),Zr=(t,e)=>(n,r,i)=>n&&n.set(t,e.get(i)),Qr=(t,e)=>(n,r,i)=>n&&n.set(t,e.get(r.name)),ti=(t,e)=>(n,r,i)=>n&&n.set(t,e[r.name]),ei=(t,e,n)=>{const r=n instanceof Map?Qr(e,n):n instanceof P?Zr(e,n):Array.isArray(n)?Jr(e,n):ti(e,n);t.type.children.forEach((e,n)=>r(t.getChildAt(n),e,n))},ni=(t,e,n)=>{t.type.mode===z.Dense?ri(t,e,n):ii(t,e,n)},ri=(t,e,n)=>{const r=t.typeIdToChildIndex[t.typeIds[e]],i=t.getChildAt(r);i&&i.set(t.valueOffsets[e],n)},ii=(t,e,n)=>{const r=t.typeIdToChildIndex[t.typeIds[e]],i=t.getChildAt(r);i&&i.set(e,n)},oi=(t,e,n)=>{const r=t.getKey(e);null!==r&&t.setValue(r,n)},si=(t,e,n)=>{t.type.unit===Y.DAY_TIME?ai(t,e,n):ci(t,e,n)},ai=({values:t},e,n)=>{t.set(n.subarray(0,2),2*e)},ci=({values:t},e,n)=>{t[e]=12*n[0]+n[1]%12},ui=(t,e,n)=>{const r=t.getChildAt(0),{stride:i}=t;for(let o=-1,s=e*i;++o<i;)r.set(s+o,n.get(o))};_r.prototype.visitBool=Ar,_r.prototype.visitInt=Nr,_r.prototype.visitInt8=Er,_r.prototype.visitInt16=Er,_r.prototype.visitInt32=Er,_r.prototype.visitInt64=Cr,_r.prototype.visitUint8=Er,_r.prototype.visitUint16=Er,_r.prototype.visitUint32=Er,_r.prototype.visitUint64=Cr,_r.prototype.visitFloat=Lr,_r.prototype.visitFloat16=Br,_r.prototype.visitFloat32=Er,_r.prototype.visitFloat64=Er,_r.prototype.visitUtf8=Dr,_r.prototype.visitBinary=kr,_r.prototype.visitFixedSizeBinary=Mr,_r.prototype.visitDate=Pr,_r.prototype.visitDateDay=Tr,_r.prototype.visitDateMillisecond=jr,_r.prototype.visitTimestamp=$r,_r.prototype.visitTimestampSecond=Fr,_r.prototype.visitTimestampMillisecond=Ur,_r.prototype.visitTimestampMicrosecond=Rr,_r.prototype.visitTimestampNanosecond=Vr,_r.prototype.visitTime=Gr,_r.prototype.visitTimeSecond=zr,_r.prototype.visitTimeMillisecond=Yr,_r.prototype.visitTimeMicrosecond=Hr,_r.prototype.visitTimeNanosecond=Wr,_r.prototype.visitDecimal=qr,_r.prototype.visitList=Kr,_r.prototype.visitStruct=ei,_r.prototype.visitUnion=ni,_r.prototype.visitDenseUnion=ri,_r.prototype.visitSparseUnion=ii,_r.prototype.visitDictionary=oi,_r.prototype.visitInterval=si,_r.prototype.visitIntervalDayTime=ai,_r.prototype.visitIntervalYearMonth=ci,_r.prototype.visitFixedSizeList=ui,_r.prototype.visitMap=Xr;const li=new _r;class fi extends nt{visitNull(){return Se}visitBool(){return Oe}visitInt(){return qe}visitInt8(){return Ke}visitInt16(){return Xe}visitInt32(){return Je}visitInt64(){return Ze}visitUint8(){return Qe}visitUint16(){return tn}visitUint32(){return en}visitUint64(){return nn}visitFloat(){return De}visitFloat16(){return Ne}visitFloat32(){return Le}visitFloat64(){return Pe}visitUtf8(){return _n}visitBinary(){return gn}visitFixedSizeBinary(){return Ee}visitDate(){return xe}visitDateDay(){return Ie}visitDateMillisecond(){return Ae}visitTimestamp(){return ln}visitTimestampSecond(){return fn}visitTimestampMillisecond(){return dn}visitTimestampMicrosecond(){return hn}visitTimestampNanosecond(){return pn}visitTime(){return on}visitTimeSecond(){return sn}visitTimeMillisecond(){return an}visitTimeMicrosecond(){return cn}visitTimeNanosecond(){return un}visitDecimal(){return Te}visitList(){return hr}visitStruct(){return vr}visitUnion(){return br}visitDenseUnion(){return gr}visitSparseUnion(){return mr}visitDictionary(){return je}visitInterval(){return yn}visitIntervalDayTime(){return vn}visitIntervalYearMonth(){return bn}visitFixedSizeList(){return pr}visitMap(){return yr}}const di=new fi;function hi(t){const e=t.type,n=new(di.getVisitFn(e)())(t);if(e.children&&e.children.length>0){const r=t["children"]||[],i={nullValues:t["nullValues"]},o=Array.isArray(r)?(t,e)=>r[e]||i:({name:t})=>r[t]||i;e.children.forEach((t,e)=>{const{type:r}=t,i=o(t,e);n.children.push(hi({...i,type:r}))})}return n}var pi;be.new=hi,Object.keys(F).map(t=>F[t]).filter(t=>"number"===typeof t&&t!==F.NONE).forEach(t=>{const e=di.visit(t);e.prototype._setValue=li.getVisitFn(t)}),_n.prototype._setValue=li.visitBinary,function(t){(function(e){(function(e){(function(e){class n{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}static getRootAsFooter(t,e){return(e||new n).__init(t.readInt32(t.position())+t.position(),t)}version(){let t=this.bb.__offset(this.bb_pos,4);return t?this.bb.readInt16(this.bb_pos+t):D.apache.arrow.flatbuf.MetadataVersion.V1}schema(t){let e=this.bb.__offset(this.bb_pos,6);return e?(t||new D.apache.arrow.flatbuf.Schema).__init(this.bb.__indirect(this.bb_pos+e),this.bb):null}dictionaries(e,n){let r=this.bb.__offset(this.bb_pos,8);return r?(n||new t.apache.arrow.flatbuf.Block).__init(this.bb.__vector(this.bb_pos+r)+24*e,this.bb):null}dictionariesLength(){let t=this.bb.__offset(this.bb_pos,8);return t?this.bb.__vector_len(this.bb_pos+t):0}recordBatches(e,n){let r=this.bb.__offset(this.bb_pos,10);return r?(n||new t.apache.arrow.flatbuf.Block).__init(this.bb.__vector(this.bb_pos+r)+24*e,this.bb):null}recordBatchesLength(){let t=this.bb.__offset(this.bb_pos,10);return t?this.bb.__vector_len(this.bb_pos+t):0}static startFooter(t){t.startObject(4)}static addVersion(t,e){t.addFieldInt16(0,e,D.apache.arrow.flatbuf.MetadataVersion.V1)}static addSchema(t,e){t.addFieldOffset(1,e,0)}static addDictionaries(t,e){t.addFieldOffset(2,e,0)}static startDictionariesVector(t,e){t.startVector(24,e,8)}static addRecordBatches(t,e){t.addFieldOffset(3,e,0)}static startRecordBatchesVector(t,e){t.startVector(24,e,8)}static endFooter(t){let e=t.endObject();return e}static finishFooterBuffer(t,e){t.finish(e)}static createFooter(t,e,r,i,o){return n.startFooter(t),n.addVersion(t,e),n.addSchema(t,r),n.addDictionaries(t,i),n.addRecordBatches(t,o),n.endFooter(t)}}e.Footer=n})(e.flatbuf||(e.flatbuf={}))})(e.arrow||(e.arrow={}))})(t.apache||(t.apache={}))}(pi||(pi={})),function(t){(function(t){(function(t){(function(t){class e{constructor(){this.bb=null,this.bb_pos=0}__init(t,e){return this.bb_pos=t,this.bb=e,this}offset(){return this.bb.readInt64(this.bb_pos)}metaDataLength(){return this.bb.readInt32(this.bb_pos+8)}bodyLength(){return this.bb.readInt64(this.bb_pos+16)}static createBlock(t,e,n,r){return t.prep(8,24),t.writeInt64(r),t.pad(4),t.writeInt32(n),t.writeInt64(e),t.offset()}}t.Block=e})(t.flatbuf||(t.flatbuf={}))})(t.arrow||(t.arrow={}))})(t.apache||(t.apache={}))}(pi||(pi={}));var yi=n("a6b2"),vi=yi["a"].Long,bi=yi["a"].Builder,mi=yi["a"].ByteBuffer,gi=pi.apache.arrow.flatbuf.Block,_i=pi.apache.arrow.flatbuf.Footer;class wi{constructor(t,e=W.V4,n,r){this.schema=t,this.version=e,n&&(this._recordBatches=n),r&&(this._dictionaryBatches=r)}static decode(t){t=new mi(Object(ne["toUint8Array"])(t));const e=_i.getRootAsFooter(t),n=ur.decode(e.schema());return new Oi(n,e)}static encode(t){const e=new bi,n=ur.encode(e,t.schema);_i.startRecordBatchesVector(e,t.numRecordBatches),[...t.recordBatches()].slice().reverse().forEach(t=>Si.encode(e,t));const r=e.endVector();_i.startDictionariesVector(e,t.numDictionaries),[...t.dictionaryBatches()].slice().reverse().forEach(t=>Si.encode(e,t));const i=e.endVector();return _i.startFooter(e),_i.addSchema(e,n),_i.addVersion(e,W.V4),_i.addRecordBatches(e,r),_i.addDictionaries(e,i),_i.finishFooterBuffer(e,_i.endFooter(e)),e.asUint8Array()}get numRecordBatches(){return this._recordBatches.length}get numDictionaries(){return this._dictionaryBatches.length}*recordBatches(){for(let t,e=-1,n=this.numRecordBatches;++e<n;)(t=this.getRecordBatch(e))&&(yield t)}*dictionaryBatches(){for(let t,e=-1,n=this.numDictionaries;++e<n;)(t=this.getDictionaryBatch(e))&&(yield t)}getRecordBatch(t){return t>=0&&t<this.numRecordBatches&&this._recordBatches[t]||null}getDictionaryBatch(t){return t>=0&&t<this.numDictionaries&&this._dictionaryBatches[t]||null}}class Oi extends wi{constructor(t,e){super(t,e.version()),this._footer=e}get numRecordBatches(){return this._footer.recordBatchesLength()}get numDictionaries(){return this._footer.dictionariesLength()}getRecordBatch(t){if(t>=0&&t<this.numRecordBatches){const e=this._footer.recordBatches(t);if(e)return Si.decode(e)}return null}getDictionaryBatch(t){if(t>=0&&t<this.numDictionaries){const e=this._footer.dictionaries(t);if(e)return Si.decode(e)}return null}}class Si{static decode(t){return new Si(t.metaDataLength(),t.bodyLength(),t.offset())}static encode(t,e){const{metaDataLength:n}=e,r=new vi(e.offset,0),i=new vi(e.bodyLength,0);return gi.createBlock(t,r,n,i)}constructor(t,e,n){this.metaDataLength=t,this.offset="number"===typeof n?n:n.low,this.bodyLength="number"===typeof e?e:e.low}}var xi=n("f673");class Ii extends xi["b"]{write(t){if((t=Object(ne["toUint8Array"])(t)).byteLength>0)return super.write(t)}toString(t=!1){return t?Object(mn["a"])(this.toUint8Array(!0)):this.toUint8Array(!1).then(mn["a"])}toUint8Array(t=!1){return t?Object(ne["joinUint8Arrays"])(this._values)[0]:(async()=>{let t=[],e=0;for await(const n of this)t.push(n),e+=n.byteLength;return Object(ne["joinUint8Arrays"])(t,e)[0]})()}}class Ai{constructor(t){t&&(this.source=new ji(L["a"].fromIterable(t)))}[Symbol.iterator](){return this}next(t){return this.source.next(t)}throw(t){return this.source.throw(t)}return(t){return this.source.return(t)}peek(t){return this.source.peek(t)}read(t){return this.source.read(t)}}class Ti{constructor(t){t instanceof Ti?this.source=t.source:t instanceof Ii?this.source=new Ei(L["a"].fromAsyncIterable(t)):Object(ae["n"])(t)?this.source=new Ei(L["a"].fromNodeStream(t)):Object(ae["m"])(t)?this.source=new Ei(L["a"].fromDOMStream(t)):Object(ae["g"])(t)?this.source=new Ei(L["a"].fromDOMStream(t.body)):Object(ae["i"])(t)?this.source=new Ei(L["a"].fromIterable(t)):(Object(ae["l"])(t)||Object(ae["f"])(t))&&(this.source=new Ei(L["a"].fromAsyncIterable(t)))}[Symbol.asyncIterator](){return this}next(t){return this.source.next(t)}throw(t){return this.source.throw(t)}return(t){return this.source.return(t)}get closed(){return this.source.closed}cancel(t){return this.source.cancel(t)}peek(t){return this.source.peek(t)}read(t){return this.source.read(t)}}class ji{constructor(t){this.source=t}cancel(t){this.return(t)}peek(t){return this.next(t,"peek").value}read(t){return this.next(t,"read").value}next(t,e="read"){return this.source.next({cmd:e,size:t})}throw(t){return Object.create(this.source.throw&&this.source.throw(t)||xi["c"])}return(t){return Object.create(this.source.return&&this.source.return(t)||xi["c"])}}class Ei{constructor(t){this.source=t,this._closedPromise=new Promise(t=>this._closedPromiseResolve=t)}async cancel(t){await this.return(t)}get closed(){return this._closedPromise}async read(t){return(await this.next(t,"read")).value}async peek(t){return(await this.next(t,"peek")).value}async next(t,e="read"){return await this.source.next({cmd:e,size:t})}async throw(t){const e=this.source.throw&&await this.source.throw(t)||xi["c"];return this._closedPromiseResolve&&this._closedPromiseResolve(),this._closedPromiseResolve=void 0,Object.create(e)}async return(t){const e=this.source.return&&await this.source.return(t)||xi["c"];return this._closedPromiseResolve&&this._closedPromiseResolve(),this._closedPromiseResolve=void 0,Object.create(e)}}class Bi extends Ai{constructor(t,e){super(),this.position=0,this.buffer=Object(ne["toUint8Array"])(t),this.size="undefined"===typeof e?this.buffer.byteLength:e}readInt32(t){const{buffer:e,byteOffset:n}=this.readAt(t,4);return new DataView(e,n).getInt32(0,!0)}seek(t){return this.position=Math.min(t,this.size),t<this.size}read(t){const{buffer:e,size:n,position:r}=this;return e&&r<n?("number"!==typeof t&&(t=1/0),this.position=Math.min(n,r+Math.min(n-r,t)),e.subarray(r,this.position)):null}readAt(t,e){const n=this.buffer,r=Math.min(this.size,t+e);return n?n.subarray(t,r):new Uint8Array(e)}close(){this.buffer&&(this.buffer=null)}throw(t){return this.close(),{done:!0,value:t}}return(t){return this.close(),{done:!0,value:t}}}class Ci extends Ti{constructor(t,e){super(),this.position=0,this._handle=t,"number"===typeof e?this.size=e:this._pending=(async()=>{this.size=(await t.stat()).size,delete this._pending})()}async readInt32(t){const{buffer:e,byteOffset:n}=await this.readAt(t,4);return new DataView(e,n).getInt32(0,!0)}async seek(t){return this._pending&&await this._pending,this.position=Math.min(t,this.size),t<this.size}async read(t){this._pending&&await this._pending;const{_handle:e,size:n,position:r}=this;if(e&&r<n){"number"!==typeof t&&(t=1/0);let i=r,o=0,s=0,a=Math.min(n,i+Math.min(n-i,t)),c=new Uint8Array(Math.max(0,(this.position=a)-i));while((i+=s)<a&&(o+=s)<c.byteLength)({bytesRead:s}=await e.read(c,o,c.byteLength-o,i));return c}return null}async readAt(t,e){this._pending&&await this._pending;const{_handle:n,size:r}=this;if(n&&t+e<r){const i=Math.min(r,t+e),o=new Uint8Array(i-t);return(await n.read(o,0,e,t)).buffer}return new Uint8Array(e)}async close(){const t=this._handle;this._handle=null,t&&await t.close()}async throw(t){return await this.close(),{done:!0,value:t}}async return(t){return await this.close(),{done:!0,value:t}}}const Mi=65536;function ki(t){return t<0&&(t=4294967295+t+1),"0x"+t.toString(16)}const Di=8,Ni=[1,10,100,1e3,1e4,1e5,1e6,1e7,1e8];class Li{constructor(t){this.buffer=t}high(){return this.buffer[1]}low(){return this.buffer[0]}_times(t){const e=new Uint32Array([this.buffer[1]>>>16,65535&this.buffer[1],this.buffer[0]>>>16,65535&this.buffer[0]]),n=new Uint32Array([t.buffer[1]>>>16,65535&t.buffer[1],t.buffer[0]>>>16,65535&t.buffer[0]]);let r=e[3]*n[3];this.buffer[0]=65535&r;let i=r>>>16;return r=e[2]*n[3],i+=r,r=e[3]*n[2]>>>0,i+=r,this.buffer[0]+=i<<16,this.buffer[1]=i>>>0<r?Mi:0,this.buffer[1]+=i>>>16,this.buffer[1]+=e[1]*n[3]+e[2]*n[2]+e[3]*n[1],this.buffer[1]+=e[0]*n[3]+e[1]*n[2]+e[2]*n[1]+e[3]*n[0]<<16,this}_plus(t){const e=this.buffer[0]+t.buffer[0]>>>0;this.buffer[1]+=t.buffer[1],e<this.buffer[0]>>>0&&++this.buffer[1],this.buffer[0]=e}lessThan(t){return this.buffer[1]<t.buffer[1]||this.buffer[1]===t.buffer[1]&&this.buffer[0]<t.buffer[0]}equals(t){return this.buffer[1]===t.buffer[1]&&this.buffer[0]==t.buffer[0]}greaterThan(t){return t.lessThan(this)}hex(){return`${ki(this.buffer[1])} ${ki(this.buffer[0])}`}}class Pi extends Li{times(t){return this._times(t),this}plus(t){return this._plus(t),this}static from(t,e=new Uint32Array(2)){return Pi.fromString("string"===typeof t?t:t.toString(),e)}static fromNumber(t,e=new Uint32Array(2)){return Pi.fromString(t.toString(),e)}static fromString(t,e=new Uint32Array(2)){const n=t.length;let r=new Pi(e);for(let i=0;i<n;){const e=Di<n-i?Di:n-i,o=new Pi(new Uint32Array([parseInt(t.substr(i,e),10),0])),s=new Pi(new Uint32Array([Ni[e],0]));r.times(s),r.plus(o),i+=e}return r}static convertArray(t){const e=new Uint32Array(2*t.length);for(let n=-1,r=t.length;++n<r;)Pi.from(t[n],new Uint32Array(e.buffer,e.byteOffset+2*n*4,2));return e}static multiply(t,e){let n=new Pi(new Uint32Array(t.buffer));return n.times(e)}static add(t,e){let n=new Pi(new Uint32Array(t.buffer));return n.plus(e)}}class Fi extends Li{negate(){return this.buffer[0]=1+~this.buffer[0],this.buffer[1]=~this.buffer[1],0==this.buffer[0]&&++this.buffer[1],this}times(t){return this._times(t),this}plus(t){return this._plus(t),this}lessThan(t){const e=this.buffer[1]<<0,n=t.buffer[1]<<0;return e<n||e===n&&this.buffer[0]<t.buffer[0]}static from(t,e=new Uint32Array(2)){return Fi.fromString("string"===typeof t?t:t.toString(),e)}static fromNumber(t,e=new Uint32Array(2)){return Fi.fromString(t.toString(),e)}static fromString(t,e=new Uint32Array(2)){const n=t.startsWith("-"),r=t.length;let i=new Fi(e);for(let o=n?1:0;o<r;){const e=Di<r-o?Di:r-o,n=new Fi(new Uint32Array([parseInt(t.substr(o,e),10),0])),s=new Fi(new Uint32Array([Ni[e],0]));i.times(s),i.plus(n),o+=e}return n?i.negate():i}static convertArray(t){const e=new Uint32Array(2*t.length);for(let n=-1,r=t.length;++n<r;)Fi.from(t[n],new Uint32Array(e.buffer,e.byteOffset+2*n*4,2));return e}static multiply(t,e){let n=new Fi(new Uint32Array(t.buffer));return n.times(e)}static add(t,e){let n=new Fi(new Uint32Array(t.buffer));return n.plus(e)}}class Ui{constructor(t){this.buffer=t}high(){return new Fi(new Uint32Array(this.buffer.buffer,this.buffer.byteOffset+8,2))}low(){return new Fi(new Uint32Array(this.buffer.buffer,this.buffer.byteOffset,2))}negate(){return this.buffer[0]=1+~this.buffer[0],this.buffer[1]=~this.buffer[1],this.buffer[2]=~this.buffer[2],this.buffer[3]=~this.buffer[3],0==this.buffer[0]&&++this.buffer[1],0==this.buffer[1]&&++this.buffer[2],0==this.buffer[2]&&++this.buffer[3],this}times(t){const e=new Pi(new Uint32Array([this.buffer[3],0])),n=new Pi(new Uint32Array([this.buffer[2],0])),r=new Pi(new Uint32Array([this.buffer[1],0])),i=new Pi(new Uint32Array([this.buffer[0],0])),o=new Pi(new Uint32Array([t.buffer[3],0])),s=new Pi(new Uint32Array([t.buffer[2],0])),a=new Pi(new Uint32Array([t.buffer[1],0])),c=new Pi(new Uint32Array([t.buffer[0],0]));let u=Pi.multiply(i,c);this.buffer[0]=u.low();let l=new Pi(new Uint32Array([u.high(),0]));u=Pi.multiply(r,c),l.plus(u),u=Pi.multiply(i,a),l.plus(u),this.buffer[1]=l.low(),this.buffer[3]=l.lessThan(u)?1:0,this.buffer[2]=l.high();let f=new Pi(new Uint32Array(this.buffer.buffer,this.buffer.byteOffset+8,2));return f.plus(Pi.multiply(n,c)).plus(Pi.multiply(r,a)).plus(Pi.multiply(i,s)),this.buffer[3]+=Pi.multiply(e,c).plus(Pi.multiply(n,a)).plus(Pi.multiply(r,s)).plus(Pi.multiply(i,o)).low(),this}plus(t){let e=new Uint32Array(4);return e[3]=this.buffer[3]+t.buffer[3]>>>0,e[2]=this.buffer[2]+t.buffer[2]>>>0,e[1]=this.buffer[1]+t.buffer[1]>>>0,e[0]=this.buffer[0]+t.buffer[0]>>>0,e[0]<this.buffer[0]>>>0&&++e[1],e[1]<this.buffer[1]>>>0&&++e[2],e[2]<this.buffer[2]>>>0&&++e[3],this.buffer[3]=e[3],this.buffer[2]=e[2],this.buffer[1]=e[1],this.buffer[0]=e[0],this}hex(){return`${ki(this.buffer[3])} ${ki(this.buffer[2])} ${ki(this.buffer[1])} ${ki(this.buffer[0])}`}static multiply(t,e){let n=new Ui(new Uint32Array(t.buffer));return n.times(e)}static add(t,e){let n=new Ui(new Uint32Array(t.buffer));return n.plus(e)}static from(t,e=new Uint32Array(4)){return Ui.fromString("string"===typeof t?t:t.toString(),e)}static fromNumber(t,e=new Uint32Array(4)){return Ui.fromString(t.toString(),e)}static fromString(t,e=new Uint32Array(4)){const n=t.startsWith("-"),r=t.length;let i=new Ui(e);for(let o=n?1:0;o<r;){const e=Di<r-o?Di:r-o,n=new Ui(new Uint32Array([parseInt(t.substr(o,e),10),0,0,0])),s=new Ui(new Uint32Array([Ni[e],0,0,0]));i.times(s),i.plus(n),o+=e}return n?i.negate():i}static convertArray(t){const e=new Uint32Array(4*t.length);for(let n=-1,r=t.length;++n<r;)Ui.from(t[n],new Uint32Array(e.buffer,e.byteOffset+16*n,4));return e}}class Ri extends nt{constructor(t,e,n,r){super(),this.nodesIndex=-1,this.buffersIndex=-1,this.bytes=t,this.nodes=e,this.buffers=n,this.dictionaries=r}visit(t){return super.visit(t instanceof lr?t.type:t)}visitNull(t,{length:e}=this.nextFieldNode()){return ie.Null(t,0,e)}visitBool(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Bool(t,0,e,n,this.readNullBitmap(t,n),this.readData(t))}visitInt(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Int(t,0,e,n,this.readNullBitmap(t,n),this.readData(t))}visitFloat(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Float(t,0,e,n,this.readNullBitmap(t,n),this.readData(t))}visitUtf8(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Utf8(t,0,e,n,this.readNullBitmap(t,n),this.readOffsets(t),this.readData(t))}visitBinary(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Binary(t,0,e,n,this.readNullBitmap(t,n),this.readOffsets(t),this.readData(t))}visitFixedSizeBinary(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.FixedSizeBinary(t,0,e,n,this.readNullBitmap(t,n),this.readData(t))}visitDate(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Date(t,0,e,n,this.readNullBitmap(t,n),this.readData(t))}visitTimestamp(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Timestamp(t,0,e,n,this.readNullBitmap(t,n),this.readData(t))}visitTime(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Time(t,0,e,n,this.readNullBitmap(t,n),this.readData(t))}visitDecimal(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Decimal(t,0,e,n,this.readNullBitmap(t,n),this.readData(t))}visitList(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.List(t,0,e,n,this.readNullBitmap(t,n),this.readOffsets(t),this.visit(t.children[0]))}visitStruct(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Struct(t,0,e,n,this.readNullBitmap(t,n),this.visitMany(t.children))}visitUnion(t){return t.mode===z.Sparse?this.visitSparseUnion(t):this.visitDenseUnion(t)}visitDenseUnion(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Union(t,0,e,n,this.readNullBitmap(t,n),this.readTypeIds(t),this.readOffsets(t),this.visitMany(t.children))}visitSparseUnion(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Union(t,0,e,n,this.readNullBitmap(t,n),this.readTypeIds(t),this.visitMany(t.children))}visitDictionary(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Dictionary(t,0,e,n,this.readNullBitmap(t,n),this.readData(t.indices),this.readDictionary(t))}visitInterval(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Interval(t,0,e,n,this.readNullBitmap(t,n),this.readData(t))}visitFixedSizeList(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.FixedSizeList(t,0,e,n,this.readNullBitmap(t,n),this.visit(t.children[0]))}visitMap(t,{length:e,nullCount:n}=this.nextFieldNode()){return ie.Map(t,0,e,n,this.readNullBitmap(t,n),this.readOffsets(t),this.visit(t.children[0]))}nextFieldNode(){return this.nodes[++this.nodesIndex]}nextBufferRange(){return this.buffers[++this.buffersIndex]}readNullBitmap(t,e,n=this.nextBufferRange()){return e>0&&this.readData(t,n)||new Uint8Array(0)}readOffsets(t,e){return this.readData(t,e)}readTypeIds(t,e){return this.readData(t,e)}readData(t,{length:e,offset:n}=this.nextBufferRange()){return this.bytes.subarray(n,n+e)}readDictionary(t){return this.dictionaries.get(t.id)}}class Vi extends Ri{constructor(t,e,n,r){super(new Uint8Array(0),e,n,r),this.sources=t}readNullBitmap(t,e,{offset:n}=this.nextBufferRange()){return e<=0?new Uint8Array(0):J(this.sources[n])}readOffsets(t,{offset:e}=this.nextBufferRange()){return Object(ne["toArrayBufferView"])(Uint8Array,Object(ne["toArrayBufferView"])(Int32Array,this.sources[e]))}readTypeIds(t,{offset:e}=this.nextBufferRange()){return Object(ne["toArrayBufferView"])(Uint8Array,Object(ne["toArrayBufferView"])(t.ArrayType,this.sources[e]))}readData(t,{offset:e}=this.nextBufferRange()){const{sources:n}=this;return Ot.isTimestamp(t)||(Ot.isInt(t)||Ot.isTime(t))&&64===t.bitWidth||Ot.isDate(t)&&t.unit===R.MILLISECOND?Object(ne["toArrayBufferView"])(Uint8Array,Fi.convertArray(n[e])):Ot.isDecimal(t)?Object(ne["toArrayBufferView"])(Uint8Array,Ui.convertArray(n[e])):Ot.isBinary(t)||Ot.isFixedSizeBinary(t)?$i(n[e]):Ot.isBool(t)?J(n[e]):Ot.isUtf8(t)?Object(mn["b"])(n[e].join("")):Object(ne["toArrayBufferView"])(Uint8Array,Object(ne["toArrayBufferView"])(t.ArrayType,n[e].map(t=>+t)))}}function $i(t){const e=t.join(""),n=new Uint8Array(e.length/2);for(let r=0;r<e.length;r+=2)n[r>>1]=parseInt(e.substr(r,2),16);return n}var zi=yi["a"].Long,Yi=D.apache.arrow.flatbuf.Null,Hi=D.apache.arrow.flatbuf.Int,Wi=D.apache.arrow.flatbuf.FloatingPoint,Gi=D.apache.arrow.flatbuf.Binary,qi=D.apache.arrow.flatbuf.Bool,Ki=D.apache.arrow.flatbuf.Utf8,Xi=D.apache.arrow.flatbuf.Decimal,Ji=D.apache.arrow.flatbuf.Date,Zi=D.apache.arrow.flatbuf.Time,Qi=D.apache.arrow.flatbuf.Timestamp,to=D.apache.arrow.flatbuf.Interval,eo=D.apache.arrow.flatbuf.List,no=D.apache.arrow.flatbuf.Struct_,ro=D.apache.arrow.flatbuf.Union,io=D.apache.arrow.flatbuf.DictionaryEncoding,oo=D.apache.arrow.flatbuf.FixedSizeBinary,so=D.apache.arrow.flatbuf.FixedSizeList,ao=D.apache.arrow.flatbuf.Map;class co extends nt{visit(t,e){return null==t||null==e?void 0:super.visit(t,e)}visitNull(t,e){return Yi.startNull(e),Yi.endNull(e)}visitInt(t,e){return Hi.startInt(e),Hi.addBitWidth(e,t.bitWidth),Hi.addIsSigned(e,t.isSigned),Hi.endInt(e)}visitFloat(t,e){return Wi.startFloatingPoint(e),Wi.addPrecision(e,t.precision),Wi.endFloatingPoint(e)}visitBinary(t,e){return Gi.startBinary(e),Gi.endBinary(e)}visitBool(t,e){return qi.startBool(e),qi.endBool(e)}visitUtf8(t,e){return Ki.startUtf8(e),Ki.endUtf8(e)}visitDecimal(t,e){return Xi.startDecimal(e),Xi.addScale(e,t.scale),Xi.addPrecision(e,t.precision),Xi.endDecimal(e)}visitDate(t,e){return Ji.startDate(e),Ji.addUnit(e,t.unit),Ji.endDate(e)}visitTime(t,e){return Zi.startTime(e),Zi.addUnit(e,t.unit),Zi.addBitWidth(e,t.bitWidth),Zi.endTime(e)}visitTimestamp(t,e){const n=t.timezone&&e.createString(t.timezone)||void 0;return Qi.startTimestamp(e),Qi.addUnit(e,t.unit),void 0!==n&&Qi.addTimezone(e,n),Qi.endTimestamp(e)}visitInterval(t,e){return to.startInterval(e),to.addUnit(e,t.unit),to.endInterval(e)}visitList(t,e){return eo.startList(e),eo.endList(e)}visitStruct(t,e){return no.startStruct_(e),no.endStruct_(e)}visitUnion(t,e){ro.startTypeIdsVector(e,t.typeIds.length);const n=ro.createTypeIdsVector(e,t.typeIds);return ro.startUnion(e),ro.addMode(e,t.mode),ro.addTypeIds(e,n),ro.endUnion(e)}visitDictionary(t,e){const n=this.visit(t.indices,e);return io.startDictionaryEncoding(e),io.addId(e,new zi(t.id,0)),io.addIsOrdered(e,t.isOrdered),void 0!==n&&io.addIndexType(e,n),io.endDictionaryEncoding(e)}visitFixedSizeBinary(t,e){return oo.startFixedSizeBinary(e),oo.addByteWidth(e,t.byteWidth),oo.endFixedSizeBinary(e)}visitFixedSizeList(t,e){return so.startFixedSizeList(e),so.addListSize(e,t.listSize),so.endFixedSizeList(e)}visitMap(t,e){return ao.startMap(e),ao.addKeysSorted(e,t.keysSorted),ao.endMap(e)}}const uo=new co;function lo(t,e=new Map){return new ur(po(t,e),_o(t["customMetadata"]),e)}function fo(t){return new Po(t["count"],vo(t["columns"]),bo(t["columns"]))}function ho(t){return new Fo(fo(t["data"]),t["id"],t["isDelta"])}function po(t,e){return(t["fields"]||[]).filter(Boolean).map(t=>lr.fromJSON(t,e))}function yo(t,e){return(t["children"]||[]).filter(Boolean).map(t=>lr.fromJSON(t,e))}function vo(t){return(t||[]).reduce((t,e)=>[...t,new Ro(e["count"],mo(e["VALIDITY"])),...vo(e["children"])],[])}function bo(t,e=[]){for(let n=-1,r=(t||[]).length;++n<r;){const r=t[n];r["VALIDITY"]&&e.push(new Uo(e.length,r["VALIDITY"].length)),r["TYPE"]&&e.push(new Uo(e.length,r["TYPE"].length)),r["OFFSET"]&&e.push(new Uo(e.length,r["OFFSET"].length)),r["DATA"]&&e.push(new Uo(e.length,r["DATA"].length)),e=bo(r["children"],e)}return e}function mo(t){return(t||[]).reduce((t,e)=>t+ +(0===e),0)}function go(t,e){let n,r,i,o,s,a;return e&&(o=t["dictionary"])?e.has(n=o["id"])?(r=(r=o["indexType"])?wo(r):new Tt,a=new te(e.get(n),r,n,o["isOrdered"]),i=new lr(t["name"],a,t["nullable"],_o(t["customMetadata"]))):(r=(r=o["indexType"])?wo(r):new Tt,e.set(n,s=Oo(t,yo(t,e))),a=new te(s,r,n,o["isOrdered"]),i=new lr(t["name"],a,t["nullable"],_o(t["customMetadata"]))):(s=Oo(t,yo(t,e)),i=new lr(t["name"],s,t["nullable"],_o(t["customMetadata"]))),i||null}function _o(t){return new Map(Object.entries(t||{}))}function wo(t){return new xt(t["isSigned"],t["bitWidth"])}function Oo(t,e){const n=t["type"]["name"];switch(n){case"NONE":return new St;case"null":return new St;case"binary":return new Pt;case"utf8":return new Ft;case"bool":return new Ut;case"list":return new Gt((e||[])[0]);case"struct":return new qt(e||[]);case"struct_":return new qt(e||[])}switch(n){case"int":{const e=t["type"];return new xt(e["isSigned"],e["bitWidth"])}case"floatingpoint":{const e=t["type"];return new kt($[e["precision"]])}case"decimal":{const e=t["type"];return new Rt(e["scale"],e["precision"])}case"date":{const e=t["type"];return new Vt(R[e["unit"]])}case"time":{const e=t["type"];return new Yt(V[e["unit"]],e["bitWidth"])}case"timestamp":{const e=t["type"];return new Ht(V[e["unit"]],e["timezone"])}case"interval":{const e=t["type"];return new Wt(Y[e["unit"]])}case"union":{const n=t["type"];return new Kt(z[n["mode"]],n["typeIds"]||[],e||[])}case"fixedsizebinary":{const e=t["type"];return new Xt(e["byteWidth"])}case"fixedsizelist":{const n=t["type"];return new Jt(n["listSize"],(e||[])[0])}case"map":{const n=t["type"];return new Zt((e||[])[0],n["keysSorted"])}}throw new Error(`Unrecognized type: "${n}"`)}var So=yi["a"].Long,xo=yi["a"].Builder,Io=yi["a"].ByteBuffer,Ao=D.apache.arrow.flatbuf.Type,To=D.apache.arrow.flatbuf.Field,jo=D.apache.arrow.flatbuf.Schema,Eo=D.apache.arrow.flatbuf.Buffer,Bo=N.apache.arrow.flatbuf.Message,Co=D.apache.arrow.flatbuf.KeyValue,Mo=N.apache.arrow.flatbuf.FieldNode,ko=D.apache.arrow.flatbuf.Endianness,Do=N.apache.arrow.flatbuf.RecordBatch,No=N.apache.arrow.flatbuf.DictionaryBatch;class Lo{constructor(t,e,n,r){this._version=e,this._headerType=n,this.body=new Uint8Array(0),r&&(this._createHeader=()=>r),this._bodyLength="number"===typeof t?t:t.low}static fromJSON(t,e){const n=new Lo(0,W.V4,e);return n._createHeader=Vo(t,e),n}static decode(t){t=new Io(Object(ne["toUint8Array"])(t));const e=Bo.getRootAsMessage(t),n=e.bodyLength(),r=e.version(),i=e.headerType(),o=new Lo(n,r,i);return o._createHeader=$o(e,i),o}static encode(t){let e=new xo,n=-1;return t.isSchema()?n=ur.encode(e,t.header()):t.isRecordBatch()?n=Po.encode(e,t.header()):t.isDictionaryBatch()&&(n=Fo.encode(e,t.header())),Bo.startMessage(e),Bo.addVersion(e,W.V4),Bo.addHeader(e,n),Bo.addHeaderType(e,t.headerType),Bo.addBodyLength(e,new So(t.bodyLength,0)),Bo.finishMessageBuffer(e,Bo.endMessage(e)),e.asUint8Array()}static from(t,e=0){if(t instanceof ur)return new Lo(0,W.V4,H.Schema,t);if(t instanceof Po)return new Lo(e,W.V4,H.RecordBatch,t);if(t instanceof Fo)return new Lo(e,W.V4,H.DictionaryBatch,t);throw new Error("Unrecognized Message header: "+t)}get type(){return this.headerType}get version(){return this._version}get headerType(){return this._headerType}get bodyLength(){return this._bodyLength}header(){return this._createHeader()}isSchema(){return this.headerType===H.Schema}isRecordBatch(){return this.headerType===H.RecordBatch}isDictionaryBatch(){return this.headerType===H.DictionaryBatch}}class Po{get nodes(){return this._nodes}get length(){return this._length}get buffers(){return this._buffers}constructor(t,e,n){this._nodes=e,this._buffers=n,this._length="number"===typeof t?t:t.low}}class Fo{get id(){return this._id}get data(){return this._data}get isDelta(){return this._isDelta}get length(){return this.data.length}get nodes(){return this.data.nodes}get buffers(){return this.data.buffers}constructor(t,e,n=!1){this._data=t,this._isDelta=n,this._id="number"===typeof e?e:e.low}}class Uo{constructor(t,e){this.offset="number"===typeof t?t:t.low,this.length="number"===typeof e?e:e.low}}class Ro{constructor(t,e){this.length="number"===typeof t?t:t.low,this.nullCount="number"===typeof e?e:e.low}}function Vo(t,e){return()=>{switch(e){case H.Schema:return ur.fromJSON(t);case H.RecordBatch:return Po.fromJSON(t);case H.DictionaryBatch:return Fo.fromJSON(t)}throw new Error(`Unrecognized Message type: { name: ${H[e]}, type: ${e} }`)}}function $o(t,e){return()=>{switch(e){case H.Schema:return ur.decode(t.header(new jo));case H.RecordBatch:return Po.decode(t.header(new Do),t.version());case H.DictionaryBatch:return Fo.decode(t.header(new No),t.version())}throw new Error(`Unrecognized Message type: { name: ${H[e]}, type: ${e} }`)}}function zo(t,e=new Map){const n=Xo(t,e);return new ur(n,Qo(t),e)}function Yo(t,e=W.V4){return new Po(t.length(),qo(t),Ko(t,e))}function Ho(t,e=W.V4){return new Fo(Po.decode(t.data(),e),t.id(),t.isDelta())}function Wo(t){return new Uo(t.offset(),t.length())}function Go(t){return new Ro(t.length(),t.nullCount())}function qo(t){const e=[];for(let n,r=-1,i=-1,o=t.nodesLength();++r<o;)(n=t.nodes(r))&&(e[++i]=Ro.decode(n));return e}function Ko(t,e){const n=[];for(let r,i=-1,o=-1,s=t.buffersLength();++i<s;)(r=t.buffers(i))&&(e<W.V4&&(r.bb_pos+=8*(i+1)),n[++o]=Uo.decode(r));return n}function Xo(t,e){const n=[];for(let r,i=-1,o=-1,s=t.fieldsLength();++i<s;)(r=t.fields(i))&&(n[++o]=lr.decode(r,e));return n}function Jo(t,e){const n=[];for(let r,i=-1,o=-1,s=t.childrenLength();++i<s;)(r=t.children(i))&&(n[++o]=lr.decode(r,e));return n}function Zo(t,e){let n,r,i,o,s,a;return e&&(a=t.dictionary())?e.has(n=a.id().low)?(o=(o=a.indexType())?ts(o):new Tt,s=new te(e.get(n),o,n,a.isOrdered()),r=new lr(t.name(),s,t.nullable(),Qo(t))):(o=(o=a.indexType())?ts(o):new Tt,e.set(n,i=es(t,Jo(t,e))),s=new te(i,o,n,a.isOrdered()),r=new lr(t.name(),s,t.nullable(),Qo(t))):(i=es(t,Jo(t,e)),r=new lr(t.name(),i,t.nullable(),Qo(t))),r||null}function Qo(t){const e=new Map;if(t)for(let n,r,i=-1,o=0|t.customMetadataLength();++i<o;)(n=t.customMetadata(i))&&null!=(r=n.key())&&e.set(r,n.value());return e}function ts(t){return new xt(t.isSigned(),t.bitWidth())}function es(t,e){const n=t.typeType();switch(n){case Ao.NONE:return new St;case Ao.Null:return new St;case Ao.Binary:return new Pt;case Ao.Utf8:return new Ft;case Ao.Bool:return new Ut;case Ao.List:return new Gt((e||[])[0]);case Ao.Struct_:return new qt(e||[])}switch(n){case Ao.Int:{const e=t.type(new D.apache.arrow.flatbuf.Int);return new xt(e.isSigned(),e.bitWidth())}case Ao.FloatingPoint:{const e=t.type(new D.apache.arrow.flatbuf.FloatingPoint);return new kt(e.precision())}case Ao.Decimal:{const e=t.type(new D.apache.arrow.flatbuf.Decimal);return new Rt(e.scale(),e.precision())}case Ao.Date:{const e=t.type(new D.apache.arrow.flatbuf.Date);return new Vt(e.unit())}case Ao.Time:{const e=t.type(new D.apache.arrow.flatbuf.Time);return new Yt(e.unit(),e.bitWidth())}case Ao.Timestamp:{const e=t.type(new D.apache.arrow.flatbuf.Timestamp);return new Ht(e.unit(),e.timezone())}case Ao.Interval:{const e=t.type(new D.apache.arrow.flatbuf.Interval);return new Wt(e.unit())}case Ao.Union:{const n=t.type(new D.apache.arrow.flatbuf.Union);return new Kt(n.mode(),n.typeIdsArray()||[],e||[])}case Ao.FixedSizeBinary:{const e=t.type(new D.apache.arrow.flatbuf.FixedSizeBinary);return new Xt(e.byteWidth())}case Ao.FixedSizeList:{const n=t.type(new D.apache.arrow.flatbuf.FixedSizeList);return new Jt(n.listSize(),(e||[])[0])}case Ao.Map:{const n=t.type(new D.apache.arrow.flatbuf.Map);return new Zt((e||[])[0],n.keysSorted())}}throw new Error(`Unrecognized type: "${Ao[n]}" (${n})`)}function ns(t,e){const n=e.fields.map(e=>lr.encode(t,e));jo.startFieldsVector(t,n.length);const r=jo.createFieldsVector(t,n),i=e.metadata&&e.metadata.size>0?jo.createCustomMetadataVector(t,[...e.metadata].map(([e,n])=>{const r=t.createString(""+e),i=t.createString(""+n);return Co.startKeyValue(t),Co.addKey(t,r),Co.addValue(t,i),Co.endKeyValue(t)})):-1;return jo.startSchema(t),jo.addFields(t,r),jo.addEndianness(t,cs?ko.Little:ko.Big),-1!==i&&jo.addCustomMetadata(t,i),jo.endSchema(t)}function rs(t,e){let n=-1,r=-1,i=-1,o=e.type,s=e.typeId;Ot.isDictionary(o)?(s=o.dictionary.typeId,i=uo.visit(o,t),r=uo.visit(o.dictionary,t)):r=uo.visit(o,t);const a=(o.children||[]).map(e=>lr.encode(t,e)),c=To.createChildrenVector(t,a),u=e.metadata&&e.metadata.size>0?To.createCustomMetadataVector(t,[...e.metadata].map(([e,n])=>{const r=t.createString(""+e),i=t.createString(""+n);return Co.startKeyValue(t),Co.addKey(t,r),Co.addValue(t,i),Co.endKeyValue(t)})):-1;return e.name&&(n=t.createString(e.name)),To.startField(t),To.addType(t,r),To.addTypeType(t,s),To.addChildren(t,c),To.addNullable(t,!!e.nullable),-1!==n&&To.addName(t,n),-1!==i&&To.addDictionary(t,i),-1!==u&&To.addCustomMetadata(t,u),To.endField(t)}function is(t,e){const n=e.nodes||[],r=e.buffers||[];Do.startNodesVector(t,n.length),n.slice().reverse().forEach(e=>Ro.encode(t,e));const i=t.endVector();Do.startBuffersVector(t,r.length),r.slice().reverse().forEach(e=>Uo.encode(t,e));const o=t.endVector();return Do.startRecordBatch(t),Do.addLength(t,new So(e.length,0)),Do.addNodes(t,i),Do.addBuffers(t,o),Do.endRecordBatch(t)}function os(t,e){const n=Po.encode(t,e.data);return No.startDictionaryBatch(t),No.addId(t,new So(e.id,0)),No.addIsDelta(t,e.isDelta),No.addData(t,n),No.endDictionaryBatch(t)}function ss(t,e){return Mo.createFieldNode(t,new So(e.length,0),new So(e.nullCount,0))}function as(t,e){return Eo.createBuffer(t,new So(e.offset,0),new So(e.length,0))}lr["encode"]=rs,lr["decode"]=Zo,lr["fromJSON"]=go,ur["encode"]=ns,ur["decode"]=zo,ur["fromJSON"]=lo,Po["encode"]=is,Po["decode"]=Yo,Po["fromJSON"]=fo,Fo["encode"]=os,Fo["decode"]=Ho,Fo["fromJSON"]=ho,Ro["encode"]=ss,Ro["decode"]=Go,Uo["encode"]=as,Uo["decode"]=Wo;const cs=function(){const t=new ArrayBuffer(2);return new DataView(t).setInt16(0,256,!0),256===new Int16Array(t)[0]}();var us=yi["a"].ByteBuffer;const ls=t=>`Expected ${H[t]} Message in stream, but was null or length 0.`,fs=t=>`Header pointer of flatbuffer-encoded ${H[t]} Message is null or length 0.`,ds=(t,e)=>`Expected to read ${t} metadata bytes, but only read ${e}.`,hs=(t,e)=>`Expected to read ${t} bytes for message body, but only read ${e}.`;class ps{constructor(t){this.source=t instanceof Ai?t:new Ai(t)}[Symbol.iterator](){return this}next(){let t;return(t=this.readMetadataLength()).done||-1===t.value&&(t=this.readMetadataLength()).done||(t=this.readMetadata(t.value)).done?xi["c"]:t}throw(t){return this.source.throw(t)}return(t){return this.source.return(t)}readMessage(t){let e;if((e=this.next()).done)return null;if(null!=t&&e.value.headerType!==t)throw new Error(ls(t));return e.value}readMessageBody(t){if(t<=0)return new Uint8Array(0);const e=Object(ne["toUint8Array"])(this.source.read(t));if(e.byteLength<t)throw new Error(hs(t,e.byteLength));return e.byteOffset%8===0&&e.byteOffset+e.byteLength<=e.buffer.byteLength?e:e.slice()}readSchema(t=!1){const e=H.Schema,n=this.readMessage(e),r=n&&n.header();if(t&&!r)throw new Error(fs(e));return r}readMetadataLength(){const t=this.source.read(bs),e=t&&new us(t),n=e&&e.readInt32(0)||0;return{done:0===n,value:n}}readMetadata(t){const e=this.source.read(t);if(!e)return xi["c"];if(e.byteLength<t)throw new Error(ds(t,e.byteLength));return{done:!1,value:Lo.decode(e)}}}class ys{constructor(t,e){this.source=t instanceof Ti?t:Object(ae["h"])(t)?new Ci(t,e):new Ti(t)}[Symbol.asyncIterator](){return this}async next(){let t;return(t=await this.readMetadataLength()).done||-1===t.value&&(t=await this.readMetadataLength()).done||(t=await this.readMetadata(t.value)).done?xi["c"]:t}async throw(t){return await this.source.throw(t)}async return(t){return await this.source.return(t)}async readMessage(t){let e;if((e=await this.next()).done)return null;if(null!=t&&e.value.headerType!==t)throw new Error(ls(t));return e.value}async readMessageBody(t){if(t<=0)return new Uint8Array(0);const e=Object(ne["toUint8Array"])(await this.source.read(t));if(e.byteLength<t)throw new Error(hs(t,e.byteLength));return e.byteOffset%8===0&&e.byteOffset+e.byteLength<=e.buffer.byteLength?e:e.slice()}async readSchema(t=!1){const e=H.Schema,n=await this.readMessage(e),r=n&&n.header();if(t&&!r)throw new Error(fs(e));return r}async readMetadataLength(){const t=await this.source.read(bs),e=t&&new us(t),n=e&&e.readInt32(0)||0;return{done:0===n,value:n}}async readMetadata(t){const e=await this.source.read(t);if(!e)return xi["c"];if(e.byteLength<t)throw new Error(ds(t,e.byteLength));return{done:!1,value:Lo.decode(e)}}}class vs extends ps{constructor(t){super(new Uint8Array(0)),this._schema=!1,this._body=[],this._batchIndex=0,this._dictionaryIndex=0,this._json=t instanceof xi["a"]?t:new xi["a"](t)}next(){const{_json:t}=this;if(!this._schema){this._schema=!0;const e=Lo.fromJSON(t.schema,H.Schema);return{done:!1,value:e}}if(this._dictionaryIndex<t.dictionaries.length){const e=t.dictionaries[this._dictionaryIndex++];this._body=e["data"]["columns"];const n=Lo.fromJSON(e,H.DictionaryBatch);return{done:!1,value:n}}if(this._batchIndex<t.batches.length){const e=t.batches[this._batchIndex++];this._body=e["columns"];const n=Lo.fromJSON(e,H.RecordBatch);return{done:!1,value:n}}return this._body=[],xi["c"]}readMessageBody(t){return e(this._body);function e(t){return(t||[]).reduce((t,n)=>[...t,...n["VALIDITY"]&&[n["VALIDITY"]]||[],...n["TYPE"]&&[n["TYPE"]]||[],...n["OFFSET"]&&[n["OFFSET"]]||[],...n["DATA"]&&[n["DATA"]]||[],...e(n["children"])],[])}}readMessage(t){let e;if((e=this.next()).done)return null;if(null!=t&&e.value.headerType!==t)throw new Error(ls(t));return e.value}readSchema(){const t=H.Schema,e=this.readMessage(t),n=e&&e.header();if(!e||!n)throw new Error(fs(t));return n}}const bs=4,ms="ARROW1",gs=new Uint8Array(ms.length);for(let cl=0;cl<ms.length;cl+=1)gs[cl]=ms.charCodeAt(cl);function _s(t,e=0){for(let n=-1,r=gs.length;++n<r;)if(gs[n]!==t[e+n])return!1;return!0}const ws=gs.length,Os=ws+bs,Ss=2*ws+bs;class xs extends nt{constructor(){super(),this._byteLength=0,this._nodes=[],this._buffers=[],this._bufferRegions=[]}static assemble(...t){const e=new xs,n=rr(lu,t),[r=e]=e.visitMany(n);return r}visit(t){if(!Ot.isDictionary(t.type)){const{data:e,length:n,nullCount:r}=t;if(n>2147483647)throw new RangeError("Cannot write arrays larger than 2^31 - 1 in length");Ot.isNull(t.type)||Is.call(this,r<=0?new Uint8Array(0):X(e.offset,n,e.nullBitmap)),this.nodes.push(new Ro(n,r))}return super.visit(t)}visitNull(t){return this}visitDictionary(t){return this.visit(t.indices)}get nodes(){return this._nodes}get buffers(){return this._buffers}get byteLength(){return this._byteLength}get bufferRegions(){return this._bufferRegions}}function Is(t){const e=t.byteLength+7&-8;return this.buffers.push(t),this.bufferRegions.push(new Uo(this._byteLength,e)),this._byteLength+=e,this}function As(t){const{type:e,length:n,typeIds:r,valueOffsets:i}=t;if(Is.call(this,r),e.mode===z.Sparse)return Cs.call(this,t);if(e.mode===z.Dense){if(t.offset<=0)return Is.call(this,i),Cs.call(this,t);{const o=r.reduce((t,e)=>Math.max(t,e),r[0]),s=new Int32Array(o+1),a=new Int32Array(o+1).fill(-1),c=new Int32Array(n),u=Object(ne["rebaseValueOffsets"])(-i[0],n,i);for(let t,e,i=-1;++i<n;)-1===(e=a[t=r[i]])&&(e=a[t]=u[t]),c[i]=u[i]-e,++s[t];Is.call(this,c);for(let r,i=-1,l=e.children.length;++i<l;)if(r=t.getChildAt(i)){const t=e.typeIds[i],o=Math.min(n,s[t]);this.visit(r.slice(a[t],o))}}}return this}function Ts(t){let e;return t.nullCount>=t.length?Is.call(this,new Uint8Array(0)):(e=t.values)instanceof Uint8Array?Is.call(this,X(t.offset,t.length,e)):Is.call(this,J(t))}function js(t){return Is.call(this,t.values.subarray(0,t.length*t.stride))}function Es(t){const{length:e,values:n,valueOffsets:r}=t,i=r[0],o=r[e],s=Math.min(o-i,n.byteLength-i);return Is.call(this,Object(ne["rebaseValueOffsets"])(-r[0],e,r)),Is.call(this,n.subarray(i,i+s)),this}function Bs(t){const{length:e,valueOffsets:n}=t;return n&&Is.call(this,Object(ne["rebaseValueOffsets"])(n[0],e,n)),this.visit(t.getChildAt(0))}function Cs(t){return this.visitMany(t.type.children.map((e,n)=>t.getChildAt(n)).filter(Boolean))[0]}xs.prototype.visitBool=Ts,xs.prototype.visitInt=js,xs.prototype.visitFloat=js,xs.prototype.visitUtf8=Es,xs.prototype.visitBinary=Es,xs.prototype.visitFixedSizeBinary=js,xs.prototype.visitDate=js,xs.prototype.visitTimestamp=js,xs.prototype.visitTime=js,xs.prototype.visitDecimal=js,xs.prototype.visitList=Bs,xs.prototype.visitStruct=Cs,xs.prototype.visitUnion=As,xs.prototype.visitInterval=js,xs.prototype.visitFixedSizeList=Bs,xs.prototype.visitMap=Bs;class Ms extends xi["d"]{constructor(t){super(),this._position=0,this._started=!1,this._sink=new Ii,this._schema=null,this._dictionaryBlocks=[],this._recordBatchBlocks=[],this._dictionaryDeltaOffsets=new Map,Object(ae["k"])(t)||(t={autoDestroy:!0,writeLegacyIpcFormat:!1}),this._autoDestroy="boolean"!==typeof t.autoDestroy||t.autoDestroy,this._writeLegacyIpcFormat="boolean"===typeof t.writeLegacyIpcFormat&&t.writeLegacyIpcFormat}static throughNode(t){throw new Error('"throughNode" not available in this environment')}static throughDOM(t,e){throw new Error('"throughDOM" not available in this environment')}toString(t=!1){return this._sink.toString(t)}toUint8Array(t=!1){return this._sink.toUint8Array(t)}writeAll(t){return Object(ae["l"])(t)?t.then(t=>this.writeAll(t)):Object(ae["f"])(t)?Ls(this,t):Ns(this,t)}get closed(){return this._sink.closed}[Symbol.asyncIterator](){return this._sink[Symbol.asyncIterator]()}toDOMStream(t){return this._sink.toDOMStream(t)}toNodeStream(t){return this._sink.toNodeStream(t)}close(){return this.reset()._sink.close()}abort(t){return this.reset()._sink.abort(t)}finish(){return this._autoDestroy?this.close():this.reset(this._sink,this._schema),this}reset(t=this._sink,e=null){return t===this._sink||t instanceof Ii?this._sink=t:(this._sink=new Ii,t&&Object(ae["o"])(t)?this.toDOMStream({type:"bytes"}).pipeTo(t):t&&Object(ae["p"])(t)&&this.toNodeStream({objectMode:!1}).pipe(t)),this._started&&this._schema&&this._writeFooter(this._schema),this._started=!1,this._dictionaryBlocks=[],this._recordBatchBlocks=[],this._dictionaryDeltaOffsets=new Map,e&&e.compareTo(this._schema)||(null===e?(this._position=0,this._schema=null):(this._started=!0,this._schema=e,this._writeSchema(e))),this}write(t){let e=null;if(!this._sink)throw new Error("RecordBatchWriter is closed");if(null===t||void 0===t)return this.finish()&&void 0;if(t instanceof au&&!(e=t.schema))return this.finish()&&void 0;if(t instanceof lu&&!(e=t.schema))return this.finish()&&void 0;if(e&&!e.compareTo(this._schema)){if(this._started&&this._autoDestroy)return this.close();this.reset(this._sink,e)}t instanceof lu?t instanceof fu||this._writeRecordBatch(t):t instanceof au?this.writeAll(t.chunks):Object(ae["i"])(t)&&this.writeAll(t)}_writeMessage(t,e=8){const n=e-1,r=Lo.encode(t),i=r.byteLength,o=this._writeLegacyIpcFormat?4:8,s=i+o+n&~n,a=s-i-o;return t.headerType===H.RecordBatch?this._recordBatchBlocks.push(new Si(s,t.bodyLength,this._position)):t.headerType===H.DictionaryBatch&&this._dictionaryBlocks.push(new Si(s,t.bodyLength,this._position)),this._writeLegacyIpcFormat||this._write(Int32Array.of(-1)),this._write(Int32Array.of(s-o)),i>0&&this._write(r),this._writePadding(a)}_write(t){if(this._started){const e=Object(ne["toUint8Array"])(t);e&&e.byteLength>0&&(this._sink.write(e),this._position+=e.byteLength)}return this}_writeSchema(t){return this._writeMessage(Lo.from(t))}_writeFooter(t){return this._writeLegacyIpcFormat?this._write(Int32Array.of(0)):this._write(Int32Array.of(-1,0))}_writeMagic(){return this._write(gs)}_writePadding(t){return t>0?this._write(new Uint8Array(t)):this}_writeRecordBatch(t){const{byteLength:e,nodes:n,bufferRegions:r,buffers:i}=xs.assemble(t),o=new Po(t.length,n,r),s=Lo.from(o,e);return this._writeDictionaries(t)._writeMessage(s)._writeBodyBuffers(i)}_writeDictionaryBatch(t,e,n=!1){this._dictionaryDeltaOffsets.set(e,t.length+(this._dictionaryDeltaOffsets.get(e)||0));const{byteLength:r,nodes:i,bufferRegions:o,buffers:s}=xs.assemble(t),a=new Po(t.length,i,o),c=new Fo(a,e,n),u=Lo.from(c,r);return this._writeMessage(u)._writeBodyBuffers(s)}_writeBodyBuffers(t){let e,n,r;for(let i=-1,o=t.length;++i<o;)(e=t[i])&&(n=e.byteLength)>0&&(this._write(e),(r=(n+7&-8)-n)>0&&this._writePadding(r));return this}_writeDictionaries(t){for(let[e,n]of t.dictionaries){let t=this._dictionaryDeltaOffsets.get(e)||0;if(0===t||(n=n.slice(t)).length>0){const r="chunks"in n?n.chunks:[n];for(const n of r)this._writeDictionaryBatch(n,e,t>0),t+=n.length}}return this}}class ks extends Ms{static writeAll(t,e){const n=new ks(e);return Object(ae["l"])(t)?t.then(t=>n.writeAll(t)):Object(ae["f"])(t)?Ls(n,t):Ns(n,t)}}class Ds extends Ms{constructor(){super(),this._autoDestroy=!0}static writeAll(t){const e=new Ds;return Object(ae["l"])(t)?t.then(t=>e.writeAll(t)):Object(ae["f"])(t)?Ls(e,t):Ns(e,t)}_writeSchema(t){return this._writeMagic()._writePadding(2)}_writeFooter(t){const e=wi.encode(new wi(t,W.V4,this._recordBatchBlocks,this._dictionaryBlocks));return super._writeFooter(t)._write(e)._write(Int32Array.of(e.byteLength))._writeMagic()}}function Ns(t,e){let n=e;e instanceof au&&(n=e.chunks,t.reset(void 0,e.schema));for(const r of n)t.write(r);return t.finish()}async function Ls(t,e){for await(const n of e)t.write(n);return t.finish()}const Ps=new Uint8Array(0),Fs=t=>[Ps,Ps,new Uint8Array(t),Ps];function Us(t,e,n=e.reduce((t,e)=>Math.max(t,e.length),0)){let r,i,o=-1,s=e.length;const a=[...t.fields],c=[],u=(n+63&-64)>>3;while(++o<s)(r=e[o])&&r.length===n?c[o]=r:((i=a[o]).nullable||(a[o]=a[o].clone({nullable:!0})),c[o]=r?r._changeLengthAndBackfillNullBitmap(n):ie.new(i.type,0,n,n,Fs(u)));return[new ur(a),n,c]}function Rs(t){return Vs(new ur(t.map(({field:t})=>t)),t)}function Vs(t,e){return $s(t,e.map(t=>t instanceof Wn?t.chunks.map(t=>t.data):[t.data]))}function $s(t,e){const n=[...t.fields],r=[],i={numBatches:e.reduce((t,e)=>Math.max(t,e.length),0)};let o,s=0,a=0,c=-1,u=e.length,l=[];while(i.numBatches-- >0){for(a=Number.POSITIVE_INFINITY,c=-1;++c<u;)l[c]=o=e[c].shift(),a=Math.min(a,o?o.length:a);isFinite(a)&&(l=zs(n,a,l,e,i),a>0&&(r[s++]=[a,l.slice()]))}return[t=new ur(n,t.metadata),r.map(e=>new lu(t,...e))]}function zs(t,e,n,r,i){let o,s,a=0,c=-1,u=r.length;const l=(e+63&-64)>>3;while(++c<u)(o=n[c])&&(a=o.length)>=e?a===e?n[c]=o:(n[c]=o.slice(0,e),o=o.slice(e,a-e),i.numBatches=Math.max(i.numBatches,r[c].unshift(o))):((s=t[c]).nullable||(t[c]=s.clone({nullable:!0})),n[c]=o?o._changeLengthAndBackfillNullBitmap(e):ie.new(s.type,0,e,e,Fs(l)));return n}class Ys extends P{constructor(t,e){super(),this._children=e,this.numChildren=t.childData.length,this._bindDataAccessors(this.data=t)}get type(){return this.data.type}get typeId(){return this.data.typeId}get length(){return this.data.length}get offset(){return this.data.offset}get stride(){return this.data.stride}get nullCount(){return this.data.nullCount}get byteLength(){return this.data.byteLength}get VectorName(){return F[this.typeId]+"Vector"}get ArrayType(){return this.type.ArrayType}get values(){return this.data.values}get typeIds(){return this.data.typeIds}get nullBitmap(){return this.data.nullBitmap}get valueOffsets(){return this.data.valueOffsets}get[Symbol.toStringTag](){return`${this.VectorName}<${this.type[Symbol.toStringTag]}>`}clone(t,e=this._children){return P.new(t,e)}concat(...t){return Wn.concat(this,...t)}slice(t,e){return Dn(this,t,e,this._sliceInternal)}isValid(t){if(this.nullCount>0){const e=this.offset+t,n=this.nullBitmap[e>>3],r=n&1<<e%8;return 0!==r}return!0}getChildAt(t){return t<0||t>=this.numChildren?null:(this._children||(this._children=[]))[t]||(this._children[t]=P.new(this.data.childData[t]))}toJSON(){return[...this]}_sliceInternal(t,e,n){return t.clone(t.data.slice(e,n-e),null)}_bindDataAccessors(t){}}Ys.prototype[Symbol.isConcatSpreadable]=!0;class Hs extends Ys{asUtf8(){return P.new(this.data.clone(new Ft))}}class Ws extends Ys{static from(t){return eu(()=>new Ut,t)}}class Gs extends Ys{static from(...t){return 2===t.length?eu(()=>t[1]===R.DAY?new $t:new zt,t[0]):eu(()=>new zt,t[0])}}class qs extends Gs{}class Ks extends Gs{}class Xs extends Ys{}class Js extends Ys{constructor(t){super(t),this.indices=P.new(t.clone(this.type.indices))}static from(...t){if(3===t.length){const[e,n,r]=t,i=new te(e.type,n,null,null);return P.new(ie.Dictionary(i,0,r.length,0,null,r,e))}return eu(()=>t[0].type,t[0])}get dictionary(){return this.data.dictionary}reverseLookup(t){return this.dictionary.indexOf(t)}getKey(t){return this.indices.get(t)}getValue(t){return this.dictionary.get(t)}setKey(t,e){return this.indices.set(t,e)}setValue(t,e){return this.dictionary.set(t,e)}}Js.prototype.indices=null;class Zs extends Ys{}class Qs extends Ys{}class ta extends Ys{static from(t){let e=sa(this);if(t instanceof ArrayBuffer||ArrayBuffer.isView(t)){let n=oa(t.constructor)||e;if(null===e&&(e=n),e&&e===n){let n=new e,r=t.byteLength/n.ArrayType.BYTES_PER_ELEMENT;if(!ia(e,t.constructor))return P.new(ie.Float(n,0,r,0,null,t))}}if(e)return eu(()=>new e,t);if(t instanceof DataView||t instanceof ArrayBuffer)throw new TypeError("Cannot infer float type from instance of "+t.constructor.name);throw new TypeError("Unrecognized FloatVector input")}}class ea extends ta{toFloat32Array(){return new Float32Array(this)}toFloat64Array(){return new Float64Array(this)}}class na extends ta{}class ra extends ta{}const ia=(t,e)=>t===Dt&&e!==Uint16Array,oa=t=>{switch(t){case Uint16Array:return Dt;case Float32Array:return Nt;case Float64Array:return Lt;default:return null}},sa=t=>{switch(t){case ea:return Dt;case na:return Nt;case ra:return Lt;default:return null}};class aa extends Ys{}class ca extends aa{}class ua extends aa{}class la extends Ys{static from(...t){let[e,n=!1]=t,r=wa(this,n);if(e instanceof ArrayBuffer||ArrayBuffer.isView(e)){let t=_a(e.constructor,n)||r;if(null===r&&(r=t),r&&r===t){let t=new r,n=e.byteLength/t.ArrayType.BYTES_PER_ELEMENT;return ga(r,e.constructor)&&(n*=.5),P.new(ie.Int(t,0,n,0,null,e))}}if(r)return eu(()=>new r,e);if(e instanceof DataView||e instanceof ArrayBuffer)throw new TypeError("Cannot infer integer type from instance of "+e.constructor.name);throw new TypeError("Unrecognized IntVector input")}}class fa extends la{}class da extends la{}class ha extends la{}class pa extends la{toBigInt64Array(){return Object(ne["toBigInt64Array"])(this.values)}get values64(){return this._values64||(this._values64=this.toBigInt64Array())}}class ya extends la{}class va extends la{}class ba extends la{}class ma extends la{toBigUint64Array(){return Object(ne["toBigUint64Array"])(this.values)}get values64(){return this._values64||(this._values64=this.toBigUint64Array())}}const ga=(t,e)=>(t===jt||t===Mt)&&(e===Int32Array||e===Uint32Array),_a=(t,e)=>{switch(t){case Int8Array:return It;case Int16Array:return At;case Int32Array:return e?jt:Tt;case ae["b"]:return jt;case Uint8Array:return Et;case Uint16Array:return Bt;case Uint32Array:return e?Mt:Ct;case ae["d"]:return Mt;default:return null}},wa=(t,e)=>{switch(t){case fa:return It;case da:return At;case ha:return e?jt:Tt;case pa:return jt;case ya:return Et;case va:return Bt;case ba:return e?Mt:Ct;case ma:return Mt;default:return null}};class Oa extends Ys{}class Sa extends Ys{asList(){const t=this.type.children[0];return P.new(this.data.clone(new Gt(t)))}bind(t){const e=this.getChildAt(0),{[t]:n,[t+1]:r}=this.valueOffsets;return new jn(e.slice(n,r))}}class xa extends Ys{}const Ia=Symbol.for("rowIndex");class Aa extends Ys{bind(t){const e=this._row||(this._row=new En(this)),n=Object.create(e);return n[Ia]=t,n}}class Ta extends Ys{}class ja extends Ta{}class Ea extends Ta{}class Ba extends Ta{}class Ca extends Ta{}class Ma extends Ys{}class ka extends Ma{}class Da extends Ma{}class Na extends Ma{}class La extends Ma{}class Pa extends Ys{get typeIdToChildIndex(){return this.data.type.typeIdToChildIndex}}class Fa extends Pa{get valueOffsets(){return this.data.valueOffsets}}class Ua extends Pa{}class Ra extends Ys{static from(t){return eu(()=>new Ft,t)}asBinary(){return P.new(this.data.clone(new Pt))}}function Va(t){return function(){return t(this)}}function $a(t){return function(e){return t(this,e)}}function za(t){return function(e,n){return t(this,e,n)}}class Ya extends nt{}const Ha=(t,e)=>864e5*t[e],Wa=(t,e)=>4294967296*t[e+1]+(t[e]>>>0),Ga=(t,e)=>t[e+1]/1e3*4294967296+(t[e]>>>0)/1e3,qa=(t,e)=>t[e+1]/1e6*4294967296+(t[e]>>>0)/1e6,Ka=t=>new Date(t),Xa=(t,e)=>Ka(Ha(t,e)),Ja=(t,e)=>Ka(Wa(t,e)),Za=(t,e)=>null,Qa=(t,e,n)=>{const{[n]:r,[n+1]:i}=e;return null!=r&&null!=i?t.subarray(r,i):null},tc=({offset:t,values:e},n)=>{const r=t+n,i=e[r>>3];return 0!==(i&1<<r%8)},ec=({values:t},e)=>Xa(t,e),nc=({values:t},e)=>Ja(t,2*e),rc=({stride:t,values:e},n)=>e[t*n],ic=({stride:t,values:e},n)=>Me(e[t*n]),oc=({stride:t,values:e,type:n},r)=>Ge.new(e.subarray(t*r,t*(r+1)),n.isSigned),sc=({stride:t,values:e},n)=>e.subarray(t*n,t*(n+1)),ac=({values:t,valueOffsets:e},n)=>Qa(t,e,n),cc=({values:t,valueOffsets:e},n)=>{const r=Qa(t,e,n);return null!==r?Object(mn["a"])(r):null},uc=(t,e)=>t.type.bitWidth<64?rc(t,e):oc(t,e),lc=(t,e)=>t.type.precision!==$.HALF?rc(t,e):ic(t,e),fc=(t,e)=>t.type.unit===R.DAY?ec(t,e):nc(t,e),dc=({values:t},e)=>1e3*Wa(t,2*e),hc=({values:t},e)=>Wa(t,2*e),pc=({values:t},e)=>Ga(t,2*e),yc=({values:t},e)=>qa(t,2*e),vc=(t,e)=>{switch(t.type.unit){case V.SECOND:return dc(t,e);case V.MILLISECOND:return hc(t,e);case V.MICROSECOND:return pc(t,e);case V.NANOSECOND:return yc(t,e)}},bc=({values:t,stride:e},n)=>t[e*n],mc=({values:t,stride:e},n)=>t[e*n],gc=({values:t},e)=>Ge.signed(t.subarray(2*e,2*(e+1))),_c=({values:t},e)=>Ge.signed(t.subarray(2*e,2*(e+1))),wc=(t,e)=>{switch(t.type.unit){case V.SECOND:return bc(t,e);case V.MILLISECOND:return mc(t,e);case V.MICROSECOND:return gc(t,e);case V.NANOSECOND:return _c(t,e)}},Oc=({values:t},e)=>Ge.decimal(t.subarray(4*e,4*(e+1))),Sc=(t,e)=>{const n=t.getChildAt(0),{valueOffsets:r,stride:i}=t;return n.slice(r[e*i],r[e*i+1])},xc=(t,e)=>t.bind(e),Ic=(t,e)=>t.bind(e),Ac=(t,e)=>t.type.mode===z.Dense?Tc(t,e):jc(t,e),Tc=(t,e)=>{const n=t.typeIdToChildIndex[t.typeIds[e]],r=t.getChildAt(n);return r?r.get(t.valueOffsets[e]):null},jc=(t,e)=>{const n=t.typeIdToChildIndex[t.typeIds[e]],r=t.getChildAt(n);return r?r.get(e):null},Ec=(t,e)=>t.getValue(t.getKey(e)),Bc=(t,e)=>t.type.unit===Y.DAY_TIME?Cc(t,e):Mc(t,e),Cc=({values:t},e)=>t.subarray(2*e,2*(e+1)),Mc=({values:t},e)=>{const n=t[e],r=new Int32Array(2);return r[0]=n/12|0,r[1]=n%12|0,r},kc=(t,e)=>{const n=t.getChildAt(0),{stride:r}=t;return n.slice(e*r,(e+1)*r)};Ya.prototype.visitNull=Za,Ya.prototype.visitBool=tc,Ya.prototype.visitInt=uc,Ya.prototype.visitInt8=rc,Ya.prototype.visitInt16=rc,Ya.prototype.visitInt32=rc,Ya.prototype.visitInt64=oc,Ya.prototype.visitUint8=rc,Ya.prototype.visitUint16=rc,Ya.prototype.visitUint32=rc,Ya.prototype.visitUint64=oc,Ya.prototype.visitFloat=lc,Ya.prototype.visitFloat16=ic,Ya.prototype.visitFloat32=rc,Ya.prototype.visitFloat64=rc,Ya.prototype.visitUtf8=cc,Ya.prototype.visitBinary=ac,Ya.prototype.visitFixedSizeBinary=sc,Ya.prototype.visitDate=fc,Ya.prototype.visitDateDay=ec,Ya.prototype.visitDateMillisecond=nc,Ya.prototype.visitTimestamp=vc,Ya.prototype.visitTimestampSecond=dc,Ya.prototype.visitTimestampMillisecond=hc,Ya.prototype.visitTimestampMicrosecond=pc,Ya.prototype.visitTimestampNanosecond=yc,Ya.prototype.visitTime=wc,Ya.prototype.visitTimeSecond=bc,Ya.prototype.visitTimeMillisecond=mc,Ya.prototype.visitTimeMicrosecond=gc,Ya.prototype.visitTimeNanosecond=_c,Ya.prototype.visitDecimal=Oc,Ya.prototype.visitList=Sc,Ya.prototype.visitStruct=Ic,Ya.prototype.visitUnion=Ac,Ya.prototype.visitDenseUnion=Tc,Ya.prototype.visitSparseUnion=jc,Ya.prototype.visitDictionary=Ec,Ya.prototype.visitInterval=Bc,Ya.prototype.visitIntervalDayTime=Cc,Ya.prototype.visitIntervalYearMonth=Mc,Ya.prototype.visitFixedSizeList=kc,Ya.prototype.visitMap=xc;const Dc=new Ya;class Nc extends nt{}function Lc(t,e){return null===e&&t.length>0?0:-1}function Pc(t,e){const{nullBitmap:n}=t;if(!n||t.nullCount<=0)return-1;let r=0;for(const i of Z(n,t.data.offset+(e||0),t.length,n,G)){if(!i)return r;++r}return-1}function Fc(t,e,n){if(void 0===e)return-1;if(null===e)return Pc(t,n);const r=Pn(e);for(let i=(n||0)-1,o=t.length;++i<o;)if(r(t.get(i)))return i;return-1}function Uc(t,e,n){const r=Pn(e);for(let i=(n||0)-1,o=t.length;++i<o;)if(r(t.get(i)))return i;return-1}Nc.prototype.visitNull=Lc,Nc.prototype.visitBool=Fc,Nc.prototype.visitInt=Fc,Nc.prototype.visitInt8=Fc,Nc.prototype.visitInt16=Fc,Nc.prototype.visitInt32=Fc,Nc.prototype.visitInt64=Fc,Nc.prototype.visitUint8=Fc,Nc.prototype.visitUint16=Fc,Nc.prototype.visitUint32=Fc,Nc.prototype.visitUint64=Fc,Nc.prototype.visitFloat=Fc,Nc.prototype.visitFloat16=Fc,Nc.prototype.visitFloat32=Fc,Nc.prototype.visitFloat64=Fc,Nc.prototype.visitUtf8=Fc,Nc.prototype.visitBinary=Fc,Nc.prototype.visitFixedSizeBinary=Fc,Nc.prototype.visitDate=Fc,Nc.prototype.visitDateDay=Fc,Nc.prototype.visitDateMillisecond=Fc,Nc.prototype.visitTimestamp=Fc,Nc.prototype.visitTimestampSecond=Fc,Nc.prototype.visitTimestampMillisecond=Fc,Nc.prototype.visitTimestampMicrosecond=Fc,Nc.prototype.visitTimestampNanosecond=Fc,Nc.prototype.visitTime=Fc,Nc.prototype.visitTimeSecond=Fc,Nc.prototype.visitTimeMillisecond=Fc,Nc.prototype.visitTimeMicrosecond=Fc,Nc.prototype.visitTimeNanosecond=Fc,Nc.prototype.visitDecimal=Fc,Nc.prototype.visitList=Fc,Nc.prototype.visitStruct=Fc,Nc.prototype.visitUnion=Fc,Nc.prototype.visitDenseUnion=Uc,Nc.prototype.visitSparseUnion=Uc,Nc.prototype.visitDictionary=Fc,Nc.prototype.visitInterval=Fc,Nc.prototype.visitIntervalDayTime=Fc,Nc.prototype.visitIntervalYearMonth=Fc,Nc.prototype.visitFixedSizeList=Fc,Nc.prototype.visitMap=Fc;const Rc=new Nc;class Vc extends nt{}function $c(t){const e=Dc.getVisitFn(t);return Z(t.nullBitmap,t.offset,t.length,t,(t,n,r,i)=>0!==(r&1<<i)?e(t,n):null)}function zc(t){if(t.nullCount>0)return $c(t);const{type:e,typeId:n,length:r}=t;return 1===t.stride&&(n===F.Timestamp||n===F.Int&&64!==e.bitWidth||n===F.Time&&64!==e.bitWidth||n===F.Float&&e.precision>0)?t.values.subarray(0,r)[Symbol.iterator]():function*(e){for(let n=-1;++n<r;)yield e(t,n)}(Dc.getVisitFn(t))}Vc.prototype.visitNull=zc,Vc.prototype.visitBool=zc,Vc.prototype.visitInt=zc,Vc.prototype.visitInt8=zc,Vc.prototype.visitInt16=zc,Vc.prototype.visitInt32=zc,Vc.prototype.visitInt64=zc,Vc.prototype.visitUint8=zc,Vc.prototype.visitUint16=zc,Vc.prototype.visitUint32=zc,Vc.prototype.visitUint64=zc,Vc.prototype.visitFloat=zc,Vc.prototype.visitFloat16=zc,Vc.prototype.visitFloat32=zc,Vc.prototype.visitFloat64=zc,Vc.prototype.visitUtf8=zc,Vc.prototype.visitBinary=zc,Vc.prototype.visitFixedSizeBinary=zc,Vc.prototype.visitDate=zc,Vc.prototype.visitDateDay=zc,Vc.prototype.visitDateMillisecond=zc,Vc.prototype.visitTimestamp=zc,Vc.prototype.visitTimestampSecond=zc,Vc.prototype.visitTimestampMillisecond=zc,Vc.prototype.visitTimestampMicrosecond=zc,Vc.prototype.visitTimestampNanosecond=zc,Vc.prototype.visitTime=zc,Vc.prototype.visitTimeSecond=zc,Vc.prototype.visitTimeMillisecond=zc,Vc.prototype.visitTimeMicrosecond=zc,Vc.prototype.visitTimeNanosecond=zc,Vc.prototype.visitDecimal=zc,Vc.prototype.visitList=zc,Vc.prototype.visitStruct=zc,Vc.prototype.visitUnion=zc,Vc.prototype.visitDenseUnion=zc,Vc.prototype.visitSparseUnion=zc,Vc.prototype.visitDictionary=zc,Vc.prototype.visitInterval=zc,Vc.prototype.visitIntervalDayTime=zc,Vc.prototype.visitIntervalYearMonth=zc,Vc.prototype.visitFixedSizeList=zc,Vc.prototype.visitMap=zc;const Yc=new Vc;class Hc extends nt{}function Wc(t){const{type:e,length:n,stride:r}=t;switch(e.typeId){case F.Int:case F.Float:case F.Decimal:case F.Time:case F.Timestamp:return t.values.subarray(0,n*r)}return[...Yc.visit(t)]}Hc.prototype.visitNull=Wc,Hc.prototype.visitBool=Wc,Hc.prototype.visitInt=Wc,Hc.prototype.visitInt8=Wc,Hc.prototype.visitInt16=Wc,Hc.prototype.visitInt32=Wc,Hc.prototype.visitInt64=Wc,Hc.prototype.visitUint8=Wc,Hc.prototype.visitUint16=Wc,Hc.prototype.visitUint32=Wc,Hc.prototype.visitUint64=Wc,Hc.prototype.visitFloat=Wc,Hc.prototype.visitFloat16=Wc,Hc.prototype.visitFloat32=Wc,Hc.prototype.visitFloat64=Wc,Hc.prototype.visitUtf8=Wc,Hc.prototype.visitBinary=Wc,Hc.prototype.visitFixedSizeBinary=Wc,Hc.prototype.visitDate=Wc,Hc.prototype.visitDateDay=Wc,Hc.prototype.visitDateMillisecond=Wc,Hc.prototype.visitTimestamp=Wc,Hc.prototype.visitTimestampSecond=Wc,Hc.prototype.visitTimestampMillisecond=Wc,Hc.prototype.visitTimestampMicrosecond=Wc,Hc.prototype.visitTimestampNanosecond=Wc,Hc.prototype.visitTime=Wc,Hc.prototype.visitTimeSecond=Wc,Hc.prototype.visitTimeMillisecond=Wc,Hc.prototype.visitTimeMicrosecond=Wc,Hc.prototype.visitTimeNanosecond=Wc,Hc.prototype.visitDecimal=Wc,Hc.prototype.visitList=Wc,Hc.prototype.visitStruct=Wc,Hc.prototype.visitUnion=Wc,Hc.prototype.visitDenseUnion=Wc,Hc.prototype.visitSparseUnion=Wc,Hc.prototype.visitDictionary=Wc,Hc.prototype.visitInterval=Wc,Hc.prototype.visitIntervalDayTime=Wc,Hc.prototype.visitIntervalYearMonth=Wc,Hc.prototype.visitFixedSizeList=Wc,Hc.prototype.visitMap=Wc;const Gc=new Hc,qc=(t,e)=>t+e,Kc=t=>"Cannot compute the byte width of variable-width column "+t;class Xc extends nt{visitNull(t){return 0}visitInt(t){return t.bitWidth/8}visitFloat(t){return t.ArrayType.BYTES_PER_ELEMENT}visitBinary(t){throw new Error(Kc(t))}visitUtf8(t){throw new Error(Kc(t))}visitBool(t){return 1/8}visitDecimal(t){return 16}visitDate(t){return 4*(t.unit+1)}visitTime(t){return t.bitWidth/8}visitTimestamp(t){return t.unit===V.SECOND?4:8}visitInterval(t){return 4*(t.unit+1)}visitList(t){throw new Error(Kc(t))}visitStruct(t){return this.visitFields(t.children).reduce(qc,0)}visitUnion(t){return this.visitFields(t.children).reduce(qc,0)}visitFixedSizeBinary(t){return t.byteWidth}visitFixedSizeList(t){return t.listSize*this.visitFields(t.children).reduce(qc,0)}visitMap(t){return this.visitFields(t.children).reduce(qc,0)}visitDictionary(t){return this.visit(t.indices)}visitFields(t){return(t||[]).map(t=>this.visit(t.type))}visitSchema(t){return this.visitFields(t.fields).reduce(qc,0)}}const Jc=new Xc;class Zc extends nt{visitNull(){return xa}visitBool(){return Ws}visitInt(){return la}visitInt8(){return fa}visitInt16(){return da}visitInt32(){return ha}visitInt64(){return pa}visitUint8(){return ya}visitUint16(){return va}visitUint32(){return ba}visitUint64(){return ma}visitFloat(){return ta}visitFloat16(){return ea}visitFloat32(){return na}visitFloat64(){return ra}visitUtf8(){return Ra}visitBinary(){return Hs}visitFixedSizeBinary(){return Zs}visitDate(){return Gs}visitDateDay(){return qs}visitDateMillisecond(){return Ks}visitTimestamp(){return Ta}visitTimestampSecond(){return ja}visitTimestampMillisecond(){return Ea}visitTimestampMicrosecond(){return Ba}visitTimestampNanosecond(){return Ca}visitTime(){return Ma}visitTimeSecond(){return ka}visitTimeMillisecond(){return Da}visitTimeMicrosecond(){return Na}visitTimeNanosecond(){return La}visitDecimal(){return Xs}visitList(){return Oa}visitStruct(){return Aa}visitUnion(){return Pa}visitDenseUnion(){return Fa}visitSparseUnion(){return Ua}visitDictionary(){return Js}visitInterval(){return aa}visitIntervalDayTime(){return ca}visitIntervalYearMonth(){return ua}visitFixedSizeList(){return Qs}visitMap(){return Sa}}const Qc=new Zc;function tu(t,...e){return new(Qc.getVisitFn(t)())(t,...e)}function eu(t,e){if(Object(ae["i"])(e))return P.from({nullValues:[null,void 0],type:t(),values:e});if(Object(ae["f"])(e))return P.from({nullValues:[null,void 0],type:t(),values:e});const{values:n=[],type:r=t(),nullValues:i=[null,void 0]}={...e};return Object(ae["i"])(n),P.from({nullValues:i,...e,type:r})}function nu(t){const{values:e=[],...n}={nullValues:[null,void 0],...t};if(Object(ae["i"])(e)){const t=[...be.throughIterable(n)(e)];return 1===t.length?t[0]:Wn.concat(t)}return(async t=>{const r=be.throughAsyncIterable(n);for await(const n of r(e))t.push(n);return 1===t.length?t[0]:Wn.concat(t)})([])}function ru(t){return function(){return t(this.type)}}function iu(t){return function(e){return this.isValid(e)?t.call(this,e):null}}function ou(t){return function(e,n){K(this.nullBitmap,this.offset+e,!(null===n||void 0===n))&&t.call(this,e,n)}}function su(){const t=this.nullBitmap;t&&t.byteLength>0&&(this.get=iu(this.get),this.set=ou(this.set))}P.new=tu,P.from=nu,Ys.prototype.get=function(t){return Dc.visit(this,t)},Ys.prototype.set=function(t,e){return li.visit(this,t,e)},Ys.prototype.indexOf=function(t,e){return Rc.visit(this,t,e)},Ys.prototype.toArray=function(){return Gc.visit(this)},Ys.prototype.getByteWidth=function(){return Jc.visit(this.type)},Ys.prototype[Symbol.iterator]=function(){return Yc.visit(this)},Ys.prototype._bindDataAccessors=su,Object.keys(F).map(t=>F[t]).filter(t=>"number"===typeof t).filter(t=>t!==F.NONE).forEach(t=>{const e=Qc.visit(t);e.prototype["get"]=$a(Dc.getVisitFn(t)),e.prototype["set"]=za(li.getVisitFn(t)),e.prototype["indexOf"]=za(Rc.getVisitFn(t)),e.prototype["toArray"]=Va(Gc.getVisitFn(t)),e.prototype["getByteWidth"]=ru(Jc.getVisitFn(t)),e.prototype[Symbol.iterator]=Va(Yc.getVisitFn(t))});class au extends Wn{constructor(...t){let e=null;t[0]instanceof ur&&(e=t.shift());let n=Qn(lu,t);if(!e&&!(e=n[0]&&n[0].schema))throw new TypeError("Table must be initialized with a Schema or at least one RecordBatch");n[0]||(n[0]=new fu(e)),super(new qt(e.fields),n),this._schema=e,this._chunks=n}static empty(t=new ur([])){return new au(t,[])}static from(t){if(!t)return au.empty();if("object"===typeof t){let e=Object(ae["i"])(t["values"])?cu(t):Object(ae["f"])(t["values"])?uu(t):null;if(null!==e)return e}let e=hu.from(t);return Object(ae["l"])(e)?(async()=>await au.from(await e))():e.isSync()&&(e=e.open())?e.schema?new au(e.schema,[...e]):au.empty():(async t=>{const e=await t,n=e.schema,r=[];if(n){for await(let t of e)r.push(t);return new au(n,r)}return au.empty()})(e.open())}static async fromAsync(t){return await au.from(t)}static fromStruct(t){return au.new(t.data.childData,t.type.children)}static new(...t){return new au(...Rs(tr(t)))}get schema(){return this._schema}get length(){return this._length}get chunks(){return this._chunks}get numCols(){return this._numChildren}clone(t=this._chunks){return new au(this._schema,t)}getColumn(t){return this.getColumnAt(this.getColumnIndex(t))}getColumnAt(t){return this.getChildAt(t)}getColumnIndex(t){return this._schema.fields.findIndex(e=>e.name===t)}getChildAt(t){if(t<0||t>=this.numChildren)return null;let e,n;const r=this._schema.fields,i=this._children||(this._children=[]);if(n=i[t])return n;if(e=r[t]){const n=this._chunks.map(e=>e.getChildAt(t)).filter(t=>null!=t);if(n.length>0)return i[t]=new Xn(e,n)}return null}serialize(t="binary",e=!0){const n=e?ks:Ds;return n.writeAll(this).toUint8Array(!0)}count(){return this._length}select(...t){const e=this._schema.fields.reduce((t,e,n)=>t.set(e.name,n),new Map);return this.selectAt(...t.map(t=>e.get(t)).filter(t=>t>-1))}selectAt(...t){const e=this._schema.selectAt(...t);return new au(e,this._chunks.map(({length:n,data:{childData:r}})=>new lu(e,n,t.map(t=>r[t]).filter(Boolean))))}assign(t){const e=this._schema.fields,[n,r]=t.schema.fields.reduce((t,n,r)=>{const[i,o]=t,s=e.findIndex(t=>t.name===n.name);return~s?o[s]=r:i.push(r),t},[[],[]]),i=this._schema.assign(t.schema),o=[...e.map((e,n,i,o=r[n])=>void 0===o?this.getColumnAt(n):t.getColumnAt(o)),...n.map(e=>t.getColumnAt(e))].filter(Boolean);return new au(...Vs(i,o))}}function cu(t){const{type:e}=t;return e instanceof qt?au.fromStruct(Aa.from(t)):null}function uu(t){const{type:e}=t;return e instanceof qt?Aa.from(t).then(t=>au.fromStruct(t)):null}class lu extends Aa{constructor(...t){let e,n,r=t[0];if(t[1]instanceof ie)[,e,n]=t;else{const n=r.fields,[,i,o]=t;e=ie.Struct(new qt(n),0,i,0,null,o)}super(e,n),this._schema=r}static from(t){return Object(ae["i"])(t["values"]),au.from(t)}static new(...t){const[e,n]=er(t),r=n.filter(t=>t instanceof P);return new lu(...Us(new ur(e),r.map(t=>t.data)))}clone(t,e=this._children){return new lu(this._schema,t,e)}concat(...t){const e=this._schema,n=Wn.flatten(this,...t);return new au(e,n.map(({data:t})=>new lu(e,t)))}get schema(){return this._schema}get numCols(){return this._schema.fields.length}get dictionaries(){return this._dictionaries||(this._dictionaries=du.collect(this))}select(...t){const e=this._schema.fields.reduce((t,e,n)=>t.set(e.name,n),new Map);return this.selectAt(...t.map(t=>e.get(t)).filter(t=>t>-1))}selectAt(...t){const e=this._schema.selectAt(...t),n=t.map(t=>this.data.childData[t]).filter(Boolean);return new lu(e,this.length,n)}}class fu extends lu{constructor(t){super(t,0,t.fields.map(t=>ie.new(t.type,0,0,0)))}}class du extends nt{constructor(){super(...arguments),this.dictionaries=new Map}static collect(t){return(new du).visit(t.data,new qt(t.schema.fields)).dictionaries}visit(t,e){return Ot.isDictionary(e)?this.visitDictionary(t,e):(t.childData.forEach((t,n)=>this.visit(t,e.children[n].type)),this)}visitDictionary(t,e){const n=t.dictionary;return n&&n.length>0&&this.dictionaries.set(e.id,n),this}}class hu extends xi["d"]{constructor(t){super(),this._impl=t}get closed(){return this._impl.closed}get schema(){return this._impl.schema}get autoDestroy(){return this._impl.autoDestroy}get dictionaries(){return this._impl.dictionaries}get numDictionaries(){return this._impl.numDictionaries}get numRecordBatches(){return this._impl.numRecordBatches}get footer(){return this._impl.isFile()?this._impl.footer:null}isSync(){return this._impl.isSync()}isAsync(){return this._impl.isAsync()}isFile(){return this._impl.isFile()}isStream(){return this._impl.isStream()}next(){return this._impl.next()}throw(t){return this._impl.throw(t)}return(t){return this._impl.return(t)}cancel(){return this._impl.cancel()}reset(t){return this._impl.reset(t),this._DOMStream=void 0,this._nodeStream=void 0,this}open(t){const e=this._impl.open(t);return Object(ae["l"])(e)?e.then(()=>this):this}readRecordBatch(t){return this._impl.isFile()?this._impl.readRecordBatch(t):null}[Symbol.iterator](){return this._impl[Symbol.iterator]()}[Symbol.asyncIterator](){return this._impl[Symbol.asyncIterator]()}toDOMStream(){return L["a"].toDOMStream(this.isSync()?{[Symbol.iterator]:()=>this}:{[Symbol.asyncIterator]:()=>this})}toNodeStream(){return L["a"].toNodeStream(this.isSync()?{[Symbol.iterator]:()=>this}:{[Symbol.asyncIterator]:()=>this},{objectMode:!0})}static throughNode(t){throw new Error('"throughNode" not available in this environment')}static throughDOM(t,e){throw new Error('"throughDOM" not available in this environment')}static from(t){return t instanceof hu?t:Object(ae["e"])(t)?Tu(t):Object(ae["h"])(t)?Bu(t):Object(ae["l"])(t)?(async()=>await hu.from(await t))():Object(ae["g"])(t)||Object(ae["m"])(t)||Object(ae["n"])(t)||Object(ae["f"])(t)?Eu(new Ti(t)):ju(new Ai(t))}static readAll(t){return t instanceof hu?t.isSync()?Iu(t):Au(t):Object(ae["e"])(t)||ArrayBuffer.isView(t)||Object(ae["i"])(t)||Object(ae["j"])(t)?Iu(t):Au(t)}}class pu extends hu{constructor(t){super(t),this._impl=t}[Symbol.iterator](){return this._impl[Symbol.iterator]()}async*[Symbol.asyncIterator](){yield*this[Symbol.iterator]()}}class yu extends hu{constructor(t){super(t),this._impl=t}[Symbol.iterator](){throw new Error("AsyncRecordBatchStreamReader is not Iterable")}[Symbol.asyncIterator](){return this._impl[Symbol.asyncIterator]()}}class vu extends pu{constructor(t){super(t),this._impl=t}}class bu extends yu{constructor(t){super(t),this._impl=t}}class mu{constructor(t=new Map){this.closed=!1,this.autoDestroy=!0,this._dictionaryIndex=0,this._recordBatchIndex=0,this.dictionaries=t}get numDictionaries(){return this._dictionaryIndex}get numRecordBatches(){return this._recordBatchIndex}isSync(){return!1}isAsync(){return!1}isFile(){return!1}isStream(){return!1}reset(t){return this._dictionaryIndex=0,this._recordBatchIndex=0,this.schema=t,this.dictionaries=new Map,this}_loadRecordBatch(t,e){return new lu(this.schema,t.length,this._loadVectors(t,e,this.schema.fields))}_loadDictionaryBatch(t,e){const{id:n,isDelta:r,data:i}=t,{dictionaries:o,schema:s}=this,a=o.get(n);if(r||!a){const t=s.dictionaries.get(n);return a&&r?a.concat(P.new(this._loadVectors(i,e,[t])[0])):P.new(this._loadVectors(i,e,[t])[0])}return a}_loadVectors(t,e,n){return new Ri(e,t.nodes,t.buffers,this.dictionaries).visitMany(n)}}class gu extends mu{constructor(t,e){super(e),this._reader=Object(ae["e"])(t)?new vs(this._handle=t):new ps(this._handle=t)}isSync(){return!0}isStream(){return!0}[Symbol.iterator](){return this}cancel(){!this.closed&&(this.closed=!0)&&(this.reset()._reader.return(),this._reader=null,this.dictionaries=null)}open(t){return this.closed||(this.autoDestroy=xu(this,t),this.schema||(this.schema=this._reader.readSchema())||this.cancel()),this}throw(t){return!this.closed&&this.autoDestroy&&(this.closed=!0)?this.reset()._reader.throw(t):xi["c"]}return(t){return!this.closed&&this.autoDestroy&&(this.closed=!0)?this.reset()._reader.return(t):xi["c"]}next(){if(this.closed)return xi["c"];let t,{_reader:e}=this;while(t=this._readNextMessageAndValidate())if(t.isSchema())this.reset(t.header());else{if(t.isRecordBatch()){this._recordBatchIndex++;const n=t.header(),r=e.readMessageBody(t.bodyLength),i=this._loadRecordBatch(n,r);return{done:!1,value:i}}if(t.isDictionaryBatch()){this._dictionaryIndex++;const n=t.header(),r=e.readMessageBody(t.bodyLength),i=this._loadDictionaryBatch(n,r);this.dictionaries.set(n.id,i)}}return this.schema&&0===this._recordBatchIndex?(this._recordBatchIndex++,{done:!1,value:new fu(this.schema)}):this.return()}_readNextMessageAndValidate(t){return this._reader.readMessage(t)}}class _u extends mu{constructor(t,e){super(e),this._reader=new ys(this._handle=t)}isAsync(){return!0}isStream(){return!0}[Symbol.asyncIterator](){return this}async cancel(){!this.closed&&(this.closed=!0)&&(await this.reset()._reader.return(),this._reader=null,this.dictionaries=null)}async open(t){return this.closed||(this.autoDestroy=xu(this,t),this.schema||(this.schema=await this._reader.readSchema())||await this.cancel()),this}async throw(t){return!this.closed&&this.autoDestroy&&(this.closed=!0)?await this.reset()._reader.throw(t):xi["c"]}async return(t){return!this.closed&&this.autoDestroy&&(this.closed=!0)?await this.reset()._reader.return(t):xi["c"]}async next(){if(this.closed)return xi["c"];let t,{_reader:e}=this;while(t=await this._readNextMessageAndValidate())if(t.isSchema())await this.reset(t.header());else{if(t.isRecordBatch()){this._recordBatchIndex++;const n=t.header(),r=await e.readMessageBody(t.bodyLength),i=this._loadRecordBatch(n,r);return{done:!1,value:i}}if(t.isDictionaryBatch()){this._dictionaryIndex++;const n=t.header(),r=await e.readMessageBody(t.bodyLength),i=this._loadDictionaryBatch(n,r);this.dictionaries.set(n.id,i)}}return this.schema&&0===this._recordBatchIndex?(this._recordBatchIndex++,{done:!1,value:new fu(this.schema)}):await this.return()}async _readNextMessageAndValidate(t){return await this._reader.readMessage(t)}}class wu extends gu{constructor(t,e){super(t instanceof Bi?t:new Bi(t),e)}get footer(){return this._footer}get numDictionaries(){return this._footer?this._footer.numDictionaries:0}get numRecordBatches(){return this._footer?this._footer.numRecordBatches:0}isSync(){return!0}isFile(){return!0}open(t){if(!this.closed&&!this._footer){this.schema=(this._footer=this._readFooter()).schema;for(const t of this._footer.dictionaryBatches())t&&this._readDictionaryBatch(this._dictionaryIndex++)}return super.open(t)}readRecordBatch(t){if(this.closed)return null;this._footer||this.open();const e=this._footer&&this._footer.getRecordBatch(t);if(e&&this._handle.seek(e.offset)){const t=this._reader.readMessage(H.RecordBatch);if(t&&t.isRecordBatch()){const e=t.header(),n=this._reader.readMessageBody(t.bodyLength),r=this._loadRecordBatch(e,n);return r}}return null}_readDictionaryBatch(t){const e=this._footer&&this._footer.getDictionaryBatch(t);if(e&&this._handle.seek(e.offset)){const t=this._reader.readMessage(H.DictionaryBatch);if(t&&t.isDictionaryBatch()){const e=t.header(),n=this._reader.readMessageBody(t.bodyLength),r=this._loadDictionaryBatch(e,n);this.dictionaries.set(e.id,r)}}}_readFooter(){const{_handle:t}=this,e=t.size-Os,n=t.readInt32(e),r=t.readAt(e-n,n);return wi.decode(r)}_readNextMessageAndValidate(t){if(this._footer||this.open(),this._footer&&this._recordBatchIndex<this.numRecordBatches){const e=this._footer&&this._footer.getRecordBatch(this._recordBatchIndex);if(e&&this._handle.seek(e.offset))return this._reader.readMessage(t)}return null}}class Ou extends _u{constructor(t,...e){const n="number"!==typeof e[0]?e.shift():void 0,r=e[0]instanceof Map?e.shift():void 0;super(t instanceof Ci?t:new Ci(t,n),r)}get footer(){return this._footer}get numDictionaries(){return this._footer?this._footer.numDictionaries:0}get numRecordBatches(){return this._footer?this._footer.numRecordBatches:0}isFile(){return!0}isAsync(){return!0}async open(t){if(!this.closed&&!this._footer){this.schema=(this._footer=await this._readFooter()).schema;for(const t of this._footer.dictionaryBatches())t&&await this._readDictionaryBatch(this._dictionaryIndex++)}return await super.open(t)}async readRecordBatch(t){if(this.closed)return null;this._footer||await this.open();const e=this._footer&&this._footer.getRecordBatch(t);if(e&&await this._handle.seek(e.offset)){const t=await this._reader.readMessage(H.RecordBatch);if(t&&t.isRecordBatch()){const e=t.header(),n=await this._reader.readMessageBody(t.bodyLength),r=this._loadRecordBatch(e,n);return r}}return null}async _readDictionaryBatch(t){const e=this._footer&&this._footer.getDictionaryBatch(t);if(e&&await this._handle.seek(e.offset)){const t=await this._reader.readMessage(H.DictionaryBatch);if(t&&t.isDictionaryBatch()){const e=t.header(),n=await this._reader.readMessageBody(t.bodyLength),r=this._loadDictionaryBatch(e,n);this.dictionaries.set(e.id,r)}}}async _readFooter(){const{_handle:t}=this;t._pending&&await t._pending;const e=t.size-Os,n=await t.readInt32(e),r=await t.readAt(e-n,n);return wi.decode(r)}async _readNextMessageAndValidate(t){if(this._footer||await this.open(),this._footer&&this._recordBatchIndex<this.numRecordBatches){const e=this._footer.getRecordBatch(this._recordBatchIndex);if(e&&await this._handle.seek(e.offset))return await this._reader.readMessage(t)}return null}}class Su extends gu{constructor(t,e){super(t,e)}_loadVectors(t,e,n){return new Vi(e,t.nodes,t.buffers,this.dictionaries).visitMany(n)}}function xu(t,e){return e&&"boolean"===typeof e["autoDestroy"]?e["autoDestroy"]:t["autoDestroy"]}function*Iu(t){const e=hu.from(t);try{if(!e.open({autoDestroy:!1}).closed)do{yield e}while(!e.reset().open().closed)}finally{e.cancel()}}async function*Au(t){const e=await hu.from(t);try{if(!(await e.open({autoDestroy:!1})).closed)do{yield e}while(!(await e.reset().open()).closed)}finally{await e.cancel()}}function Tu(t){return new pu(new Su(t))}function ju(t){const e=t.peek(ws+7&-8);return e&&e.byteLength>=4?_s(e)?new vu(new wu(t.read())):new pu(new gu(t)):new pu(new gu(function*(){}()))}async function Eu(t){const e=await t.peek(ws+7&-8);return e&&e.byteLength>=4?_s(e)?new vu(new wu(await t.read())):new yu(new _u(t)):new yu(new _u(async function*(){}()))}async function Bu(t){const{size:e}=await t.stat(),n=new Ci(t,e);return e>=Ss&&_s(await n.readAt(0,ws+7&-8))?new bu(new Ou(n)):new yu(new _u(n))}function Cu(t,e){if(Object(ae["f"])(t))return ku(t,e);if(Object(ae["i"])(t))return Mu(t,e);throw new Error("toDOMStream() must be called with an Iterable or AsyncIterable")}function Mu(t,e){let n=null;const r=e&&"bytes"===e.type||!1,i=e&&e.highWaterMark||2**24;return new ReadableStream({...e,start(e){o(e,n||(n=t[Symbol.iterator]()))},pull(t){n?o(t,n):t.close()},cancel(){n&&n.return&&n.return(),n=null}},{highWaterMark:r?i:void 0,...e});function o(t,e){let n,i=null,o=t.desiredSize||null;while(!(i=e.next(r?o:null)).done)if(ArrayBuffer.isView(i.value)&&(n=Object(ne["toUint8Array"])(i.value))&&(null!=o&&r&&(o=o-n.byteLength+1),i.value=n),t.enqueue(i.value),null!=o&&--o<=0)return;t.close()}}function ku(t,e){let n=null;const r=e&&"bytes"===e.type||!1,i=e&&e.highWaterMark||2**24;return new ReadableStream({...e,async start(e){await o(e,n||(n=t[Symbol.asyncIterator]()))},async pull(t){n?await o(t,n):t.close()},async cancel(){n&&n.return&&await n.return(),n=null}},{highWaterMark:r?i:void 0,...e});async function o(t,e){let n,i=null,o=t.desiredSize||null;while(!(i=await e.next(r?o:null)).done)if(ArrayBuffer.isView(i.value)&&(n=Object(ne["toUint8Array"])(i.value))&&(null!=o&&r&&(o=o-n.byteLength+1),i.value=n),t.enqueue(i.value),null!=o&&--o<=0)return;t.close()}}function Du(t){return new Nu(t)}class Nu{constructor(t){this._numChunks=0,this._finished=!1,this._bufferedSize=0;const{["readableStrategy"]:e,["writableStrategy"]:n,["queueingStrategy"]:r="count",...i}=t;this._controller=null,this._builder=be.new(i),this._getSize="bytes"!==r?Lu:Pu;const{["highWaterMark"]:o=("bytes"===r?16384:1e3)}={...e},{["highWaterMark"]:s=("bytes"===r?16384:1e3)}={...n};this["readable"]=new ReadableStream({["cancel"]:()=>{this._builder.clear()},["pull"]:t=>{this._maybeFlush(this._builder,this._controller=t)},["start"]:t=>{this._maybeFlush(this._builder,this._controller=t)}},{highWaterMark:o,size:"bytes"!==r?Lu:Pu}),this["writable"]=new WritableStream({["abort"]:()=>{this._builder.clear()},["write"]:()=>{this._maybeFlush(this._builder,this._controller)},["close"]:()=>{this._maybeFlush(this._builder.finish(),this._controller)}},{highWaterMark:s,size:t=>this._writeValueAndReturnChunkSize(t)})}_writeValueAndReturnChunkSize(t){const e=this._bufferedSize;return this._bufferedSize=this._getSize(this._builder.append(t)),this._bufferedSize-e}_maybeFlush(t,e){null!==e&&(this._bufferedSize>=e.desiredSize&&++this._numChunks&&this._enqueue(e,t.toVector()),t.finished&&((t.length>0||0===this._numChunks)&&++this._numChunks&&this._enqueue(e,t.toVector()),!this._finished&&(this._finished=!0)&&this._enqueue(e,null)))}_enqueue(t,e){this._bufferedSize=0,this._controller=null,null===e?t.close():t.enqueue(e)}}const Lu=t=>t.length,Pu=t=>t.byteLength;function Fu(t,e){const n=new Ii;let r=null;const i=new ReadableStream({async cancel(){await n.close()},async start(t){await s(t,r||(r=await o()))},async pull(t){r?await s(t,r):t.close()}});return{writable:new WritableStream(n,{highWaterMark:16384,...t}),readable:i};async function o(){return await(await hu.from(n)).open(e)}async function s(t,e){let n=t.desiredSize,r=null;while(!(r=await e.next()).done)if(t.enqueue(r.value),null!=n&&--n<=0)return;t.close()}}function Uu(t,e){const n=new this(t),r=new Ti(n),i=new ReadableStream({type:"bytes",async cancel(){await r.cancel()},async pull(t){await o(t)},async start(t){await o(t)}},{highWaterMark:16384,...e});return{writable:new WritableStream(n,t),readable:i};async function o(t){let e=null,n=t.desiredSize;while(e=await r.read(n||null))if(t.enqueue(e),null!=n&&(n-=e.byteLength)<=0)return;t.close()}}class Ru{eq(t){return t instanceof Ru||(t=new Vu(t)),new qu(this,t)}le(t){return t instanceof Ru||(t=new Vu(t)),new Ku(this,t)}ge(t){return t instanceof Ru||(t=new Vu(t)),new Xu(this,t)}lt(t){return new Ju(this.ge(t))}gt(t){return new Ju(this.le(t))}ne(t){return new Ju(this.eq(t))}}class Vu extends Ru{constructor(t){super(),this.v=t}}class $u extends Ru{constructor(t){super(),this.name=t}bind(t){if(!this.colidx){this.colidx=-1;const e=t.schema.fields;for(let t=-1;++t<e.length;)if(e[t].name===this.name){this.colidx=t;break}if(this.colidx<0)throw new Error(`Failed to bind Col "${this.name}"`)}const e=this.vector=t.getChildAt(this.colidx);return t=>e.get(t)}}class zu{and(...t){return new Wu(this,...t)}or(...t){return new Gu(this,...t)}not(){return new Ju(this)}}class Yu extends zu{constructor(t,e){super(),this.left=t,this.right=e}bind(t){return this.left instanceof Vu?this.right instanceof Vu?this._bindLitLit(t,this.left,this.right):this._bindLitCol(t,this.left,this.right):this.right instanceof Vu?this._bindColLit(t,this.left,this.right):this._bindColCol(t,this.left,this.right)}}class Hu extends zu{constructor(...t){super(),this.children=t}}Hu.prototype.children=Object.freeze([]);class Wu extends Hu{constructor(...t){t=t.reduce((t,e)=>t.concat(e instanceof Wu?e.children:e),[]),super(...t)}bind(t){const e=this.children.map(e=>e.bind(t));return(t,n)=>e.every(e=>e(t,n))}}class Gu extends Hu{constructor(...t){t=t.reduce((t,e)=>t.concat(e instanceof Gu?e.children:e),[]),super(...t)}bind(t){const e=this.children.map(e=>e.bind(t));return(t,n)=>e.some(e=>e(t,n))}}class qu extends Yu{_bindLitLit(t,e,n){const r=e.v==n.v;return()=>r}_bindColCol(t,e,n){const r=e.bind(t),i=n.bind(t);return(t,e)=>r(t,e)==i(t,e)}_bindColLit(t,e,n){const r=e.bind(t);if(e.vector instanceof Js){let t;const r=e.vector;return r.dictionary!==this.lastDictionary?(t=r.reverseLookup(n.v),this.lastDictionary=r.dictionary,this.lastKey=t):t=this.lastKey,-1===t?()=>!1:e=>r.getKey(e)===t}return(t,e)=>r(t,e)==n.v}_bindLitCol(t,e,n){return this._bindColLit(t,n,e)}}class Ku extends Yu{_bindLitLit(t,e,n){const r=e.v<=n.v;return()=>r}_bindColCol(t,e,n){const r=e.bind(t),i=n.bind(t);return(t,e)=>r(t,e)<=i(t,e)}_bindColLit(t,e,n){const r=e.bind(t);return(t,e)=>r(t,e)<=n.v}_bindLitCol(t,e,n){const r=n.bind(t);return(t,n)=>e.v<=r(t,n)}}class Xu extends Yu{_bindLitLit(t,e,n){const r=e.v>=n.v;return()=>r}_bindColCol(t,e,n){const r=e.bind(t),i=n.bind(t);return(t,e)=>r(t,e)>=i(t,e)}_bindColLit(t,e,n){const r=e.bind(t);return(t,e)=>r(t,e)>=n.v}_bindLitCol(t,e,n){const r=n.bind(t);return(t,n)=>e.v>=r(t,n)}}class Ju extends zu{constructor(t){super(),this.child=t}bind(t){const e=this.child.bind(t);return(t,n)=>!e(t,n)}}au.prototype.countBy=function(t){return new Zu(this.chunks).countBy(t)},au.prototype.scan=function(t,e){return new Zu(this.chunks).scan(t,e)},au.prototype.scanReverse=function(t,e){return new Zu(this.chunks).scanReverse(t,e)},au.prototype.filter=function(t){return new Zu(this.chunks).filter(t)};class Zu extends au{filter(t){return new tl(this.chunks,t)}scan(t,e){const n=this.chunks,r=n.length;for(let i=-1;++i<r;){const r=n[i];e&&e(r);for(let e=-1,n=r.length;++e<n;)t(e,r)}}scanReverse(t,e){const n=this.chunks,r=n.length;for(let i=r;--i>=0;){const r=n[i];e&&e(r);for(let e=r.length;--e>=0;)t(e,r)}}countBy(t){const e=this.chunks,n=e.length,r="string"===typeof t?new $u(t):t;r.bind(e[n-1]);const i=r.vector;if(!Ot.isDictionary(i.type))throw new Error("countBy currently only supports dictionary-encoded columns");const o=Math.ceil(Math.log(i.length)/Math.log(256)),s=4==o?Uint32Array:o>=2?Uint16Array:Uint8Array,a=new s(i.dictionary.length);for(let c=-1;++c<n;){const t=e[c];r.bind(t);const n=r.vector.indices;for(let e=-1,r=t.length;++e<r;){let t=n.get(e);null!==t&&a[t]++}}return new Qu(i.dictionary,la.from(a))}}class Qu extends au{constructor(t,e){const n=new ur([new lr("values",t.type),new lr("counts",e.type)]);super(new lu(n,e.length,[t,e]))}toJSON(){const t=this.getColumnAt(0),e=this.getColumnAt(1),n={};for(let r=-1;++r<this.length;)n[t.get(r)]=e.get(r);return n}}class tl extends Zu{constructor(t,e){super(t),this._predicate=e}scan(t,e){const n=this._chunks,r=n.length;for(let i=-1;++i<r;){const r=n[i],o=this._predicate.bind(r);let s=!1;for(let n=-1,i=r.length;++n<i;)o(n,r)&&(e&&!s&&(e(r),s=!0),t(n,r))}}scanReverse(t,e){const n=this._chunks,r=n.length;for(let i=r;--i>=0;){const r=n[i],o=this._predicate.bind(r);let s=!1;for(let n=r.length;--n>=0;)o(n,r)&&(e&&!s&&(e(r),s=!0),t(n,r))}}count(){let t=0;const e=this._chunks,n=e.length;for(let r=-1;++r<n;){const n=e[r],i=this._predicate.bind(n);for(let e=-1,r=n.length;++e<r;)i(e,n)&&++t}return t}*[Symbol.iterator](){const t=this._chunks,e=t.length;for(let n=-1;++n<e;){const e=t[n],r=this._predicate.bind(e);for(let t=-1,n=e.length;++t<n;)r(t,e)&&(yield e.get(t))}}filter(t){return new tl(this._chunks,this._predicate.and(t))}countBy(t){const e=this._chunks,n=e.length,r="string"===typeof t?new $u(t):t;r.bind(e[n-1]);const i=r.vector;if(!Ot.isDictionary(i.type))throw new Error("countBy currently only supports dictionary-encoded columns");const o=Math.ceil(Math.log(i.length)/Math.log(256)),s=4==o?Uint32Array:o>=2?Uint16Array:Uint8Array,a=new s(i.dictionary.length);for(let c=-1;++c<n;){const t=e[c],n=this._predicate.bind(t);r.bind(t);const i=r.vector.indices;for(let e=-1,r=t.length;++e<r;){let r=i.get(e);null!==r&&n(e,t)&&a[r]++}}return new Qu(i.dictionary,la.from(a))}}L["a"].toDOMStream=Cu,be["throughDOM"]=Du,hu["throughDOM"]=Fu,Ms["throughDOM"]=Uu;
/**
 * @license
 * Copyright 2018-2021 Streamlit Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
var el,nl=function(){function t(t,e,n,r){var i=this;this.getCell=function(t,e){var n=t<i.headerRows&&e<i.headerColumns,r=t>=i.headerRows&&e<i.headerColumns,o=t<i.headerRows&&e>=i.headerColumns;if(n){var s=["blank"];return e>0&&s.push("level"+t),{type:"blank",classNames:s.join(" "),content:""}}if(o){var a=e-i.headerColumns;s=["col_heading","level"+t,"col"+a];return{type:"columns",classNames:s.join(" "),content:i.getContent(i.columnsTable,a,t)}}if(r){var c=t-i.headerRows;s=["row_heading","level"+e,"row"+c];return{type:"index",id:"T_"+i.uuid+"level"+e+"_row"+c,classNames:s.join(" "),content:i.getContent(i.indexTable,c,e)}}c=t-i.headerRows,a=e-i.headerColumns,s=["data","row"+c,"col"+a];var u=i.styler?i.getContent(i.styler.displayValuesTable,c,a):i.getContent(i.dataTable,c,a);return{type:"data",id:"T_"+i.uuid+"row"+c+"_col"+a,classNames:s.join(" "),content:u}},this.getContent=function(t,e,n){var r=t.getColumnAt(n);if(null===r)return"";var o=i.getColumnTypeId(t,n);switch(o){case F.Timestamp:return i.nanosToDate(r.get(e));default:return r.get(e)}},this.dataTable=au.from(t),this.indexTable=au.from(e),this.columnsTable=au.from(n),this.styler=r?{caption:r.caption,displayValuesTable:au.from(r.displayValues),styles:r.styles,uuid:r.uuid}:void 0}return Object.defineProperty(t.prototype,"rows",{get:function(){return this.indexTable.length+this.columnsTable.numCols},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"columns",{get:function(){return this.indexTable.numCols+this.columnsTable.length},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"headerRows",{get:function(){return this.rows-this.dataRows},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"headerColumns",{get:function(){return this.columns-this.dataColumns},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dataRows",{get:function(){return this.dataTable.length},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"dataColumns",{get:function(){return this.dataTable.numCols},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"uuid",{get:function(){return this.styler&&this.styler.uuid},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"caption",{get:function(){return this.styler&&this.styler.caption},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"styles",{get:function(){return this.styler&&this.styler.styles},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"table",{get:function(){return this.dataTable},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"index",{get:function(){return this.indexTable},enumerable:!0,configurable:!0}),Object.defineProperty(t.prototype,"columnTable",{get:function(){return this.columnsTable},enumerable:!0,configurable:!0}),t.prototype.serialize=function(){return{data:this.dataTable.serialize(),index:this.indexTable.serialize(),columns:this.columnsTable.serialize()}},t.prototype.getColumnTypeId=function(t,e){return t.schema.fields[e].type.typeId},t.prototype.nanosToDate=function(t){return new Date(t/1e6)},t}(),rl=function(){return rl=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n],e)Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t},rl.apply(this,arguments)};
/**
 * @license
 * Copyright 2018-2021 Streamlit Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */(function(t){t["COMPONENT_READY"]="streamlit:componentReady",t["SET_COMPONENT_VALUE"]="streamlit:setComponentValue",t["SET_FRAME_HEIGHT"]="streamlit:setFrameHeight"})(el||(el={}));var il=function(){function t(){}return t.API_VERSION=1,t.RENDER_EVENT="streamlit:render",t.events=new k,t.registeredMessageListener=!1,t.setComponentReady=function(){t.registeredMessageListener||(window.addEventListener("message",t.onMessageEvent),t.registeredMessageListener=!0),t.sendBackMsg(el.COMPONENT_READY,{apiVersion:t.API_VERSION})},t.setFrameHeight=function(e){void 0===e&&(e=document.body.scrollHeight),e!==t.lastFrameHeight&&(t.lastFrameHeight=e,t.sendBackMsg(el.SET_FRAME_HEIGHT,{height:e}))},t.setComponentValue=function(e){var n;e instanceof nl?(n="dataframe",e=e.serialize()):sl(e)?(n="bytes",e=new Uint8Array(e.buffer)):e instanceof ArrayBuffer?(n="bytes",e=new Uint8Array(e)):n="json",t.sendBackMsg(el.SET_COMPONENT_VALUE,{value:e,dataType:n})},t.onMessageEvent=function(e){var n=e.data["type"];switch(n){case t.RENDER_EVENT:t.onRenderMessage(e.data);break}},t.onRenderMessage=function(e){var n=e["args"];null==n&&(console.error("Got null args in onRenderMessage. This should never happen"),n={});var r=e["dfs"]&&e["dfs"].length>0?t.argsDataframeToObject(e["dfs"]):{};n=rl(rl({},n),r);var i=Boolean(e["disabled"]),o=e["theme"];o&&ol(o);var s={disabled:i,args:n,theme:o},a=new CustomEvent(t.RENDER_EVENT,{detail:s});t.events.dispatchEvent(a)},t.argsDataframeToObject=function(e){var n=e.map((function(e){var n=e.key,r=e.value;return[n,t.toArrowTable(r)]}));return Object.fromEntries(n)},t.toArrowTable=function(t){var e=t.data,n=e.data,r=e.index,i=e.columns,o=e.styler;return new nl(n,r,i,o)},t.sendBackMsg=function(t,e){window.parent.postMessage(rl({isStreamlitMessage:!0,type:t},e),"*")},t}(),ol=function(t){var e=document.createElement("style");document.head.appendChild(e),e.innerHTML="\n    :root {\n      --primary-color: "+t.primaryColor+";\n      --background-color: "+t.backgroundColor+";\n      --secondary-background-color: "+t.secondaryBackgroundColor+";\n      --text-color: "+t.textColor+";\n      --font: "+t.font+";\n    }\n\n    body {\n      background-color: var(--background-color);\n      color: var(--text-color);\n    }\n  "};function sl(t){var e=!1;try{e=t instanceof BigInt64Array||t instanceof BigUint64Array}catch(n){}return t instanceof Int8Array||t instanceof Uint8Array||t instanceof Uint8ClampedArray||t instanceof Int16Array||t instanceof Uint16Array||t instanceof Int32Array||t instanceof Uint32Array||t instanceof Float32Array||t instanceof Float64Array||e}
/**
 * @license
 * Copyright 2018-2021 Streamlit Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */var al=function(){var t=function(e,n){return t=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])},t(e,n)};return function(e,n){function r(){this.constructor=e}t(e,n),e.prototype=null===n?Object.create(n):(r.prototype=n.prototype,new r)}}();(function(t){function e(){return null!==t&&t.apply(this,arguments)||this}al(e,t),e.prototype.componentDidMount=function(){il.setFrameHeight()},e.prototype.componentDidUpdate=function(){il.setFrameHeight()}})(u.a.PureComponent)},d1e7:function(t,e,n){"use strict";var r={}.propertyIsEnumerable,i=Object.getOwnPropertyDescriptor,o=i&&!r.call({1:2},1);e.f=o?function(t){var e=i(this,t);return!!e&&e.enumerable}:r},d2bb:function(t,e,n){var r=n("e330"),i=n("825a"),o=n("3bbe");t.exports=Object.setPrototypeOf||("__proto__"in{}?function(){var t,e=!1,n={};try{t=r(Object.getOwnPropertyDescriptor(Object.prototype,"__proto__").set),t(n,[]),e=n instanceof Array}catch(s){}return function(n,r){return i(n),o(r),e?t(n,r):n.__proto__=r,n}}():void 0)},d3b7:function(t,e,n){var r=n("00ee"),i=n("6eeb"),o=n("b041");r||i(Object.prototype,"toString",o,{unsafe:!0})},d44e:function(t,e,n){var r=n("9bf2").f,i=n("1a2d"),o=n("b622"),s=o("toStringTag");t.exports=function(t,e,n){t&&!i(t=n?t:t.prototype,s)&&r(t,s,{configurable:!0,value:e})}},d4c3:function(t,e,n){var r=n("342f"),i=n("da84");t.exports=/ipad|iphone|ipod/i.test(r)&&void 0!==i.Pebble},d9b5:function(t,e,n){var r=n("da84"),i=n("d066"),o=n("1626"),s=n("3a9b"),a=n("fdbf"),c=r.Object;t.exports=a?function(t){return"symbol"==typeof t}:function(t){var e=i("Symbol");return o(e)&&s(e.prototype,c(t))}},da6a:function(t,e,n){"use strict";var r=n("4cec"),i={childContextTypes:!0,contextType:!0,contextTypes:!0,defaultProps:!0,displayName:!0,getDefaultProps:!0,getDerivedStateFromError:!0,getDerivedStateFromProps:!0,mixins:!0,propTypes:!0,type:!0},o={name:!0,length:!0,prototype:!0,caller:!0,callee:!0,arguments:!0,arity:!0},s={$$typeof:!0,render:!0,defaultProps:!0,displayName:!0,propTypes:!0},a={$$typeof:!0,compare:!0,defaultProps:!0,displayName:!0,propTypes:!0,type:!0},c={};function u(t){return r.isMemo(t)?a:c[t["$$typeof"]]||i}c[r.ForwardRef]=s,c[r.Memo]=a;var l=Object.defineProperty,f=Object.getOwnPropertyNames,d=Object.getOwnPropertySymbols,h=Object.getOwnPropertyDescriptor,p=Object.getPrototypeOf,y=Object.prototype;function v(t,e,n){if("string"!==typeof e){if(y){var r=p(e);r&&r!==y&&v(t,r,n)}var i=f(e);d&&(i=i.concat(d(e)));for(var s=u(t),a=u(e),c=0;c<i.length;++c){var b=i[c];if(!o[b]&&(!n||!n[b])&&(!a||!a[b])&&(!s||!s[b])){var m=h(e,b);try{l(t,b,m)}catch(g){}}}}return t}t.exports=v},da84:function(t,e,n){(function(e){var n=function(t){return t&&t.Math==Math&&t};t.exports=n("object"==typeof globalThis&&globalThis)||n("object"==typeof window&&window)||n("object"==typeof self&&self)||n("object"==typeof e&&e)||function(){return this}()||Function("return this")()}).call(this,n("c8ba"))},dac5:function(t,e,n){!function(e,n){t.exports=n()}("undefined"!=typeof self&&self,(function(){return function(t){var e={};function n(r){if(e[r])return e[r].exports;var i=e[r]={i:r,l:!1,exports:{}};return t[r].call(i.exports,i,i.exports,n),i.l=!0,i.exports}return n.m=t,n.c=e,n.d=function(t,e,r){n.o(t,e)||Object.defineProperty(t,e,{enumerable:!0,get:r})},n.r=function(t){"undefined"!=typeof Symbol&&Symbol.toStringTag&&Object.defineProperty(t,Symbol.toStringTag,{value:"Module"}),Object.defineProperty(t,"__esModule",{value:!0})},n.t=function(t,e){if(1&e&&(t=n(t)),8&e)return t;if(4&e&&"object"==typeof t&&t&&t.__esModule)return t;var r=Object.create(null);if(n.r(r),Object.defineProperty(r,"default",{enumerable:!0,value:t}),2&e&&"string"!=typeof t)for(var i in t)n.d(r,i,function(e){return t[e]}.bind(null,i));return r},n.n=function(t){var e=t&&t.__esModule?function(){return t.default}:function(){return t};return n.d(e,"a",e),e},n.o=function(t,e){return Object.prototype.hasOwnProperty.call(t,e)},n.p="",n(n.s=2)}([function(t,e,n){"use strict";Object.defineProperty(e,"__esModule",{value:!0}),e.SequentialHook=e.Hook=e.PreventableBaklavaEvent=e.BaklavaEvent=void 0;var r=n(1),i=function(){function t(){this.listeners=new Map}return t.prototype.addListener=function(t,e){this.listeners.set(t,e)},t.prototype.removeListener=function(t){this.listeners.has(t)&&this.listeners.delete(t)},t.prototype.emit=function(t){this.listeners.forEach((function(e){return e(t)}))},t}();e.BaklavaEvent=i;var o=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return r.__extends(e,t),e.prototype.emit=function(t){var e,n;try{for(var i=r.__values(Array.from(this.listeners.values())),o=i.next();!o.done;o=i.next())if(!1===(0,o.value)(t))return!0}catch(t){e={error:t}}finally{try{o&&!o.done&&(n=i.return)&&n.call(i)}finally{if(e)throw e.error}}return!1},e}(i);e.PreventableBaklavaEvent=o;var s=function(){function t(){this.tapMap=new Map,this.taps=[]}return t.prototype.tap=function(t,e){this.tapMap.has(t)&&this.untap(t),this.tapMap.set(t,e),this.taps.push(e)},t.prototype.untap=function(t){if(this.tapMap.has(t)){var e=this.tapMap.get(t);this.tapMap.delete(t);var n=this.taps.indexOf(e);n>=0&&this.taps.splice(n,1)}},t}();e.Hook=s;var a=function(t){function e(){return null!==t&&t.apply(this,arguments)||this}return r.__extends(e,t),e.prototype.execute=function(t){var e,n,i=t;try{for(var o=r.__values(this.taps),s=o.next();!s.done;s=o.next())i=(0,s.value)(i)}catch(t){e={error:t}}finally{try{s&&!s.done&&(n=o.return)&&n.call(o)}finally{if(e)throw e.error}}return i},e}(s);e.SequentialHook=a},function(t,e,n){"use strict";n.r(e),n.d(e,"__extends",(function(){return i})),n.d(e,"__assign",(function(){return o})),n.d(e,"__rest",(function(){return s})),n.d(e,"__decorate",(function(){return a})),n.d(e,"__param",(function(){return c})),n.d(e,"__metadata",(function(){return u})),n.d(e,"__awaiter",(function(){return l})),n.d(e,"__generator",(function(){return f})),n.d(e,"__createBinding",(function(){return d})),n.d(e,"__exportStar",(function(){return h})),n.d(e,"__values",(function(){return p})),n.d(e,"__read",(function(){return y})),n.d(e,"__spread",(function(){return v})),n.d(e,"__spreadArrays",(function(){return b})),n.d(e,"__await",(function(){return m})),n.d(e,"__asyncGenerator",(function(){return g})),n.d(e,"__asyncDelegator",(function(){return _})),n.d(e,"__asyncValues",(function(){return w})),n.d(e,"__makeTemplateObject",(function(){return O})),n.d(e,"__importStar",(function(){return S})),n.d(e,"__importDefault",(function(){return x})),n.d(e,"__classPrivateFieldGet",(function(){return I})),n.d(e,"__classPrivateFieldSet",(function(){return A}));
/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
var r=function(t,e){return(r=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(t,e){t.__proto__=e}||function(t,e){for(var n in e)e.hasOwnProperty(n)&&(t[n]=e[n])})(t,e)};function i(t,e){function n(){this.constructor=t}r(t,e),t.prototype=null===e?Object.create(e):(n.prototype=e.prototype,new n)}var o=function(){return(o=Object.assign||function(t){for(var e,n=1,r=arguments.length;n<r;n++)for(var i in e=arguments[n])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t}).apply(this,arguments)};function s(t,e){var n={};for(var r in t)Object.prototype.hasOwnProperty.call(t,r)&&e.indexOf(r)<0&&(n[r]=t[r]);if(null!=t&&"function"==typeof Object.getOwnPropertySymbols){var i=0;for(r=Object.getOwnPropertySymbols(t);i<r.length;i++)e.indexOf(r[i])<0&&Object.prototype.propertyIsEnumerable.call(t,r[i])&&(n[r[i]]=t[r[i]])}return n}function a(t,e,n,r){var i,o=arguments.length,s=o<3?e:null===r?r=Object.getOwnPropertyDescriptor(e,n):r;if("object"==typeof Reflect&&"function"==typeof Reflect.decorate)s=Reflect.decorate(t,e,n,r);else for(var a=t.length-1;a>=0;a--)(i=t[a])&&(s=(o<3?i(s):o>3?i(e,n,s):i(e,n))||s);return o>3&&s&&Object.defineProperty(e,n,s),s}function c(t,e){return function(n,r){e(n,r,t)}}function u(t,e){if("object"==typeof Reflect&&"function"==typeof Reflect.metadata)return Reflect.metadata(t,e)}function l(t,e,n,r){return new(n||(n=Promise))((function(i,o){function s(t){try{c(r.next(t))}catch(t){o(t)}}function a(t){try{c(r.throw(t))}catch(t){o(t)}}function c(t){var e;t.done?i(t.value):(e=t.value,e instanceof n?e:new n((function(t){t(e)}))).then(s,a)}c((r=r.apply(t,e||[])).next())}))}function f(t,e){var n,r,i,o,s={label:0,sent:function(){if(1&i[0])throw i[1];return i[1]},trys:[],ops:[]};return o={next:a(0),throw:a(1),return:a(2)},"function"==typeof Symbol&&(o[Symbol.iterator]=function(){return this}),o;function a(o){return function(a){return function(o){if(n)throw new TypeError("Generator is already executing.");for(;s;)try{if(n=1,r&&(i=2&o[0]?r.return:o[0]?r.throw||((i=r.return)&&i.call(r),0):r.next)&&!(i=i.call(r,o[1])).done)return i;switch(r=0,i&&(o=[2&o[0],i.value]),o[0]){case 0:case 1:i=o;break;case 4:return s.label++,{value:o[1],done:!1};case 5:s.label++,r=o[1],o=[0];continue;case 7:o=s.ops.pop(),s.trys.pop();continue;default:if(i=s.trys,!((i=i.length>0&&i[i.length-1])||6!==o[0]&&2!==o[0])){s=0;continue}if(3===o[0]&&(!i||o[1]>i[0]&&o[1]<i[3])){s.label=o[1];break}if(6===o[0]&&s.label<i[1]){s.label=i[1],i=o;break}if(i&&s.label<i[2]){s.label=i[2],s.ops.push(o);break}i[2]&&s.ops.pop(),s.trys.pop();continue}o=e.call(t,s)}catch(t){o=[6,t],r=0}finally{n=i=0}if(5&o[0])throw o[1];return{value:o[0]?o[1]:void 0,done:!0}}([o,a])}}}function d(t,e,n,r){void 0===r&&(r=n),t[r]=e[n]}function h(t,e){for(var n in t)"default"===n||e.hasOwnProperty(n)||(e[n]=t[n])}function p(t){var e="function"==typeof Symbol&&Symbol.iterator,n=e&&t[e],r=0;if(n)return n.call(t);if(t&&"number"==typeof t.length)return{next:function(){return t&&r>=t.length&&(t=void 0),{value:t&&t[r++],done:!t}}};throw new TypeError(e?"Object is not iterable.":"Symbol.iterator is not defined.")}function y(t,e){var n="function"==typeof Symbol&&t[Symbol.iterator];if(!n)return t;var r,i,o=n.call(t),s=[];try{for(;(void 0===e||e-- >0)&&!(r=o.next()).done;)s.push(r.value)}catch(t){i={error:t}}finally{try{r&&!r.done&&(n=o.return)&&n.call(o)}finally{if(i)throw i.error}}return s}function v(){for(var t=[],e=0;e<arguments.length;e++)t=t.concat(y(arguments[e]));return t}function b(){for(var t=0,e=0,n=arguments.length;e<n;e++)t+=arguments[e].length;var r=Array(t),i=0;for(e=0;e<n;e++)for(var o=arguments[e],s=0,a=o.length;s<a;s++,i++)r[i]=o[s];return r}function m(t){return this instanceof m?(this.v=t,this):new m(t)}function g(t,e,n){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var r,i=n.apply(t,e||[]),o=[];return r={},s("next"),s("throw"),s("return"),r[Symbol.asyncIterator]=function(){return this},r;function s(t){i[t]&&(r[t]=function(e){return new Promise((function(n,r){o.push([t,e,n,r])>1||a(t,e)}))})}function a(t,e){try{(n=i[t](e)).value instanceof m?Promise.resolve(n.value.v).then(c,u):l(o[0][2],n)}catch(t){l(o[0][3],t)}var n}function c(t){a("next",t)}function u(t){a("throw",t)}function l(t,e){t(e),o.shift(),o.length&&a(o[0][0],o[0][1])}}function _(t){var e,n;return e={},r("next"),r("throw",(function(t){throw t})),r("return"),e[Symbol.iterator]=function(){return this},e;function r(r,i){e[r]=t[r]?function(e){return(n=!n)?{value:m(t[r](e)),done:"return"===r}:i?i(e):e}:i}}function w(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e,n=t[Symbol.asyncIterator];return n?n.call(t):(t=p(t),e={},r("next"),r("throw"),r("return"),e[Symbol.asyncIterator]=function(){return this},e);function r(n){e[n]=t[n]&&function(e){return new Promise((function(r,i){(function(t,e,n,r){Promise.resolve(r).then((function(e){t({value:e,done:n})}),e)})(r,i,(e=t[n](e)).done,e.value)}))}}}function O(t,e){return Object.defineProperty?Object.defineProperty(t,"raw",{value:e}):t.raw=e,t}function S(t){if(t&&t.__esModule)return t;var e={};if(null!=t)for(var n in t)Object.hasOwnProperty.call(t,n)&&(e[n]=t[n]);return e.default=t,e}function x(t){return t&&t.__esModule?t:{default:t}}function I(t,e){if(!e.has(t))throw new TypeError("attempted to get private field on non-instance");return e.get(t)}function A(t,e,n){if(!e.has(t))throw new TypeError("attempted to set private field on non-instance");return e.set(t,n),n}},function(t,e,n){"use strict";n.r(e),n.d(e,"Connection",(function(){return s})),n.d(e,"DummyConnection",(function(){return a})),n.d(e,"Editor",(function(){return u})),n.d(e,"Node",(function(){return d})),n.d(e,"NodeBuilder",(function(){return p})),n.d(e,"NodeInterface",(function(){return l})),n.d(e,"NodeOption",(function(){return f}));var r=0;function i(){return Date.now().toString()+(r++).toString()}var o=n(0),s=function(){function t(t,e){if(this.isInDanger=!1,this.destructed=!1,this.events={destruct:new o.BaklavaEvent},!t||!e)throw new Error("Cannot initialize connection with null/undefined for 'from' or 'to' values");this.id=i(),this.from=t,this.to=e,this.from.connectionCount++,this.to.connectionCount++}return t.prototype.destruct=function(){this.events.destruct.emit(),this.from.connectionCount--,this.to.connectionCount--,this.destructed=!0},t}(),a=function(t,e){if(!t||!e)throw new Error("Cannot initialize connection with null/undefined for 'from' or 'to' values");this.id=i(),this.from=t,this.to=e},c=n(1),u=function(){function t(){this._plugins=new Set,this._nodes=[],this._connections=[],this._nodeTypes=new Map,this._nodeCategories=new Map([["default",[]]]),this.events={beforeRegisterNodeType:new o.PreventableBaklavaEvent,registerNodeType:new o.BaklavaEvent,beforeAddNode:new o.PreventableBaklavaEvent,addNode:new o.BaklavaEvent,beforeRemoveNode:new o.PreventableBaklavaEvent,removeNode:new o.BaklavaEvent,beforeAddConnection:new o.PreventableBaklavaEvent,addConnection:new o.BaklavaEvent,checkConnection:new o.PreventableBaklavaEvent,beforeRemoveConnection:new o.PreventableBaklavaEvent,removeConnection:new o.BaklavaEvent,beforeUsePlugin:new o.PreventableBaklavaEvent,usePlugin:new o.BaklavaEvent},this.hooks={save:new o.SequentialHook,load:new o.SequentialHook}}return Object.defineProperty(t.prototype,"nodes",{get:function(){return this._nodes},enumerable:!1,configurable:!0}),Object.defineProperty(t.prototype,"connections",{get:function(){return this._connections},enumerable:!1,configurable:!0}),Object.defineProperty(t.prototype,"nodeTypes",{get:function(){return this._nodeTypes},enumerable:!1,configurable:!0}),Object.defineProperty(t.prototype,"nodeCategories",{get:function(){return this._nodeCategories},enumerable:!1,configurable:!0}),Object.defineProperty(t.prototype,"plugins",{get:function(){return this._plugins},enumerable:!1,configurable:!0}),t.prototype.registerNodeType=function(t,e,n){void 0===n&&(n="default"),this.events.beforeRegisterNodeType.emit({typeName:t,type:e,category:n})||(this._nodeTypes.set(t,e),this.nodeCategories.has(n)||this._nodeCategories.set(n,[]),this.nodeCategories.get(n).push(t),this.events.registerNodeType.emit({typeName:t,type:e,category:n}))},t.prototype.addNode=function(t){if(!this.events.beforeAddNode.emit(t))return t.registerEditor(this),this._nodes.push(t),this.events.addNode.emit(t),t},t.prototype.removeNode=function(t){var e=this;if(this.nodes.includes(t)){if(this.events.beforeRemoveNode.emit(t))return;this.connections.filter((function(e){return e.from.parent===t||e.to.parent===t})).forEach((function(t){return e.removeConnection(t)})),this._nodes.splice(this.nodes.indexOf(t),1),this.events.removeNode.emit(t)}},t.prototype.addConnection=function(t,e){var n=this.checkConnection(t,e);if(n&&!this.events.beforeAddConnection.emit({from:t,to:e})){var r=new s(n.from,n.to);return this._connections.push(r),this.events.addConnection.emit(r),r}},t.prototype.removeConnection=function(t){if(this.connections.includes(t)){if(this.events.beforeRemoveConnection.emit(t))return;t.destruct(),this._connections.splice(this.connections.indexOf(t),1),this.events.removeConnection.emit(t)}},t.prototype.checkConnection=function(t,e){if(!t||!e)return!1;if(t.parent===e.parent)return!1;if(t.isInput&&!e.isInput){var n=t;t=e,e=n}return!(t.isInput||!e.isInput)&&!this.connections.some((function(n){return n.from===t&&n.to===e}))&&!this.events.checkConnection.emit({from:t,to:e})&&new a(t,e)},t.prototype.load=function(t){for(var e,n,r,i,o=[],s=this.connections.length-1;s>=0;s--)this.removeConnection(this.connections[s]);for(s=this.nodes.length-1;s>=0;s--)this.removeNode(this.nodes[s]);try{for(var a=Object(c.__values)(t.nodes),u=a.next();!u.done;u=a.next()){var l=u.value,f=this.nodeTypes.get(l.type);if(f){var d=new f;this.addNode(d),d.load(l)}else o.push("Node type "+l.type+" is not registered")}}catch(t){e={error:t}}finally{try{u&&!u.done&&(n=a.return)&&n.call(a)}finally{if(e)throw e.error}}try{for(var h=Object(c.__values)(t.connections),p=h.next();!p.done;p=h.next()){var y=p.value,v=this.findNodeInterface(y.from),b=this.findNodeInterface(y.to);if(v)if(b){var m=this.addConnection(v,b);m?m.id=y.id:o.push("Unable to create connection from "+y.from+" to "+y.to)}else o.push("Could not find interface with id "+y.to);else o.push("Could not find interface with id "+y.from)}}catch(t){r={error:t}}finally{try{p&&!p.done&&(i=h.return)&&i.call(h)}finally{if(r)throw r.error}}return this.hooks.load.execute(t),o.forEach((function(t){return console.warn(t)})),o},t.prototype.save=function(){var t={nodes:this.nodes.map((function(t){return t.save()})),connections:this.connections.map((function(t){return{id:t.id,from:t.from.id,to:t.to.id}}))};return this.hooks.save.execute(t)},t.prototype.use=function(t){return!this.events.beforeUsePlugin.emit(t)&&(this._plugins.add(t),t.register(this),this.events.usePlugin.emit(t),!0)},t.prototype.generateId=function(t){return void 0===t&&(t=""),t+i()},t.prototype.findNodeInterface=function(t){var e,n,r,i;try{for(var o=Object(c.__values)(this.nodes),s=o.next();!s.done;s=o.next()){var a=s.value;try{for(var u=(r=void 0,Object(c.__values)(a.interfaces.keys())),l=u.next();!l.done;l=u.next()){var f=l.value;if(a.interfaces.get(f).id===t)return a.interfaces.get(f)}}catch(t){r={error:t}}finally{try{l&&!l.done&&(i=u.return)&&i.call(u)}finally{if(r)throw r.error}}}}catch(t){e={error:t}}finally{try{s&&!s.done&&(n=o.return)&&n.call(o)}finally{if(e)throw e.error}}},t}(),l=function(){function t(t,e){this.events={setConnectionCount:new o.BaklavaEvent,beforeSetValue:new o.PreventableBaklavaEvent,setValue:new o.BaklavaEvent,updated:new o.BaklavaEvent},this.hooks={load:new o.SequentialHook,save:new o.SequentialHook},this._connectionCount=0,this._value=null,this.parent=t,this.isInput=e,this.id="ni_"+i()}return Object.defineProperty(t.prototype,"connectionCount",{get:function(){return this._connectionCount},set:function(t){this._connectionCount=t,this.events.setConnectionCount.emit(t)},enumerable:!1,configurable:!0}),Object.defineProperty(t.prototype,"value",{get:function(){return this._value},set:function(t){this.events.beforeSetValue.emit(t)||(this._value=t,this.events.setValue.emit(t))},enumerable:!1,configurable:!0}),t.prototype.load=function(t){this.id=t.id,this.value=t.value,this.hooks.load.execute(t)},t.prototype.save=function(){var t={id:this.id,value:this.value};return this.hooks.save.execute(t)},t}(),f=function(){function t(t,e,n){this.events={beforeSetValue:new o.PreventableBaklavaEvent,setValue:new o.BaklavaEvent,updated:new o.BaklavaEvent},this.optionComponent=t,this.sidebarComponent=n,this._value=e}return Object.defineProperty(t.prototype,"value",{get:function(){return this._value},set:function(t){this.events.beforeSetValue.emit(t)||(this._value=t,this.events.setValue.emit(t))},enumerable:!1,configurable:!0}),t}(),d=function(){function t(){this.id="node_"+i(),this.interfaces=new Map,this.options=new Map,this.state={},this.events={beforeAddInterface:new o.PreventableBaklavaEvent,addInterface:new o.BaklavaEvent,beforeRemoveInterface:new o.PreventableBaklavaEvent,removeInterface:new o.BaklavaEvent,beforeAddOption:new o.PreventableBaklavaEvent,addOption:new o.BaklavaEvent,beforeRemoveOption:new o.PreventableBaklavaEvent,removeOption:new o.BaklavaEvent,update:new o.BaklavaEvent},this.hooks={load:new o.SequentialHook,save:new o.SequentialHook}}return Object.defineProperty(t.prototype,"inputInterfaces",{get:function(){var t={};return this.interfaces.forEach((function(e,n){e.isInput&&(t[n]=e)})),t},enumerable:!1,configurable:!0}),Object.defineProperty(t.prototype,"outputInterfaces",{get:function(){var t={};return this.interfaces.forEach((function(e,n){e.isInput||(t[n]=e)})),t},enumerable:!1,configurable:!0}),t.prototype.load=function(t){var e=this;this.id=t.id,this.name=t.name,this.state=t.state,t.options.forEach((function(t){var n=Object(c.__read)(t,2),r=n[0],i=n[1];e.options.has(r)&&(e.options.get(r).value=i)})),t.interfaces.forEach((function(t){var n=Object(c.__read)(t,2),r=n[0],i=n[1];e.interfaces.has(r)&&e.interfaces.get(r).load(i)})),this.hooks.load.execute(t)},t.prototype.save=function(){var t={type:this.type,id:this.id,name:this.name,options:Array.from(this.options.entries()).map((function(t){var e=Object(c.__read)(t,2);return[e[0],e[1].value]})),state:this.state,interfaces:Array.from(this.interfaces.entries()).map((function(t){var e=Object(c.__read)(t,2);return[e[0],e[1].save()]}))};return this.hooks.save.execute(t)},t.prototype.calculate=function(t){},t.prototype.addInputInterface=function(t,e,n,r){var i=this;if(void 0===n&&(n=null),!this.events.beforeAddInterface.emit({name:t,isInput:!0,option:e,defaultValue:n})){var o=this.addInterface(!0,t,e);return o.events.setValue.addListener(this,(function(){return i.events.update.emit({name:t,interface:o})})),o.value=n,Object.entries(r||{}).forEach((function(t){var e=Object(c.__read)(t,2),n=e[0],r=e[1];o[n]=r})),this.events.addInterface.emit(o),o}},t.prototype.addOutputInterface=function(t,e){if(!this.events.beforeAddInterface.emit({name:t,isInput:!1})){var n=this.addInterface(!1,t);return Object.entries(e||{}).forEach((function(t){var e=Object(c.__read)(t,2),r=e[0],i=e[1];n[r]=i})),this.events.addInterface.emit(n),n}},t.prototype.removeInterface=function(t){var e=this,n=this.getInterface(t);if(n){if(this.events.beforeRemoveInterface.emit(n))return;if(n.connectionCount>0){if(!this.editorInstance)throw new Error("Interface is connected, but no editor instance is specified. Unable to delete interface");this.editorInstance.connections.filter((function(t){return t.from===n||t.to===n})).forEach((function(t){e.editorInstance.removeConnection(t)}))}this.interfaces.delete(t),this.events.removeInterface.emit(n)}},t.prototype.addOption=function(t,e,n,r,i){var o=this;if(void 0===n&&(n=null),!this.events.beforeAddOption.emit({name:t,component:e,defaultValue:n,sidebarComponent:r})){var s=new f(e,n,r);return Object.entries(i||{}).forEach((function(t){var e=Object(c.__read)(t,2),n=e[0],r=e[1];s[n]=r})),s.events.setValue.addListener(this,(function(){o.events.update.emit({name:t,option:s})})),this.options.set(t,s),this.events.addOption.emit({name:t,option:s}),s}},t.prototype.removeOption=function(t){if(this.options.has(t)){var e=this.options.get(t);if(this.events.beforeRemoveOption.emit({name:t,option:e}))return;this.options.delete(t),this.events.removeOption.emit({name:t,option:e})}},t.prototype.getInterface=function(t){if(!this.interfaces.has(t))throw new Error("No interface named '"+t+"'");return this.interfaces.get(t)},t.prototype.getOptionValue=function(t){if(!this.options.has(t))throw new Error("No option named '"+t+"'");return this.options.get(t).value},t.prototype.setOptionValue=function(t,e){if(!this.options.has(t))throw new Error("No option named '"+t+"'");this.options.get(t).value=e},t.prototype.registerEditor=function(t){this.editorInstance=t},t.prototype.addInterface=function(t,e,n){var r=new l(this,t);return r.option=n,this.interfaces.set(e,r),r},t}();function h(t){return"function"==typeof t?t():t}var p=function(){function t(t,e){this.type="",this.name="",this.intfs=[],this.options=new Map,this.type=t,this.name=t,this.additionalProperties=e}return t.prototype.build=function(){return t=this.type,e=this.name,n=this.additionalProperties,r=this.intfs,i=this.options,o=this.calcFunction,function(s){function a(){var o,a,u=s.call(this)||this;u.type=t,u.name=e,n&&Object.assign(u,n);try{for(var l=Object(c.__values)(r),f=l.next();!f.done;f=l.next()){var d=f.value;d.isInput?u.addInputInterface(d.name,d.option,h(d.defaultValue),d.additionalProperties):u.addOutputInterface(d.name,d.additionalProperties)}}catch(t){o={error:t}}finally{try{f&&!f.done&&(a=l.return)&&a.call(l)}finally{if(o)throw o.error}}return Array.from(i.entries()).forEach((function(t){var e=Object(c.__read)(t,2),n=e[0],r=e[1];u.addOption(n,r.optionComponent,h(r.value),r.sidebarComponent,r.additionalProperties)})),u}return Object(c.__extends)(a,s),a.prototype.calculate=function(t){if(o)return o.call(this,this,t)},a}(d);var t,e,n,r,i,o},t.prototype.setName=function(t){return this.name=t,this},t.prototype.addInputInterface=function(t,e,n,r){return void 0===n&&(n=null),this.checkDefaultValue(n),this.intfs.push({isInput:!0,name:t,option:e,defaultValue:n,additionalProperties:r}),this},t.prototype.addOutputInterface=function(t,e){return this.intfs.push({isInput:!1,name:t,additionalProperties:e}),this},t.prototype.addOption=function(t,e,n,r,i){return void 0===n&&(n=null),this.checkDefaultValue(n),this.options.set(t,{value:n,optionComponent:e,sidebarComponent:r,additionalProperties:i}),this},t.prototype.onCalculate=function(t){return this.calcFunction=t,this},t.prototype.checkDefaultValue=function(t){if("object"==typeof t&&null!==t)throw new Error("If the default value is an object, provide a generator function instead of the object")},t}()}])}))},dc4a:function(t,e,n){var r=n("59ed");t.exports=function(t,e){var n=t[e];return null==n?void 0:r(n)}},df75:function(t,e,n){var r=n("ca84"),i=n("7839");t.exports=Object.keys||function(t){return r(t,i)}},e163:function(t,e,n){var r=n("da84"),i=n("1a2d"),o=n("1626"),s=n("7b0b"),a=n("f772"),c=n("e177"),u=a("IE_PROTO"),l=r.Object,f=l.prototype;t.exports=c?l.getPrototypeOf:function(t){var e=s(t);if(i(e,u))return e[u];var n=e.constructor;return o(n)&&e instanceof n?n.prototype:e instanceof l?f:null}},e177:function(t,e,n){var r=n("d039");t.exports=!r((function(){function t(){}return t.prototype.constructor=null,Object.getPrototypeOf(new t)!==t.prototype}))},e260:function(t,e,n){"use strict";var r=n("fc6a"),i=n("44d2"),o=n("3f8c"),s=n("69f3"),a=n("7dd0"),c="Array Iterator",u=s.set,l=s.getterFor(c);t.exports=a(Array,"Array",(function(t,e){u(this,{type:c,target:r(t),index:0,kind:e})}),(function(){var t=l(this),e=t.target,n=t.kind,r=t.index++;return!e||r>=e.length?(t.target=void 0,{value:void 0,done:!0}):"keys"==n?{value:r,done:!1}:"values"==n?{value:e[r],done:!1}:{value:[r,e[r]],done:!1}}),"values"),o.Arguments=o.Array,i("keys"),i("values"),i("entries")},e2cc:function(t,e,n){var r=n("6eeb");t.exports=function(t,e,n){for(var i in e)r(t,i,e[i],n);return t}},e330:function(t,e){var n=Function.prototype,r=n.bind,i=n.call,o=r&&r.bind(i);t.exports=r?function(t){return t&&o(i,t)}:function(t){return t&&function(){return i.apply(t,arguments)}}},e3db:function(t,e){var n={}.toString;t.exports=Array.isArray||function(t){return"[object Array]"==n.call(t)}},e667:function(t,e){t.exports=function(t){try{return{error:!1,value:t()}}catch(e){return{error:!0,value:e}}}},e6cf:function(t,e,n){"use strict";var r,i,o,s,a=n("23e7"),c=n("c430"),u=n("da84"),l=n("d066"),f=n("c65b"),d=n("fea9"),h=n("6eeb"),p=n("e2cc"),y=n("d2bb"),v=n("d44e"),b=n("2626"),m=n("59ed"),g=n("1626"),_=n("861d"),w=n("19aa"),O=n("8925"),S=n("2266"),x=n("1c7e"),I=n("4840"),A=n("2cf4").set,T=n("b575"),j=n("cdf9"),E=n("44de"),B=n("f069"),C=n("e667"),M=n("69f3"),k=n("94ca"),D=n("b622"),N=n("6069"),L=n("605d"),P=n("2d00"),F=D("species"),U="Promise",R=M.getterFor(U),V=M.set,$=M.getterFor(U),z=d&&d.prototype,Y=d,H=z,W=u.TypeError,G=u.document,q=u.process,K=B.f,X=K,J=!!(G&&G.createEvent&&u.dispatchEvent),Z=g(u.PromiseRejectionEvent),Q="unhandledrejection",tt="rejectionhandled",et=0,nt=1,rt=2,it=1,ot=2,st=!1,at=k(U,(function(){var t=O(Y),e=t!==String(Y);if(!e&&66===P)return!0;if(c&&!H["finally"])return!0;if(P>=51&&/native code/.test(t))return!1;var n=new Y((function(t){t(1)})),r=function(t){t((function(){}),(function(){}))},i=n.constructor={};return i[F]=r,st=n.then((function(){}))instanceof r,!st||!e&&N&&!Z})),ct=at||!x((function(t){Y.all(t)["catch"]((function(){}))})),ut=function(t){var e;return!(!_(t)||!g(e=t.then))&&e},lt=function(t,e){if(!t.notified){t.notified=!0;var n=t.reactions;T((function(){var r=t.value,i=t.state==nt,o=0;while(n.length>o){var s,a,c,u=n[o++],l=i?u.ok:u.fail,d=u.resolve,h=u.reject,p=u.domain;try{l?(i||(t.rejection===ot&&pt(t),t.rejection=it),!0===l?s=r:(p&&p.enter(),s=l(r),p&&(p.exit(),c=!0)),s===u.promise?h(W("Promise-chain cycle")):(a=ut(s))?f(a,s,d,h):d(s)):h(r)}catch(y){p&&!c&&p.exit(),h(y)}}t.reactions=[],t.notified=!1,e&&!t.rejection&&dt(t)}))}},ft=function(t,e,n){var r,i;J?(r=G.createEvent("Event"),r.promise=e,r.reason=n,r.initEvent(t,!1,!0),u.dispatchEvent(r)):r={promise:e,reason:n},!Z&&(i=u["on"+t])?i(r):t===Q&&E("Unhandled promise rejection",n)},dt=function(t){f(A,u,(function(){var e,n=t.facade,r=t.value,i=ht(t);if(i&&(e=C((function(){L?q.emit("unhandledRejection",r,n):ft(Q,n,r)})),t.rejection=L||ht(t)?ot:it,e.error))throw e.value}))},ht=function(t){return t.rejection!==it&&!t.parent},pt=function(t){f(A,u,(function(){var e=t.facade;L?q.emit("rejectionHandled",e):ft(tt,e,t.value)}))},yt=function(t,e,n){return function(r){t(e,r,n)}},vt=function(t,e,n){t.done||(t.done=!0,n&&(t=n),t.value=e,t.state=rt,lt(t,!0))},bt=function(t,e,n){if(!t.done){t.done=!0,n&&(t=n);try{if(t.facade===e)throw W("Promise can't be resolved itself");var r=ut(e);r?T((function(){var n={done:!1};try{f(r,e,yt(bt,n,t),yt(vt,n,t))}catch(i){vt(n,i,t)}})):(t.value=e,t.state=nt,lt(t,!1))}catch(i){vt({done:!1},i,t)}}};if(at&&(Y=function(t){w(this,H),m(t),f(r,this);var e=R(this);try{t(yt(bt,e),yt(vt,e))}catch(n){vt(e,n)}},H=Y.prototype,r=function(t){V(this,{type:U,done:!1,notified:!1,parent:!1,reactions:[],rejection:!1,state:et,value:void 0})},r.prototype=p(H,{then:function(t,e){var n=$(this),r=n.reactions,i=K(I(this,Y));return i.ok=!g(t)||t,i.fail=g(e)&&e,i.domain=L?q.domain:void 0,n.parent=!0,r[r.length]=i,n.state!=et&&lt(n,!1),i.promise},catch:function(t){return this.then(void 0,t)}}),i=function(){var t=new r,e=R(t);this.promise=t,this.resolve=yt(bt,e),this.reject=yt(vt,e)},B.f=K=function(t){return t===Y||t===o?new i(t):X(t)},!c&&g(d)&&z!==Object.prototype)){s=z.then,st||(h(z,"then",(function(t,e){var n=this;return new Y((function(t,e){f(s,n,t,e)})).then(t,e)}),{unsafe:!0}),h(z,"catch",H["catch"],{unsafe:!0}));try{delete z.constructor}catch(mt){}y&&y(z,H)}a({global:!0,wrap:!0,forced:at},{Promise:Y}),v(Y,U,!1,!0),b(U),o=l(U),a({target:U,stat:!0,forced:at},{reject:function(t){var e=K(this);return f(e.reject,void 0,t),e.promise}}),a({target:U,stat:!0,forced:c||at},{resolve:function(t){return j(c&&this===o?Y:this,t)}}),a({target:U,stat:!0,forced:ct},{all:function(t){var e=this,n=K(e),r=n.resolve,i=n.reject,o=C((function(){var n=m(e.resolve),o=[],s=0,a=1;S(t,(function(t){var c=s++,u=!1;a++,f(n,e,t).then((function(t){u||(u=!0,o[c]=t,--a||r(o))}),i)})),--a||r(o)}));return o.error&&i(o.value),n.promise},race:function(t){var e=this,n=K(e),r=n.reject,i=C((function(){var i=m(e.resolve);S(t,(function(t){f(i,e,t).then(n.resolve,r)}))}));return i.error&&r(i.value),n.promise}})},e893:function(t,e,n){var r=n("1a2d"),i=n("56ef"),o=n("06cf"),s=n("9bf2");t.exports=function(t,e){for(var n=i(e),a=s.f,c=o.f,u=0;u<n.length;u++){var l=n[u];r(t,l)||a(t,l,c(e,l))}}},e8b5:function(t,e,n){var r=n("c6b6");t.exports=Array.isArray||function(t){return"Array"==r(t)}},e95a:function(t,e,n){var r=n("b622"),i=n("3f8c"),o=r("iterator"),s=Array.prototype;t.exports=function(t){return void 0!==t&&(i.Array===t||s[o]===t)}},f069:function(t,e,n){"use strict";var r=n("59ed"),i=function(t){var e,n;this.promise=new t((function(t,r){if(void 0!==e||void 0!==n)throw TypeError("Bad Promise constructor");e=t,n=r})),this.resolve=r(e),this.reject=r(n)};t.exports.f=function(t){return new i(t)}},f36a:function(t,e,n){var r=n("e330");t.exports=r([].slice)},f5df:function(t,e,n){var r=n("da84"),i=n("00ee"),o=n("1626"),s=n("c6b6"),a=n("b622"),c=a("toStringTag"),u=r.Object,l="Arguments"==s(function(){return arguments}()),f=function(t,e){try{return t[e]}catch(n){}};t.exports=i?s:function(t){var e,n,r;return void 0===t?"Undefined":null===t?"Null":"string"==typeof(n=f(e=u(t),c))?n:l?s(e):"Object"==(r=s(e))&&o(e.callee)?"Arguments":r}},f673:function(t,e,n){"use strict";n.d(e,"c",(function(){return i})),n.d(e,"a",(function(){return o})),n.d(e,"d",(function(){return s})),n.d(e,"b",(function(){return a}));var r=n("6bfb");const i=Object.freeze({done:!0,value:void 0});class o{constructor(t){this._json=t}get schema(){return this._json["schema"]}get batches(){return this._json["batches"]||[]}get dictionaries(){return this._json["dictionaries"]||[]}}class s{tee(){return this._getDOMStream().tee()}pipe(t,e){return this._getNodeStream().pipe(t,e)}pipeTo(t,e){return this._getDOMStream().pipeTo(t,e)}pipeThrough(t,e){return this._getDOMStream().pipeThrough(t,e)}_getDOMStream(){return this._DOMStream||(this._DOMStream=this.toDOMStream())}_getNodeStream(){return this._nodeStream||(this._nodeStream=this.toNodeStream())}}class a extends s{constructor(){super(),this._values=[],this.resolvers=[],this._closedPromise=new Promise(t=>this._closedPromiseResolve=t)}get closed(){return this._closedPromise}async cancel(t){await this.return(t)}write(t){this._ensureOpen()&&(this.resolvers.length<=0?this._values.push(t):this.resolvers.shift().resolve({done:!1,value:t}))}abort(t){this._closedPromiseResolve&&(this.resolvers.length<=0?this._error={error:t}:this.resolvers.shift().reject({done:!0,value:t}))}close(){if(this._closedPromiseResolve){const{resolvers:t}=this;while(t.length>0)t.shift().resolve(i);this._closedPromiseResolve(),this._closedPromiseResolve=void 0}}[Symbol.asyncIterator](){return this}toDOMStream(t){return r["a"].toDOMStream(this._closedPromiseResolve||this._error?this:this._values,t)}toNodeStream(t){return r["a"].toNodeStream(this._closedPromiseResolve||this._error?this:this._values,t)}async throw(t){return await this.abort(t),i}async return(t){return await this.close(),i}async read(t){return(await this.next(t,"read")).value}async peek(t){return(await this.next(t,"peek")).value}next(...t){return this._values.length>0?Promise.resolve({done:!1,value:this._values.shift()}):this._error?Promise.reject({done:!0,value:this._error.error}):this._closedPromiseResolve?new Promise((t,e)=>{this.resolvers.push({resolve:t,reject:e})}):Promise.resolve(i)}_ensureOpen(){if(this._closedPromiseResolve)return!0;throw new Error(this+" is closed")}}},f772:function(t,e,n){var r=n("5692"),i=n("90e3"),o=r("keys");t.exports=function(t){return o[t]||(o[t]=i(t))}},fc6a:function(t,e,n){var r=n("44ad"),i=n("1d80");t.exports=function(t){return r(i(t))}},fdbc:function(t,e){t.exports={CSSRuleList:0,CSSStyleDeclaration:0,CSSValueList:0,ClientRectList:0,DOMRectList:0,DOMStringList:0,DOMTokenList:1,DataTransferItemList:0,FileList:0,HTMLAllCollection:0,HTMLCollection:0,HTMLFormElement:0,HTMLSelectElement:0,MediaList:0,MimeTypeArray:0,NamedNodeMap:0,NodeList:1,PaintRequestList:0,Plugin:0,PluginArray:0,SVGLengthList:0,SVGNumberList:0,SVGPathSegList:0,SVGPointList:0,SVGStringList:0,SVGTransformList:0,SourceBufferList:0,StyleSheetList:0,TextTrackCueList:0,TextTrackList:0,TouchList:0}},fdbf:function(t,e,n){var r=n("4930");t.exports=r&&!Symbol.sham&&"symbol"==typeof Symbol.iterator},fea9:function(t,e,n){var r=n("da84");t.exports=r.Promise}}]);
//# sourceMappingURL=chunk-vendors.eb7fca41.js.map